package com.application.objects.threeD;

import com.application.basic.Vector;

/**
 * Object3D.java
 * 
 * An abstract class representing an OpenGL 3D graphical object
 * 
 * @author Derek Dupuis
 * @date December 3, 2012
 * 
 *       Edited by Abinaya Pitchai
 */

public abstract class Object3D {
	protected float xLoc, yLoc, zLoc;
	protected float xScale, yScale, zScale;
	protected float xRot, yRot, zRot;
	protected float color[];

	protected int id;

	protected float red;
	protected float green;
	protected float blue;

	// Sphere
	public float sphere_radius;

	// Cylinder
	public float baseRadius;
	public float topRadius;
	public float height;
	public int slices;
	public int stacks;

	// Angle of 3D Object
	public float radians;
	public Vector dr = new Vector(1.0f, 0.0f, 0.0f);

	public Object3D() {
		setLocation(0, 0, 0);
		setScale(1, 1, 1);
	}

	public void setLocation(float x, float y, float z) {
		this.xLoc = x;
		this.yLoc = y;
		this.zLoc = z;
	}

	public void setScale(float x, float y, float z) {
		this.xScale = x;
		this.yScale = y;
		this.zScale = z;
	}

	public void setOrientation(float x, float y, float z) {
		this.xRot = x;
		this.yRot = y;
		this.zRot = z;
	}

	public void setColor(float r, float g, float b) {
		this.red = r;
		this.green = g;
		this.blue = b;
	}

	public void setParametersSphere(float radius, float radians, Vector dor) {

		this.sphere_radius = radius;
		this.radians = radians;
		this.dr = dor;
	}

	public void setParametersCylinder(float bRadius, float tRadius,
			float height, int stacks, int slices, float radians, Vector dor) {
		this.baseRadius = bRadius;
		this.topRadius = tRadius;
		this.height = height;
		this.slices = slices;
		this.stacks = stacks;
		this.radians = radians;
		this.dr = dor;

	}

	public abstract void redraw();

}
