/*
 * ARBuilder.java: comments by rdb 02/09/12
 * 
 * Edit history (recent)
 * 02/09/12 rdb: - changed code in processData to call new ARDataSource method
 *                 to write the AR xfdl file.
 *               - stored the ARDataSource that is built in an instance variable 
 *               - added a public getDataSource() method to return it. This 
 *                 makes it easy to avoid building the ARDataSource twice,
 *                 like we have been doing in MREBuilder. This is probably never
 *                 going to be used anyway since I later decided it was better to
 *                 move the writing of the fdl file to ARDataSource instead of
 *                 doing it in processData. Now there is no reason for another
 *                 class to build an ARDataSource via ARBuilder.processData.
 *               
 */

package ar;

import java.io.IOException;
import org.xml.sax.SAXException;

import edu.unh.sdb.datasource.AxisOrdering;
import edu.unh.sdb.datasource.DataBlock;
import edu.unh.sdb.datasource.DataSource;
import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.datasource.IntegerIndexSpaceID;

/**
 * Builds an AR tree representation file given an MR hierarchy of data, and an
 * error tolerance. Produces the ARDL xml file that describes the octree AR
 * representation
 * 
 * Is this just a class to test ARDataSource and ARWriter? It seems that
 * MREBuilder does all this and more???
 * 
 * @author svohr
 */
public class ARBuilder {
	static private String input = null;
	static private String output = null;
	static private ARFloatError tolerance;

	// Non-static for the above - above is used for testing
	// -------------------- instance variables ------------------------------
	private ARDataSource ards = null;
	private ARFloatError errtolerance;

	// Abi
	private static ARDataSource arDataSource = null;
	private static AxisOrdering axisOrdering = null;

	public AxisOrdering getAxisOrdering() {
		return axisOrdering;
	}

	public static void setAxisOrdering(AxisOrdering axisOrdering) {
		ARBuilder.axisOrdering = axisOrdering;
	}

	public ARDataSource getArDataSource() {
		return arDataSource;
	}

	public ARBuilder(){
		
		
	}
	public ARBuilder(String[] test) {
		// String[] test = new String[2];
		// test[0] = "SolarDataVxb/vx-4x4x4-b-bin." ;
		// test[1] = "xfdl" ;
		// main(test);

	}
	

	public MRHierarchy generateMRHierarchy(String[] test) {

		DataBlock[] blocks = new DataBlock[3];
		MRHierarchy mr_abi = new MRHierarchy(3);

		String prefix_1 = test[0];
		String suffix_1 = test[1];

		DataSource ds_mr0 = DataSource.create("ds-mr0", prefix_1 + suffix_1);
		ds_mr0.activate();
		setAxisOrdering(ds_mr0.getFDLDescriptor().getAxisOrdering());
		blocks[0] = ds_mr0.subblock(ds_mr0.getBounds());

		DataSource ds_mr1 = DataSource
				.create("ds-mr1", prefix_1 + "bin.d1.fdl");
		ds_mr1.activate();
		blocks[1] = ds_mr1.subblock(ds_mr1.getBounds());

		DataSource ds_mr2 = DataSource
				.create("ds-mr2", prefix_1 + "bin.d2.fdl");
		ds_mr2.activate();
		blocks[2] = ds_mr2.subblock(ds_mr2.getBounds());

		DataBlock[] errBlocks = new DataBlock[3];
		DataSource ds_er1 = DataSource.create("ds-er1", prefix_1
				+ "bin.maxabs.d1.e1.fdl");
		ds_er1.activate();
		errBlocks[0] = null;
		errBlocks[1] = ds_er1.subblock(ds_er1.getBounds());

		DataSource ds_er2 = DataSource.create("ds-er2", prefix_1
				+ "bin.maxabs.d2.e2.fdl");
		ds_er2.activate();
		errBlocks[2] = ds_er2.subblock(ds_er2.getBounds());

		mr_abi.setData(blocks);
		mr_abi.setDataFDLNames(prefix_1 + suffix_1, 0);
		mr_abi.setDataFDLNames(prefix_1 + "bin.d1.fdl", 1);
		mr_abi.setDataFDLNames(prefix_1 + "bin.d2.fdl", 2);

		mr_abi.setError(errBlocks);
		mr_abi.setErrorFDLNames("This shouldn't be called", 0);
		mr_abi.setErrorFDLNames(prefix_1 + "bin.maxabs.d1.e1.fdl", 1);
		mr_abi.setErrorFDLNames(prefix_1 + "bin.maxabs.d2.e2.fdl", 2);

		mr_abi.setLevels(1);

		return mr_abi;

	}

	public void processData(MRHierarchy mr, float errortolerance) 
	{
		ards = new ARDataSource(mr);
		this.errtolerance = new ARFloatError(errortolerance);

		ARNode root = ards.getRoot();
		root.prune(errtolerance);
		root.condense();

		// Abi : old version lines
		// ards.writeARfdl( output );
		// ards.writeARfdl( "testsend_abi-ar.xml" ); // Bochichio's hack

//		try {
//			ARWriter out = new ARWriter();
//			// Abi : old version lines
//			// output = "testsend0-ar"; //rdb: looks like a Bochichio hack to
//			// test
//			// // simulation time version, this should
//			// // be set by a parameter to the method
//			output = "ParticleTracer-Tornado-1.5f.xml";
//			ards.writeARfdl(output);
//
//			if (output != null)
//				out.setOutputFile(output);
//
//			out.writeAR(ards);
//			out.close();
//		} catch (IOException e) {
//			System.err.println(e.getMessage());
//		}
		/********************/
		
		try
		 {
		 output = "ARTreeResTest-Wavelet-0.1.xml" ;
		 ARWriter out = new ARWriter();
		
		 if ( output != null )
		 out.setOutputFile( output );
		
		 out.writeAR( ards );
		 out.close();
		 }
		 catch ( IOException e )
		 {
		 System.err.println( e.getMessage() );
		 }

		return;
	}

	// ------------- getDataSource() ---------------------------
	/**
	 * Returns the ARDataSource that was built if this class is invoked by the
	 * processData method. If processData has not been called, null is returned.
	 * 
	 * @return
	 */
	public ARDataSource getDataSource() {
		return ards;
	}

	// ----------------------------------------------------------------------------------------------------------------------------
	// ----------------------------------------------------------------------------------------------------------------------------

	static private void parseArguments(String[] args) {
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("-o")) {
				if (i + 1 < args.length)
					output = args[i + 1];
				else
					System.err.println("Missing arg for -o option.");

				i += 1;
			} else if (args[i].equals("-f")) {
				if (i + 1 < args.length)
					input = args[i + 1];
				else
					System.err.println("Missing arg for -f option.");

				i += 1;
			} else if (args[i].equals("-e")) {
				if (i + 1 < args.length) {
					float e = Float.parseFloat(args[i + 1]);
					tolerance = new ARFloatError(e);
				} else
					System.err.println("Missing arg for -e option.");

				i += 1;
			} else if (args[i].equals("-h") || args[i].equals("-?")) {
				printUsage();
				System.exit(0);
			} else if (input != null)
				input = args[i];
			else {
				System.err.println("Unrecognized argument: " + args[i]);
			}
		}
		if (input == null) {
			printUsage();
			System.exit(1);
		}
	}

	static private void printUsage() {
		System.err
				.println("USAGE\n java ar.ARBuilder MRinputFile [OPTIONS]\n"
						+ "    MRinputFile\n"
						+ "        The MR representation used to build the tree (required).\n"
						+ "    -o OUT_FILE\n"
						+ "        The output file for the ARDL information. STDOUT is default.\n"
						+ "    -e VALUE\n"
						+ "        The error tolerance for the tree. Full tree built by default.\n"
						+ "    -h -?\n" + "        Prints this message.");
	}

	public static void main(String[] args) {
		/*------- hard coded tests not needed any more ----*
		String [] data = new String [3];
		String [] error = new String [3];
		
		data[0] = "data_0.fdl";
		data[1] = "data_1.fdl";
		data[2] = "data_2.fdl";
		
		error[1] = "error_1_2.fdl";
		error[2] = "error_2_3.fdl";
		/*********************/

		// parse the command line arguments
		// parseArguments( args );

		// Read MR from file.
		MRHierarchy mr = new MRHierarchy();

		// try
		// {
		// MRDLReader.readMRDLfile( mr, input );
		// }
		// catch ( SAXException e )
		// {
		// }
		// catch ( IOException e )
		// {
		// System.err.println( e.getMessage() );
		// }
		// // ************************************************
		DataBlock[] blocks = new DataBlock[3];
		MRHierarchy mr_abi = new MRHierarchy(3);

		String prefix_1 = args[0];
		String suffix_1 = args[1];

		DataSource ds_mr0 = DataSource.create("ds-mr0", prefix_1 + suffix_1);
		ds_mr0.activate();
		setAxisOrdering(ds_mr0.getFDLDescriptor().getAxisOrdering());
		blocks[0] = ds_mr0.subblock(ds_mr0.getBounds());

		DataSource ds_mr1 = DataSource
				.create("ds-mr1", prefix_1 + "bin.d1.fdl");
		ds_mr1.activate();
		blocks[1] = ds_mr1.subblock(ds_mr1.getBounds());

		DataSource ds_mr2 = DataSource
				.create("ds-mr2", prefix_1 + "bin.d2.fdl");
		ds_mr2.activate();
		blocks[2] = ds_mr2.subblock(ds_mr2.getBounds());

		DataBlock[] errBlocks = new DataBlock[3];
		DataSource ds_er1 = DataSource.create("ds-er1", prefix_1
				+ "bin.maxabs.d1.e1.fdl");
		ds_er1.activate();
		errBlocks[0] = null;
		errBlocks[1] = ds_er1.subblock(ds_er1.getBounds());

		DataSource ds_er2 = DataSource.create("ds-er2", prefix_1
				+ "bin.maxabs.d2.e2.fdl");
		ds_er2.activate();
		errBlocks[2] = ds_er2.subblock(ds_er2.getBounds());

		mr_abi.setData(blocks);
		mr_abi.setDataFDLNames(prefix_1 + suffix_1, 0);
		mr_abi.setDataFDLNames(prefix_1 + "bin.d1.fdl", 1);
		mr_abi.setDataFDLNames(prefix_1 + "bin.d2.fdl", 2);

		mr_abi.setError(errBlocks);
		mr_abi.setErrorFDLNames("This shouldn't be called", 0);
		mr_abi.setErrorFDLNames(prefix_1 + "bin.maxabs.d1.e1.fdl", 1);
		mr_abi.setErrorFDLNames(prefix_1 + "bin.maxabs.d2.e2.fdl", 2);

		mr_abi.setLevels(1);
		//
		// *******
		// build AR representation

		// ARDataSource ar = new ARDataSource( mr_abi );
		// arDataSource = ar ;

		// ARNode ar_root = ar.getRoot();
		//
		// ar_root.prune( new ARFloatError( error ));
		// ar_root.condense();

		float error = 750.0f;
		// processData(mr_abi, error) ;

		// ************************************************

		/** Tried for ar dump **/
		// tolerance = new ARFloatError(error);
		// ar.dumpBlocks( tolerance );
		// ----------------------------------------------------------------------------------------
		//
		// DataBlock[] blocks = new DataBlock[3];
		// MRHierarchy mr_abi = new MRHierarchy(3);
		//
		// DataSource ds_mr0 = DataSource.create("ds-mr0",
		// "DataSet/vx-1800-res3.xfdl");
		// ds_mr0.activate();
		// blocks[0] = ds_mr0.subblock(ds_mr0.getBounds());
		//
		// DataSource ds_mr1 = DataSource.create("ds-mr1",
		// "DataSet/raf1024x512x512-512node.vx.001800.res3.d1.fdl");
		// ds_mr1.activate();
		// blocks[1] = ds_mr1.subblock(ds_mr1.getBounds());
		//
		// DataSource ds_mr2 = DataSource.create("ds-mr2",
		// "DataSet/raf1024x512x512-512node.vx.001800.res3.d2.fdl");
		// ds_mr2.activate();
		// blocks[2] = ds_mr2.subblock(ds_mr2.getBounds());
		//
		// DataBlock[] errBlocks = new DataBlock[3];
		// DataSource ds_er1 = DataSource.create("ds-er1",
		// "DataSet/raf1024x512x512-512node.vx.001800.res3.maxabs.d1.e1.fdl");
		// ds_er1.activate();
		// errBlocks[0] = ds_er1.subblock(ds_er1.getBounds()); // check this
		// errBlocks[1] = ds_er1.subblock(ds_er1.getBounds());
		//
		// DataSource ds_er2 = DataSource.create("ds-er2",
		// "DataSet/raf1024x512x512-512node.vx.001800.res3.maxabs.d2.e2.fdl");
		// ds_er2.activate();
		// errBlocks[2] = ds_er2.subblock(ds_er2.getBounds());
		//
		// mr_abi.setData(blocks);
		// mr_abi.setDataFDLNames("DataSet/vx-1800-res3.xfdl", 0);
		// mr_abi.setDataFDLNames("DataSet/raf1024x512x512-512node.vx.001800.res3.d1.fdl",
		// 1);
		// mr_abi.setDataFDLNames("DataSet/raf1024x512x512-512node.vx.001800.res3.d2.fdl",
		// 2);
		//
		// mr_abi.setError(errBlocks);
		// mr_abi.setErrorFDLNames("DataSet/raf1024x512x512-512node.vx.001800.res3.maxabs.d1.e1.fdl",
		// 0);
		// mr_abi.setErrorFDLNames("DataSet/raf1024x512x512-512node.vx.001800.res3.maxabs.d1.e1.fdl",
		// 1);
		// mr_abi.setErrorFDLNames("DataSet/raf1024x512x512-512node.vx.001800.res3.maxabs.d2.e2.fdl",
		// 2);
		//
		//
		//
		// mr_abi.setLevels(1);
		//
		// ARDataSource ar = new ARDataSource( mr_abi );
		// arDataSource = ar ;

		// ----------------------------------------------------------------------------------------

		// write the ARDL file
		// try
		// {
		// output = "ARTreeResTest" ;
		// ARWriter out = new ARWriter();
		//
		// if ( output != null )
		// out.setOutputFile( output );
		//
		// out.writeAR( ar );
		// out.close();
		// }
		// catch ( IOException e )
		// {
		// System.err.println( e.getMessage() );
		// }

		return;
	}
}
