package ar;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.LinkedList;

import javax.print.attribute.HashAttributeSet;

import edu.unh.sdb.common.Datum;
import edu.unh.sdb.datasource.AxisOrdering;
import edu.unh.sdb.datasource.ConversionUtilities;
import edu.unh.sdb.datasource.DataBlock;
import edu.unh.sdb.datasource.DataSource;
import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.datasource.IntegerIndexSpaceID;
import edu.unh.sdb.datasource.RegularISPartitioning;
import edu.unh.sdb.datasource.RegularPartitionISID;

public class ARStorage { // ARSparseStorage
	

	private ArrayList<ARNodeBlock> blockList  ; 
    private ARDataSource arDataSource ; 
    private DataSource[] dsArr ; // instead have array of SparseStorage
    private boolean tableCreated;
   

	

	private Hashtable<Integer, ARStorageBlock > arStorageBlockTable  ; 
    private AxisOrdering axisOrdering ;
    private RegularISPartitioning[] blockPartioning  ; 
	




	// Variables created during impl for understanding purposes - not used in the actual implementation of  class anywhere.
	static File file = new File("Debug-run.txt");
    static FileWriter writer  ; 
	static int number = 0 ;
	
	 public static ARStorage[] arStorage = new ARStorage[3]; 
  

	public ARStorage(String[] filePrefixSuffixArr) {
    	
    	ARBuilder arBuilder = new ARBuilder(filePrefixSuffixArr);
		arDataSource = arBuilder.getArDataSource();
		blockList = new ArrayList<ARNodeBlock>() ; 
		axisOrdering = arBuilder.getAxisOrdering();
		arStorageBlockTable = new Hashtable<Integer, ARStorageBlock >() ; 
		blockPartioning = new RegularISPartitioning[arDataSource.getResolutions()];
		
		createDataSourceArray() ; 
		createPartioningArray(new IntegerIndexSpaceID(3, 3, 3)) ; 
		//createValidARNodeBlockList(arDataSource.getRoot(), blockList);
	
	}
    
    public AxisOrdering getAxisOrdering() {
		return this.axisOrdering;
	}
    
	public RegularISPartitioning getBlockPartioning(int axis) {
		return blockPartioning[axis];
	}
	
	 public DataSource getDsArr(int res) {
			return dsArr[res];
		}

	
	 public boolean isTableCreated() {
			return tableCreated;
		}

		public void setTableCreated(boolean tableCreated) {
			this.tableCreated = tableCreated;
		}
	 
    private void  createDataSourceArray() { 
    	int nResolutions = arDataSource.getResolutions();
    	String[] dsNames = arDataSource.getDataFDLNames(); 
    	dsArr = new DataSource[nResolutions];
    	// Note : Need to Replace with SparseStorage class 
    	for(int m = 0 ; m < nResolutions ; m++){
    		dsArr[m] = DataSource.create("dsResolutions_" + m, dsNames[m]);
    		dsArr[m].activate();
    	}
    	
    }

    private void createPartioningArray(IntegerIndexSpaceID spaceId){
    	// each res DB will be the space bounds, not sure about the partition space in each res. 
    	// Need to handle it later. For now setting only blockPartioning[of res 1]
    	blockPartioning[0] = 
    		new RegularISPartitioning(spaceId , dsArr[0].getBounds(),  dsArr[0].getFDLDescriptor().getAxisOrdering());
    	
    	
    }
    public ArrayList<ARNodeBlock> getBlockList() {
 		return blockList;
 	}


 	public ARDataSource getArDataSource() {
 		return arDataSource;
 	}
	
 	  public Hashtable<Integer, ARStorageBlock> getArStorageBlockTable() {
 			return arStorageBlockTable;
 		}
	
	/**
	 * This method traverse through each children of the node passing and 
	 * 		adds the leaf Node's ARNodeBlock to the BlockList
	 * @param n
	 * 			ARNode variable
	 * @param list
	 * 		  An ArrayList of ARNodeBlock where the leaf ARNodeBlocks are added
	 */
	private void createValidARNodeBlockList(ARNode n, ArrayList<ARNodeBlock> list){
		if(n == null){
			return ;
		}
	
		for(int i = 0 ; i < 7 ; i ++){
			ARNode child = (ARNode) n.children[i] ;
			if(child!= null &&  child.isLeaf()){
				list.add(child.getBlock());
			    
			}
			createValidARNodeBlockList((ARNode)n.children[i], list);
		}
	}
	/**
	 *  This method creates a Hash table ARStorageBlock class. 
	 *  Extracts block data with respective  block bounds and resolution Datasource 
	 *  	adds in the datablock variable of ARStorageBlock
	 * @param list
	 * 		ArrayList of ARNodeBlock that contains the ARNode data
	 */
	public void createARDataBlocksAndAddToARTable(){
		
		for(int i = 0 ; i < blockList.size() ; i++ ){
			ARNodeBlock blk = blockList.get(i);
			ARStorageBlock arBlock = new ARStorageBlock() ;
			int res = blk.getResolution(); 
			int[] dimArr = blk.getBounds().getDimensionsArray();
			
			// Treating this ARStorage as cache block so not adding now.
			DataBlock dataBlock = dsArr[res].subblock(new ISBounds(dimArr));
			//System.out.println("count : " + dataBlock.getValues().getFloats().length ); 
			dsArr[res].subblock(dataBlock, blk.getBounds());
			
			arBlock.setArDataBlock(dataBlock);
			arBlock.setArBlockBounds(blk.getBounds());
			
			
			//arStorageBlockTable.put(ConversionUtilities.arrayToString(arBlock.getArBlockBounds().getLowerArray()), arBlock);
			//System.out.println(" Hash bounds " + blk.getBounds());
			arStorageBlockTable.put(i, arBlock); 
		}
		this.setTableCreated(true);
	}
	
	public static void main(String args[]) throws IOException{
		    /**
		    * For pruning according to given error tolerance 
		    */
		   // Float error = 50.0f ; 
		   // ar_root.prune( new ARFloatError( error ));
		   // ar_root.condense();
	     
//			String[] testArr = null ; 
//		  // need to assign the test Arr values - just added now to avoid compilation errors
//		   ARStorage arStorage = new ARStorage(testArr) ;
//		   arStorage.createValidARNodeBlockList(arStorage.getArDataSource().getRoot(), arStorage.getBlockList());
//		   System.out.println("List size " + arStorage.getBlockList().size() ) ;
//		   
//		   
//		   arStorage.createARDataBlocksAndAddToARTable(arStorage.getBlockList());
		    
		    /**
		     *  Lines of code written for easy understanding of code
		     */
		    
//		     file.createNewFile();
//			 writer = new FileWriter(file); 
//			
//		     writeNode(arDataSource_x.getRoot()); 
//		     
//		     writer.flush();
//		     writer.close();
//			
		
		
		 String[] file_x = new String[2];
	       file_x[0] = "ARData/Solar_vx/vx-64x64x64-bin." ; 
	       file_x[1] = "xfdl" ; 
	       arStorage[0] = new ARStorage(file_x);
	      
	       String[] file_y = new String[2];
	       file_y[0] = "ARData/Solar_vy/vy-64x64x64-bin." ; 
	       file_y[1] = "xfdl" ; 
	       arStorage[1] = new ARStorage(file_y) ;
	       
	       String[] file_z = new String[2];
	       file_z[0] = "ARData/Solar_vz/vz-64x64x64-bin." ; 
	       file_z[1] = "xfdl" ; 
	       arStorage[2] = new ARStorage(file_z) ; 
	       
	       for(int y = 0 ; y < 64; y++ ){
	        	 for(int x = 0 ; x < 64; x++ ){ 
	        		
	        		  System.out.println("x, y : " + x + ", " + y);
	        		  IntegerIndexSpaceID id = new IntegerIndexSpaceID(x, y, 0);  
	        		     float dx = getPosValueInARDatasource(id, 0);
	        		     float dy = getPosValueInARDatasource(id, 1);
	        		     
	        		     System.out.println("dx : " + dx ) ; 
	        		     System.out.println("dy : " + dy ) ; 
	        		     
	        		
	        	}
	        }
	      
		
		
		}

	 private static float getPosValueInARDatasource(IntegerIndexSpaceID id, int index){

		 // Code used for Regular Partitioning - not used now since the AR block partitions were not equal.
//		 RegularPartitionISID  rpISID ; 
//		 RegularISPartitioning partitioning = arStorage[index].getBlockPartioning(0) ; 
//	      
//		 ISBounds containingBounds  = partitioning.getContainingPartitionBounds(id) ; 
//		 System.out.println("containingBounds " + containingBounds ) ; 
//		 ARStorageBlock arblock = arStorage[index].getArStorageBlockTable().get(index); 
//		 DataBlock dBlock =   arblock.getArDataBlock() ; 
//		 Datum dtm = arStorage[index].dsArr[index].createDatum();
//		 rpISID = new RegularPartitionISID(arStorage[index].getBlockPartioning(0), id );
//		 dBlock.datum( dtm, rpISID.getOffsetISID() );
//		 return  dtm.getFloat(0) ; // one only attribute available in current test data

		 
	       // Data accessing based on the position
	      
	       ArrayList<ARNodeBlock> list = arStorage[index].getBlockList();
	       System.out.println("list.size " + list.size() );
	       int h=0; 
	       for(h = 0 ; h < list.size(); h++){
	    	   boolean withinBlock = false ; 
	    	   ISBounds blkbounds = list.get(h).getBounds();
	    	   // index space dimensions for loop
	    	   for(int d = 0; d < id.dim() ; d++){
	    		   if(id.getCoord(d) >=  blkbounds.getLower(d) && id.getCoord(d) <= blkbounds.getUpper(d)){
	    			   withinBlock = true ; 
	    		   }else {
	    			  withinBlock = false ; 
	    			  break;
	    		   }
	    	   }
	    	  
	    	   if(withinBlock == true){
	    		   break ;
	    	   }       
	      
	       }
	     
	         ARStorageBlock arBlock = arStorage[index].getArStorageBlockTable().get(h);
	         ISBounds arBounds = arBlock.getArBlockBounds();
			 DataBlock dBlock =   arBlock.getArDataBlock() ; 
			 
			 int[] exacLoc = new int[3];
			 exacLoc[0] = id.getCoord(0) - arBounds.getLower(0) ;
			 exacLoc[1] = id.getCoord(1) - arBounds.getLower(1) ;
			 exacLoc[2] = id.getCoord(2) - arBounds.getLower(2) ;
			 
			 IntegerIndexSpaceID offsetCalcuID = new IntegerIndexSpaceID(exacLoc[0], exacLoc[1], exacLoc[2]);
			 
			 return dBlock.getFloat(offsetCalcuID, 0);
	    
	}
	 
	 /**
	  * 
	  */
	 public ISBounds traverseAndGetBlockBounds(ARNode node, IntegerIndexSpaceID id ){
	
		 for(int i=0; i < 8; i++){
			 ARNode child = (ARNode) node.children[i] ; 
			 if(child!= null){
				 if(child.getNumChildren() > 0){
					 // this is the node we need
					 boolean withinBlock = false ; 
					 ISBounds blkbounds = child.getBounds();
					 for(int d = 0; d < id.dim() ; d++){
						 if(id.getCoord(d) >=  blkbounds.getLower(d) && id.getCoord(d) <= blkbounds.getUpper(d)){
							 withinBlock = true ; 
						 }else {
							 withinBlock = false ; 
							 break;
						 }
					 }
					 if(withinBlock == true){
						 traverseAndGetBlockBounds(child, id) ; 
					 }
				 }else{
					 // Found the exact block ; 
					 return child.getBounds();
				 }
		    }
		 }
	    	return node.getBounds();
	 }
	
	/**
	 *  Write method written for easy understanding of ARTree during the developement
	 * @param n
	 * @throws IOException
	 */
	private static void writeNode(ARNode n) throws IOException{
		
		if(n == null){
			writer.write("\nNode is null\n");
			return ;
		}
		
		 int currNode = 0 ;
	     writer.write("START NODE :" + number  + "\n") ; 
		 
	     currNode = number ; 
	     number++;
		 String bounds = "\tBounds : " + n.getBounds() + "\n" ; 
		 
		 writer.write(bounds);

	     ARNodeBlock arBlk = n.getBlock();
	     String blk_bounds = "\tBlock : \n \t\t Bounds :" + arBlk.getBounds() + "\n" ; 
	     writer.write(blk_bounds);
	    
	     String res = "\t\t Res: " + arBlk.getResolution() + "\n";
	     writer.write(res); 
	     
	     String err = "\t\t Error: type -" + arBlk.getError().typeString() + "\t value - " + arBlk.getError().valueString() + "\n" ; 
	     writer.write(err);
	     
	     // write the tags for the child nodes
	      OctreeNode[] children = n.getChildren();
	      for (int i = 0; i < children.length; i++) {
	    	  if((ARNode) (children[i]) != null ){
	    		  writer.write("\nSTART OF CHILD: " + i + " OF NODE:" + currNode + "\n");
	    		  writeNode((ARNode) (children[i]));
	    	  }
	      }
		
	      writer.write("\nEND OF NODE :" + currNode + "\n");
	      
		
		
	}
	
	

		
		
	}


