///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//          U N I V E R S I T Y    O F    N E W    H A M P S H I R E         //
//                                                                           //
//                       COMPUTER  SCIENCE  DEPARTMENT                       //
//                                                                           //
//                    Graphics And Visualization Research                    //
//                                                                           //
//          --------------------------------------------------------         //
//                                                                           //
//  Professor:   Dan Bergeron <rdb@cs.unh.edu>                               //
//  Programmer:  Andy Foulks <rafoulks@cs.unh.edu>                           //
//  File:        RemoteConnection.h                                          //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef _REMOTE_CONNECTION_H_
#define _REMOTE_CONNECTION_H_

#include "StarObject.h"



/**
 *  A RemoteConnection encapsulates a socket connection to a remote
 *  server.  Messages can be sent through the socket.  To disconnect,
 *  delete the RemoteConnection object.
 *
 *  Messages are sent to the server using the methods
 *      sendStringGetXXX()
 *
 *  which send a printf-formatted variable length string to the
 *  server and cast the return value to the expected type.  Error
 *  handling is performed by underlying communication methods
 *  and the return value is set appropriately if the message
 *  did not get through.  For details, see
 *
 *      sendStringGetInt(const char* format, ...);
 *      sendStringGetFloat(const char* format, ...);
 *      sendStringGetRawData(void* buf, int bufsize, const char* format, ...)
 */

class RemoteConnection : public StarObject
{
private: // constants
    static const int BUFSIZE = 1024;

public:  // constructors
            RemoteConnection(const char* hostname, int port);
    virtual ~RemoteConnection();

public:  // api
    virtual int     sendStringGetInt        (const char* format, ...);
    virtual float   sendStringGetFloat      (const char* format, ...);
    virtual void    sendStringGetRawData    (void* buf, int bufsize, 
                                             const char* format, ...);
    virtual char*   sendStringGetString     (const char* format, ...);

    void enableByteSwapping()               { byteSwapFlag = true; };
    void disableByteSwapping()              { byteSwapFlag = false; };
    bool isConnected() { return mConnectionEstablished;};

private: // server communication methods
    int             connectToServer         (const char* hostname, int port);
    void            sendMessage             (int socketId, const char* message);
    void            receiveReply            (void* buf, int size, int socketId);
    char*           receiveStringReply      (int socketId);
    int             readbytes               (int sd,char *buf,int nbytes);
    int             writebytes              (int sd, const char *buf, int size);

private:  // members
    int             mSocketId;
    bool            mConnectionEstablished;
    bool            byteSwapFlag;

private: // disable copy constructor and operator= 
    /** copy constructor disabled by making it private */
    RemoteConnection(RemoteConnection&);
    /** operator= disabled by making it private */
    RemoteConnection& operator=(RemoteConnection&);
};



#endif // _REMOTE_CONNECTION_H_

