#include <stdio.h>
#include <string.h>

#include "RemoteConnection.h"
#include "GraniteRemoteIF.h"


int main(int argc, char** argv){

    // default server/port
    char server[1024];
    strcpy(server,"132.177.4.137");

    int port = 50000;

    // check arguments for server/port
    for (int i=1; i<argc; i++) {
   
      if (strstr(argv[i],"-server") != NULL) {
        if ( (i+1) < argc)
          strcpy(server,argv[i+1]);
      }
      else if (strstr(argv[i],"-port") != NULL) {
        if ( (i+1) < argc)
          port = atoi(argv[i+1]);
      }

    }


    // Create Granite Remote IF
    GraniteRemoteIF* myIF    = new GraniteRemoteIF(server, port);

    // Create Data Source
    int dsID = myIF->createDataSource("test",
                                      "/local/data/jpo39/Increment64.xfdl");
    fprintf(stderr," Datasource created with ID: %d\n", dsID);
     
    // retrieve a float value - expect 0.0
    float val = myIF->getFloat(dsID, 0, 0, 0, 0);
    fprintf(stderr, "  val(%d 0 0 0 0 ) = %f (expect 0.0)\n", dsID, val);

    // retrieve a float value - expect 27.0
    val = myIF->getFloat(dsID, 0, 1, 2, 3);
    fprintf(stderr, "  val(%d 0 1 2 3 ) = %f (expect 27.0)\n", dsID, val);

    // retrieve all floats for data source
    // - expect incrementing values from 0.0 to 63.0
    float arr[64];
    myIF->getFloatsPointOrder(dsID, arr, 64*(sizeof(float)));

    printf(" Floats returned :  (expect 0.0 to 63.0 in 1.0 increments)\n\n");
    for (int i=0; i<64; i++) {
      if (i%4 == 0)
        printf("\n");
      printf("%6.2f ", arr[i]);
    }
    printf("\n\n");
   

    //------------------------------------------------------------
    //  SECOND DATA SOURCE
    //------------------------------------------------------------

    // create a subblock of data from original data source
    int subID = myIF->subblock(0, 1, 1, 1, 3, 3, 3);
    fprintf(stderr," Subblock created with ID: %d\n", subID);

    // retrieve a float value - expect 21.0
    val = myIF->getFloat(subID, 0, 0, 0, 0);
    fprintf(stderr, "  val(%d 0 0 0 0 ) = %f (expect 21.0)\n", subID, val);
      

    // retrieve all floats for data source
    // Expect:
    //
    //  21.00  22.00  23.00 
    //  25.00  26.00  27.00 
    //  29.00  30.00  31.00 
    //  37.00  38.00  39.00 
    //  41.00  42.00  43.00 
    //  45.00  46.00  47.00 
    //  53.00  54.00  55.00 
    //  57.00  58.00  59.00 
    //  61.00  62.00  63.00 

    float arr2[27];
    myIF->getFloatsPointOrder(subID, arr2, 27*(sizeof(float)));

    printf(" Floats returned :  (expect 21.0 22.0 23.0 25.0 26.0 27.0 29.0 "
           "30.0 31.0 37.0 ...\n\n");
    for (int i=0; i<27; i++) {
      if (i%3 == 0)
        printf("\n");
      printf("%6.2f ", arr2[i]);
    }
    printf("\n\n");



    //------------------------------------------------------------
    //  THIRD DATA SOURCE
    //------------------------------------------------------------

    // Create a multi attribute datasource
    // - using same binary data as first data source but now its
    //   contents have two attributes at each point for a total of
    //   32 data locations rather than 64
    int dsMultiID = myIF->createDataSource("test2",
                                "/local/data/jpo39/Increment64_2attr.xfdl");
    fprintf(stderr," Dual Attribute Datasource created with ID: %d\n",
                   dsMultiID);

    // retrieve a float value - expect 0.0
    val = myIF->getFloat(dsMultiID, 0, 0, 0, 0);
    fprintf(stderr, "  val(%d 0 0 0 0 ) = %f (expect 0.0)\n", dsMultiID, val);

    // retrieve a float value - expect 1.0
    val = myIF->getFloat(dsMultiID, 1, 0, 0, 0);
    fprintf(stderr, "  val(%d 1 0 0 0 ) = %f (expect 1.0)\n", dsMultiID, val);

   // retrieve a float value - expect 0.0
    val = myIF->getFloat(dsMultiID, 0, 1, 1, 7);
    fprintf(stderr, "  val(%d 0 1 1 7 ) = %f (expect 62.0)\n", dsMultiID, val);

    // retrieve a float value - expect 1.0
    val = myIF->getFloat(dsMultiID, 1, 1, 1, 7);
    fprintf(stderr, "  val(%d 1 1 1 7 ) = %f (expect 63.0)\n", dsMultiID, val);

    // retrieve all floats for data source in point order
    // - expect incrementing values from 0.0 to 63.0
    float arr3[64];
    myIF->getFloatsPointOrder(dsMultiID, arr3, 64*(sizeof(float)));

    printf(" Floats returned :  (expect 0.0 to 63.0 in 1.0 increments)\n\n");
    for (int i=0; i<64; i++) {
      if (i%4 == 0)
        printf("\n");
      printf("%6.2f ", arr3[i]);
    }
    printf("\n\n");


    // retrieve all floats for data source in attribute order
    // - expect incrementing values from 0.0 to 62.0 in increments of 2.0
    // - followed by 1.0 to 63.0 in  increments of 2.0
    float arr4[64];
    myIF->getFloatsAttrOrder(dsMultiID, arr4, 64*(sizeof(float)));

    printf(" Floats returned :  (expect 0.0 to 62.0 in 2.0 increments)\n");
    printf("                    (followed by 1.0 to 63.0 in increments of 2.0)");
    for (int i=0; i<64; i++) {
      if (i%4 == 0)
        printf("\n");
      printf("%6.2f ", arr4[i]);
    }
    printf("\n\n");


    // Testing get byte order
    int order = myIF->getByteOrder(dsID);
    printf(" Byte order of Datasource with id: %d  is ",dsID);
    if (order == 0)
       printf("BIG ENDIAN\n");
    else if (order == 1)
       printf("LITTLE ENDIAN\n");
    else
       printf("<ERROR!! unrecognized value of %d>\n",order);
 


    // Testing set byte order
    printf(" Setting DS order to LITTLE_ENDIAN\n");

    myIF->setByteOrder(dsID,"LITTLE_ENDIAN\n");

    order = myIF->getByteOrder(dsID);
    printf(" Byte order of Datasource with id: %d  is ",dsID);
    if (order == 0)
       printf("BIG ENDIAN\n");
    else if (order == 1)
       printf("LITTLE ENDIAN\n");
    else
       printf("<ERROR!! unrecognized value of %d>\n",order);


    printf(" Setting DS order to BIG_ENDIAN\n");

    myIF->setByteOrder(dsID,"BIG_ENDIAN");

    order = myIF->getByteOrder(dsID);
    printf(" Byte order of Datasource with id: %d  is ",dsID);
    if (order == 0)
       printf("BIG ENDIAN\n");
    else if (order == 1)
       printf("LITTLE ENDIAN\n");
    else
       printf("<ERROR!! unrecognized value of %d>\n",order);

    // Testing GetDimensions
    int* arr9;
    int size;

    arr9 = myIF->getDimensions(dsID, &size);

    fprintf(stderr, " DS %d has:  %d dimensions", dsID, size);

    return 0;   
}
