#ifndef WIN32
#include <arpa/inet.h>
#endif
#include <fstream>
#include <iostream>
#include <cstdlib>
#include <udt.h>
#include <jni.h>
#include "edu_unh_sdb_remote_UDTReceiver.h"

using namespace std;
extern "C"{
JNIEXPORT jint JNICALL 
Java_edu_unh_sdb_remote_UDTReceiver_nativeReceiveData(JNIEnv *env, jclass cls, jbyteArray jarr, jstring servIP, jint port, jint size)
{
    const char *ip;
	 int time_out = 20*1000;  
    
    ip = env->GetStringUTFChars(servIP, NULL);
    if (ip == NULL) {
        return 1; /* OutOfMemoryError already thrown */
    }
    /*
    printf("In native method, %s\n", ip); 
		printf("In native method, %d\n", port);
		printf("In native method, %d\n", size);
		*/
   UDTSOCKET fhandle;
	if (UDT::INVALID_SOCK == ( fhandle = UDT::socket(AF_INET, SOCK_STREAM, 0)))
	{
		cout << "socket: " << UDT::getlasterror().getErrorMessage() << endl;
 	   env->ReleaseStringUTFChars(servIP, ip);
		return 1;
	}

    //printf("--------------------------------------------\n"); 
	 //printf("In native method, UDTSOCKET is created\n");

   sockaddr_in serv_addr;
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_port = htons(short(port));

		//printf("In native method, serv_addr is filled\n");

#ifndef WIN32
   if (inet_pton(AF_INET, ip, &serv_addr.sin_addr) <= 0)
#else
   if (INADDR_NONE == (serv_addr.sin_addr.s_addr = inet_addr(ip)))
#endif
   {
      cout << "incorrect network address.\n";
      return 1;
   }
   memset(&(serv_addr.sin_zero), '\0', 8);
    //printf("In native method, before connect ... \n"); 

   UDT::setsockopt(fhandle, 0, UDT_RCVTIMEO, &time_out, sizeof(int));
   if (UDT::ERROR == UDT::connect(fhandle, (sockaddr*)&serv_addr, sizeof(serv_addr)))
   {
   	cout << "UDT::ERROR: connect() " << UDT::getlasterror().getErrorMessage() << endl;
   	cout << "Error code = " << UDT::getlasterror().getErrorCode() << endl;
 	   env->ReleaseStringUTFChars(servIP, ip);
      UDT::close(fhandle);
      return 1;
   }

    //printf("In native method, connected ! \n"); 

   // get size information
//    int datasize;
   //int recvSize = 0;

//    if (UDT::ERROR == (recvSize=UDT::recv(fhandle, (char*)&datasize, sizeof(int), 0)))
//    {
//       cout << "recv: " << UDT::getlasterror().getErrorMessage() << endl;
//       return 0;
//    }
//    int targetsize = *((int*)&datasize);
   
   //cout << "Planned recv file size: " << datasize << "-" << targetsize << endl;
   	
	 int recvsize = 0;
	 
	 //char buf[datasize]; //can not just use a char pointer or a segment fault will be generated
   jbyte *cppPtr = env->GetByteArrayElements(jarr, 0); //not (env, buf)
   if(cppPtr == NULL){
      cout << "recv: GetByteArrayElements() returned NULL" << endl;
    	return 1; /* Exception occurred */
   }
   char* buf = (char*)(&(*cppPtr));
   //char* buf;
   //buf = new char[datasize];
	 int handle;
   // receive the stream
	//datasize = 9894;
	
   //cout << "In native method, (cout) size - " << size << endl;
   //printf("In native method, receiving data...");
   fflush(stdout);
   
	UDT::setsockopt(fhandle, 0, UDT_RCVTIMEO, &time_out, sizeof(int));
	   
   if (UDT::ERROR == (recvsize = UDT::recv(fhandle, buf, size, 0, &handle)))
   {
	   jclass newExcCls;
      cout << "Error in native method, recv(): " << UDT::getlasterror().getErrorMessage() << endl;
      fflush(stdout);
		///////////////////////////////////////////////////
		//////// Newly added to throw exception ///////////

	   newExcCls = (env)->FindClass("java/io/IOException");
	   if (newExcCls == NULL) {
			/* Unable to find the exception class, give up. */
			env->ReleaseByteArrayElements(jarr, cppPtr, 0);
			env->ReleaseStringUTFChars(servIP, ip);
			UDT::close(fhandle);
			return 1;
	   }

		///////////////////////////////////////////////////
		//if the sender failed, this will eventually time out
		//if the sender succeeded, this will be received by sender
		//namely, this scheme makes sure the server does not close before
		//the receiver close or else it is highly likely the receiver will
		//get connection broken error
		int sendSize;
		int notok = 1;
	   UDT::setsockopt(fhandle, 0, UDT_SNDTIMEO, &time_out, sizeof(int));
		if (UDT::ERROR == (sendSize=UDT::send(fhandle, (char*)&notok, sizeof(int), 0, &handle)))
		{
			env->ReleaseByteArrayElements(jarr, cppPtr, 0);
			env->ReleaseStringUTFChars(servIP, ip);
			UDT::close(fhandle);
			return 1;
		}

		env->ReleaseByteArrayElements(jarr, cppPtr, 0);
		env->ReleaseStringUTFChars(servIP, ip);
		UDT::close(fhandle);
	   env->ThrowNew(newExcCls, "Receive Timeout Exception from Native Code!");

      return 1;
   }
	
	//the receive operation is successful, the send must be successful
	//and the sender must be waiting for the confirmation
	int sendSize;
	int ok = 0;
	UDT::setsockopt(fhandle, 0, UDT_LINGER, &time_out, sizeof(int)); // PJR
   UDT::setsockopt(fhandle, 0, UDT_SNDTIMEO, &time_out, sizeof(int));
   if (UDT::ERROR == (sendSize=UDT::send(fhandle, (char*)&ok, sizeof(int), 0, &handle)))
   {
		env->ReleaseByteArrayElements(jarr, cppPtr, 0);
		env->ReleaseStringUTFChars(servIP, ip);
		UDT::close(fhandle);
      return 1;
   }

   //cout << "In native method, (cout) recvsize - " << recvsize << endl;
   //printf("In native method, received %d bytes\n", recvsize);
	 	 
   env->ReleaseByteArrayElements(jarr, cppPtr, 0);
   env->ReleaseStringUTFChars(servIP, ip);
   UDT::close(fhandle);
   //printf("In native method, done and returning to Java!!! \n");
   //printf("--------------------------------------------\n"); 
   //printf("done.\n");
	fflush(stdout);
   return 0;
}
};
