//----------------------- ARRSCopyModelBuilder class ----------------------------
//
// ARRSCopyModelBuilder class
//    builds an arrs tree step by step from an input file
//    This is an abstract class created from Li Ye's concret ARRSCopyModelBuilder class that
//       is now named ARRSCopyModelBuilder.
// 
// Created 10/24/05 by rdb
//
//------------------------- Edit history -----------------------
// 01/02/06 rdb: moved constructor code based on Copy version of ARDS to its Builder
// 10/24/05 rdb: renamed getBaseBounds to makeBaseBounds
//-------------------------------------------------------------------------------

package edu.unh.sdb.ARRSTreeBuilder;

import java.io.*;

import edu.unh.sdb.datasource.*;

public abstract class ARRSBuilder
{

   protected int dim = 0;
   protected int nodeNum = 0;
   protected CellIndexable arrSource;
   protected InputStreamReader inFile = null;
   protected ARRSTree tree = null;
   protected ARCellParser parser = null;

   public ARRSBuilder( String command, CellIndexable arrSource)
   {
      // Some day may be able to share some code for all forms of Cell DS??
   }

   public boolean isDim( int i )
   {
        if ( this.dim == 0 )
           this.dim = i;
        return ( this.dim == i );
     }

   public boolean addKdNode( int parent, int bounds[] )
   {
        /*System.out.println( "kdTreeNode: " + nodeNum +
                            " child of: " + parent +
                            " at " + aToString( bounds ));
        */
        if ((bounds.length < 2)||(bounds.length%this.dim != 0)
            ||(parent < -1)) {
           return false;
        } 
        ISBounds baseb = new ISBounds( bounds );
        //ISBounds baseb = this.getBaseBounds(bounds);
        KDNode kdnode = new KDNode(baseb, this.arrSource);
        tree.addNode( kdnode, parent );
        nodeNum++;
        return true;
     }

   public boolean addNtreeNode( int parent, int bounds[] )
   {
       /* System.out.println( "nTreeNode:  " + nodeNum +
                            " child of: " + parent +
                            " at " + aToString( bounds ));
       */
        if (   ( bounds.length < 2 ) 
            || ( bounds.length % this.dim != 0 )
            || ( parent < -1 ) ) 
        {
           return false;
        } 
        ISBounds baseb = this.makeBaseBounds( bounds );
        NTreeNode kdnode = new NTreeNode( baseb, this.arrSource );
        tree.addNode( kdnode, parent );
        nodeNum++;
        return true;
     }

   /**
    * Abstract method to create a leaf node with data. Children classes must define
    * this method since this is what distinguishes the child implementations.
    * @param parent parent node for this leaf
    * @param bounds 2*dim array; 1st dim are lower bounds of leaf block in base space
    *                            2nd dim are upper bounds of leaf block in base space
    * @param size   array of length dim defining size of leaf block in points (?)
    */
   abstract public boolean addLeaf( int parent, int bounds[], int size[] );
   
   /**
    * Return an ISBounds representing the base bounds for this AR Tree
    * @param bounds is a 2*dim array where 1st dim entries are lower bounds, 
    *               and last dim entries are upper bounds.
    * @return
    */
   protected ISBounds makeBaseBounds( int[] bounds )
   {
        int[] lower = new int[this.dim];
        int[] upper = new int[this.dim];
        for ( int i = 0; i < this.dim; i++ ) 
        {
           lower[i] = bounds[i];
           upper[i] = bounds[i+this.dim];
        }
   
        return new ISBounds( new IntegerIndexSpaceID( lower ), 
                             new IntegerIndexSpaceID( upper ));
     }

   /** Get the ARRS tree */
   public ARRSTree getARRSTree()
   {
      //rdb: really should only buildTreePath if it is needed
      this.tree.buildTreePath();
      return this.tree;
   }
}
