//----------------------- ARRSCopyModelBuilder class ----------------------------
//
// ARRSCopyModelBuilder class
//    builds an arrs tree step by step from an input file
//
//------------------------- Edit history (since 10/21/05) -----------------------
// 01/02/06 rdb: Moved copy-dependent code from ARRSBuilder constructor to here.
// 10/21/05 rdb: Cleaned up unused variables and some reformatting.
//-------------------------------------------------------------------------------

package edu.unh.sdb.ARRSTreeBuilder;

//import java.io.*;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;
import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;

public class ARRSCopyModelBuilder extends ARRSBuilder
{
  public float[] indexValues = null;
  private int indexposition = 0;
 
  //--------------------------------------------------------------
  public ARRSCopyModelBuilder( String command, String indexFDL, CellIndexable arrSource )
  {
     super( command, arrSource ); // This doesn't do anything right now.
     ISBounds b;
     //rdb int[] numValues;
     float[] fvalues;  
     /*set index value array*/
     DataSource indexDS=DataSource.create( " ", indexFDL );
     indexDS.activate();
  // System.out.println("Beginning of ARRSCopyModelBuilder: "); //-------
  // System.out.println("indesDS: \n"); //-------
   //System.out.println(indexDS.toString()); //-------
  
     b = indexDS.getBounds();
   
     //System.out.println("indexDS bounds: "+b.toString()); //-------
     //numValues = new int[b.volume()*indexDS.getNumAttributes()];
     fvalues = new float[b.volume()*indexDS.getNumAttributes()];
     //IntValues values = new IntValues(numValues);
     FloatValues values = new FloatValues(fvalues);
     BasicBlock indexDB = new BasicBlock( indexDS.getBounds(), 
                                          indexDS.getRecordDescriptor(), values ); 
     indexDS.subblock(indexDB, indexDS.getBounds(), indexDS.getBounds()); 
     indexDS.deactivate();
     //System.out.println("indexDB: \n"); //-------
     //System.out.println(indexDB.toString()); //-------
        
     this.indexValues = ((FloatValues)indexDB.getValues()).getDataReference();
     //System.out.println("indexValues len: " + this.indexValues.length); //-------

        /* set the ARRS tree */
      this.tree = new ARRSTree();
      /* set the ARRS source */
      this.arrSource = arrSource;

      /* set inFile */
      if ( command == null )
         this.inFile = new InputStreamReader( System.in );
      // this.inFile = new InputStreamReader(new FileInputStream(System.in));
      else try
      {
         this.inFile = new InputStreamReader( new FileInputStream( command ) );
      } catch ( IOException e )
      {
         System.err.println( "File not found: " + command );
         System.err.println( "Using standard input" );
         this.inFile = new InputStreamReader( System.in );
         // this.inFile = new InputStreamReader(new FileInputStream(System.in));
      }

      /* set the parser */
      try
      {
         this.parser = new ARCellParser( this.inFile, this );
         this.parser.readAllCommands();
      } 
      catch ( IOException e )
      {
      }
   }

  public boolean addLeaf( int parent, int bounds[], int size[] )
  {
    /* System.out.println( "LeafNode:   " + nodeNum +
                         " child of: " + parent +
                         " at " + aToString( bounds ) +
                         " size:  " + aToString( size ));
*/
     if (( bounds.length < 2 ) || ( bounds.length != 2 * this.dim  )
         || (parent < -1 )) {
        return false;
     } 
     ISBounds baseb = this.makeBaseBounds( bounds );
     ARRSCopyLeafNode leafnode = new ARRSCopyLeafNode( baseb, this.arrSource );
     int  len = size.length;
     int[] lower = new int[this.dim];
     int[] upper = new int[this.dim];
     for ( int i = 0; i < len; i++) 
     {
        lower[i] = 0;
        upper[i] = size[i] - 1;
//        num *= size[i];
     }
     /*set index range*/
 //    low = this.indexposition;
  //   high = low + num - 1;

     /*update the indexposition*/
   //  this.indexposition += num;

     /*create an IndexBlock*/
     //IndexBlock indexb = new IndexBlock(this.indexValues, low, high);
     
     //rdb: How can the following ever work? 12/26/05
     //rdb IndexBlock indexb = new IndexBlock( this.indexValues, 0, 0 );
     
     IndexBlock indexb = null;
     
     if ( this.indexValues == null )
        indexb = new IndexBlock( this.indexValues, 0, 0 );  // rdb: should get here
     else
        indexb = new IndexBlock( this.indexValues, 0, this.indexValues.length - 1 );         
     
     //set values to the ARRSCopyLeafNode 
     leafnode.setIndexBlock( indexb );
     leafnode.setIndexBlockBounds( new ISBounds( new IntegerIndexSpaceID( lower ), 
                                                 new IntegerIndexSpaceID( upper )) );
     
     //rdb ?????? formerly tree.addNode, but leaf nodes are not added???
     tree.addNode( leafnode, parent );
     nodeNum++;
     return true;
  }

  /* set low and high index value */
  private void setIndex()
  {
     int  len, num, low, high;
     ArrayList leaves = null;
     ARRSCopyLeafNode node;
     IndexBlock indexb;
     ISBounds b;
     this.tree.buildLeafList();
     leaves = this.tree.getLeafList();
     len = leaves.size();
     for ( int i = 0; i < len; i++ ) 
     {
        node = (ARRSCopyLeafNode)leaves.get(i);
        indexb = node.getIndexBlock();
        b = node.getIndexBlockBounds();
        num = b.volume();
        low = this.indexposition;
        high = low + num - 1;
        this.indexposition += num;
        indexb.setLowIndex(low);
        indexb.setHighIndex(high);
     }
     
  }
  /** Get the ARRS tree -- override parent to handle index information.
   * 
   */
  public ARRSTree getARRSTree()
  {
     // override parent to handle index component
     this.setIndex();
     return super.getARRSTree();  // and let parent do its thing as well
  }

  /************************** rdb delete since not used
  private StringBuffer aToString( int a[] )
  {
     StringBuffer out = new StringBuffer("[");
     for ( int i = 0; i < a.length; i++ )
        out.append( String.valueOf( a[i] ) + "," );
     out.setCharAt( out.length()-1, ']' );
     return out;
  }
   /***********************************************/
}


