/* 
    InternalFaultError.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
*/

//---------------------- Edit history as of 11/24/05 ------------------------------
// 11/24/05 rdb: revised (and simplified) to use new SDBError.logError 
//---------------------------------------------------------------------------------

package edu.unh.sdb.common;
//import java.io.*;

/** Used whenever something happens that should be impossible occurrs.
 *     For example, if a data member has an illegal value, then this error
 *     should be created.
 *       In theory, this should never happen. And even in practice, if
 *     we check constructor and method arguments carefully, we would generate
 *     an IllegalArgumentError earlier, rather than an InternalFaultError
 *     later on. 
 *        
 *     Creating an error will cause the message argument to be logged, and
 *     an exception to be raised, if appropriate. This behavior is governed
 *     by the error level given to the constructor, and by the error reporting
 *     level maintained in the SDBError class. 
 *     
 *     Invoking the constructor is sufficient to log the error. There is no
 *     need to save a reference to the error object.
    
 */

public class InternalFaultError extends SDBError
{
    /** Create an error with the specified error level and message.
        The error level can be either SDBError.Fatal or 
        SDBError.Recoverable.
    */
    
    public InternalFaultError( int errorLevel, String message )
    {
        super( errorLevel, new InternalFaultException( message ) );
    }  
} 
