/* 
    ShortDatum.java
    Copyright 2003 University of New Hampshire
    Durham, NH 03824
    ------------------------- Edit history (started 11/24/04) -----------
    10/21/05 rdb: Removed unused variables
    01/14/05 rdb: Added getStorageType() and byteSize() and getValues()
    01/11/05 rdb: -Redefined semantics of ShortDatum( short[], int start, int numattr );
                  -deprecated assign( short[], int datumIndex, int numattr )
                  -renamed assign methods to setShorts, but changed semantics of
                      setShorts( short[], int start, int numattr ) 
    
    11/24/04 rdb: 
             added missing complement of set<type> and get<type> methods
             added override of Datum.getShorts() and Datum.setShorts
             also  added try-catch blocks around attribute index usage.
*/

package edu.unh.sdb.common;
import java.util.*;


/**
 *  Represent multiple data values for a single point.
 *  This implementation stores values as an array of shorts, regardless
 *  of the orginal type of an attribute. 
 */

public class ShortDatum extends Datum{
    
    private short [] values;
    
    
    /** Creates a null datum with the specified number of values.*/
    public ShortDatum(int numvalues){
        
        this.values=new short[numvalues];
        Arrays.fill(this.values,(short)-32767);
    }
    
	/** Creates a datum with a single attribute containing the argument value. 
	 *  @deprecated 
	 */
    public ShortDatum( short value ){
    
        this.values=new short[1];
        this.values[0]=value;    
    }
    
    /** Creates a datum with attributes containing the argument values. */
    public ShortDatum( short values[])
    {
        this.values=values.clone();
    }
	
    /** Creates a datum by copying a subset of values from an array. 
     * The second argument specifies where to start copying;
     * the 3rd argument specifies how many values to copy.
     *  
     *
    	* @deprecated
    	*   Not really deprecated. Just looking for usage.
    	* 	specifies the datum index of the datum to be copied, where all datums in
    	* 	the array are assumed to be of the same length. 
    */
    public ShortDatum(short values[], int startIndex, int numAttributes)
    {
        this.values=new short[numAttributes];
		//rdb int startIndex=datumIndex*numAttributes;
		
		for(int i=0; i < numAttributes; i++)
		{
		   this.values[i] = values[startIndex + i];
		}
	}
	
	/** Creates a datum with attributes containing the argument values. */
    public ShortDatum( int values[] )
   {
      int i, n = values.length;
      this.values = new short[values.length];

      for ( i = 0; i < n; i++ )
      {
         this.values[i] = (short) values[i];
      }
   }
	
    /** Copy Constructor creates a copy of the argument datum.*/
    public ShortDatum(ShortDatum d)
    { 
        this.values= d.values.clone();
    }        

    /** Copy Constructor creates a copy of the argument datum.*/
    public ShortDatum(Datum d)
    { 
		this.values = new short[ d.size() ];
		
		for ( int i=0; i<d.size() ; i++)
			this.values[i]=d.getShort(i);
    }        
	
	/** This constructor creates a single ShortDatum from an array of datums. 
	 *	The Datums in the array are copied in the order they are given.
	*/
	public ShortDatum( ShortDatum [] datums )
	{
		int totalSize = 0;
		int datumSize;
		int index = 0;
		
		for ( int i=0; i < datums.length; i++ )
			totalSize += datums[i].size();
			
		this.values = new short [totalSize];
		
		for ( int i=0; i < datums.length; i++ )
		{
			datumSize = datums[i].size();
			for ( int j=0; j<datumSize; j++ )
				this.values[ index++ ] = datums[i].values[j];
		}
	}
	
	/** Return a copy of this datum. */
    public Object clone(){
        
        return new ShortDatum(this);
    }    
    
    /** Copy the values of this ShortDatum object into the argument datum. */
    public void copyTo(ShortDatum d){
	
		try{
		
			System.arraycopy(this.values,0,d.values,0,this.values.length);
			
		} catch ( ArrayIndexOutOfBoundsException e ) 
		  {
			new SDBError( SDBError.Recoverable, "Datum size mismatch." );
		  }
		  catch ( NullPointerException e )
		  {
			if ( d == null) 
				new IllegalArgumentError( SDBError.Recoverable,
				      "Argument Datum was null or had no data.");
			else 
				new InternalFaultError( SDBError.Recoverable, "This Datum has no data.");
		  }
    }

    /** Copy the values of this ShortDatum object into the argument datum. */
    public void copyTo(Datum d){
	
		try{
			int i,n=d.size();
			
			for(i=0; i<n; i++){
			
				d.setShort(i,this.values[i]);
			}
		} catch (ArrayIndexOutOfBoundsException e){
		
			new SDBError(SDBError.Recoverable,"Datum size mismatch.");
		}
		
		catch (NullPointerException e){
		
			if(d == null){
				new IllegalArgumentError(SDBError.Recoverable,"Argument Datum was null.");
			} else {
				new InternalFaultError(SDBError.Recoverable, "This Datum has no data.");
			}
		}
    }
        
    /** Determine if the argument is a ShortDatum with the same set of data values. */
    public boolean equals( Object o )
    {
        try 
        {
            ShortDatum d = (ShortDatum) o;
            if ( d.values.length != this.values.length )
               return false;
            for ( int i=0; i < this.values.length; i++ )
            {
    			if ( this.values[i] != d.values[i] )
                    return false;
    		}
        } catch ( Exception e ) // any exception means not equal
          {
              return false;
          }
        return true;
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setDouble( int i, double v )
    {
       try {  values[i]= (short) v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting double to short" ); }
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setFloat( int i, float v )
    {
       try {  values[i]= (short) v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting float to short" ); }
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setInt( int i, int v )
    {
       try {  values[i]= (short) v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting int to short" ); }
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setShort( int i, short v )
    {
       try {  values[i]=  v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setByte( int i, byte v )
    {
       try {  values[i] =  v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
    }
    
    /** Assigns values to an existing datum by copying a subset of values from an array. 
     *  The second argument specifies where to start copying;
     *  the 3rd argument defines how many values to copy.
     *  If this object contains too few attributes to
     * 	hold the new data, an error is created. 
     */
	public void setShorts(short values[], int startIndex, int numAttributes)
	{		
		try
		{
			for ( int i=0; i < numAttributes; i++ )
			   this.values[i] = values[startIndex + i];
		} catch (ArrayIndexOutOfBoundsException e)
		  {
		     new IllegalArgumentError(SDBError.Recoverable, 
		         "Tried to assign to a ShortDatum with too few attributes");
		  }
	}

    /** Copy the values in the argument to this datum. 
     * @deprecated
     * Use setShorts( short[] )/
     */
    public void assign( short [] data )
    {
       setShorts( data );
    }

    /** Assigns values to an existing datum by copying a subset of values from an array. 
 	 * 	2nd arg specifies the datum index of the datum to be copied, where all datums in
 	 * 	the array are assumed to be of the same length. Copying continues until 
 	 * 	numAttributes elements have been copied. If this object contains too few attributes to
 	 * 	hold the new data, an error is created. 
     * 
     * @deprecated 
     * Use setShorts( short[], int start, int numAttr );
     * Note that the semantics of the start argument of setShorts is different.
	*/
	public void assign(short values[], int startIndex, int numAttributes)
	{
	   setShorts( values, startIndex*numAttributes, numAttributes );
	}

    /** Return the field specified by the argument as a double.*/
    public double getDouble( int i )
    {
       try {  return values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       return Double.NaN;
    }

    /** Return the field specified by the argument as a float.*/
    public float getFloat( int i )
    {
       try {  return values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       return Float.NaN;
    }

    /** Return the field specified by the argument as a int.*/
    public int getInt( int i )
    {
       try {  return values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       return Integer.MAX_VALUE;
    }

    /** Return the field specified by the argument as a short.*/
    public short getShort(int i)
    {
       try {  return values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       return Short.MAX_VALUE;
    }

    /** Return the field specified by the argument as a byte.*/
    public byte getByte( int i )
    {
       try {  return (byte) values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting short to byte" ); }
       return Byte.MAX_VALUE;
    }

    /** Return all attributes of the Datum as a 1-d short array */
    public short[] getShorts()
    {  
       // should we hava a reference version?
       return this.values.clone();
    }

    /** Return all attributes in a Values object */
    public Values getValues()
    {  
       return new ShortValues( values );
    }

    /** Return a new datum with the attributes indicated by the RecordSpec. The values of these
      *  attributes are taken from this ShortDatum object. 
      */
    public Datum subset(RecordSpec spec)
    {
		int i;
		int specsize=spec.size();
		
		ShortDatum newdatum=new ShortDatum(specsize);
		
		for(i=0;i<specsize;i++){
			newdatum.values[i]=this.values[spec.getFieldNumber(i)];
		}
		return newdatum;
    }


    /** Copy the attributes specified by the RecordSpec into the argument datum. */
    public void subset(ShortDatum d, RecordSpec spec){
        

		int i;
		int specsize=spec.size();
		
		try{
			for(i=0;i<specsize;i++){
				d.values[i]=this.values[spec.getFieldNumber(i)];
			}
		} catch(ArrayIndexOutOfBoundsException e){
			new IllegalArgumentError(SDBError.Recoverable, "Argument datum was not large enough for RecordSpec.");
		}
    }



    /** Copy the attributes specified by the RecordSpec into the argument datum. */
    public void subset(Datum d, RecordSpec spec){
        

		int i;
		int specsize=spec.size();
		
		try{
			for(i=0;i<specsize;i++){
				d.setFloat(i, this.values[spec.getFieldNumber(i)]);
			}
		} catch(ArrayIndexOutOfBoundsException e){
			new IllegalArgumentError(SDBError.Recoverable, "RecordSpec was too large.");
		}
    }



    
    /** Return a string representation of the datum. */
    public String toString(){
    
        int i, size=this.values.length;
        
        String result=new String("{");
        
        for(i=0;i<size-1;i++)
        {
            result+=values[i]+", ";
        }
        
        result+=values[i]+"}";
        return result;        
    }
    
    /** Return the number of attributes in the datum. */
    public int size(){

        return this.values.length;
    }

    /** Return the size in bytes needed to store the data for the attributes. */
    public int byteSize()
    {
       return this.values.length*2;
    }

    /**
     * Return the StorageType for this Datum
     */
    public int getStorageType()
    {
       return StorageType.Short;
    }
}
