/*
 * XMLEntityResolver.java
 *
 * Author:  Gang Lu
 * University of New Hampshire
 * Durham, NH 03824
 * Created on July 18, 2003, 11:17 AM
 * 
 * --------------------------- Edit history --------------------------
 * 09/19/05 rdb: Improved previous change; I had looked for ./config/dtd/*dtd.
 *               This seemed to work before both under Eclipse and from the
 *               jar file. After moving to Java 1.5, the jar version didn't
 *               work any more. By removing the initial "./", it now works
 *               again in both cases. 
 * 09/06/05 rdb: -If user does not have a config/dtd directory or a specific
 *                .dtd file, use getResourceAsStream; this works if system
 *                being accessed via the jar file AND if the jar file 
 *                includes edu/unh/sdb/common/config/dtd/*dtd (all the 
 *                dtd files that we need). This also works with IDEs like
 *                Eclipse as long as the config directory is stored in the
 *                same location as the edu/unh/sdb/common/*.class files.
 *               -Also, remove explicit test on dtd file names. OK as long as
 *                filename ends in ".dtd" assume. Might help in future.
 */

package edu.unh.sdb.common;

import java.io.*;
import java.net.URL;

import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;

public class XMLEntityResolver implements EntityResolver
{
    public InputSource resolveEntity (String publicId, String systemId)
    {
       char slash = File.separatorChar;
       String dtdDefault = "config" + slash + "dtd" + slash; // config/dtd/
       dtdDefault = "config/dtd/";
       File path = new File( systemId );
       String fileName= path.getName();
        
       if ( publicId != null && systemId != null 
                             && fileName.endsWith( ".dtd") )
       {
           String dtdPath = new Configuration().getDTDPath();
           if ( dtdPath != null ) // user has dtd override
           {
              String searchFile= ( new File(dtdPath)).getPath() 
                                             + slash + fileName;      
              File searchPath = new File( searchFile );
              if ( searchPath.exists() )
                 return new InputSource( searchPath.getPath() );
           }
           // else use default dtd files
           InputStream temp = XMLEntityResolver.class.getResourceAsStream( 
                                     dtdDefault + fileName );
           if ( temp == null ) 
           {
              System.err.println( "XMLEntityResolver can't find config/dtd "
                                       + " Trying ./config/dtd ");
              temp = XMLEntityResolver.class.getResourceAsStream( 
                                       "./" + dtdDefault + fileName );
           }
           if ( temp == null )
           {
              System.err.println( "XMLEntityResolver unable to resolve " +
                                         dtdDefault + fileName );
              System.err.println( "Need config/dtd/*dtd files to be located in\n"
                   + "        $(HOME)\n"
                   + "    or  <build>/edu/unh/sdb/common/\n"
                   + "        where <build> is location of the granite .class files\n"
                   + "\n"
                   + "You can copy this directory structure from \n"
                   + "     <src>/edu/unh/sdb/common/config/\n" 
                   + "        where <src> is location of the granite .java files\n"
                   );
              System.err.println( "Exiting ----" );
              System.exit( -1 );             
           }
           return new InputSource( temp );
       }
           
     // use the default behaviour
         return null;
     }
    
    public static void main( String[] args )
    {
       URL dtdURL = Configuration.class.getResource( "config/dtd/fdl.dtd" );
       String dtdDir = dtdURL.getFile();
       System.out.println( "getFile: " + dtdDir );
       File dtdFile = new File( dtdDir );
       System.out.println( "Filename: " + dtdFile.getPath() + " " + dtdFile.exists());
       System.out.println( "====================================");
       dtdDir = dtdURL.getRef();
       System.out.println( "getRef: " + dtdDir );
       dtdFile = new File( dtdDir + "fdl.dtd" );
       System.out.println( "Filename: " + dtdFile.getPath() + " " + dtdFile.exists());
    }     
}
