/* 
    XMLReader.java
    Author:  Denise D. Mitchell 2002
    University of New Hampshire
    Durham, NH 03824
    
    ------------------ Edit history since 6/5/08 -----------------------------
    02/10/12 rdb: Added vectorToArray( ArrayList )
                  Vector is now considered obsolete, so am changing Vectors
                  to ArrayLists in various places. Easiest to implement this
                  version of vectorToIntArray, while leaving the name the same.
    06/05/08 rdb: Made ArrayList generic
*/

package edu.unh.sdb.common;
import org.apache.xerces.parsers.*; //xerces parser


import java.util.*;

import org.w3c.dom.*;

//added by Gang Lu to turn on the validation
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.xml.sax.InputSource;

public abstract class XMLReader{
  /** the file/url to be parsed and processed */
  protected String inputFileName;

  /** the parsing object */
   protected DOMParser parser;

  /** the key element to be processed*/
  protected String keyElement;

  /** flag to set verbose on/off */
  protected boolean verbose = false;

 //added by Gang Lu
  protected Document doc= null;
  
/** Takes an argument name of a file or URL as a String and verbose indicator*/
  public  XMLReader( String inf, boolean v) 
  {
    start (inf, v);
  }

  /** Takes an argument name of a file or URL as a String */
  public XMLReader( String inf) 
  {
        start(inf, false);
  }

  /** Sets the initial values and calls the parser*/
  protected void start( String inf, boolean v) 
  {
    setInputFileName(inf);
    verbose = v;
    setDefaults(); //instantiate the FDLDescriptor and set any default values.
    try {
      parse(); 
    } catch (Exception e)
      {
        e.printStackTrace();
      }
  }

    /** sets the input file name to the string argument passed */
  protected void setInputFileName(String inf)
  {
        if (inf != null)
            inputFileName = inf;
        else 
            inputFileName = "xml.xml";
  }

  /** Sets all pending values to null and sets the join flag to FALSE */
  protected void setDefaults()
  {
  }

    /** process the main document in the xml file */
    abstract protected boolean processDoc();

    /** converts a string to and Integer object */
    protected static Integer sToI(String s)
    {        
       if (s.length() == 0)
            return new Integer(0);
       return new Integer(s.trim());
    }

    /** converts a string to an integer */
    protected static int sToi(String s)
    {
       if (s.length() == 0)
            return 0;
       return (new Integer(s.trim()).intValue());
    }
    
    /** converts a vector to an array of integers **/
    protected static int[] vectorToIntArray(Vector v)
    {
      int[] array;
        
      if (v.size() == 0) //no bounds were specified
      {
        array = new int[1];
        array[0] = 0;
      }
      else
      {
        array = new int[v.size()];
        for (int i = 0; i < v.size(); i++)
            array[i] = ((Integer)v.elementAt(i)).intValue();
      }
      return array;
    }
    /** converts a vector to an array of integers **/
    protected static int[] vectorToIntArray( ArrayList v )
    { 
      int[] array;
        
      if (v.size() == 0) //no bounds were specified
      {
        array = new int[1];
        array[0] = 0;
      }
      else
      {
        array = new int[v.size()];
        for (int i = 0; i < v.size(); i++)
            array[i] = ((Integer)v.get(i)).intValue();
      }
      return array;
    }

    /** instantiates the main parser **/
    protected void parse() 
    {
        /**
      parser = new DOMParser();     // @XML4JV2
      try
      {
        parser.parse(inputFileName);  // @XML4JV2
        processDoc();
      } catch (Exception e)
      {
        e.printStackTrace();
      }
      */
        
        // added by Gang Lu for turn on the dtd validation
        DocumentBuilderFactory dbf= DocumentBuilderFactory.newInstance();
        dbf.setValidating(true);
        dbf.setNamespaceAware(true);
        dbf.setIgnoringElementContentWhitespace(true);

      try
      {
        DocumentBuilder builder = dbf.newDocumentBuilder();
        builder.setErrorHandler(new XMLErrorHandler());
        builder.setEntityResolver(new XMLEntityResolver());
        
        InputSource is = new InputSource(inputFileName);
        
        doc = builder.parse(is);
        
        processDoc();

      } catch (Exception e)
      {
        // a new XMLError?
        e.printStackTrace();
        new SDBError(SDBError.Recoverable, e.toString());
      }
  }

  /** processes a child node from the XML parsed tree int an Integer object */
  protected static Integer makeChildrenInteger (Node node)
  {
     String s = makeChildrenText(node); //get the text as a string
     if (s != null)
          return sToI(s); //convert it to an integer.
     return new Integer("0");  //default will be 0 if there is no text.
  }

    /** converts a string to a Float object */
    protected static Float sToF(String s)
    {
       if (s.length() == 0)
            return (new Float("0"));
       return (new Float(s));
    }

    /** converts a vector to an array of floats and defaults to a 1 item array of 0 if 
    vector is empty **/
    protected static float[] vectorToFloatArray(Vector v)
    {
      float[] array;

      if (v.size() == 0) //no bounds were specified
      {
        array = new float[1];
        array[0] = 0;
      }
      else
      {
        array = new float[v.size()];
        for (int i = 0; i < v.size(); i++)
            array[i] = ((Float)v.elementAt(i)).floatValue();
      }
      return array;
    }

  /** converts a node from the XML parsed tree into a Float object **/
  protected static Float makeChildrenFloat (Node node)
  {
     String s = makeChildrenText(node); //get the text as a string
     if (s != null)
          return sToF(s); //convert it to an float.
     return new Float("0");  //default will be 0 if there is no text.
  }

  /** converts a node from the XML parsed tree into a String **/
  protected static String makeChildrenText (Node node)
  {
     // Create a StringBuffer to store the result.
     // StringBuffer is more efficient than String
    if (node == null)
      return new String();
      StringBuffer buffer = new StringBuffer();
    return makeChildrenText1 (node, buffer);
  }

  /** recursive method to support makeChildrenText(Node node) method */
  protected static String makeChildrenText1 (Node node, StringBuffer buffer)
  {
     // Visit all the child nodes
     for (Node ch = node.getFirstChild(); ch != null; ch = ch.getNextSibling())
       {
          // Recursively call if the child may have children
         if (ch instanceof Element || ch instanceof EntityReference)
          {
             buffer.append(makeChildrenText(ch));
            // If the child is a text, append it to the result buffer
           } else if (ch instanceof Text)
             {
               buffer.append(ch.getNodeValue());
             }
       }
     return buffer.toString();
  }
  
   public int[] processArrayOfInt(String axisString)
    {   
        int[] axis = null;
        
        if ( axisString != null )
        {
            int len;
            Vector<String> axisList = new Vector<String>();
            StringTokenizer st= new StringTokenizer(axisString);
        
            while ( st.hasMoreTokens() )
            {
                axisList.add( st.nextToken() );
            }
        
            len = axisList.size();
            if ( len > 0 )
            {
            axis = new int[len];
        
            for(int i=0; i < len; i++)
            {
                try{
                    axis[i] = sToi(axisList.elementAt(i).trim());
                
                }
                catch(NumberFormatException e)
                {
                    String message = "Error: The value for Axis " + i + " is not an integer: " + axisList.elementAt(i);
                    new XMLReaderError(SDBError.Recoverable, message);
                    axis=null;
                    break;
                }
            }
            }
        }
            return axis;
   }
}