/* 
    AJDataSourceBuilder.java
    Author:  Denise D. Mitchell 2002
    University of New Hampshire
    Durham, NH 03824
*/
package edu.unh.sdb.datasource;
//import java.io.*;
import edu.unh.sdb.common.*;

//--------------------------AJDataSourceBuilder class ----------------------
/** Manufactures Logical AttributeJoinDataSource objects.  Holds relevant data and 
 * validates before instantiating the appropriate DataSource objects
 */
public class AJDataSourceBuilder extends DataSourceBuilder
{
    
    

  /** Default Constructor sets the bounds and recordDesc members to default values.
  */
  public AJDataSourceBuilder(DataSourceCollection d, String n)
  {
    super(d,n);
//        axisOrdering = null; 
  }

  /** returns a FieldIDMapper for the current component **/
  protected FieldIDMapper makeComponentFieldIDMapper()
  {
    if (componentFieldPositions.isEmpty() ||
        compositeFieldPositions.isEmpty())  //user didn't specify fields
    { 
      //make up a FieldIDMapper with all fields from the component datasource
      if (getCurrentDataSource() != null)
      { //we have a datasource, get record descriptor
        //get record descriptor
        RecordDescriptor rd = getCurrentDataSource().getRecordDescriptor();
        int[] array; 
    
        if (rd != null)
        {// we have a record descriptor, make array for mapping
          if (rd.size() < 1)
          {
            array = new int[1];
            array[0] = 0;
          } 
          else 
          {
            array = new int[rd.size()];
            for (int i = 0; i < rd.size(); i++)
              array[i] = i;
          }
          return new FieldIDMapper(array, array);
        }
      }
        System.err.println("Could not make a FieldIDMapper.");
        return null;  //this shouldn't happen
    }
    else  //we have field positions specified;
    { 
      return new FieldIDMapper(
                vectorToIntArray(componentFieldPositions),
                vectorToIntArray(compositeFieldPositions));
    }
  }
  /** instantiate an actual datasource object**/
  public DataSource makeDataSource()
  {
    if ( discardFlag == true )
      return null;
    DataSource ajds = new AttributeJoinDataSource( getDSName(), makeComponentInfos());
       
    return ajds; 
  }

 }
