/*
    ARRSCellBlock.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: Li Ye
*/
//-------------------- Edit history as of 10/24/05 ---------------------------
// 06/05/08 rdb: Made ArrayLists generic
// 11/25/05 rdb: Replace references to CopyLeaf methods with Leaf methods.
// 11/14/05 rdb: deleted activate, deactivate methods; this a block not a ds
// 11/13/05 rdb: **** This is still specific for Copy version, since only a single
//               pds is provided. This could be an MRDS in future, but need to think
//               about those implications.
//               **** Also there should be a parent ARRSCellDataCollection ****
// 11/13/05 rdb:- Deleted all references to CompositeBlock
//              - deprecated constructor from DS files input and re-did all uses.
//                The uses were all in test programs; I replaced with a creation
//                 of the DS and then take full subblock.
// 10/24/05 rdb: added  DataBlock urCellBlock( ISBounds, ISBounds ) convenience method
//               cleaned up unused variables and deprecated code
//-----------------------------------------------------------------------------

/**
 * The ARRSCellBlock class implements an interface to access blocks of cell data
 * in an ARRS cell space. It contains an ARRS tree corresponding to the ARRS
 * cell space.
 */
                                                                                

package edu.unh.sdb.datasource;                                                                                
import edu.unh.sdb.common.*;
import edu.unh.sdb.ARRSTreeBuilder.*;
//import java.io.*;
import java.util.*;

public class ARRSCellBlock extends RectCellBlock implements CellIndexable 
{
    private DataSource pointDS; //store 1-D point datasource
    private ARRSTree tree; //store an ARRS tree
    private ARRSTree subtree; //store an ARRS tree with a subblock data

    /** Construct a ARRSCellBlock without arguments */
    public ARRSCellBlock()
    {
       super();
       this.pointDS = null;
       this.tree = null;
       this.subtree = null;
    }
                                                                                
    /** Construct a ARRSCellBlock with an ARRSTree*/
    public ARRSCellBlock( ARRSTree tree, DataSource pointds )
    {
       super();
       this.pointDS = pointds;
       this.tree = tree;
       this.subtree = null;
       this.init();
    }

    /** Construct a ARRSCellBlock with the given base bounds and number of 
      * attributes
      */
    public ARRSCellBlock( ISBounds basebounds, RecordDescriptor recordDesc )
    {
       super( basebounds, recordDesc );
       this.pointDS = null;
       this.tree = null;
       this.subtree = null;
    }

    /** Construct a ARRSCellBlock with command file,indexFDL,and ARRSCellBlock
     * @deprecated A Block should not be created from DS specs; create DS and subblock.
     *          This is also "copy" version specific.
     */
    public ARRSCellBlock( String command, String indexFDL, String pdsFDL )
    {
       DataSource pds = DataSource.create( " ", pdsFDL);
       if (pds == null) {
          new IllegalArgumentError(SDBError.Recoverable,"The pdsFDL file is invalid, fail to create a point datasource.");
       }
       //pds.activate();
       this.pointDS = pds;
       ARRSCopyModelBuilder builder = new ARRSCopyModelBuilder(command,indexFDL,this);
       ARRSTree t = builder.getARRSTree();
       this.tree = t;
       if (t == null) {
          new IllegalArgumentError(SDBError.Recoverable,"The command file is invalid, fail to create a ARRS tree.");
       }
       else
          this.init();
    }

    /** Set the base bounds */
    public void setBaseBounds(ISBounds basebounds)
    {
       this.bounds = (ISBounds)basebounds.clone();
    }

    /** Set the point datasource */
    public void setPointDS(DataSource pds)
    {
       this.pointDS = pds;
    }

    /** Set an ARRS tree */
    public void setARRSTree(ARRSTree tree)
    {
       if (this.tree != null)
          this.tree.treeClear();

       this.tree = tree;
       if (this.pointDS != null)
          this.init();
    }

    /* set base bounds */
    private void init()
    {
       ARRSNode node = this.tree.getNode(0);
       this.bounds = node.getBaseBounds(); 
       this.dimensionality = this.bounds.dim();
       this.recordDesc = this.pointDS.getRecordDescriptor();
       this.numAttributes = this.pointDS.getNumAttributes();
    }
 
    /** Get an ARRS tree */
    public ARRSTree getARRSTree()
    {
       return this.tree;
    }

    /** Get the point datasource */
    public DataSource getPointDS()
    {
       return this.pointDS;
    }

    /** Get the base bounds */
    public ISBounds getBaseBounds()
    {
       return this.bounds;
    }

    /** Given an IndexSpaceID, return the corresponding datum.*/
    public  Datum datum( IndexSpaceID did)
    {
       System.err.println( "ARRSCellBlock.datum: Unimplemented function." );
       return null;
    }

    /** Given an array of coordinates, copy the corresponding datum
        into the first parameter.*/
    public void datum( Datum d, int[] coordArray ) 
    { 
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                                     + this.getClass().getName() + ".datum(...)");
       return;
    }


    /** Given an IndexSpaceID, return the corresponding datum with attributes
        specified in the RecordSpec. */
    public  Datum datum( IndexSpaceID did, RecordSpec spec )
    { 
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                                     + this.getClass().getName() + ".datum(...)");
       return null;
    }

    /** Set the location specified by the IndexSpaceID to the given datum.*/
    public void setDatum( IndexSpaceID isid, Datum d )
    {
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                                     + this.getClass().getName() + ".setDatum(...)");
       return;
    }

    /** Return a new DataBlock that is a subset of this datablock. The Data
        is shared between the two DataBlocks.*/
    public  DataBlock subblock( ISBounds db ) 
    {
      CellIndexable cellblock;
      ARRSNode node;
      ISBounds basebounds;
      this.subtree = new ARRSTree();

      if ( this.tree.getNodeList().isEmpty() )
         return null;

      node = this.tree.getNode( 0 );
      cellblock = node.getCellIndexable();
      basebounds = node.getBaseBounds();
      //System.out.println("basebounds: " + basebounds.toString()); //-----
      if ( !basebounds.contains( db ) )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "The passed ISBoundshas a illegal range" );
         return null;
      }

      if ( node instanceof ARRSCopyLeafNode )
      {
         this.buildLeafNodes( node, db, -1 );
         return new ARRSCellBlock( this.subtree, this.pointDS );
      }

      if ( node instanceof NTreeNode )
      {
         // System.out.println("NTreeNode in subblock: "); //-----
         // System.out.println("basebounds of NTreeNode: " +
         // node.getBaseBounds().toString()); //-----
         //NTreeNode qnode = new NTreeNode(db, cellblock);
         NTreeNode qnode = new NTreeNode( basebounds, cellblock );
         this.subtree.addNode( qnode, -1 );
      }

      if ( node instanceof KDNode )
      {
         // System.out.println("KdNode in subblock: "); //-----
         // System.out.println("basebounds of KdNode: " +
         // node.getBaseBounds().toString()); //-----
         //KDNode kdnode = new KDNode(db, cellblock);
         KDNode kdnode = new KDNode( basebounds, cellblock );
         this.subtree.addNode( kdnode, -1 );
      }

      this.buildInternalNode( node, db, 0 );
      this.subtree.buildTreePath();
      return new ARRSCellBlock( this.subtree, this.pointDS );
    }
 
    /* build leaf nodes */ 
    private void buildLeafNodes(ARRSNode node, ISBounds db, int nodeid)
    { 
       //System.out.println("beginning of buildLeafNode"); //-----
       CellIndexable cds = ((ARRSLeafNode)node).getCellIndexable();
       ISBounds baseb = ((ARRSLeafNode)node).getBaseBounds();
       ARRSClippedCopyLeafNode cnode = new ARRSClippedCopyLeafNode( baseb, cds );
       cnode.setSubBounds( db );
       cnode.setIndexBlock(((ARRSCopyLeafNode)node).getIndexBlock());
       cnode.setIndexBlockBounds( ((ARRSCopyLeafNode)node).getIndexBlockBounds() );
       
       //rdb ??? formerly addNode but leaf nodes are not added
       this.subtree.addNode( cnode, nodeid );
    }
   
    /* build an internal node */
    private void buildInternalNode(ARRSNode node, ISBounds db, int nodeid)
    {
      ARRSNode[] nodes;
      ARRSNode tempnode;
      ISBounds baseb;
      int numnodes;
      // System.out.println("beginning of buildInternalNode"); //-----
      nodes = this.tree.getChildren(node);
      if ( nodes == null || nodes.length == 0 )
         return;
      
      for ( int i = 0; i < nodes.length; i++ ) 
      {
         baseb = nodes[i].getBaseBounds();
         //System.out.println( "baseb: " +baseb.toString());
         //System.out.println( "db: " +db.toString());
         if ( baseb.contains( db )) 
         {
            //System.out.println( "contains of buildInternalNode"); //-----
            //if ( nodes[i] instanceof ARRSCopyLeafNode ) 
            if ( nodes[i] instanceof ARRSLeafNode ) 
            {
               //System.out.println("Leaf nodes"); //-----
               this.buildLeafNodes( nodes[i], db, nodeid );   
               //rdb break;
            }
            else if ( nodes[i] instanceof NTreeNode ) 
            {
               //System.out.println("quad nodes"); //-----
               tempnode = nodes[i];
               //NTreeNode qnode = new NTreeNode(db,((NTreeNode)tempnode).getCellIndexable());
               NTreeNode qnode = new NTreeNode( baseb,((NTreeNode)tempnode).getCellIndexable());
               this.subtree.addNode( qnode, nodeid );
               numnodes = this.subtree.getNumOfNodes();
               //System.out.println("numnodes: " + numnodes); //-----
               this.buildInternalNode( nodes[i], db, numnodes );
               //rdb break;
            } 
            else if ( nodes[i] instanceof KDNode ) 
            {
               //System.out.println("Kd nodes"); //-----
               tempnode = nodes[i];
               //KDNode kdnode = new KDNode(db,((KDNode)tempnode).getCellIndexable());
               KDNode kdnode = new KDNode( baseb, ((KDNode)tempnode).getCellIndexable());
               this.subtree.addNode( kdnode, nodeid );
               numnodes = this.subtree.getNumOfNodes();
                //System.out.println("numnodes: " + numnodes); //-----
               this.buildInternalNode( nodes[i], db, numnodes );
               //rdb break;
            }
            else
               new SDBError( SDBError.Warning, "ARRSCellBlock.buildInternalNodes: "
                     + " unknown node type: " + nodes[i].getClass().getName() );
         }
         else if ( baseb.intersection( db ) != null )
         {   
            //System.out.println("intersects of buildInternalNode"); //-----
            ISBounds interbounds = baseb.intersection( db );
            //System.out.println("intersection: "+interbounds.toString()); //-----
            //if ( nodes[i] instanceof ARRSCopyLeafNode ) 
            if ( nodes[ i ] instanceof ARRSLeafNode ) 
            {
               //System.out.println("Leaf nodes"); //-----
               this.buildLeafNodes( nodes[ i ], interbounds, nodeid );
            }
            else if ( nodes[ i ] instanceof NTreeNode ) 
            {
               //System.out.println("Quad nodes"); //-----
               tempnode = nodes[ i ];
               //NTreeNode qnode = new NTreeNode(interbounds, 
               NTreeNode qnode = new NTreeNode(baseb, 
                                     ((NTreeNode)tempnode).getCellIndexable());
               this.subtree.addNode( qnode, nodeid );
               numnodes = this.subtree.getNumOfNodes();
               //System.out.println("numnodes: " + numnodes); //-----
               this.buildInternalNode( nodes[ i ], interbounds, numnodes );
            }
            else if ( nodes[ i ] instanceof KDNode ) 
            {
               //System.out.println("Kd nodes"); //-----
               tempnode = nodes[ i ];
               //KDNode kdnode = new KDNode(interbounds, 
               KDNode kdnode = new KDNode( baseb, 
                                    ((KDNode)tempnode).getCellIndexable());
               this.subtree.addNode( kdnode, nodeid );
               numnodes = this.subtree.getNumOfNodes();
               //System.out.println("numnodes: " + numnodes); //-----
               this.buildInternalNode( nodes[ i ], interbounds, numnodes );
            }
            else
               new SDBError( SDBError.Warning, "ARRSCellBlock.buildInternalNodes: "
                     + " unknown node type: " + nodes[ i ].getClass().getName() );

         }
      }

    }

    /** Reads a subblock defined by the specified bounds into the DataBlock
        argument */
    public void subblock( DataBlock theBlock, ISBounds bd )
    {
        //System.out.println("Beginning of subblock(datablock, bounds)");
       if ( theBlock instanceof URCellBlock ) 
       {
         //System.out.println("Yes, URCellBlock");
         //System.out.println("this.bounds in subblock: "+this.bounds.toString());
         this.urCellBlock( theBlock, bd, this.bounds.copyReducedBy(1) );
         /*URCellBlock urBlock = (URCellBlock)this.urCellBlock(bd, this.bounds.reduceByOne());
         ((URCellBlock)theBlock).setPointDataBlock(urBlock.getPointDataBlock());
         */
       }
       else if ( theBlock instanceof ARRSCellBlock ) 
       {
        //System.out.println("Yes, ARRSCellBlock");
         ARRSCellBlock arBlock = (ARRSCellBlock) this.subblock( bd ); 
        //DataSource p = arBlock.getPointDS();
         // System.out.println(p.toString());
         ((ARRSCellBlock)theBlock).setPointDS( arBlock.getPointDS() );
         ((ARRSCellBlock)theBlock).setARRSTree( arBlock.getARRSTree() );
       }
       else
          new SDBError( SDBError.Warning, "Unimplemented ARRS function." );
    }

    /** Reads a subblock defined by srcBounds into the region of the Datablock
        argument defined by dstBounds. */
    public void subblock( DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds)
    {
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                      + this.getClass().getName() + ".subblock( blk, bnds, bnds)");
        return;
    }

    /** Reads a subblock defined by srcBounds into the region of the Datablock 
        argument defined by dstBounds. The FieldIDMapper argument determines 
        the mapping between datums in this DataBlock and datums in the argument.
        DataBlock. A null FieldIDMapper argument implies that all fields are used.
      */
    public void subblock( DataBlock theBlock, ISBounds dstBounds,
                        ISBounds srcBounds, FieldIDMapper mp )
   {
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                  + this.getClass().getName() + ".subblock( blk, bnds, bnds, fm )");
   }
   
    /**
     * Reads a subblock defined by srcBounds into the region of the Datablock
     * argument defined by dstBounds. The RecordSpec argument specifies a subset
     * of the attributes of this DataSource to be written into the DataBlock
     * datums. A null RecordSpec argument implies that all fields are used.
     */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, 
                                    ISBounds srcBounds, RecordSpec spec)
    {           
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                    + this.getClass().getName() + ".subblock( blk, bnds, bnds, rs )");
    }

    /** Return a reference to the actual array containing the data. */
    public Datum[] data()
    { 
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                   + this.getClass().getName() + ".data()");
       return null; 
    }

    /** Return by reference the Datum corresponding to the given
     *  IndexSpaceID.
     */
    public void datum( Datum dtm, IndexSpaceID did ) 
    { 
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                                     + this.getClass().getName() + ".datum(...)");
    }

    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
    { 
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                                     + this.getClass().getName() + ".datum(...)");
    }
   

    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
    public void datum( Datum dtm, IndexSpaceID did, RecordSpec recSpec )
    { 
       new SDBError( SDBError.Warning, "Unimplemented ARRS function: " 
                                     + this.getClass().getName() + ".datum(...)");
    }

    /** Return a cell block with the specified bounds and the specified
     *  resolution 
     */
    public DataBlock urCellBlock( ISBounds b, ISBounds res )
    {
       // create a Cell Block of the right size and call the in place version
       DataBlock newCB = new URCellBlock( res, this.recordDesc );
       this.urCellBlock( newCB, b, res );
       return newCB;
    }

    /** Fill in a cell block with the specified bounds and the specified
     *  resolution. 
     *  This version should: 1) check that the passed in block is the right
     *  kind and the right size; 2) if it's not correct, it should allocate a
     *  correct one; 3) in all cases it should return the block actually filled;
     *  and 4) it should accept null.
     */
    public void urCellBlock( DataBlock urBlock, ISBounds b, ISBounds res )
    {
       //rdb CompositeBlock pblock;
       DataBlock pblock;
       ISBounds bd;
       //System.out.println("start urCellBlock in ARRSCellBlock");   
       /*get the info of component data sources with the specified resolution*/
       ComponentInfo[] compInfo = this.createPointCompInfo( b, res );
       
       if ( compInfo == null || compInfo.length == 0 )
       {
          new SDBError( SDBError.Warning, "No leaf nodes?? " );
          return;
       }
      
       /*create a OverlapBlockedDataSource with compoInfo list*/
       OverlapBlockedDataSource bds = new OverlapBlockedDataSource("", compInfo);
       /*get the point bounds corresponding to the specified cell bounds*/
       bd = b.copyIncreasedBy( 1 );
       bd.moveToOrigin();
       /*create a point data block*/
       //rdb pblock = new CompositeBlock(bd, bds.getRecordDescriptor());
       pblock = bds.getRecordDescriptor().createDataBlock( bd );
       bds.subblock( pblock, bd, bd );
       //System.out.println(pblock.toString());
       /*create a uniform cell block containing the created point datablock*/
       //urBlock = new URCellBlock(b, pblock.getRecordDescriptor());
       ((URCellBlock)urBlock).setPointDataBlock( pblock );
    } 

    /* Create the list of ComponentInfo of point component data source */
    private ComponentInfo[] createPointCompInfo( ISBounds b, ISBounds res )
    {
       int length;
       ISIDMapper isMap, isMap2;
       ISBounds baseb, baseR, compositeb, pointb;
       ArrayList<ARRSLeafNode> array; 
       ArrayList<ComponentInfo>componentArray;
       ComponentInfo[] compInfo;
       URCellBlock urBlock;
       //rdb CompositeBlock block, block2, componentblock;
       DataBlock block, block2, componentblock;
       DSApproximator approx = new DSApproximator();

     // System.out.println("Beginning of createPointCompInfo"); //-----------

       /*create a ISIDMapper for the base resolution bounds and 
         the specified resolution bounds*/
       pointb = b.copyIncreasedBy( 1 );
       baseR = res.copyIncreasedBy( 1 );
       isMap = new ISIDMapper( this.bounds, baseR );

       /*get the leaf cell blocks in a ARRS tree*/
       array = this.tree.getLeafList();
       length = array.size();

       ISBounds isb1, isb2;
	   ARRSNode node;
       componentArray = new ArrayList<ComponentInfo>();
       IndexSpaceID offset = pointb.copyLower();
       offset.negate();
       for ( int i = 0; i < length; i++ ) 
       {
          node = array.get(i);
          /*get the base resolution bounds for the component*/
          baseb = node.getBaseBounds();

          /*get the resolution bounds for the component*/
          if (this.bounds.equals(baseR))
              compositeb = baseb;
          else 
              compositeb = isMap.toComposite(baseb); 

          /*get the intersection between compositeb and the specified bound*/ 
          isb1 = compositeb.intersection(pointb);
          if ( isb1 != null ) 
          {
             isb2 = isb1.copy();
             isb2.move(offset); //composite bounds in the specified bounds

             IndexSpaceID offset2 = compositeb.copyLower();
             offset2.negate();
             isb1.move(offset2);

             /*get a component cell data source*/
             urBlock = (URCellBlock)((ARRSLeafNode)node).getRegionData();

             //get the point data source included in the component cell datasource
             //rdb block = (CompositeBlock)urBlock.getPointDataBlock(); 
             block = urBlock.getPointDataBlock(); 

             /*get the resolution bounds for the component*/
             if ( block.getBounds().volume() == compositeb.volume() )
                componentblock = block;
             else
                componentblock = (DataBlock) approx.approximate( block, compositeb );
                //rdb componentblock = (CompositeBlock)approx.approximate( block, compositeb);

             if ( componentblock.getBounds().volume() == isb2.volume() )
				block2 = componentblock;
		     else 
             {	
                /*get component data contained in the specified bounds*/
                //rdb block2 = new CompositeBlock( isb2, block.getRecordDescriptor() ); 
                block2 = block.getRecordDescriptor().createDataBlock( isb2 );
                componentblock.subblock( block2, block2.getBounds(), isb1 ); 
             }

             /*create ISIDMapper object with the base bounds of the component 
               and its specified resolution bounds*/
             isMap2 = new ISIDMapper( block2.getBounds(), isb2 );
             /*create the info for the component data source*/
             ComponentInfo info = new ComponentInfo( block2, null, isMap2 );
             componentArray.add(info); 
          }
       }
       
       compInfo = new ComponentInfo[ componentArray.size() ];
       for ( int i = 0;i < compInfo.length; i++)
       {
          compInfo[i] = componentArray.get( i );
       }

       return compInfo;
    } 
}

