/*
    ARRSCellCopyDS.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
    
*/
//------------------------- Edit history (since 10/21/05) -----------------------
// 11/13/05 rdb: -For all subblock calls, now return an ARRSCellBlock rather than
//                an ARRSCellCopyDS (which used to be a child of RectBlock)
//               -Remove all  references to CompositeBlock
// 10/21/05 rdb: Cleaned up unused variables and some reformatting.
//-------------------------------------------------------------------------------
/**
 * The ARRSCellCopyDS class implements an interface to access an adaptive
 * resolution cell data source. It contains an ARRS tree corresponding to 
 * the ARRS cell space.
 */
                                                                                

package edu.unh.sdb.datasource;                                                                                
import edu.unh.sdb.common.*;
import edu.unh.sdb.ARRSTreeBuilder.*;
//import java.io.*;
import java.util.*;

public class ARRSCellCopyDS extends ARRSCellDataSource implements CellIndexable
{
   /** Construct a ARRSCellCopyDS without arguments */
    public ARRSCellCopyDS()
    {
       super();
    }
                                                                                
    /** Construct a ARRSCellCopyDS with an ARRSTree*/
    public ARRSCellCopyDS( ARRSTree tree, DataSource pointds )
    {
       super( tree, pointds );
    }

    /** Construct a ARRSCellCopyDS with the given base bounds and number of 
      * attributes
      * 
      * ****** rdb delete ***********************
    public ARRSCellCopyDS( ISBounds basebounds, RecordDescriptor recordDesc )
    {
       super( basebounds, recordDesc );
    }
      ***************************************/

    /** Construct a ARRSCellCopyDS with command file,indexFDL,and ARRSCellCopyDS*/
    public ARRSCellCopyDS( String command, String indexFDL, String pdsFDL )
    {
       DataSource pds = DataSource.create( " ", pdsFDL );
       if ( pds == null ) 
          new IllegalArgumentError( SDBError.Recoverable, 
                "The pdsFDL file is invalid, fail to create a point datasource.");

       //pds.activate();
       this.pointDS = pds;
       ARRSCopyModelBuilder builder = 
                               new ARRSCopyModelBuilder( command, indexFDL, this );
       ARRSTree t = builder.getARRSTree();
       this.tree = t;
       if ( t == null )
          new IllegalArgumentError( SDBError.Recoverable, 
                        "The command file is invalid, fail to create a ARRS tree.");
       else
          this.init();
    }

    public void activate() 
    { 
       super.activate(); 
    }

    /** Return a new DataBlock that is a subset of this datablock. The Data
        is shared between the two DataBlocks.
        */
    //rdb public  CompositeBlock subblock(ISBounds db) 
    public  DataBlock subblock( ISBounds db ) 
    {  
       CellIndexable cellblock;
       ARRSNode node;
       ISBounds basebounds;
       this.subtree = new ARRSTree();

       if ( this.tree.getNodeList().isEmpty() ) 
          return null;
                                                                                
       node = this.tree.getNode( 0 );
       cellblock = node.getCellIndexable();
       basebounds = node.getBaseBounds();

       if ( !basebounds.contains( db )) 
       { 
          new IllegalArgumentError( SDBError.Recoverable, 
                                        "The passed ISBounds has illegal range");
          return null;
       }

       if ( node instanceof ARRSLeafNode ) 
       {
          this.buildLeafNodes( node, db, -1 );
          //rdb return new ARRSCellCopyDS( this.subtree, this.pointDS );
          return new ARRSCellBlock( this.subtree, this.pointDS );
       } 
                                                                                
       else if ( node instanceof NTreeNode )
       {
          // System.out.println("NTreeNode in subblock: "); 
          // System.out.println("basebounds of NTreeNode: " + node.getBaseBounds().toString()); 
          //NTreeNode qnode = new NTreeNode(db, cellblock);
          NTreeNode qnode = new NTreeNode( basebounds, cellblock );
          this.subtree.addNode( qnode, -1 );
       }
 
       else if ( node instanceof KDNode ) 
       {
          // System.out.println("KdNode in subblock: "); 
          // System.out.println("basebounds of KdNode: " + node.getBaseBounds().toString());
          //KDNode kdnode = new KDNode(db, cellblock);
          KDNode kdnode = new KDNode( basebounds, cellblock );
          this.subtree.addNode( kdnode, -1 );
       }

       this.buildInternalNode( node, db, 0 );
       this.subtree.buildTreePath();
       //rdb return new ARRSCellCopyDS( this.subtree, this.pointDS );
       return new ARRSCellBlock( this.subtree, this.pointDS );
    }
 
    /* build leaf nodes */ 
    private void buildLeafNodes( ARRSNode node, ISBounds db, int nodeid )
    { 
       //System.out.println("beginning of buildLeafNode"); //-----
       ARRSCopyLeafNode  cnode = null;
       ISBounds baseb = ((ARRSCopyLeafNode)node).getBaseBounds();
       CellIndexable cds = ((ARRSCopyLeafNode)node).getCellIndexable();

       if ( baseb.volume() == cds.getBounds().volume() )
          cnode = (ARRSCopyLeafNode) node.copy();
       else
       {
          cnode = new ARRSClippedCopyLeafNode( baseb, cds );
          ((ARRSClippedCopyLeafNode)cnode).setSubBounds( db );
          cnode.setIndexBlock( ((ARRSCopyLeafNode)node).getIndexBlock() );
          cnode.setIndexBlockBounds( ((ARRSCopyLeafNode)node).getIndexBlockBounds() );
       }
       
       //rdb ??? formerly addNode, but leaf nodes were not added ???
       this.subtree.addNode( cnode, nodeid );
    }
   
    /* build an internal node */
    private void buildInternalNode( ARRSNode node, ISBounds db, int nodeid )
    {
      ARRSNode[] nodes;
      ARRSNode tempnode;
      ISBounds baseb;
      int numnodes;
      // System.out.println("beginning of buildInternalNode"); //-----
      nodes = this.tree.getChildren( node );
      if ( nodes == null || nodes.length == 0 )
         return;
      for ( int i = 0; i < nodes.length; i++ ) 
      {
         baseb = nodes[ i ].getBaseBounds();
         //System.out.println("baseb: " +baseb.toString());
         //System.out.println("db: " +db.toString());
         if ( baseb.contains( db ) ) 
         {
            //System.out.println("contains of buildInternalNode"); //-----
            if ( nodes[i] instanceof ARRSCopyLeafNode )
            {
               //System.out.println("Leaf nodes"); //-----
               this.buildLeafNodes( nodes[i], db, nodeid );   
               break;
            }
            if ( nodes[i] instanceof NTreeNode ) 
            {
               //System.out.println("quad nodes"); //-----
               tempnode = nodes[ i ];
               //NTreeNode qnode = new NTreeNode(db,((NTreeNode)tempnode).getCellIndexable());
               NTreeNode qnode = new NTreeNode( baseb, ((NTreeNode)tempnode).getCellIndexable() );
               this.subtree.addNode( qnode, nodeid );
               numnodes = this.subtree.getNumOfNodes();
               //System.out.println("numnodes: " + numnodes); //-----
               this.buildInternalNode( nodes[ i ], db, numnodes );
               break;
            }

            if ( nodes[ i ] instanceof KDNode ) 
            {
               //System.out.println("Kd nodes"); //-----
               tempnode = nodes[ i ];
               //KDNode kdnode = new KDNode(db,((KDNode)tempnode).getCellIndexable());
               KDNode kdnode = new KDNode( baseb, ((KDNode)tempnode).getCellIndexable());
               this.subtree.addNode( kdnode, nodeid );
               numnodes = this.subtree.getNumOfNodes();
               //System.out.println("numnodes: " + numnodes); //-----
               this.buildInternalNode( nodes[i], db, numnodes );
               break;
            }
         }
        
         if ( baseb.intersection( db ) != null )
         {   
            //System.out.println("intersects of buildInternalNode"); //-----
            ISBounds interbounds = baseb.intersection( db );
            //System.out.println("intersection: "+interbounds.toString()); //-----
            if ( nodes[i] instanceof ARRSLeafNode ) 
            {
               //System.out.println("Leaf nodes"); //-----
               this.buildLeafNodes( nodes[i], interbounds, nodeid );
            }
            else if ( nodes[i] instanceof NTreeNode ) 
            {
               tempnode = nodes[ i ];
               //NTreeNode qnode = new NTreeNode(interbounds, 
               NTreeNode qnode = new NTreeNode( baseb, 
                                     ((NTreeNode)tempnode).getCellIndexable() );
               this.subtree.addNode( qnode, nodeid );
               numnodes = this.subtree.getNumOfNodes();
               this.buildInternalNode( nodes[i], interbounds, numnodes );
            }
            else if ( nodes[i] instanceof KDNode )  
            {
               tempnode = nodes[ i ];
               //KDNode kdnode = new KDNode(interbounds, 
               KDNode kdnode = new KDNode( baseb, 
                                    ((KDNode)tempnode).getCellIndexable() );
               this.subtree.addNode( kdnode, nodeid);
               numnodes = this.subtree.getNumOfNodes();
               this.buildInternalNode( nodes[ i ], interbounds, numnodes );
            }
         }
      }

    }

    /** Reads a subblock defined by the specified bounds into the DataBlock
        argument */
    public void subblock( DataBlock theBlock, ISBounds bd )
    {
       //System.out.println("Beginning of subblock(datablock, bounds)");
       if ( theBlock instanceof URCellBlock) 
       {
         //System.out.println("Yes, URCellBlock");
         //System.out.println("this.bounds in subblock: "+this.bounds.toString());
         this.urCellBlock( theBlock, bd, this.bounds.copyReducedBy( 1 ));
         //URCellBlock urBlock = (URCellBlock)this.urCellBlock(bd, this.bounds.reduceByOne());
         //((URCellBlock)theBlock).setPointDataBlock(urBlock.getPointDataBlock());
       }
       //rdb else if ( theBlock instanceof ARRSCellCopyDS ) 
       else if ( theBlock instanceof ARRSCellBlock ) 
       {
          //System.out.println("Yes, ARRSCellCopyDS");
          //rdb ARRSCellCopyDS arBlock = (ARRSCellCopyDS)this.subblock( bd ); 
          ARRSCellBlock arBlock = (ARRSCellBlock)this.subblock( bd ); 

          //DataSource p = arBlock.getPointDS();
          // System.out.println(p.toString());
          //rdb ((ARRSCellCopyDS)theBlock).setPointDS(arBlock.getPointDS());
          //rdb ((ARRSCellCopyDS)theBlock).setARRSTree(arBlock.getARRSTree());
          ((ARRSCellBlock)theBlock).setPointDS( arBlock.getPointDS() );
          ((ARRSCellBlock)theBlock).setARRSTree( arBlock.getARRSTree() );
       }
       else
       {
          new SDBError( SDBError.AlwaysThrow, 
                       "ARRSCellCopyDS.subblock: Passed DataBlock type not valid:"  +
                       theBlock.getClass().getName() );
       }
    }


    /** Return a cell block with the specified bounds and the specified
     *  resolution 
     */
    public void urCellBlock( DataBlock urBlock, ISBounds b, ISBounds res )
    {
       //rdb CompositeBlock pblock;
       DataBlock pblock;
       ISBounds bd;
       //System.out.println("start urCellBlock in ARRSCellCopyDS");   
       /*get the info of component data sources with the specified resolution*/
       ComponentInfo[] compInfo = this.createPointCompInfo( b, res ); 
       /*create a OverlapBlockedDataSource with compoInfo list*/
       OverlapBlockedDataSource bds = new OverlapBlockedDataSource( "", compInfo );
       /*get the point bounds corresponding to the specified cell bounds*/
       bd = b.copyIncreasedBy( 1 );
       bd.moveToOrigin();
       /*create a point data block*/
       //rdb pblock = new CompositeBlock(bd, bds.getRecordDescriptor());
       pblock = bds.getRecordDescriptor().createDataBlock( bd );
       bds.subblock( pblock, bd, bd );
       //System.out.println(pblock.toString());
       /*create a uniform cell block containing the created point datablock*/
       //urBlock = new URCellBlock(b, pblock.getRecordDescriptor());
       ((URCellBlock)urBlock).setPointDataBlock( pblock );
    } 

    /* Create the list of ComponentInfo of point component data source */
    private ComponentInfo[] createPointCompInfo(ISBounds b, ISBounds res)
    {
       int length;
       ISIDMapper isMap, isMap2;
       ISBounds baseb, baseR, compositeb, pointb;
       ArrayList<ARRSLeafNode> array; 
       ArrayList<ComponentInfo> componentArray;
       ComponentInfo[] compInfo;
       URCellBlock urBlock;
       //rdb CompositeBlock block, block2, componentblock;
       DataBlock block, block2, componentblock;
       DSApproximator approx = new DSApproximator();

       // System.out.println("Beginning of createPointCompInfo"); 

       // create ISIDMapper for base resolution bounds and specified resolution bounds
       pointb = b.copyIncreasedBy( 1 );
       baseR = res.copyIncreasedBy( 1 );
       isMap = new ISIDMapper( this.bounds, baseR );

       // get the leaf cell blocks in a ARRS tree
       array = this.tree.getLeafList();
       length = array.size();

       ISBounds isb1, isb2;
	   ARRSNode node;
       componentArray = new ArrayList<ComponentInfo>();
       IndexSpaceID offset = pointb.copyLower();
       offset.negate();
       
       for ( int i = 0; i < length; i++) 
       {
          node = array.get( i );
          // get the base resolution bounds for the component
          baseb = node.getBaseBounds();

          // get the resolution bounds for the component
          if ( this.bounds.equals( baseR ))
              compositeb = baseb;
          else 
              compositeb = isMap.toComposite( baseb ); 

          /*get the intersection between compositeb and the specified bound*/ 
          isb1 = compositeb.intersection( pointb );
          if ( isb1 != null ) 
          {
             isb2 = isb1.copy();
             isb2.move( offset ); //composite bounds in the specified bounds

             IndexSpaceID offset2 = compositeb.copyLower();
             offset2.negate();
             isb1.move( offset2 );

             /*get a component cell data source*/
             urBlock = (URCellBlock)((ARRSCopyLeafNode)node).getRegionData();

             /*get the point data source included in the component 
               cell datasource*/
             block = urBlock.getPointDataBlock(); 

             /*get the resolution bounds for the component*/
             if ( block.getBounds().volume() == compositeb.volume() )
                componentblock = block;
             else
                componentblock = (DataBlock)approx.approximate( block, compositeb );
                //rdb componentblock = (CompositeBlock)approx.approximate(block,compositeb);

             if ( componentblock.getBounds().volume() == isb2.volume() )
				block2 = componentblock;
		     else 
             {	
                /*get component data contained in the specified bounds*/
                //rdb block2 = new CompositeBlock( isb2, block.getRecordDescriptor() ); 
                block2 = block.getRecordDescriptor().createDataBlock( isb2 );
                componentblock.subblock( block2, block2.getBounds(), isb1); 
             }

             /*create ISIDMapper object with the base bounds of the component 
               and its specified resolution bounds*/
             isMap2 = new ISIDMapper( block2.getBounds(), isb2 );
             /*create the info for the component data source*/
              ComponentInfo info = new ComponentInfo( block2, null, isMap2 );
              componentArray.add(info); 
          }
       }
       
       compInfo = new ComponentInfo[componentArray.size()];
       for ( int i = 0; i < compInfo.length; i++ )
       {
          compInfo[ i ] = componentArray.get( i );
       }

       return compInfo;
    } 
}

