/*
    ARRSCellDataSource.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
*/
    
// ------------------------ Edit history (since 10/21/05) -----------------------
// 11/13/05 rdb: Inherited from CellDataSource rather than RectCellBlock
// 11/09/05 rdb: Made abstract class to be parent of ARRSCellCopyDS and ARRSRefCellDS
// 10/21/05 rdb: Cleaned up unused variables and some reformatting.
//--------------------------------------------------------------------------------

/**
 * The ARRSCellDataSource class implements an interface to access an adaptive
 * resolution cell data source. It contains an ARRS tree corresponding to 
 * the ARRS cell space. 
 */
                                                                                

package edu.unh.sdb.datasource;                                                                                
import edu.unh.sdb.common.*;
//import java.io.*;

//abstract public class ARRSCellDataSource extends RectCellBlock implements CellIndexable 
abstract public class ARRSCellDataSource extends CellDataSource 
{
    protected DataSource pointDS; //store 1-D point datasource
    protected ARRSTree tree; //store an ARRS tree
    protected ARRSTree subtree; //store an ARRS tree with a subblock data

    /** Construct a ARRSCellDataSource without arguments */
    public ARRSCellDataSource()
    {
       super();
       this.pointDS = null;
       this.tree = null;
       this.subtree = null;
    }
                                                                                
    /** Construct a ARRSCellDataSource with an ARRSTree*/
    public ARRSCellDataSource( ARRSTree tree, DataSource pointds )
    {
       super();
       this.pointDS = pointds;
       this.tree = tree;
       this.subtree = null;
       this.init();
    }

    /** Construct a ARRSCellDataSource with the given base bounds and number of 
      * attributes
      *
      * ***************** rdb delete ****************
    public ARRSCellDataSource( ISBounds basebounds, RecordDescriptor recordDesc )
    {
       super( basebounds, recordDesc );
       this.pointDS = null;
       this.tree = null;
       this.subtree = null;
    }
    ***************************************/

    public void activate() 
    { 
       this.pointDS.activate(); 
    }

    public void deactivate() { this.pointDS.deactivate(); }


    /** Set the base bounds */
    public void setBaseBounds(ISBounds basebounds)
    {
       this.bounds = (ISBounds)basebounds.clone();
    }

    /** Set the point datasource */
    public void setPointDS(DataSource pds)
    {
       this.pointDS = pds;
    }

    /** Set an ARRS tree */
    public void setARRSTree(ARRSTree tree)
    {
       if (this.tree != null)
          this.tree.treeClear();

       this.tree = tree;
       if (this.pointDS != null)
          this.init();
    }

    /* set base bounds */
    protected void init()
    {
       ARRSNode node = null;
       if ( this.tree != null )
       {
          node = this.tree.getNode( 0 );
          this.bounds = node.getBaseBounds(); 
          this.dimensionality = this.bounds.dim();
          this.recordDesc = this.pointDS.getRecordDescriptor();
          this.numAttributes = this.pointDS.getNumAttributes();
       }
    }
 
    /** Get an ARRS tree */
    public ARRSTree getARRSTree()
    {
       return this.tree;
    }

    /** Get the point datasource */
    public DataSource getPointDS()
    {
       return this.pointDS;
    }

    /**
     * Size of the ARRSCellDataSource  -- not implemented
     */
    public int volume()
    {
       return -1;
    }

    /**
     * @deprecated Use volume instead
     * @return 
     */
    public int size()
    {
       return volume();
    }

    /** Get the base bounds */
    public ISBounds getBaseBounds()
    {
       return this.bounds;
    }

    /** Set the location specified by the IndexSpaceID to the given datum.*/
    public void setDatum(IndexSpaceID isid, Datum d)
    { 
       new SDBError( SDBError.Warning, "Unimplemented AR Cell functionality." );
    }


    /** Return a new DataBlock that is a subset of this datablock. The Data
        is shared between the two DataBlocks.
        */
    abstract public  DataBlock subblock( ISBounds db );
 

    /** Reads a subblock defined by the specified bounds into the DataBlock
        argument */
    abstract public void subblock( DataBlock theBlock, ISBounds bd );

    /** Reads a subblock defined by srcBounds into the region of the Datablock 
     argument defined by dstBounds. The FieldIDMapper argument determines 
     the mapping between datums in this DataBlock and datums in the argument         
     DataBlock. 
     */
    public void subblock( DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds)
    { 
       new SDBError( SDBError.Warning, "Unimplemented AR Cell functionality." );
    }
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock 
        argument defined by dstBounds. The FieldIDMapper argument determines 
        the mapping between datums in this DataBlock and datums in the argument        
        DataBlock. */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, 
                         ISBounds srcBounds, FieldIDMapper mp)
    { 
       new SDBError( SDBError.Warning, "Unimplemented AR Cell functionality." );
    }
   
    /** Return a reference to the actual array containing the data. */
    public Datum [] data()
    { 
       new SDBError( SDBError.Warning, "Unimplemented AR Cell functionality." );
       return null;
    }

    /** Return by reference the Datum corresponding to the given
     *  IndexSpaceID.
     */
    public void datum( Datum dtm, IndexSpaceID did ) 
    { 
       new SDBError( SDBError.Warning, "Unimplemented AR Cell functionality." );
    }
   
    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
    { 
       new SDBError( SDBError.Warning, "Unimplemented AR Cell functionality." );
    }

    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec)
    { 
       new SDBError( SDBError.Warning, "Unimplemented AR Cell functionality." );
    }

    /** Return a cell block with the specified bounds and the specified
     *  resolution 
     */
    abstract public void urCellBlock(DataBlock urBlock, ISBounds b, ISBounds res);
 }

