/*
 * ARRSCopyDataBlock
 * 
 * --------------------------------------------------
 * History since 6/5/08
 * 06/05/08 rdb: commented out unused variables
 */
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.RecordDescriptor;

/**
 * ARRSCopyDataBlock.java
 * 
 * Used to hold the definition of a DataBlock (for the points) in an ARRSCellBlock
 * or an ARRSCellDS using the Copy model.
 */
public class ARRSCopyDataBlock extends DataBlockProxy
{
   private DataCollection pointDC = null;       // DataCollection for point data
   private DataCollection indexPointDC = null;  // DataCollection for the index info
   private int lowIndex = 0;                    // low index for block
   private int highIndex = 0;                   // high index for block
   /**
    * 
    */
   private ARRSCopyDataBlock()
   {
      super();
   }
   /**
    * @param isb
    * @param recordDesc
    * @param axisOrdering
    */
   public ARRSCopyDataBlock( ISBounds isb, RecordDescriptor recordDesc,
                                 AxisOrdering ordering )
   {
      super( isb, recordDesc, ordering );
   }

   /**
    * @param isb
    * @param recordDesc
    */
   public ARRSCopyDataBlock( ISBounds isb, RecordDescriptor recordDesc )
   {
      super( isb, recordDesc );
   }
   
   /**
    * setIndexPDS
    */
   public void setIndexData( DataCollection indexDC, int low, int high )
   {
      this.indexPointDC = indexDC;
      this.lowIndex = low;
      this.highIndex = high;
   }
   
   /**
    * setPointPDS
    */
   public void setPointData( DataCollection pointDC )
   {
      this.pointDC = pointDC;
   }
   
   /**
    * getPointPDS
    */
   public DataCollection getPointData( )
   {
      return this.pointDC;
   }
   
   /**
    * getIndexDC
    */
   public DataCollection getIndexDC()
   {
      return this.indexPointDC;
   }
   
   /**
    * getIndexHigh
    */
   public int getIndexHigh()
   {
      return this.highIndex;
   }
   
   /**
    * getIndexLow
    */
   public int getIndexLow()
   {
      return this.lowIndex;
   }
   
   /**
    * loadData() - responsibility of the child class to implement code to load the
    * true DataBlock object.
    * @return DataBlock
    */
   protected DataBlock loadData()
   {
      return null;
   }
}
