/*
    ARRSCopyLeafNode.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
*/
//------------------------ Edit History since 11/09/05 -------------------------
// 12/26/05 rdb: Added DS case in setRegionData()
//               Added error throw if cellSource not valid type
// 12/10/05 rdb: Added copy constructor
// 11/13/05 rdb: Removed use of CompositeBlock
// 11/09/05 rdb: Renamed from ARRSLeafNode; made ARRSLeafNode the parent.
//------------------------------------------------------------------------------
/**
 * The ARRSCopyLeafNode class develop an ARRS leaf node structure.
 * It is extended from the ARRSNode class.
 */
                                                                               

package edu.unh.sdb.datasource;

//import java.io.*;

import edu.unh.sdb.common.*;

public class ARRSCopyLeafNode extends ARRSLeafNode 
{
   private IndexBlock indexb;
   private ISBounds indexBlockBounds;
   //private RecordDescriptor recordDesc;   

   /** Constructor without argument */
   public ARRSCopyLeafNode()
   {
      this( (ISBounds) null );
   }
                                                                                
   /** Constructor takes a single ISbounds */
   public ARRSCopyLeafNode(ISBounds bounds)
   {
      this( bounds, null );
   }
                                                                                
   /** Constructor takes a ISbounds and a cell block */
   public ARRSCopyLeafNode(ISBounds bounds, CellIndexable cellSource)
   {
      super( bounds, cellSource );
      this.indexb = null;
      this.indexBlockBounds = null;
   }
  
   /** Copy constructor  */
   public ARRSCopyLeafNode( ARRSCopyLeafNode src )
   {
      super( src.baseBounds, src.cellSource );
      this.indexb = src.indexb;
      this.indexBlockBounds = src.indexBlockBounds;
   }
   
   /** 
    * copy method creates references to all components, not copies.
    */
   public ARRSNode copy()
   {
      return new ARRSCopyLeafNode( this );
   }
                                                                                
   /** Set an IndexBlock */  
   final public void setIndexBlock( IndexBlock indexb )
   {
      this.indexb = indexb;
   }

   /** Get the IndexBlock */
   final public IndexBlock getIndexBlock()
   {
      return this.indexb;
   }


   /** Set the index block bounds */
   final public void setIndexBlockBounds(ISBounds b)
   {
      this.indexBlockBounds = b;
   }                                                                            
 
   /** Get the index block bounds */
   final public ISBounds getIndexBlockBounds() 
   { 
      return this.indexBlockBounds; 
   }

   /** Set an uniform region cell data*/
   protected void setRegionData()
   {
      int d = this.baseBounds.dim();
      int[] lower, upper, up, bsize, csize;
      IntegerIndexSpaceID baseposition, baseshape, cshape;
      DataSource ds = null;
      if ( this.cellSource instanceof ARRSCellCopyDS )
         ds = ((ARRSCellCopyDS)this.cellSource ).getPointDS();
      else if ( this.cellSource instanceof ARRSCellBlock )  
         ds = ((ARRSCellBlock)this.cellSource).getPointDS();
      else
         new SDBError( SDBError.Fatal, "Unknown cellSource: " 
                                 + this.cellSource.getClass().getName() );

      RecordDescriptor recordDesc = ds.getRecordDescriptor(); 
      ISBounds b = this.indexBlockBounds.copyReducedBy(1);
      b.moveToOrigin();

      /*create a DimResolution */
      //lower = (int [])(this.baseBounds.getLower().getCoords().clone());
      //upper = (int [])(this.baseBounds.getUpper().getCoords().clone());
      lower = this.baseBounds.copyLower().refCoords();
      upper = this.baseBounds.copyUpper().refCoords();
      up = b.copyUpper().refCoords();
      bsize = new int[d];
      csize = new int[d];
      for ( int i = 0; i < d; i++) 
      {
         bsize[i] = upper[i] - lower[i];
         csize[i] = up[i] + 1;
      }
      baseposition = this.baseBounds.copyLower();
      baseshape = new IntegerIndexSpaceID( bsize );
      cshape = new IntegerIndexSpaceID( csize );
      DimResolution dresolution = new DimResolution(cshape, baseposition,
                                                    baseshape);

      // rdb: added case for DataSource. I don't understand why it wasn't there before
      //      nor how code could have worked at all for this case. Also, 
      //      not entirely sure about the correctness or elegance of this code.
      
      URCellBlock cblock = null;
      DataBlock   pblock = null;
      if ( this.cellSource instanceof ARRSCellBlock 
           || this.cellSource instanceof ARRSCellCopyDS )
      {
         /*create a point data block */
         //rdb CompositeBlock pblock = new CompositeBlock( this.indexBlockBounds,
         //rdb                                             recordDesc );
         pblock = recordDesc.createDataBlock( this.indexBlockBounds );
         this.indexb.getCellData( ds, pblock );
         //System.out.println("pblock in setRegionData: \n");
         //System.out.println(pblock.toString());
         // create a URCellBlock
         cblock = new URCellBlock( b, recordDesc );
         cblock.setPointDataBlock( pblock );
         cblock.setDimResolution( dresolution );
         cblock.setPath( this.path );
         this.theRegionData = cblock;
      }
      else
         new SDBError( SDBError.Fatal, "System error: unknown type: " 
                                       + this.cellSource.getClass().getName() );
   } 
}
  
