//
//  BlockCacheDataSource.java
//  CacheTest
//
//  Created by Philip Rhodes on Sun Nov 16 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//
//    ----------------- Edit History (as of 10/21/05  ----------------------------
//  10/21/05 rdb: removed unused variables
//                commented out unused private method (attributeMatch)

package edu.unh.sdb.datasource;

//import java.io.*;


public abstract class BlockCacheDataSource extends CacheDataSource {

    IndexSpaceID blockDimensions;


    /** Create a BlockCacheDataSource backed by the DataSource given as the second argument.
      * The third argument specifies the dimensions of the blocks used.
      */
	public BlockCacheDataSource(String name, DataSource ds, ISBounds iterationSpace, IndexSpaceID blockDimensions){
	
		super(name, ds, iterationSpace);
        
        this.blockDimensions=blockDimensions;
	}


	/** This method should be 
      * overridden by child classes to create the cache object.
      */	
	public void activate(){
	
        super.activate();

        //rdb revised so doesn't use a local variable that is never used
        if ( !( theSource instanceof PhysicalDataSource ))
            System.out.println( "Caching a non-Physical DataSource...hope that's okay!");
	}


    /** Return an ISBounds representing the dimensions of the CacheDataBlocks used in
     *  in this BlockCacheDataSource.
     */  
    public ISBounds getCacheBlockBounds(){
    
        return new ISBounds(this.blockDimensions);
    }

    /** Return an IndexSpaceID representing the dimensions of the CacheDataBlocks used in
     *  in this BlockCacheDataSource.
     */
    public IndexSpaceID getCacheBlockDimensions(){
    
        return this.blockDimensions;
    }

    /** Returns a new ISBounds representing the bounds of the cacheBlock that contains the given id. */
    public ISBounds getCacheBlockBounds(IndexSpaceID id){
    
        return ((BlockCache)this.theCache).getCacheBlockBounds(id);
    }


    
    /** Sets the ISBounds argument to the bounds of the cacheBlock that contains the given id. */
    public void getCacheBlockBounds(IndexSpaceID id, ISBounds pBounds){
    
        ((BlockCache)this.theCache).getCacheBlockBounds(id, pBounds);
    }

    /** Return the number of bytes consumed by the cache when allocated. This method 
     *  does not check to see if the cache has actually been allocated by the
     *  activate() method. 
     */
    public long getCacheMemoryUsage(){
    
        return ((BlockCache)this.theCache).getCacheMemoryUsage();
    }
    
    /** Returns the ISPartitioning for the cache asscociated with this BlockCacheDataSource.
     *  This is package access only for the moment.
     */
    ISPartitioning getCachePartitioning(){
        
        //return ((BlockCache)this.theCache).getCachePartitioning();
        
        // this works even when the datasource is not yet activated
        return new RegularISPartitioning(this.blockDimensions, this.iterationSpace);
    }

    /** Returns an ISBoundsIterator that produces bounds corresponding to the cache blocks.*/
    public ISBoundsIterator getCacheBlockBoundsIterator(){
        
        return ((BlockCache)this.theCache).getCacheBlockBoundsIterator();
    }


    /** Sets the ISBoundsIterator argument so it produces bounds corresponding to the cache blocks.*/
    public void getCacheBlockBoundsIterator(ISBoundsIterator iter){
        
        ((BlockCache)this.theCache).getCacheBlockBoundsIterator(iter);
    }


}
