/* 
    BlockedDataSourceBuilder.java
    Author:  Denise D. Mitchell 2002
    University of New Hampshire
    Durham, NH 03824
*/
package edu.unh.sdb.datasource;
//import java.io.*;
import edu.unh.sdb.common.*;

//--------------------------BlockedDataSourceBuilder class ----------------------
/** Manufactures Logical Blocked DataSource objects.  Holds relevant data and validates before instantiating the appropriate
DataSource objects*/


public class BlockedDataSourceBuilder extends DataSourceBuilder
{

    /** Default Constructor sets the bounds and recordDesc members to default values.
    */
    public BlockedDataSourceBuilder(DataSourceCollection d, String n)
    {
        super(d,n);
    }
    
  /** returns a FieldIDMapper for the current component **/
  protected FieldIDMapper makeComponentFieldIDMapper()
  {
    if (componentFieldPositions.isEmpty() ||
        compositeFieldPositions.isEmpty())  //user didn't specify fields
    { 
      //make up a FieldIDMapper with all fields from the component datasource
      if (getCurrentDataSource() != null)
      { //we have a datasource, get record descriptor
        //get record descriptor
        RecordDescriptor rd = getCurrentDataSource().getRecordDescriptor();
        int[] array; 
    
        if (rd != null)
        {// we have a record descriptor, make array for mapping
          if (rd.size() < 1)
          {
            array = new int[1];
            array[0] = 0;
          } 
          else 
          {
            array = new int[rd.size()];
            for (int i = 0; i < rd.size(); i++)
              array[i] = i;
          }
          return new FieldIDMapper(array, array);
        }
      }
        System.err.println("Could not make a FieldIDMapper.");
        return null;  //this shouldn't happen
    }
    else  //we have a field positions specified;
    { 
      return new FieldIDMapper(
                vectorToIntArray(componentFieldPositions),
                vectorToIntArray(compositeFieldPositions));
    }
  }
  
    
    /**completes a new FieldIDMapper for current component.  This method
    currently does nothing since Blocked Data Sources does not
    allow field mapping */
    
  public void processFinalComponentMap()
  {
        componentMaps.addElement( makeComponentFieldIDMapper() );
  } 
    
  /** instantiates an actual datasource **/
  public DataSource makeDataSource()
  {
    if (discardFlag == true)
      return  null;
       
    
    return new BlockedDataSource(
            getDSName(),  
            makeComponentInfos());

  }

 }
