//
//  BufferedOneDDataSink.java
//  DataSinkTest
//
//  Created by Philip Rhodes on Mon May 05 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//
//------------------- Edit history as of 10/21/05 -------------------------
// 10/21/05: rdb Deleted unused variables to avoid compiler warnings
//-------------------------------------------------------------------------

package edu.unh.sdb.datasource;

import java.io.*;

import edu.unh.sdb.common.*;
import java.nio.*;

public class BufferedOneDDataSink extends OneDDataSink{


	//protected BasicBlock buffer;
    protected byte [] bytes;
    protected int byteArrayLength;
	
	//protected int arrayType;
    protected ByteBuffer buffer;
	
	/** Constructs a one dimensional Datasink without specifying a bounds for the index space. If this
	 *	constructor is used, then an IndexSpaceID cannot be passed to the write() method.
     *  Datums are simply written one after the other. This DataSink uses a buffer to store data in
     *  memory until an entire block of data can be written to disk. The bufferSize argument determines 	 
	 *  the size of this buffer, expressed as the number of datums than can be held.
	*/
	
	
	public BufferedOneDDataSink(String filename, RecordDescriptor rd, int bufferSize){
	
		super(filename, rd);
/*		
		int [][] typeCounts=rd.getStorageTypeCounts();
		
		
		
		if(typeCounts.length > 1 ) {
		
			new IllegalArgumentError(SDBError.Fatal, "Data with more than one Storage Type is not currently supported.");
		}
		
        if(typeCounts[0][1] != StorageType.Int){
        
            new IllegalArgumentError(SDBError.Fatal, "Data types other than int not currently supported.");

        }
        
		this.arrayType=typeCounts[0][1];
		
*/		
		this.byteArrayLength=(bufferSize/datumSize)*datumSize;
		
	}
	
	
	public void activate(){

		super.activate();
		super.deactivate();
		
		
		this.bytes=new byte[this.byteArrayLength];
        this.buffer=ByteBuffer.wrap(this.bytes);

	}



	public void deactivate(){
	
        this.writeBuffer();
		
		this.bytes=null;
		this.buffer=null;
	}



	public void write(Datum d, IndexSpaceID id){
	
		new SDBError(SDBError.Fatal,"This form of write() is not supported by BufferedOneDDataSink.");
	}
	


	public void write(Datum d){
	
		
		if( ! this.buffer.hasRemaining()){
		
			writeBuffer();
		}
		
		
		int i,n=this.numFields;
		float value;
		int storageType;
		
		for(i=0;i<n;i++){
		
			storageType=recordDescriptor.getAttribute(i).getStorageType();
			
			value=d.getFloat(i);
			
			switch(storageType){
			
				case StorageType.Float:
					this.buffer.putFloat(value);
					break;
					
				case StorageType.Double:
					this.buffer.putDouble(value);
					break;
					
				case StorageType.Int:
					this.buffer.putInt( (int) value);
					break;
					
				case StorageType.Short:
					this.buffer.putShort( (short) value);
					break;

				case StorageType.Byte:
					this.buffer.put( (byte) value);
					break;
					
				default:
					new InternalFaultError(SDBError.Fatal,
				"Unsupported storage type code in BufferedOneDDataSink.write(Datum):"+storageType);
					break;
			}
		}
		
	}

	
	
	
	protected void  writeBuffer()
    {
		try {
		
			int position=this.buffer.position();
			
			if(position !=0){
					
				this.open();
				
				this.raFile.seek(this.currentOffset);
				this.raFile.write(this.bytes, 0, position);
				
				this.close();	
					
					
				this.buffer.clear();
				this.currentOffset += position;			
				
			}
			
		} catch (IOException e){
			new FileError(SDBError.Recoverable, "IO Error in writeBuffer().");
		}	
		
	
	}


}
