//
//  CacheMaker.java
//  CacheTest
//
//  Created by Philip Rhodes on Wed Dec 04 2002.
//  Copyright (c) 2002 University of New Hampshire. All rights reserved.
//
//------------------ Edit history --------------------------------------
// 03/29/06 dlipsa: Using ObliquePlaneIteratorInterface in CacheMaker
//                  I can use the cache for SliceIterator as well.
// 10/21/05 rdb: Deleted unused variables to avoid compiler warnings
// 11/23/04 rdb: replaced deprecated iter.getAxisOrdering
//
//----------------------------------------------------------------------
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

public class CacheMaker 
{

   // This flag is just for performance testing. It forces the use of cubic
   // cache blocks with the oblique iterator.
   public static boolean useCubicForOblique = false;
   
   // This flag is just for performance testing. It forces the use of cubic
   // cache blocks with the oblique iterator that are the same volume as
   // would normally be used.
   public static boolean useIsometricCubicForOblique = false;
   

    public static DataSource createCacheDS( DataSource source, 
                                            ISIterator iter, 
                                            long       memAvailable)
    {
        AxisOrdering iteratorOrdering=iter.getAxisOrdering();
        IndexSpaceID cacheBlockDimensions=computeBestBlockShape(
                        iteratorOrdering,
                        iter.getAxisLengthArray(),
                        source.getRecordDescriptor().byteSize(),
                        memAvailable );


        if ( checkCacheBlockBounds(cacheBlockDimensions.refCoords(), 
                  ((PhysicalDataSource)source).getAxisOrdering()))
        {
            return new UniBlockCacheDataSource(
                                       "UniBlockCacheDS", source, 
                                       iter.getBounds(), 
                                       cacheBlockDimensions);
        } 
        else 
        {
            System.out.println("Not enough memory to make a useful cache. " +
                                "Returning original datasource.");
            return source;
        }
    }


    public static DataSource createCacheDS( DataSource   source, 
                                            ISBounds     iterationSpace, 
                                            AxisOrdering iteratorOrdering, 
                                            long memAvailable )
    {
        IndexSpaceID cacheBlockDimensions=computeBestBlockShape(
                        iteratorOrdering,
                        iterationSpace.getDimensions(new int[iterationSpace.dim()]),
                        source.getRecordDescriptor().byteSize(),
                        memAvailable );


        if( checkCacheBlockBounds( cacheBlockDimensions.refCoords(),
                                   ((PhysicalDataSource)source).getAxisOrdering()))
        {
            return new UniBlockCacheDataSource( "UniBlockCacheDS", 
                                                source, iterationSpace, cacheBlockDimensions);
        } 
        else 
        {
            System.out.println("Not enough memory to make a useful cache. " +
                               "Returning original datasource.");
            return source;
        }
    }
    
	public static DataSource createCacheDS( DataSource       source, 
                                           ISBoundsIterator iter, 
                                           long             memAvailable)
   {
        AxisOrdering iteratorOrdering=iter.getAxisOrdering();
        int [] gaps=iter.getGapsArray();
        
        IndexSpaceID cacheBlockDimensions=
					           computeBestBlockShape(
					                        iteratorOrdering,
					                        iter.getSpaceDimensionsArray(),
					                        source.getRecordDescriptor().byteSize(),
					                        memAvailable,
					                        iter.getBlockShapeArray(),
					                        gaps );
					

        if ( checkCacheBlockBounds(cacheBlockDimensions.refCoords(),
                            ((PhysicalDataSource)source).getAxisOrdering()))
        {
            return new UniBlockCacheDataSource( "UniBlockCacheDS", source, 
                                                iter.getSpaceBounds(), cacheBlockDimensions);
        } 
        else 
        {
            System.out.println( "Not enough memory to make a useful cache. "+
                                "Returning original datasource.");
            return source;
        }

    }



    public static DataSource createCacheDS( ChunkedDataSource source, 
                                            ISIterator iter, 
                                            long memAvailable)
    {
        AxisOrdering iteratorOrdering = iter.getAxisOrdering();
        int d = source.dim();
        
        int [] chunkDimensions = source.getChunkDimensions().refCoords();
        int [] chunkAxisLengthArray = new int[d];
        int [] elementAxisLengthArray = iter.getAxisLengthArray();
        
        for ( int i=0; i<d; i++)
        {
            chunkAxisLengthArray[i] = elementAxisLengthArray[i] / chunkDimensions[i];
        }
        
        IndexSpaceID cacheBlockDimensions = 
							           computeBestBlockShape(
							                        iteratorOrdering,
							                        chunkAxisLengthArray,
							                        source.getChunkByteSize(),
							                        memAvailable );
        
        for ( int i=0; i<d; i++)
        {
            cacheBlockDimensions.setCoord( i, 
                            cacheBlockDimensions.getCoord(i) * chunkDimensions[i]);
        }

        if ( checkCacheBlockBounds( cacheBlockDimensions.refCoords(), 
                                  ((PhysicalDataSource)source).getAxisOrdering()))
        { 
            return new UniBlockCacheDataSource( "UniBlockCacheDS", source, 
                                            iter.getBounds(), cacheBlockDimensions);
        } 
        else 
        {
            System.out.println( "Not enough memory to make a useful cache. " +
                                "Returning original datasource.");
            return source;
        }
    }

   public static DataSource createCacheDS( ObliquePlaneIteratorInterface iter, long memAvailable){

     return createCacheDS(iter, memAvailable, true);
   }
    
   public static DataSource createCacheDS( ObliquePlaneIteratorInterface iter, long memAvailable, boolean threaded){
   
      int d = 3;
      long memRequired = 0;
      int [] cacheBlockDimensions = new int[d];
      int [] cacheArrayDimensions = new int[d];
      int [] planeShape = iter.getPlaneShapeArray();
      float [] normal = iter.getNormalArray(); // We might change this one.
      float [] truePlaneNormal = normal.clone();
      int [] iterationSpaceDimensions = iter.getSpaceShapeArray();
      //int [] spaceSizes = new int[d];
      int projectionAxis = ObliquePlaneIterator.chooseProjectionAxis(normal);
      //int max2Axis = findSecondMax(normal, projectionAxis);
      int elementByteSize = iter.getDataSource().getRecordDescriptor().byteSize();
      //boolean sizeExceeded = false;
      DataSource source = iter.getDataSource();
      
      // Sometimes there is already a cache, and we are being
      // asked to revise it, perhaps because of new memory constraints.
      if( source instanceof CacheDataSource) {
      
         source = ((CacheDataSource)source).getSource();
      }
      
      
      massage(normal);
      
//      for(int i=0; i<d; i++){
//       
//         if (normal[i] == 0){
//         
//            spaceSizes[i] = iterationSpaceDimensions[i];
//            //spaceSizes[i] = planeShape[i];  // should choose appropriate 3D axis
//         }
//      }


      // First, let's see if we can just use a single cache block that contains the 
      // whole iteration space.
      
      memRequired = elementByteSize;
      
      for(int i=0; i<d; i++){
         
         memRequired *= iterationSpaceDimensions[i];
      }
      
      if(memRequired < memAvailable) {
      
         iter.setDataSource(new MultiBlockCacheDataSource("mbcds",
                                                          source,
                                                          iter.getSpaceBounds(),
                                                          new IntegerIndexSpaceID(iterationSpaceDimensions),
                                                          new IntegerIndexSpaceID(new int[] {1,1,1})
                                                          )
                            );
         
         return iter.getDataSource();
      
      }
      
      boolean foundSolution = computeDimensions(cacheBlockDimensions, cacheArrayDimensions, normal, truePlaneNormal, planeShape, iterationSpaceDimensions, projectionAxis, d, elementByteSize, memAvailable, threaded);
      boolean failed=false;
      
      
      // If that didn't work, let's enter into a loop that searches for a solution
      // with smaller cache blocks.
      
      while(!foundSolution && !failed) {
      
         failed = reduceNormal(normal);
      
         foundSolution = computeDimensions(cacheBlockDimensions, cacheArrayDimensions, normal, truePlaneNormal, planeShape, iterationSpaceDimensions, projectionAxis, d, elementByteSize, memAvailable, threaded);
   
      } 
      
      
      if(useIsometricCubicForOblique) {      
      
         int volume = cacheBlockDimensions[0] * cacheBlockDimensions[1] * cacheBlockDimensions[2];
         
         int dimension = (int) (Math.pow(volume, 1.0f/3) );
         
         cacheBlockDimensions[0] = cacheBlockDimensions[1] = cacheBlockDimensions[2] = dimension;
         //computeCacheArrayDimensions(truePlaneNormal, planeShape, cacheBlockDimensions,  cacheArrayDimensions, projectionAxis, findSecondMax(normal, projectionAxis));
         computeCacheArrayDimensions(normal, planeShape, cacheBlockDimensions,  cacheArrayDimensions, projectionAxis, findSecondMax(normal, projectionAxis), threaded);
         
      }
      
      
      
      
      if(failed) {
         System.out.println( "memAvailable is too small to make a well formed cache. " +
                             "Returning original datasource.");
         return source;
      }
      
      
      
      
           
      if ( (source instanceof ChomboRemoteClient) || (source instanceof RemoteBFDS) || (source instanceof StatsDataSource) || checkCacheBlockBounds( cacheBlockDimensions, 
                                  ((PhysicalDataSource)source).getAxisOrdering()))
      { 
      
      
         // If the line below calls the constructor for MultiBlockCacheDataSourceGL, 
         // then delete the "GL" and commit the change. That class is meant for
         // debugging/visualization only, and is not in the repository.
         iter.setDataSource(new MultiBlockCacheDataSource("mbcds",
                                                          source,
                                                          iter.getSpaceBounds(),
                                                          new IntegerIndexSpaceID(cacheBlockDimensions),
                                                          new IntegerIndexSpaceID(cacheArrayDimensions)
                                                          )
                            );

         iter.setSPCacheBlocks( ! (CacheMaker.useIsometricCubicForOblique || CacheMaker.useCubicForOblique));
         
         return iter.getDataSource();
      } else {
      
         System.out.println( "memAvailable is too small to make an effective cache. " +
                             "Returning original datasource.");
         return source;
      }

      
   }

   /** Create a MultiBlockCache for the given ObliquePlaneIterator using cache blocks
    *  of the specified dimensions. This 
    *  method is meant for evaluation purposes, and would not normally be used.
    */
   public static DataSource createCacheDS( ObliquePlaneIteratorInterface iter, int [] cacheBlockDimensions, long memAvailable){

      return createCacheDS(iter, cacheBlockDimensions, memAvailable, true);
   }

   /** Create a MultiBlockCache for the given ObliquePlaneIterator using cache blocks
    *  of the specified dimensions. This 
    *  method is meant for evaluation purposes, and would not normally be used.
    */
   public static DataSource createCacheDS( ObliquePlaneIteratorInterface iter, int [] cacheBlockDimensions, long memAvailable, boolean threaded){
   
      int [] cacheArrayDimensions = new int[3];
      float [] normal = iter.getNormalArray();
      int [] planeShape = iter.getPlaneShapeArray();
      int projectionAxis = ObliquePlaneIterator.chooseProjectionAxis(normal);
      int max2Axis = findSecondMax(normal, projectionAxis);
      DataSource source = iter.getDataSource();
      
      massage(normal);
      
      computeCacheArrayDimensions(normal, planeShape, cacheBlockDimensions, cacheArrayDimensions, projectionAxis, max2Axis, threaded);

      // Sometimes there is already a cache, and we are being
      // asked to revise it, perhaps because of new memory constraints.
      if( source instanceof CacheDataSource) {
      
         source = ((CacheDataSource)source).getSource();
      }

      iter.setDataSource(new MultiBlockCacheDataSource("mbcds",
                                                          source,
                                                          // let's cache the entire DS index space
                                                          //iter.getSpaceBounds(), // or not, if this is uncommented
                                                          new IntegerIndexSpaceID(cacheBlockDimensions),
                                                          new IntegerIndexSpaceID(cacheArrayDimensions)
                                                          )
                            );
     
      return iter.getDataSource();
   }
   
   
   
   
   public static boolean reduceNormal(float[] normal){
   
      int maxIndex = findMax(normal);
        
      if( normal[maxIndex] - 1 <= 0){

         return true;
      } else {
          
         normal[maxIndex]--;
         return false;
      }
    
   }
   
   
   public static int findMax(float[] normal){
   
      int i, d=normal.length;
      float a;
      int maxIndex = 0;
      float max = Math.abs(normal[0]);

      for(i=1; i<d; i++){

         a = Math.abs(normal[i]);
         if(a > max){
             max = a;
             maxIndex = i;
         }
         
      }
      
      return maxIndex;
   }
   
   

   public static int findMin(float[] normal){
   
      int i, d=normal.length;
      float a;
      int minIndex = 0;
      float min = Math.abs(normal[0]);

      for(i=1; i<d; i++){

         a = Math.abs(normal[i]);
         if(a < min){
             min = a;
             minIndex = i;
         }
         
      }
      
      return minIndex;
   }


   /** Return the index of the element with second largest magnitude */
    protected static int findSecondMax(float vector[], int maxindex){
    
        int i, d=vector.length;
        float max2 = 0;
        float a;
        int max2index = 0;
        
        for(i=0; i<d; i++){
        
            if(i != maxindex){
               a = Math.abs(vector[i]);
               if(a > max2){
                   max2 = a;
                   max2index = i;
               }
            }
        }
             
        return max2index;
   }


   
   
   public static boolean computeDimensions(
                                          int [] cacheBlockDimensions, 
                                          int [] cacheArrayDimensions, 
                                          float [] normal,
                                          float [] truePlaneNormal,
                                          int [] planeShape,
                                          int [] boundingBoxDimensions,
                                          int projectionAxis, 
                                          int dimensionality, 
                                          int elementByteSize, 
                                          long memAvailable,
                                          boolean threaded) { 
                                          
                                          
      int s = computeMaxScaleFactor(projectionAxis, boundingBoxDimensions);
      int max2Axis = findSecondMax(normal, projectionAxis);


      long memRequired = 0;


      do {
         
         
         computeCacheBlockDimensions(s, cacheBlockDimensions, normal, boundingBoxDimensions, dimensionality);
         
         //computeCacheArrayDimensions(truePlaneNormal, planeShape, cacheBlockDimensions, cacheArrayDimensions, projectionAxis, max2Axis);
         computeCacheArrayDimensions(normal, planeShape, cacheBlockDimensions, cacheArrayDimensions, projectionAxis, max2Axis, threaded);
         
         
         memRequired = computeRequiredMemory(cacheBlockDimensions, cacheArrayDimensions, elementByteSize);

         //System.out.println("memRequired =" + memRequired + " memAvailable=" + memAvailable);
         
         
         s--;

      } while (memRequired > memAvailable && s > 0);
      
      
      if(memRequired <= memAvailable){
      
//         s++;
//         
//         computeCacheBlockDimensions(s, cacheBlockDimensions, normal, boundingBoxDimensions, dimensionality);
//                  
//         computeCacheArrayDimensions(normal, planeShape, cacheBlockDimensions, cacheArrayDimensions, projectionAxis, max2Axis);
//         
//         
//         memRequired = computeRequiredMemory(cacheBlockDimensions, cacheArrayDimensions, elementByteSize);
         
         return true;

      } else {
      
         return false;
      }

   }
   
   
   
   
   public static long computeRequiredMemory(
                                          int [] cacheBlockDimensions, 
                                          int [] cacheArrayDimensions, 
                                          int elementByteSize){
                                          
    
      long memRequired = elementByteSize;
      int d = cacheBlockDimensions.length;
         
      for(int i=0; i<d; i++){
         
         memRequired *= cacheArrayDimensions[i] * cacheBlockDimensions[i];
      }
                                                                                                                                                            
      return memRequired;
   }
   
   
   public static int computeMaxScaleFactor(int projectionAxis, int [] spaceDimensions){
   
//      int minIndex = findMin(massagedNormal);
//      
//      
//      return spaceDimensions[minIndex];


      return spaceDimensions[projectionAxis];
   
   }
   
   
   
   
   public static void computeCacheBlockDimensions(int s, int [] cacheBlockDimensions, float [] normal, int [] spaceDimensions, int dimensionality) { 

         for(int i=0; i < dimensionality; i++){
         
         
            if(useCubicForOblique) {
            
               cacheBlockDimensions[i] = s;
            } else {
            
               if (normal[i] != 0){   
               
   //               if (useCubicForOblique) {
   //               
   //                  cacheBlockDimensions[i] = s;
   //               } else {
                  
                     cacheBlockDimensions[i] = (int) Math.ceil(s * normal[i]);
   //                  cacheBlockDimensions[i] = (int) Math.floor(s * normal[i]);
   //               }
               } else {
               
                  //cacheBlockDimensions[i] = (int) s;
                  
                  cacheBlockDimensions[i] = spaceDimensions[i];

                  //cacheBlockDimensions[i] = spaceSizes[i];
               }
            }
            
            cacheBlockDimensions[i] = (cacheBlockDimensions[i] > spaceDimensions[i]) ? spaceDimensions[i] : cacheBlockDimensions[i];
            
            
         }
   
   }

   public static void computeCacheArrayDimensions(float [] normal, int[] planeShape,  int [] cacheBlockDimensions, int [] cacheArrayDimensions, int projectionAxis, int max2Axis, boolean threaded){
      
//      if(!threaded){
         switch(projectionAxis){    
            
            case 0:
                  
                  cacheArrayDimensions[0] = (int) Math.ceil( (Math.abs(  cacheBlockDimensions[1] * normal[1] / normal[0] + cacheBlockDimensions[2] * normal[2] / normal[0])) / cacheBlockDimensions[0] + 1);
                  cacheArrayDimensions[1] = (int) Math.ceil((float) planeShape[0]                   / cacheBlockDimensions[1]) + (normal[1] == 0 ? 0:1);
                  cacheArrayDimensions[2] = (int) Math.ceil((float) planeShape[1]                   / cacheBlockDimensions[2]) + (normal[2] == 0 ? 0:1);
                  
                  break;
            
            case 1:

                  cacheArrayDimensions[0] = (int) Math.ceil((float) planeShape[0]                   / cacheBlockDimensions[0]) + (normal[0] == 0 ? 0:1);
                  cacheArrayDimensions[1] = (int) Math.ceil( (Math.abs(  cacheBlockDimensions[0] * normal[0] / normal[1] + cacheBlockDimensions[2] * normal[2] / normal[1])) / cacheBlockDimensions[1] + 1);
                  cacheArrayDimensions[2] = (int) Math.ceil((float) planeShape[1]                   / cacheBlockDimensions[2]) + (normal[2] == 0 ? 0:1);
                  
                  break;
                  
            case 2:

                  cacheArrayDimensions[0] = (int) Math.ceil((float) planeShape[0]                   / cacheBlockDimensions[0]) + (normal[0] == 0 ? 0:1);
                  cacheArrayDimensions[1] = (int) Math.ceil((float) planeShape[1]                   / cacheBlockDimensions[1]) + (normal[1] == 0 ? 0:1);
                  cacheArrayDimensions[2] = (int) Math.ceil( (Math.abs(  cacheBlockDimensions[0] * normal[0] / normal[2] + cacheBlockDimensions[1] * normal[1] / normal[2])) / cacheBlockDimensions[2] + 1);
                  
                  break;
         }
//      } else {
//         switch(projectionAxis){    
//            
//            case 0:
//                  
//                  cacheArrayDimensions[0] = (int) Math.ceil( 2 * (Math.abs( cacheBlockDimensions[1] * normal[1] / normal[0] + cacheBlockDimensions[2] * normal[2] / normal[0])) / cacheBlockDimensions[0] + 2);
//                  cacheArrayDimensions[1] = (int) Math.ceil((float) planeShape[0]                   / cacheBlockDimensions[1]) + (normal[1] == 0 ? 0:2);
//                  cacheArrayDimensions[2] = (int) Math.ceil((float) planeShape[1]                   / cacheBlockDimensions[2]) + (normal[2] == 0 ? 0:2);
//                  
//                  break;
//            
//            case 1:
//                     // pegging axis 1 to 5 seems to work, yielding 9,5,9
//                  cacheArrayDimensions[0] = (int) Math.ceil((float) planeShape[0]                   / cacheBlockDimensions[0]) + (normal[0] == 0 ? 0:2);
//                  cacheArrayDimensions[1] = (int) Math.ceil( 2 * (Math.abs( cacheBlockDimensions[0] * normal[0] / normal[1] + cacheBlockDimensions[2] * normal[2] / normal[1])) / cacheBlockDimensions[1] + 2);
//                  cacheArrayDimensions[2] = (int) Math.ceil((float) planeShape[1]                   / cacheBlockDimensions[2]) + (normal[2] == 0 ? 0:2);
//
//                  break;
//                  
//            case 2:
//
//                  cacheArrayDimensions[0] = (int) Math.ceil((float) planeShape[0]                   / cacheBlockDimensions[0]) + (normal[0] == 0 ? 0:2);
//                  cacheArrayDimensions[1] = (int) Math.ceil((float) planeShape[1]                   / cacheBlockDimensions[1]) + (normal[1] == 0 ? 0:2);
//                  cacheArrayDimensions[2] = (int) Math.ceil( 2 * (Math.abs( cacheBlockDimensions[0] * normal[0] / normal[2] + 2 * cacheBlockDimensions[1] * normal[1] / normal[2])) / cacheBlockDimensions[2] + 2);
//                  
//                  break;
//         }
//      }
      
      //Hopefully, increasing the "thickness" of the cache in the 
      // projection axis will allow us to prefetch blocks needed in 
      // the near future without disturbing blocks currently needed.
      //
      // We also need to adjust the "width" of the cache in the subordinate
      // axes.
      if(threaded){
      
         for(int i=0; i < cacheArrayDimensions.length; i++){
         
            if(i == projectionAxis){
            
               cacheArrayDimensions[i] *= 2;
               cacheArrayDimensions[i]++;
               
            } else {
               if (normal[i] != 0) { // Should be safe because of setEffectiveZeros()
               
                  cacheArrayDimensions[i] += 1; 
               }
            }
         }
      }
   }
   
   
   private static double length(float [] c){
   
      int d = c.length;
      float sum = 0;
      
      for(int i=0; i<d; i++){
         sum += c[i] * c[i];
      }
      
      return Math.sqrt(sum);
   }
   
   
      
   private static float [] dilate(float [] v, int [] dimensions, float [] dv){
   
      int d = v.length;
      
      for(int i=0; i < d; i++){
      
         dv[i] = v[i] / dimensions[i];
      }
   
      return dv;
   }
   
   
   private static float max(float v1, float v2) {
   
      if (v1 > v2) {
      
         return v1;
      } else {
      
         return v2;
      }
   }

   
//   public static void oldcomputeCacheArrayDimensions(float [] normal, int[] planeShape,  int [] cacheBlockDimensions, int [] cacheArrayDimensions, int projectionAxis, int max2Axis){
//      
//      //int [] planeShape = iter.getPlaneShapeArray();
////      int [] iterationSpaceDimensions = iter.getSpaceShapeArray();
//      
//      int d = normal.length;
//      float [] psNormal = new float[d];  // Normal mapped to partition space
//                                       // described by cacheBlockDimensions
//      
//      int max2AxisPlane = 0;
//      
//      int j=0;
//      for(int i=0; i<3; i++){
//
//         psNormal[i] = (float) normal[i] / cacheBlockDimensions[i];
//
//         if(i != projectionAxis){
//                  
//            if (normal[i] == 0){
//               
//               cacheArrayDimensions[i] = 1;
//            } else {
//               
//               //cacheArrayDimensions[i] =  (int) Math.ceil( (float) planeShape[j] / cacheBlockDimensions[i]) + 1; 
//               
//               //cacheArrayDimensions[i] =  (int) Math.ceil( Math.abs(normal[i] * ((float)cacheBlockDimensions[i] / cacheBlockDimensions[projectionAxis]))) + 1;
//               
//               cacheArrayDimensions[i] =  (int) Math.ceil(Math.abs(normal[i] * ((float) cacheBlockDimensions[projectionAxis] / cacheBlockDimensions[i]))) + 1;
//            }
//            
//            j++;
//         }
//      }
//
//      //massage(psNormal);
//      
//      switch(projectionAxis){
//         
//         case 0:
//               max2AxisPlane = max2Axis - 1;
//               break;
//         
//         case 1:
//               max2AxisPlane = (max2Axis == 0) ? 0: 1;
//               break;
//               
//         case 2:
//               max2AxisPlane = max2Axis;
//               break;
//      }
//      
////      System.out.println("Slope is "+ (- psNormal[max2Axis] / psNormal[projectionAxis]));
////      System.out.println("plane extent is " + Math.ceil((- psNormal[max2Axis] / psNormal[projectionAxis]) * ((float )planeShape[max2AxisPlane]/cacheBlockDimensions[max2Axis])));
//      
//      //cacheArrayDimensions[projectionAxis] = (int) Math.ceil( Math.abs( (- psNormal[max2Axis] / psNormal[projectionAxis]) * ((float )planeShape[max2AxisPlane]/cacheBlockDimensions[max2Axis]) )) + 1;
//      
////      cacheArrayDimensions[projectionAxis] = (int) Math.ceil( Math.abs( ((float )planeShape[max2AxisPlane]/cacheBlockDimensions[max2Axis]) / (cacheArrayDimensions[max2Axis] ) )) + 1;
//
////      j=0;
////      cacheArrayDimensions[projectionAxis] = 1;
////      for(int i=0; i<d; i++){
////      
////         if(i != projectionAxis) {
////         
////            cacheArrayDimensions[projectionAxis] *= (int) Math.ceil( Math.abs( ((float )planeShape[j] / cacheBlockDimensions[i]) / (cacheArrayDimensions[i]) )) + 1;
////            
////            
////            
////            
////            j++;
////         }
////      }
//      //cacheArrayDimensions[projectionAxis]++;
//      
//
//      j=0;
//      cacheArrayDimensions[projectionAxis] = 1;
//      
//      float [] a = new float [3];
//      float [] b = new float [3];
//      float [] c = new float [3];
//      
//      switch(projectionAxis){
//         
//         case 0:
//               
//               a[0] = (normal[1] / normal[0] / planeShape[0])  / cacheBlockDimensions[0];
//               a[1] = (float) planeShape[0]                    / cacheBlockDimensions[1];
//               a[2] = 0;
//         
//
//               b[0] = (normal[1] / normal[0] / planeShape[1]) / cacheBlockDimensions[0] ;
//               b[1] = 0;
//               b[2] = (float) planeShape[1]                   / cacheBlockDimensions[2];
//               
//               break;
//         
//         case 1:
//               a[0] = 0;
//               a[1] = (normal[2] / normal[1] / planeShape[0])  / cacheBlockDimensions[1];
//               a[2] = (float) planeShape[0]                    / cacheBlockDimensions[2];
//         
//
//               b[0] = (float) planeShape[1]                    / cacheBlockDimensions[0];
//               b[1] = (normal[0] / normal[1] / planeShape[1])  / cacheBlockDimensions[1];
//               b[2] = 0;
//               break;
//               
//         case 2:
//               a[0] = (float) planeShape[0]                    / cacheBlockDimensions[0];
//               a[1] = 0;
//               a[2] = (normal[0] / normal[2] / planeShape[0])  / cacheBlockDimensions[2];
//         
//
//               b[0] = 0;
//               b[1] = (float) planeShape[1]                    / cacheBlockDimensions[1];
//               b[2] = (normal[1] / normal[2] / planeShape[1])  / cacheBlockDimensions[2];
//               
//               break;
//      }
//      
//      c[0] = a[1] * b[2] - a[2] * b[1];
//      c[1] = a[2] * b[0] - a[0] * b[2];
//      c[2] = a[0] * b[1] - a[1] * b[0];
//      
//      float sum = 0;
//      
//      for(int i=0; i<d; i++){
//         sum += c[i] * c[i];
//      }
//      
//      float  area = (int) Math.ceil(Math.sqrt(sum));
//
//      
//      cacheArrayDimensions[projectionAxis] = (int) area;
//            
//   }
   
   
//    /** Return the index of the element with maximum magnitude */
//    protected static int findMax(float vector[]){
//    
//        int i, d=vector.length;
//        float max = Math.abs(vector[0]);
//        float a;
//        int maxindex = 0;
//        
//        for(i=0; i<d; i++){
//        
//            a = Math.abs(vector[i]);
//            if(a > max){
//                max = a;
//                maxindex = i;
//            }
//            
//        }
//             
//        return maxindex;
//    }

//    /** Return the index of the element with maximum magnitude */
//    protected static int findMin(float vector[]){
//    
//        int i, d=vector.length;
//        float min = Math.abs(vector[0]);
//        float a;
//        int minindex = 0;
//        
//        for(i=0; i<d; i++){
//        
//            a = Math.abs(vector[i]);
//            if(a < min){
//                min = a;
//                minindex = i;
//            }
//            
//        }
//             
//        return minindex;
//    }


 
  
   /** This method finds the smallest magnitude non-zero coordinate and
    *  divides the other coordinates by this value. 
    */
   protected static void massage(float vector[]) {
   
      int i, d=vector.length;
      float min = Float.MAX_VALUE;
      float a, scale;
      int minindex = -1;
      //float EPSILON = 1.0e-7f;

      for(i=0; i<d; i++){
      
//         if(Math.abs(vector[i]) < EPSILON){
//            
//            vector[i] = 0;
//         }
            
      

         a = Math.abs(vector[i]);
         if(a != 0 && a < min){
             min = a;
             minindex = i;
         }
         
      }

      scale = 1.0f / vector[minindex];
      
      for(i=0; i<d; i++){
      
         vector[i] = Math.abs( scale * vector[i] );
      }
   
   }
   
   
   
  
   
    public static DataSource createThreadedCacheDS( DataSource source, 
                                                    ISIterator iter, 
                                                    long       memAvailable)
    {
        AxisOrdering iteratorOrdering=iter.getAxisOrdering();
        IndexSpaceID cacheBlockDimensions=computeBestBlockShape(
                        iteratorOrdering,
                        iter.getAxisLengthArray(),
                        source.getRecordDescriptor().byteSize(),
                        memAvailable / 2 );


        if ( checkThreadedCacheBlockBounds(cacheBlockDimensions ))
        {
            return new ThreadedBlockCacheDataSource("ThreadedBlockCacheDS", source, iter.getBounds(), cacheBlockDimensions, iteratorOrdering);
        } 
        else 
        {
            System.out.println( "Not enough memory to make a useful ThreadedBlockCache. "
                              + "Trying UniBlockCache." );
            return CacheMaker.createCacheDS(source, iter, memAvailable);
        }    
    }


	public static DataSource createThreadedCacheDS(DataSource source, 
	                                     ISBoundsIterator iter, long memAvailable)
	{
        AxisOrdering iteratorOrdering=iter.getAxisOrdering();
        int [] gaps=iter.getGapsArray();
        
        IndexSpaceID cacheBlockDimensions=computeBestBlockShape(
                        iteratorOrdering,
                        iter.getSpaceDimensionsArray(),
                        source.getRecordDescriptor().byteSize(),
                        memAvailable / 2,
                        iter.getBlockShapeArray(),
                        gaps );


        if ( checkThreadedCacheBlockBounds( cacheBlockDimensions ))
        {
            return new ThreadedBlockCacheDataSource( "ThreadedBlockCacheDS", source, 
                                   iter.getSpaceBounds(), cacheBlockDimensions, iteratorOrdering);
        } 
        else 
        {
            System.out.println( "Not enough memory to make a useful ThreadedBlockCache."
                                + " Trying UniBlockCache.");
            return CacheMaker.createCacheDS(source, iter, memAvailable);
        }
    }
    
    public static DataSource createThreadedSliceCacheDS( DataSource source, 
                                                ISIterator iter, long memAvailable)
    {
        int sliceAxis=iter.getAxisOrdering().getMostSignificantAxis();
        ISBounds iterationSpace=iter.getBounds();
        
        return CacheMaker.createThreadedSliceCacheDS(source, iterationSpace, 
                                                     sliceAxis, memAvailable);
    }
    

	public static DataSource createThreadedSliceCacheDS( DataSource source, 
	      ISBounds iterationSpace, int sliceAxis, long memAvailable)
	{
		int d=source.dim();
		AxisOrdering ordering=new AxisOrdering(d);
		ordering.setMostSignificantAxis(sliceAxis);
		
		int elementByteSize=source.getRecordDescriptor().byteSize();
		int[] spaceDimensions=iterationSpace.getDimensions(new int[d]);
		
		IndexSpaceID cacheBlockDimensions=
		   CacheMaker.computeBestBlockShape(ordering, spaceDimensions, 
		                                    elementByteSize, memAvailable/2);
        
//        int [] cacheBlockDimensionsArray = cacheBlockDimensions.getCoords();
//        
//        if( (spaceDimensions[sliceAxis] / cacheBlockDimensionsArray[sliceAxis]) % 2 != 0){
//        
//            // If their are an odd number of partitions along the slice axis, we
//            // should adjust the cache block size to make it an even number
//            cacheBlockDimensionsArray[sliceAxis] --;
//        }
        
		if ( iterationSpace.isSlice( cacheBlockDimensions, sliceAxis ))
		{
			return new ThreadedSliceCacheDataSource("ThreadedSliceCache", 
			      source, iterationSpace, sliceAxis, cacheBlockDimensions);
		} 
		else 
		{
			System.out.println("Not enough memory to make a useful ThreadedSliceCache.");
			return source;
		}
	}


	public static DataSource createThreadedSliceCacheDS( DataSource       source, 
	      											  ISBoundsIterator iter, 
	      											  long             memAvailable)
	{
        System.out.println("Calling ctscds: memAvailable="+memAvailable);
        
        AxisOrdering iteratorOrdering=iter.getAxisOrdering();
        int sliceAxis=iteratorOrdering.getMostSignificantAxis();
        ISBounds iterationSpace = iter.getSpaceBounds();
        int [] spaceDimensions=iterationSpace.getDimensions(new int[source.dim()]);
        
        IndexSpaceID cacheBlockDimensions=computeBestBlockShape(
                        iteratorOrdering,
                        spaceDimensions,
                        source.getRecordDescriptor().byteSize(),
                        memAvailable / 2, 
                        iter.getBlockShapeArray(),
                        iter.getGapsArray());
        
        
        if ( !(iterationSpace.isSlice(cacheBlockDimensions, sliceAxis))) 
        {
            new SDBError(SDBError.Recoverable, "createThreadedSliceCacheDS: " +
            		      "Not enough memory to make a useful ThreadedSliceCache. " +
            		      "Trying UniBlockCache.");
            return CacheMaker.createCacheDS(source, iter, memAvailable);
         } 
        else 
        {
            int sliceDimension = cacheBlockDimensions.getCoord( sliceAxis );
            
            if ( sliceDimension == spaceDimensions[ sliceAxis ] ) 
            {
                // We have blocks with large extent along the slice axis
                new SDBError(SDBError.Recoverable,"createThreadedSliceCacheDS: " +
                		"The block shape and iterator axis axisOrdering are not suitable " +
                		"for a ThreadedSliceCache. Trying UniBlockCache.");
                return CacheMaker.createCacheDS(source, iter, memAvailable);
            } 
            else 
            {
                return new ThreadedSliceCacheDataSource( "ThreadedSliceCache", 
                         source, iterationSpace, sliceAxis, cacheBlockDimensions );
            } 
        } 
    }

	public static IntegerIndexSpaceID computeBestBlockShape( AxisOrdering iteratorOrdering, 
	                 int [] axisLengths, int elementByteSize, long memAvailable)
	{
        int d=iteratorOrdering.dim();
        long blockBytes; 
        int axis, iteratorAxisLength;
        int[] cacheBlockDimensionsArray=new int[d];
               
        for ( int i=0; i<d; i++)        
            cacheBlockDimensionsArray[i]=1;        
            
        blockBytes    = elementByteSize;
        memAvailable -= blockBytes * 1;
    
        for ( int i=0; i<d; i++ )
        {
            axis = iteratorOrdering.getNthLeastSignificantAxis(i);
            iteratorAxisLength = axisLengths[axis];
            
            if ( blockBytes * (iteratorAxisLength - 1) <= memAvailable )
            {
                cacheBlockDimensionsArray[axis] = iteratorAxisLength;
                memAvailable -= blockBytes * (iteratorAxisLength - 1);
                blockBytes *= iteratorAxisLength;
            } 
            else 
            {
                cacheBlockDimensionsArray[axis] = (int) (memAvailable / blockBytes + 1);
                break;
            }
        }
        
        IntegerIndexSpaceID cacheBlockDimensions
                               = new IntegerIndexSpaceID( cacheBlockDimensionsArray );
        //System.out.println("computeBestBlockShape2: cacheBlockDimensions="+cacheBlockDimensions);
        return cacheBlockDimensions;
    }

	//------------ oldcomputeBestBlockShape ----------------------
	static IntegerIndexSpaceID oldcomputeBestBlockShape(AxisOrdering iteratorOrdering, 
	      int [] axisLengths, int elementByteSize, int memAvailable)
	{
        int d=iteratorOrdering.dim();
        int blockBytes, axis, blockAxisLength, iteratorAxisLength;
        int[] cacheBlockDimensionsArray=new int[d];

        blockBytes=elementByteSize;
        memAvailable -= blockBytes * 1;
    
        for ( int i=0; i<d; i++)
        {
            axis = iteratorOrdering.getNthLeastSignificantAxis(i);
            iteratorAxisLength = axisLengths[axis];
            blockAxisLength = (blockBytes * (iteratorAxisLength - 1) <= memAvailable) 
                            ? iteratorAxisLength 
                            : memAvailable / blockBytes + 1;
            
			// Make the blockAxisLength equal to the greatest power of two
			// less than or equal to the value computed above.  Perhaps we
			// can eventually remove this check.
			
			//blockAxisLength = (int) Math.pow(2, (int) (Math.log(blockAxisLength)/Math.log(2)) );
			
            
            cacheBlockDimensionsArray[axis]=blockAxisLength;
            memAvailable -= blockBytes * (blockAxisLength - 1);
            blockBytes *= blockAxisLength;
        }
        IntegerIndexSpaceID cacheBlockDimensions=new IntegerIndexSpaceID(cacheBlockDimensionsArray);
        //System.out.println("computeBestBlockShape2: cacheBlockDimensions="+cacheBlockDimensions);
    
        return cacheBlockDimensions;
    }


	/**
	 * computeBestBlockShape
	 * 
	 * @param iteratorOrdering
	 * @param axisLengths
	 * @param elementByteSize
	 * @param memAvailable
	 * @param shape
	 * @param gaps
	 * @return
	 */
	static IntegerIndexSpaceID computeBestBlockShape( AxisOrdering iteratorOrdering, 
	                                                  int [] axisLengths, 
	                                                  int elementByteSize, 
	                                                  long memAvailable, 
	                                                  int [] shape, 
	                                                  int [] gaps)
	{
        int d=iteratorOrdering.dim();
        int axis, cacheBlockAxisLength, stepsInAxis;
        long cacheBlockBytes;
        int [] cacheBlockDimensionsArray=new int[d];
        int [] steps = new int[d];
        int stepVolume=1;
        
        for (int i=0; i<d; i++)
        {
            steps[i] = Math.abs(shape[i] + gaps[i]); //PJR added abs()
            stepVolume *= steps[i];
        }
        
        int stepByteSize= stepVolume * elementByteSize;
        
        // cache block bytes, not iterator block bytes
        cacheBlockBytes = stepByteSize;
        memAvailable -= cacheBlockBytes;
    
        for (int i=0; i<d; i++)
        {
            axis = iteratorOrdering.getNthLeastSignificantAxis(i);
            stepsInAxis = axisLengths[axis] / steps[axis];
            if ( cacheBlockBytes * (stepsInAxis - 1) <= memAvailable )
                cacheBlockAxisLength = stepsInAxis * steps[axis];
            else 
            {
                /*
                bytesperstep=blockBytes*step
                numsteps = memAvailable/bytesPerStep;
                
                numSteps = memAvailable / ( blockBytes * step )
                */
                stepsInAxis  = (int)( memAvailable / cacheBlockBytes + 1);
                cacheBlockAxisLength = stepsInAxis * steps[axis];
            }
            
            cacheBlockDimensionsArray[axis]=cacheBlockAxisLength;
            memAvailable -= cacheBlockBytes * (stepsInAxis - 1);
            cacheBlockBytes *= stepsInAxis;
        }

        IntegerIndexSpaceID cacheBlockDimensions
                               = new IntegerIndexSpaceID(cacheBlockDimensionsArray);
        
        //System.out.println("computeBestBlockShape: cacheBlockDimensions="+cacheBlockDimensions);
        return cacheBlockDimensions;
    }

	/**
	 * checkCacheBLockBounds
	 * @param cacheBlockDimensions
	 * @param iteratorOrdering
	 * @param dsOrdering
	 * @return
	 */
   static boolean checkCacheBlockBounds( int [] cacheBlockDimensions, 
                                          AxisOrdering storageOrdering)
   {
      int rodAxis=storageOrdering.getLeastSignificantAxis();
      int d = storageOrdering.dim();
      
		for ( int i=0; i<d; i++)
		{
			if (cacheBlockDimensions[i] == 0 )
				return false;
		}

      // if rod axis is 1, this cache shape is unlikely to be helpful, so
      // there's no point in constructing a cache. It would result in the same
      // number of reads, so Cache overhead would give slower performance.
      // However, if we had separate threads for filling the Cache, it
      // might still be useful.
       return cacheBlockDimensions[rodAxis] != 1;
      
   }


    /**
     * checkThreadedCacheBlockBounds
     * 
     * @param cacheBlockDimensions
     * @return
     */
    static boolean checkThreadedCacheBlockBounds(IndexSpaceID cacheBlockDimensions)
    {
		int d=cacheBlockDimensions.dim();
        
		for ( int i=0; i<d; i++)
		{
			if ( cacheBlockDimensions.getCoord(i) == 0 )
				return false;
		}
        return true;
    }


	static IntegerIndexSpaceID oldoldcomputeBestBlockShape(
	                                         AxisOrdering iteratorOrdering, 
	                                         int [] axisLengths, 
	                                         int elementByteSize, 
	                                         int memAvailable)
	{
        int d=iteratorOrdering.dim();
        int blockBytes, axis, blockAxisLength, iteratorAxisLength;
        int[] cacheBlockDimensionsArray=new int[d];
            
        blockBytes=elementByteSize;
        memAvailable -= blockBytes * 1;
    
        for ( int i=0; i<d; i++)
        {
            axis = iteratorOrdering.getNthLeastSignificantAxis(i);
            iteratorAxisLength = axisLengths[axis];
            
            blockAxisLength = (blockBytes * (iteratorAxisLength - 1) <= memAvailable) 
                            ? iteratorAxisLength 
                            : memAvailable / blockBytes + 1;
            
			// Make the blockAxisLength equal to the greatest power of two
			// less than or equal to the value computed above.  Perhaps we
			// can eventually remove this check.
			
			blockAxisLength = (int) Math.pow( 2,  
			                       (int) (Math.log(blockAxisLength)/Math.log(2)) );
			
            
            cacheBlockDimensionsArray[axis] = blockAxisLength;
            memAvailable -= blockBytes * (blockAxisLength - 1);
            blockBytes *= blockAxisLength;
        }
        
        IntegerIndexSpaceID cacheBlockDimensions=new IntegerIndexSpaceID(cacheBlockDimensionsArray);
        //System.out.println("computeBestBlockShape: cacheBlockDimensions="+cacheBlockDimensions);
        return cacheBlockDimensions;
    }

	//------------------------ main --------------------------------------
    static public void main(String [] args){
    
        DataSource ds=DataSource.create("foo","simple.24.1.bin.fdl");
        ISBounds subsetBounds=new ISBounds( new IntegerIndexSpaceID( 0, 0 ), 
                                            new IntegerIndexSpaceID( 127, 4095));
        ISIterator iter=new ISIterator(subsetBounds, new AxisOrdering(new int[] {0,1}));

        System.out.println(ds);
        System.out.println(iter);

        DataSource newcds=CacheMaker.createCacheDS(ds,iter, 2097152);    
        newcds.activate();
        
        System.out.println("[autocache] cacheBlockDimensions="+((UniBlockCacheDataSource)newcds).getCacheBlockBounds().getDimensions());
        
        System.exit(0);
        
        ds=DataSource.create("bar","simple3D.24.1.bin.fdl");
        iter=new ISIterator(ds.getBounds(), new AxisOrdering(new int[] {1,2,0}));
        //ISIterator iter=new ISIterator(ds.getBounds(), new AxisOrdering(new int[] {2, 1, 0}));
        
        newcds=CacheMaker.createCacheDS(ds,iter, 16384);
        
        ISBoundsIterator iterb=new ISBoundsIterator(ds.getBounds(), 
                                                new int[]{10,10,10}, 
                                                new int[] {0,0,0},
                                                new AxisOrdering(new int[] {0,1,2}));
        
        DataSource newblockcds = CacheMaker.createCacheDS(ds, iterb, 16384);
        newblockcds.activate();  // rdb added to avoid compiler warning
    }
}
