/*
  Cell.java
  Copyright 2004 University of New Hampshire
  Durham, NH 03824
  Programmer: LI YE
  ------------------- Edit history as of 02/07/05 ---------------------
  02/07/05 rdb: renamed getDimensionality() to dim() for consistency with other classes.
*/

/**
 * The Cell class provides an interface to view the data of a cell in an 
 * N-dimensional cell space. 
 */

package edu.unh.sdb.datasource;

//import java.io.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.lattice.*;

public abstract class Cell {
   protected CellID cellID;      //store the id of a cell
   protected int dimensionality; //store the dimensionality of a cell
   protected Datum[] data;       //store the data contained in a cell
   protected Point[] vertices;   //store the vertices in a cell 

   /** Constructor without argument */
   public Cell() 
   { 
      this.cellID = null;
      this.dimensionality = 0;
      this.data = null;
      this.vertices = null;
   }

   /** Constructor takes a single CellID */
   public Cell(CellID cellid) 
   { 
      this.cellID = cellid;
      this.dimensionality = 0;
      this.data = null;
      this.vertices = null;
   }

   /** Constructor takes a CellID and the dimensionality*/
   public Cell(CellID cellid, int dim) 
   { 
      this.cellID = cellid;
      this.dimensionality = dim;
      this.data = null;
      this.vertices = null;
   }

   /** Set the cell id of a cell */
   public void setCellID(CellID cellid)
   {
      this.cellID = cellid;
   }

   /** Set the dimension of a cell */
   public void setCellDim(int dim)
   {
      this.dimensionality = dim;
   }
 
   /** Set the data of a cell */
   public void setCellData(Datum[] data)
   {
      this.data = data;
   }
                                                                                
   /** Set the vertice values of a cell */
   public void setCellVertices(Point[] vertices)
   {
      this.vertices = vertices;
   }
                                                                                

   /** Get the cell id of a cell */
   public CellID getCellID()
   {
      return this.cellID;
   }

   /** Get the dimensionality of a cell 
    * @deprecated use dim() 
    */
   public int getDimensionality()
   {
      return this.dimensionality;
   }

   /** Get the dimensionality of a cell 
    */
   public int dim()
   {
      return this.dimensionality;
   }

   /** Get the data contained in a cell */
   public Datum[] getData()
   {
      return this.data;
   }

   /** Get the corners of a cell */
   public Point[] getVertices()
   {
      return this.vertices;
   }

   /** Return a String representing the object. */
   public String toString(){
        return this.toString("");
   }

   /** Return a String representing the object indented by the argument string.
     */
   public abstract String toString(String indent);

}
        
