/*
 * DCBoundsIterator
 *    This class extends the notion of ISBoundsIterator by allowing the
 *    iterator to "know" which DataCollection is the target of the
 *    iteration. I created this for use in the GraniteServer class,
 *    but it will be useful for automatic cache creation.
 *
 * Created on May 9, 2005
 *
 * ------------------- Edit history ----------------------------------
 * 05/11/05 rdb: Added getDataBlock()
 * 05/09/05 rdb: Derive from ISBoundsIterator. Add data collection link.
 *
 */
package edu.unh.sdb.datasource;

/**
 * @author rdb
 *
 */
public class DCBoundsIterator extends ISBoundsIterator
{
   private static final long serialVersionUID = 0; // for serialization inheritance
   DataCollection baseDC = null;
   DataBlock      block  = null;    // For efficiency, pre-allocate a block that
                                    // is the right size for the iteration bounds

   /**
    * @param space
    * @param shape
    * @param gaps
    * @param axisOrdering
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, int[] shape, int[] gaps,
                                 AxisOrdering ordering )
   {
      super( space, shape, gaps, ordering );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param shape
    * @param gaps
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, int[] shape, int[] gaps )
   {
      super( space, shape, gaps );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param shape
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, int[] shape )
   {
      super( space, shape );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param shape
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, ISBounds shape )
   {
      super( space, shape );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param shape
    * @param axisOrdering
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, int[] shape, 
                                 AxisOrdering ordering )
   {
      super( space, shape, ordering );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param shape
    * @param axisOrdering
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, ISBounds shape, 
                                 AxisOrdering ordering )
   {
      super( space, shape, ordering );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param start
    * @param shape
    * @param gaps
    * @param axisOrdering
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, IntegerIndexSpaceID start,
                                 int[] shape, int[] gaps, AxisOrdering ordering )
   {
      super( space, start, shape, gaps, ordering );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param start
    * @param shape
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, 
                                 IntegerIndexSpaceID start, int[] shape )
   {
      super( space, start, shape );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param start
    * @param shape
    * @param axisOrdering
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, IntegerIndexSpaceID start, 
                                 int[] shape, AxisOrdering ordering )
   {
      super( space, start, shape, ordering );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param space
    * @param start
    * @param shape
    * @param gaps
    */
   public DCBoundsIterator( DataCollection dc, ISBounds space, IntegerIndexSpaceID start,
                                 int[] shape, int[] gaps )
   {
      super( space, start, shape, gaps );
      this.block = dc.createDataBlock( new ISBounds( shape ));
      this.baseDC = dc;
   }

   /**
    * @param d
    */
   DCBoundsIterator( DataCollection dc, int d )
   {
      super( d );
      this.baseDC = dc;
   }
   
   //+++++++++++++++++ Accessor methods ++++++++++++++++++++++++++++++++++++
   /**
    * Get the DataCollection associated with this Iterator
    */
   public DataCollection getDataCollection()
   {
      return this.baseDC;
   }

   /**
    * Set the DataCollection associated with this Iterator
    */
   public void getDataCollection( DataCollection newDC )
   {
      this.baseDC = newDC;
   }
   
   /**
    * Get a data block containing the data associated with the current
    * iteration bounds.
    */
   public DataBlock getDataBlock()
   {
      if ( !this.valid() )
         return null;
      else
      {
         this.baseDC.subblock( this.block, this );
         return this.block;
      }
   }
}
