//
//  DSStencilIterator.java
//  StencilTest
//
//  Created by Philip Rhodes on 5/25/06.
//  Copyright 2006 Philip J Rhodes. All rights reserved.
//

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;


public class DSPartitionedStencilIterator extends BasicBlock implements SDBIterator{

   DataSource source;
   DSPartitionedStencil stencil;
   
   int numStencilPartitions = 0;
   int currentPartition = 0;
   ISBounds partitionBounds;
   
   boolean valid=false;
   

   public DSPartitionedStencilIterator(DSPartitionedStencil stencil, DataSource ds){
   
      super(new ISBounds(stencil.getPartitionDimensions()), ds.getRecordDescriptor());
   
      this.stencil = stencil;
      
      this.source = ds;
      
      partitionBounds = new ISBounds(stencil.dim());
      
   }
   
   /** Return a reference to the stencil. */
   public DSPartitionedStencil getStencil(){
   
      return this.stencil;
   }

   /** Perform any necessary initialization, and set the iterator value
   *  to the first value in the iteration. If the stencil contains no
   *  elements, then the value of this iterator is undefined, but valid()
   *  will return false. However, activate() is still called on the 
   *  underlying datasource.
   */
   public void init(){
   
      this.currentPartition = 0;
      
      this.numStencilPartitions = stencil.size();
      
      if( !source.isActivated()){
      
         source.activate();
      }

      if ( this.numStencilPartitions > 0){
      
         stencil.getIndexedBounds(partitionBounds, 0);
         source.subblock(this, partitionBounds);
         this.valid=true;
      } else {
      
         this.valid=false;
      }
      
   }

   /** Advance the iterator by one step. If the iterator cannot be advanced
   *  to produce a valid value, the value of the iterator is undefined, but
   *  the next() method will return false, as will a subsequent call to
   *  valid().
   */
   public boolean next(){
   
      this.currentPartition++;
      
      if(this.currentPartition < this.numStencilPartitions){
      
         stencil.getIndexedBounds(partitionBounds, this.currentPartition);
         
         source.subblock(this, partitionBounds);
         
         
         return true;
      
      } else {
      
         this.valid = false;
      
         return false;
      }
   
   
   }

   /** Return true if the iterator currently has a valid value, false otherwise.*/
   public boolean valid(){
   
      return this.valid;
   }

   /** Advance the iterator to the last valid value in the iteration. (optional operation)*/
   public void end(){
   
      source.activate();
   
      this.numStencilPartitions = stencil.size();
      
      if(this.numStencilPartitions > 0) {
   
         this.currentPartition = this.numStencilPartitions - 1;
         stencil.getIndexedBounds(partitionBounds, this.currentPartition);
         source.subblock(this, partitionBounds);
         this.valid = true;
      } else {
      
         this.valid = false;
      }
   
   }

}
