/**
 * DataBlockProxy - an abstract class to implement the Proxy pattern for a DataBlock.
 * The idea is to allow the information describing a DataBlock to be gathered 
 * without actually loading the data until it is accessed. 
 * 
 * We expect the child classes to implement the "loadData" method to cause the 
 * data to be loaded when referenced.
 */
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

/**
 * @author rdb
 * Created 11/27/05. Not really tested. Not yet really used.
 */
abstract public class DataBlockProxy extends DataBlock
{
   private DataBlock theRealData = null;
   /**
    * 
    */
   protected DataBlockProxy()
   {
      super();
   }

   /**
    * @param isb
    * @param recordDesc
    * @param axisOrdering
    */
   public DataBlockProxy( ISBounds isb, RecordDescriptor recordDesc,
                                 AxisOrdering ordering )
   {
      super( isb, recordDesc, ordering );
   }

   /**
    * @param isb
    * @param recordDesc
    */
   public DataBlockProxy( ISBounds isb, RecordDescriptor recordDesc )
   {
      super( isb, recordDesc );
   }
   
   /**
    * loadData() - responsibility of the child class to implement code to load the
    * true DataBlock object.
    * @return
    */
   abstract protected DataBlock loadData();
   
   /**
    * setDataBlock( DataBlock blk )
    */
   protected void setDataBlock( DataBlock blk )
   {
      this.theRealData = blk;
   }

   /**
    * DataBlock getDataBlock()
    */
   protected DataBlock getDataBlock()
   {
      return this.theRealData;
   }

   /**
    * @see edu.unh.sdb.datasource.DataBlock#setDatum(edu.unh.sdb.datasource.IndexSpaceID, 
    * edu.unh.sdb.common.Datum)
    */
   public void setDatum( IndexSpaceID isid, Datum d )
   {
      if ( this.theRealData == null )
         this.theRealData = this.loadData();
      try {
         this.theRealData.setDatum( isid, d );
      } catch ( NullPointerException npe )
      {
         new SDBError( SDBError.Recoverable, "DataBlockProxy: Unable to load data" );
      }
   }

   /**
    * @see edu.unh.sdb.datasource.DataCollection#datum(edu.unh.sdb.common.Datum, 
    * edu.unh.sdb.datasource.IndexSpaceID)
    */
   public void datum( Datum dtm, IndexSpaceID did )
   {
      if ( this.theRealData == null )
         this.theRealData = this.loadData();
      try {
         this.theRealData.datum( dtm, did );
      } catch ( NullPointerException npe )
      {
         new SDBError( SDBError.Recoverable, "DataBlockProxy: Unable to load data" );
      }
   }

   /**
    * @see edu.unh.sdb.datasource.DataCollection#datum(edu.unh.sdb.common.Datum, 
    * edu.unh.sdb.datasource.IndexSpaceID, edu.unh.sdb.common.FieldIDMapper)
    */
   public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
   {
      if ( this.theRealData == null )
         this.theRealData = this.loadData();
      try {
         this.theRealData.datum( dtm, did, mp );
      } catch ( NullPointerException npe )
      {
         new SDBError( SDBError.Recoverable, "DataBlockProxy: Unable to load data" );
      }
   }

   /**
    * @see edu.unh.sdb.datasource.DataCollection#subblock(edu.unh.sdb.datasource.DataBlock, 
    * edu.unh.sdb.datasource.ISBounds, edu.unh.sdb.datasource.ISBounds)
    */
   public void subblock( DataBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds )
   {
      if ( this.theRealData == null )
         this.theRealData = this.loadData();
      try {
         this.theRealData.subblock( dstBlock, dstBounds, srcBounds );
      } catch ( NullPointerException npe )
      {
         new SDBError( SDBError.Recoverable, "DataBlockProxy: Unable to load data" );
      }
   }

   /**
    * @see edu.unh.sdb.datasource.DataCollection#subblock(edu.unh.sdb.common.DataBlock, 
    * edu.unh.sdb.datasource.ISBounds, edu.unh.sdb.datasource.FieldIDMapper)
    */
   public void subblock( DataBlock dstBlock, ISBounds srcBounds, FieldIDMapper fmap )
   {
      if ( this.theRealData == null )
         this.theRealData = this.loadData();
      try {
         this.theRealData.subblock( dstBlock, srcBounds, fmap );
      } catch ( NullPointerException npe )
      {
         new SDBError( SDBError.Recoverable, "DataBlockProxy: Unable to load data" );
      }
   }

   /**
    * @see edu.unh.sdb.datasource.DataCollection#subblock(edu.unh.sdb.common.DataBlock, 
    * edu.unh.sdb.datasource.ISBounds, edu.unh.sdb.datasource.FieldIDMapper)
    */
   public void subblock( DataBlock dstBlock, 
                                 ISBounds dstBounds, ISBounds srcBounds, 
                                 FieldIDMapper fmap )
   {
      if ( this.theRealData == null )
         this.theRealData = this.loadData();
      try {
         this.theRealData.subblock( dstBlock, dstBounds, srcBounds, fmap );
      } catch ( NullPointerException npe )
      {
         new SDBError( SDBError.Recoverable, "DataBlockProxy: Unable to load data" );
      }
   }

}
