/* 
    DatumBlock.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
    -------------- Edit history as of 12/05/04 --------------------
    01/06/05 rdb: create Datum objects from RecordDescriptor
    12/05/04 rdb: Delete size() -- inherited from DataCollection
*/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;



/** A DatumBlock is a DataBlock that stores its data as an internal array of FloatDatums. 
 *
 *  @deprecated
*/
public class DatumBlock extends DataBlock{

    private Datum[]  data;
    private Datum nullDatum; 

    /** Construct a datablock with the given data, ISBounds, axis axisOrdering and RecordDescriptor. 
        The bounds will be moved to the origin.
    */
    public DatumBlock( Datum [] data, ISBounds isb, 
                       AxisOrdering ordering, RecordDescriptor recordDesc)
    {
        super( isb, recordDesc, ordering );
        
        if ( bounds.volume() != data.length )
            new IllegalArgumentError(SDBError.Fatal,
                  "Volume of ISBounds does not match array length."); 
        this.data = data;
        nullDatum = this.recordDesc.createDatum();
    }

    /** Construct a datablock with the given data and ISBounds. The default axis axisOrdering is used, and 
        the bounds will be moved to the origin.*/    
    public DatumBlock( Datum [] data, ISBounds isb, RecordDescriptor recordDesc )
    {
        super(isb,recordDesc);
        
        if (bounds.volume() != data.length)
            new IllegalArgumentError(SDBError.Fatal,"Volume of ISBounds does not match array length."); 

        this.data=data;
        nullDatum=this.recordDesc.createDatum();
    }

    /** Construct a datablock with the given bounds and number of attributes. The data
        array for the block will be created and filled with null datums. */
    public DatumBlock( ISBounds isb, RecordDescriptor recordDesc )
    {
        super(isb,recordDesc);
        
        int i,size=isb.volume();
        this.data=new Datum[size];
        
        for(i=0;i<size;i++){
            this.data[i]=this.recordDesc.createDatum();
        }
        nullDatum=this.recordDesc.createDatum();
    }

    /** Construct a DatumBlock from the given DataBlock. Bounds, Ordering, and
        RecordDescriptor will be copied directly from the DataBlock argument. Also,
        the resulting DatumBlock will REFER to the data contained in the
        DataBlock argument. This is faster than using one of the constructors above.
    */
    DatumBlock(DataBlock b){
    
        super(b);
        
        this.data= ((DatumBlock) b).data();
        nullDatum=b.recordDesc.createDatum();
    }

    /** Construct a DatumBlock from the given DataBlock. Bounds, Ordering, and
        RecordDescriptor will be copied directly from the DataBlock argument. Also,
        the resulting DatumBlock will REFER to the data contained in the
        DataBlock argument. This is faster than using one of the constructors above.
        Note: this function is no longer working 
    */
    DatumBlock(DatumBlock b){
    
        super(b);
        
        this.data=b.data;
        nullDatum=b.nullDatum;
    }




    /** Return a new DataBlock that is a subset of this datablock.*/
    public  DataBlock subblock(ISBounds db){
    
        Datum[] data=new Datum[ db.volume() ];
        int i;
        ISIterator iter = new ISIterator( db );
        
        for(iter.init(),i=0;iter.valid();iter.next()){
            data[i++]= this.datum(iter);
        }
        
        return new DatumBlock( data, db,this.recordDesc );
    }

     //----------------------added by wenhui 5/19/2002
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds){

        ISIterator iter1=new ISIterator(srcBounds);
        ISIterator iter2=new ISIterator(dstBounds);
        
        for(iter1.init(),iter2.init(); iter1.valid(); iter1.next(),iter2.next()){
                Datum d=data[this.computeOffset(iter1)];
                theBlock.setDatum(iter2, d);
        }

    }
    //----------------------------------------------------


    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The FieldIDMapper argument determines the mapping between 
        datums in this DataBlock and datums in the argument DataBlock. */    
    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds, FieldIDMapper mp){

        new InternalFaultError(SDBError.Recoverable,"Unimplemented Method");
    }
 
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The RecordSpec argument specifies a subset of the attributes
        of this DataSource to be written into the DataBlock datums.*/
	public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds, RecordSpec spec){
    
        new InternalFaultError(SDBError.Recoverable,"Unimplemented Method");
    }
    

    /** Given an IndexSpaceID, return the corresponding datum.*/
    public  Datum datum(IndexSpaceID did)
    {
        Datum d=data[this.computeOffset(did)];
        
        if ( d==null ) 
        {
            return this.nullDatum; //return new Datum(this.numAttributes);
        } 
        else 
        {
            Datum newDatum = this.recordDesc.createDatum();
            d.copyTo( newDatum );
            return newDatum;
        }
    }

    /** Given an array of coordinates, load the corresponding datum into the datum 
        argument.*/
    public  void datum(Datum datum, IndexSpaceID id){
    
        Datum d=data[this.computeOffset(id)];
        
        if (d==null) {
            datum.copyTo(this.nullDatum);
        } else {
            datum.copyTo(d);
        }
    }


    /** Given an array of coordinates, load the corresponding datum into the datum 
        argument.*/
    public  void datum( Datum datum, int [] coords )
    {
        Datum d = data[ this.computeOffset(coords) ];
        if ( d == null ) 
           this.nullDatum.copyTo( datum) ; //rdb: 01/05: reversed assignment, was wrong!
        else 
            d.copyTo( datum ); //rdb: reversed assignment, was wrong.
    }
    
    
    /** Given an IndexSpaceID, return the corresponding datum with attributes
        specified in the RecordSpec. */    
    public  Datum datum( IndexSpaceID did, RecordSpec spec )
    {
        Datum d = data[ this.computeOffset(did) ];
        if ( d == null ) 
        {
            return this.recordDesc.createDatum( spec );
        } 
        else 
        {
           Datum retDatum = this.recordDesc.createDatum( spec );
           d.copyTo( retDatum );
           return retDatum;
        }
    }    
    
    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum(Datum dtm, IndexSpaceID did, FieldIDMapper mp){
    
        new InternalFaultError(SDBError.Recoverable,"Unimplemented method.");
    }


    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec){
    
        new InternalFaultError(SDBError.Recoverable,"Unimplemented method.");
    }

    
    
    
    /** Set the location specified by the IndexSpaceID to the given datum.*/
    public void setDatum( IndexSpaceID isid, Datum d)
    {
       Datum dst = this.data[this.computeOffset(isid)];
       if ( dst != null )
          d.copyTo( dst );
       /** rdb
		if (d instanceof FloatDatum)
		{
			this.data[this.computeOffset(isid)]= (FloatDatum) d;
		} else 
		{
			this.data[this.computeOffset(isid)]= new FloatDatum(d);
		}
       /**************/	
    }
    
    /** Return a reference to the actual array containing the data. */
    public Datum [] data(){
    
        return data;
    }
    
    public Values getValues() { return null; } //don't bother to implement
}
