/*
    DimResolution.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
*/
//-------------------- Edit history ----------------------------------------
// 10/21/05 rdb: Deleted unused variables to avoid compiler warnings
//--------------------------------------------------------------------------
                                                                                
/** 
 * The DimResolution class provide some features to view the dimensional 
 * resolution of an ARRS cell data set. The dimensional resolution 
 * represents the size of a cell in d-th dimension in base index space units.
 */                                                                                
package edu.unh.sdb.datasource;
                                                                                
import edu.unh.sdb.common.*;
//import java.io.*;

public class DimResolution {
    private IntegerIndexSpaceID cellblockshape;
               //size in cell space
    private IntegerIndexSpaceID baseposition;
               //position in base space 
    private IntegerIndexSpaceID baseshape;
               //size in base space
 
    /** Constructs a DimResolution without argument */
    public DimResolution()
    {
       this.cellblockshape = null;
       this.baseposition = null;
       this.baseshape = null;
    }
                                                                                
    /** Constructs a DimResolution with specified arguments */ 
    public DimResolution(IntegerIndexSpaceID csize,
                         IntegerIndexSpaceID bpos, IntegerIndexSpaceID bsize)
    {
       this.cellblockshape = csize;
       this.baseposition = bpos;
       this.baseshape = bsize;
    }

    /** Set the size in the cell space */
    public void setCellBlockShape(IntegerIndexSpaceID csize)
    {
       this.cellblockshape = csize;
    }                                                                           

    /** Set the position in the base space */
    public void setBasePosition(IntegerIndexSpaceID bpos)
    {
       this.baseposition = bpos;
    }                                           

    /** Set the size in the base space */
    public void setBaseShape(IntegerIndexSpaceID bsize)
    {
       this.baseshape = bsize;
    }                                                                           

    /** Get the size in the cell space */
    public IntegerIndexSpaceID getCellBlockShape()
    {
       return this.cellblockshape;
    }                                                                           

    /** Get the position in the base space */
    public IntegerIndexSpaceID getBasePosition()
    {
       return this.baseposition;
    }                                           

    /** Get the size in the base space */
    public IntegerIndexSpaceID getBaseShape()
    {
       return this.baseshape;
    }                                                                           
 
    /** Get the base bounds */
    public ISBounds getBaseBounds()
    {
       int[] low, up, size;
       int i, dim;
       
       low = this.baseposition.refCoords().clone();
       size = this.baseshape.refCoords().clone();
       dim = low.length;
       up = new int[dim];
       for (i = 0; i < dim; i++) {
          up[i] = low[i] + size[i]; 
       }

       return new ISBounds(low, up);
    }
      
    /** Get the cell bounds */
    public ISBounds getCellBlockBounds()
    {
       int[] low, up, size;
       int i, dim;
                           
       //low = (int [])(this.cellposition.getCoords().clone());
       size = this.cellblockshape.refCoords().clone();
       dim = size.length;
       low = new int[dim];
       up = new int[dim];
       for (i = 0; i < dim; i++) {
          low[i] = 0;
          up[i] = 0 + size[i] - 1;
       }
       return new ISBounds(low, up);
    }
                                
    /** Return a String representing the object. */
    public String toString(){
       return this.toString("");
    }
                                                                                
                                                                                
    /** Return a String representing the object indented by the argument string.     */
    public String toString(String indent){
       String indent2=indent+Globals.toStringIndent;
       StringBuffer result=new StringBuffer(indent+"{ DimResolution:\n");
                                                                                
       result.append(indent2+"Cellblock size: ");
       result.append(this.cellblockshape.toString()+"\n");
       result.append(indent2+"Base position: ");
       result.append(this.baseposition.toString()+"\n");
       result.append(indent2+"Base shape: ");
       result.append(this.baseshape.toString()+"\n");
       result.append(indent+"}");
       return new String(result);
    }
 
}
                                                                                
                                                                                
