/************************************************************
   GraniteRemoteClient.java
   Copyright 2005 University of Mississippi
   University, MS 38677

************************************************************/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import edu.unh.sdb.remote.*;

import java.io.*;
import java.nio.*;
import java.net.*;
import java.util.*;

/** This class will read data  from a remote datasource using UDT or UDP logic.
*/
public class GraniteRemoteClient extends PhysicalDataSource{

   private DataReceiver receiver = null;
   private boolean useUDT = true;
   private RemoteMessenger rm = null;

   private boolean networkInitialized = false;
   private Server fileServer =null;
   private String fileServerURL = null;

   private String scheme;  // udt, udp, etc.
   private int dataChannelProtocol=Globals.UDT; //UDT is default protocol
   private String remotePath; // path of fdl file on server
   private String hostname;

   private DataInputStream input;

   private String remoteXmlFile, remoteSourceName;

   private ISBounds currentSubBounds = null;

   private byte[] store = null; // used for stencils

   //for performance tests
   private long t1 =0;
   private long t2 =0;
   private long time =0;

   public static int buffer_size = 0;

   //sockets used.
   public static ServerSocket welcomeSocket;
   public static Socket connectionSocket;
   public static Socket clientAlarmSocket;
   public static InetAddress ip;
   public static String ipStr;
   public String ctrlMsg = null;
   //some global variables used.
   public static String fileName;// = "8gigbytes.xfdl";//name of file being transferred.
   public static int fileSize;//size of file being transferred.
   public static int globalVal;//variable used by the alarm thread to check the client activity.
   public static long start, end;//variables to hold start & end times of file transfer.
   public static int statusFlag = 0;//flag that tells which stage of main thread is running.
   public static byte[] dstByteArray;//byte array into which the requested subblock will get into.
   public static int[] lower, upper;//integer arrays used for storing the bounds.
   public static int[] lowerSpaceArray, upperSpaceArray;//integer arrays used for storing the space bounds.
   public static int[] lowerPartitioningArray, upperPartitioningArray;//arrays for storing partitioning bounds.
   public static String bounds;//variable used to hold a string of lower & upper bounds.

   private int allowedRetryTimes = 1;

   private boolean sentStencilConstructorInfo = false;

/** GraniteRemoteClient class constructor.
 */
public GraniteRemoteClient(String name, RecordDescriptor recordDesc, ISBounds theBounds,
        AxisOrdering ordering, String URL, String serverURL)
{
	  super(name,recordDesc,theBounds,ordering,URL );// call inherited constructor
	  this.fileServerURL = serverURL;
	  remoteSourceName = URL;
	  /*
	  System.out.println("==== In GraniteRemoteClient Constructor ====");
	  System.out.println("remoteSourceName: "+remoteSourceName);
	  System.out.println("serverURL: "+serverURL);
	  System.out.println("========================================");
	  */
}

/** Sets the filename for transfer.
*/
public static void setFileName(String s)
{
  	fileName = s;
}

public void SendFileName(String FileInfo){
		rm.PassivelySendControlInfo(welcomeSocket, FileInfo, ipStr);
}

public void SendStencilInfo(String stencilInfo)
{
		rm.PassivelySendControlInfo(welcomeSocket, stencilInfo, ipStr);
}

public void SendPartitionListLengthInfo(String partitionListLengthString)
{
     rm.PassivelySendControlInfo(welcomeSocket, partitionListLengthString, ipStr);//partition list length info.
}


public void SendBounds(String bounds)
{
     rm.PassivelySendControlInfo(welcomeSocket, bounds, ipStr);
}


public int GetBufferSize(){
		int bufSize = 0;
    ctrlMsg = rm.ActivelyReceiveControlInfo(ipStr, Globals.serverControlPort);
    try{
    	bufSize = Integer.parseInt(ctrlMsg.substring(13,ctrlMsg.length()));
    }catch(Exception e)
    {
    	e.printStackTrace();
    }
    return bufSize;
}

public int GetDataSize(){
	 int datasize = 0;
    ctrlMsg = rm.ActivelyReceiveControlInfo(ipStr, Globals.serverControlPort);
    try{
      //if(ctrlMsg!=null)
    		datasize = Integer.parseInt(ctrlMsg.substring(6,ctrlMsg.length()));
    }catch(Exception e)
    {
    	e.printStackTrace();
    }
    return datasize;
}


/** This method does the necessary initialization so that information can be
 *  passed to and from the server. Called by activate() and clearCache().
 */
private void initNetwork(){

   if(! this.networkInitialized) {
     try
     {
        URI remoteSourceURI = new URI(remoteSourceName);
        //System.out.println("Remote Source Name: "+remoteSourceName);
        this.hostname = remoteSourceURI.getHost();
        //System.out.println("Hostname: "+hostname);
        this.remotePath = remoteSourceURI.getPath();
        //System.out.println("path: "+path);
        this.scheme = remoteSourceURI.getScheme();
        //System.out.println("scheme: "+scheme);
        
        if(scheme.equalsIgnoreCase("udp")){
					receiver = new UDPReceiver();
					dataChannelProtocol = Globals.UDP;
					useUDT = false;
				}
        else if(scheme.equalsIgnoreCase("tcp")){
					receiver = new TCPReceiver();
					dataChannelProtocol = Globals.TCP;
					useUDT = false;
				}
				else{
					receiver = new UDTReceiver();
					//System.out.println("UDT Receiver is created.");
        }
        
        ip = InetAddress.getByName(hostname);//getting the ip address from URL.
        //System.out.println("Server IP : "+ip);
				ipStr = ip.getHostAddress();
				//System.out.println("Server IP String : "+ipStr);
				
				receiver.setServerIP(ipStr);
				receiver.setServerDataPort(Globals.serverDataPort);

	      rm = new RemoteMessenger();
				//System.out.println("Remote Messenger is created.");
				
        welcomeSocket = new ServerSocket();//Control channel from client to Server.
        welcomeSocket.setSoTimeout(10 * 60 * 1000);// 10 minute timeout
        welcomeSocket.bind(new InetSocketAddress(Globals.clientControlPort));

        if(scheme.equalsIgnoreCase("udp")){
            /* ###################################################################
            different from UDT receiver, UDP receiver need extra communication
            due to its own logic, to share the control socket and remote messenger
            we have two set methods to do this
            ################################################################### */
            ((UDPReceiver) receiver).setRemoteMessenger(rm);
            ((UDPReceiver) receiver).setClientCtrlSocket(welcomeSocket);
         }	
         
         this.networkInitialized = true;
         
     } catch ( Exception exc)
     {
        new SDBError(SDBError.Fatal, "Exception in initNetwork():" + exc.toString());
        System.exit(1); //newly added
     }

   } else {
   
      System.out.println("GraniteRemoteClient.initNetwork(): already initialized.");
   }

}


/** Open the server socket on the client side.
*/
public void activate()
{
  if(!this.isActivated)
  {
     super.activate();


      initNetwork();
// PJR moved the code below to initNetwork, so it can be called by clearCache().     
//     try
//     {
//        URI remoteSourceURI = new URI(remoteSourceName);
//        //System.out.println("Remote Source Name: "+remoteSourceName);
//        String hostname = remoteSourceURI.getHost();
//        //System.out.println("Hostname: "+hostname);
//        String path = remoteSourceURI.getPath();
//        //System.out.println("path: "+path);
//        String scheme = remoteSourceURI.getScheme();
//        //System.out.println("scheme: "+scheme);
//        
//        if(scheme.equalsIgnoreCase("udp")){
//					receiver = new UDPReceiver();
//					useUDT = false;
//				}
//				else{
//					receiver = new UDTReceiver();
//					//System.out.println("UDT Receiver is created.");
//        }
//        
//        ip = InetAddress.getByName(hostname);//getting the ip address from URL.
//        //System.out.println("Server IP : "+ip);
//				ipStr = ip.getHostAddress();
//				//System.out.println("Server IP String : "+ipStr);
//				
//				receiver.setServerIP(ipStr);
//				receiver.setServerDataPort(Globals.serverDataPort);
//
//	      rm = new RemoteMessenger();
//				//System.out.println("Remote Messenger is created.");
//				
//        welcomeSocket = new ServerSocket();//Control channel from client to Server.
//        welcomeSocket.setSoTimeout(10 * 60 * 1000);// 10 minute timeout
//        welcomeSocket.bind(new InetSocketAddress(Globals.clientControlPort));
//
//        if(scheme.equalsIgnoreCase("udp")){
//			/* ###################################################################
//			different from UDT receiver, UDP receiver need extra communication
//			due to its own logic, to share the control socket and remote messenger
//			we have two set methods to do this
//			################################################################### */
//			((UDPReceiver) receiver).setRemoteMessenger(rm);
//			((UDPReceiver) receiver).setClientCtrlSocket(welcomeSocket);
//		}	
		
		//System.out.println("Client Welcome Socket (for control) is created.");
		//System.out.println(welcomeSocket.getInetAddress());
		//System.out.println(welcomeSocket.getLocalPort());
		
				
        
   try{
         Thread.sleep(1000);

        //System.out.println("Requesting Server Thread");
        
        int ack = 1;
        int retrys = 0;
        while( (ack=SendControlCode((byte)'A')) != 0 && (retrys < allowedRetryTimes))//get the server to create a Server thread
        {
        	retrys++;
        	System.out.println("Server thread activation was not successful, requesting new server thread again!");
        }
        
		    /*###################
		    Added by Baoqiang Yan
		    ###################*/
        if(ack!=0){
      		System.out.println("Could not activate the server thread after " + allowedRetryTimes + " trials, exit now!!!");
      		System.out.println("Reason:");
				 switch(ack){
					  case 1:
							System.out.println("Server side activation failed!");
							System.exit(1);
							break;
						case -1:
								System.out.println("Communication problem!");
								System.exit(1);
								break;
							default:
								break;
				 }
        }
										
        this.sentStencilConstructorInfo = false;  // must send basic stencil info if  we use stencils.
				
				/*######## here we send together with file name the scheme  ######*/
        //rm.PassivelySendControlInfo(welcomeSocket, "{"+scheme+"}File Name: " + path);
        //System.out.println("Telling Server Thread File Info: "+"{"+scheme+"}File Name: " + path);
        SendFileName("{"+scheme+"}File Name: " + remotePath);
        
        /*######## wait for the buffer size from server if using UDP  ######*/
        //if(!useUDT){
        if(	dataChannelProtocol == Globals.UDP){
//	        ctrlMsg = rm.ActivelyReceiveControlInfo(ipStr, Globals.serverControlPort);
//	        buffer_size = Integer.parseInt(ctrlMsg.substring(13,ctrlMsg.length()));
						buffer_size = GetBufferSize();
	      }
     }
     catch ( Exception exc)
     {
        new SDBError(SDBError.Fatal, "Exception in activate():" + exc.toString());
        System.exit(1); //newly added
     }
  }//end of if.
}//end of activate.

/** Close the server socket on the client side.
*/
public void deactivate()
{
	  super.deactivate();
	  //System.out.println("Closing the welcome socket.");   
	  receiver.freeResource();   
	  int ack = 0;

	  try
	  {
	  	 int retrys = 0;
	     while( (ack = SendControlCode((byte)'D')) !=0 && (retrys < allowedRetryTimes)) //get the server to exit the infinite while loop.
	     {
	     		retrys++;
	     		System.out.println("The deactivate request was not successfully served, sending the request again!");
	     }
	     //System.out.println("Deactivating server");
	     //System.out.println("Activated flag: " + this.isActivated);
	     this.welcomeSocket.close();

		    /*###################
		    Added by Baoqiang Yan
		    ###################*/
       if(ack!=0){
      		System.out.println("Could not deactivate the server thread after " + allowedRetryTimes + " trials, exit now!!!");
      		System.out.println("Reason:");
				 switch(ack){
					  case 1:
							System.out.println("Server side deactivation failed!");
							//we don't exit for now since the new request would also first check if it is still alive
							//System.exit(0);
							break;
						case -1:
								System.out.println("Communication problem!");
								//Did not exit here because the next activation request will also deactivate the old server Thread
								//before creating a new one, which will probably succeed.
								//System.exit(0);
								break;
							default:
								break;
				 }
       }
       
       this.networkInitialized = false;
	  }
	  catch ( IOException exc)
	  {
	     new SDBError(SDBError.Warning, exc.toString());
	     //This is already the end of the client program
	  }
}

/**Method to tell server the bounds for basic block construction.
*/
public void setBasicBlockBounds(String s) throws IOException
{
  rm.PassivelySendControlInfo(welcomeSocket, s, ipStr);
}

/** set the xml file for the remote datasource.
*/
public void setRemoteDataSource( String xmlFile, String dsName )
{
		this.remoteXmlFile = xmlFile;
		this.remoteSourceName = dsName;
}

/** Return the Datum corresponding to the given IndexSpaceID.
*/
public Datum datum( IndexSpaceID did )
{
    Datum returnVal = this.getRecordDescriptor().createDatum();
    this.datum( returnVal, did);
    return returnVal;
}

/** Get the Datum and fill its values in the passed Datum at positions
 *  specified by the corresponding IndexSpaceID coordinates. No argument
 *  validation is performed. Package Access Only.
 */
void pdatum(Datum datum, IndexSpaceID did){

  bounds = "@,3,0,0,0";//setting bounds so as not to change the client code.
  pdatumUDP(datum, did);//call to the remote datum fetch method.
}

/** Get the Datum and fill its values in the passed Datum at positions
 *  specified by the corresponding RecordSpec.
 */
public void datum(Datum datum, IndexSpaceID did, RecordSpec recSpec){
}

/** Get the Datum and fill its values in the passed Datum at positions
 *  specified by the corresponding FieldIDMapper.
 */
public void datum(Datum datum, IndexSpaceID did, FieldIDMapper mp){
}

/** Return the Datum corresponding to the given IndexSpaceID. The
    Datum will only contain attributes specified by the RecordSpec
    argument.
*/
public Datum datum(IndexSpaceID did, RecordSpec spec){
    Datum returnVal= null;
    return returnVal;
}

/** this method will extract data from datasource on remote server using UDT or UDP logic.
 */
protected  void pdatumUDP(Datum datum, IndexSpaceID did)
{
	  int dataType = datum.getStorageType();
	  //int offset;
	
	  switch(dataType)
	  {
	     case StorageType.Float:
	
	           System.out.println("Float not implemented");
	           break;
	     case StorageType.Short:
	
	           System.out.println("Short not implemented");
	           break;
	
	     case StorageType.Int:
	
	           System.out.println("Int not implemented");
	           break;
	
	     case StorageType.Double:
	
	           System.out.println("Double not implemented");
	           break;
	
	     case StorageType.Byte:
	
	           ReceiveData();
	           datum.setBytes(dstByteArray);
	            break;
	
	     default:
	
	            new InternalFaultError(SDBError.Fatal,"Unimplemented storage type in RemoteBFDS.pssublockUDP()");
	            break;
  	}
}

/** Reads a subblock defined by srcBounds into the region of the Datablock argument
    defined by dstBounds. The FieldIDMapper argument determines the mapping between
    datums in the datasource and datums in the DataBlock. Note that the DataBlock is
    assumed to already have a valid data array. the dataBlock can be basicBlock or
    compositeBlock
*/
public void subblock( DataBlock dstBlock,
                      ISBounds dstBounds, ISBounds srcBounds,
                      FieldIDMapper mp)
{
   if ( dstBlock instanceof BasicBlock)
      subBasicBlock( (BasicBlock)dstBlock,dstBounds,srcBounds,mp);
   else if ( dstBlock instanceof CompositeBlock)
      subCompositeBlock( (CompositeBlock)dstBlock, dstBounds, srcBounds,mp);
   else  // what's the next
   {
      new SDBError( SDBError.Recoverable, "RemoteBFDS.subbblock:: " +
      " Unimplemented method. Unknown DataBlock type" );
   }
}

// compositeBlock implementation
void subCompositeBlock( CompositeBlock dstBlock,
                                ISBounds dstBounds, ISBounds srcBounds,
                                FieldIDMapper mp)
{
   new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: " +
                " Unimplemented method!" );
}

void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
{
   new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: " +
                " Unimplemented method!" );
}

void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                              RecordSpec recSpec )
{
   new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: " +
                " Unimplemented method!" );
}

/** Reads a subblock defined by srcBounds into the region of the Datablock argument
    defined by dstBounds. The RecordSpec argument specifies a subset of the DataSource
    attributes to be written into the DataBlock datums. Note that the DataBlock is
    assumed to already have a valid data array. the DataBlock can be BasicBlock or
    CompositeBlock
*/
public void subblock( DataBlock dstBlock,
                      ISBounds dstBounds, ISBounds srcBounds,
                      RecordSpec recSpec)
{
   if ( dstBlock instanceof BasicBlock)
      subBasicBlock( (BasicBlock)dstBlock,dstBounds,srcBounds, recSpec);
   else // should work for compositeBlock
      subCompositeblock( (CompositeBlock) dstBlock,dstBounds,srcBounds, recSpec);
}

/** Unimplemented
 * @param dstBlock
 * @param dstBounds
 * @param srcBounds
 * @param recSpec
 */
void subBasicBlock( BasicBlock dstBlock,
                            ISBounds dstBounds, ISBounds srcBounds,
                            RecordSpec recSpec)
{
   new SDBError( SDBError.Recoverable, "GraniteRemoteClient.subBasicBlock:: " +
                " Unimplemented method!" );
}

void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds,
                              FieldIDMapper mp )
{
//   new SDBError( SDBError.Recoverable, "GraniteRemoteClient.subBasicBlock:: " +
//                " Unimplemented method!" );


   this.subblock(dstBlock, dstBounds, srcBounds);
}

/** Return true if the stencil constructor info should be sent to the server.
*  Return false otherwise. In general, a new stencil, or a stencil that has
*  had its partitioning modified should be resent. This implementation is 
*  a weak one, meant to be augmented in future.
*
*  This method assumes that if the value true is returned, then the stencil
*  will actually be sent. 
*/
private boolean shouldSendStencilConstructorInfo(DSStencil stencil){

  // For now, we don't use more than one stencil per session, so this
  // code will do the job.

  if (this.sentStencilConstructorInfo){
     return false;
  } else {  
     this.sentStencilConstructorInfo = true;
     return true;
  }
}

private void sendStencilConstructorInfo(DSStencil stencil){
     String stencilInfo = buildStencilConstructorInfo(stencil);	     
     //System.out.println("Sending Stencil Constructor Info");
     //rm.PassivelySendControlInfo(welcomeSocket, stencilInfo);
     SendStencilInfo(stencilInfo);
}

public void subblock(DSBlockStencil stencil)
{
   DSBlockStencil pStencil = (DSBlockStencil) stencil;


   
   this.store = pStencil.getPartitionListBytes(this.store);
   //System.out.println(stencil);
   int partitionListLength = pStencil.getPartitionListBytesLength();

   //creating and populating a byte array with length preceeding the partition list.
   String partitionListLengthString = "@," + partitionListLength;


   if(shouldSendStencilConstructorInfo(stencil)){

     sendStencilConstructorInfo(stencil);
   }

   SendPartitionListLengthInfo(partitionListLengthString);

   ClientToServerStencilPartitionList(store, partitionListLength);//sending the partition list to server.

   getStencilBlocks(pStencil);
}

// composite implementation
private void subCompositeblock( CompositeBlock dstBlock,
                                ISBounds dstBounds, ISBounds srcBounds,
                                RecordSpec recSpec )
{
   new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeblock:: " +
                " Unimplemented method!" );
}

/** Reads a subblock defined by srcBounds into the region of the Datablock argument
 defined by dstBounds. The BasicBlock is assumed to be created and  a valid data array is already allocated.
 NOTE: dstBlock can be  BasicBlock or CompositeBlock. data is copied from file to BasicBlock or CompositeBlock
based on its data type.
*/
public void subblock( DataBlock dstBlock,
                   ISBounds dstBounds, ISBounds srcBounds)
{
	  buildBounds(srcBounds);
	
	  if ( dstBlock instanceof BasicBlock){
	
	     psubblockUDP((BasicBlock)dstBlock,dstBounds,srcBounds ); // PJR //changed the method name to psubblockUDP.
	  }  
	  else if ( dstBlock instanceof CompositeBlock ){
	        
	        new SDBError( SDBError.Fatal, "RemoteBFDS.subblock:: " +
	                " Unimplemented method!" );
	  }
}
   /** 
    *  Override parent abstract class; this is just psubblockUDP
    */
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
   {
      this.psubblockUDP( dstBlk, dstBnds, srcBnds );
   }

    /** this is an implementation for BasicBlock based on java 1.4
        Note:  The BasicBlock is assumed to be created and  a valid data array is already allocated.

    *  @ deprecated
    */
    /*
    private void subBasicBlock(BasicBlock dstBlock,
                               ISBounds dstBounds, ISBounds srcBounds)
    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subBasicBlock:: " +
                    " Unimplemented method!" );
    }
   */

/** 
 * this method will extract data from datasource on remote server using UDT or UDP logic.
 */
protected void psubblockUDP(BasicBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds)
{
  int dataType = dstBlock.getStorageType();
  int offset;
  this.currentSubBounds = srcBounds;
  //why are we calling this method second time besides the one in subblock function?
  buildBounds(srcBounds);
  /*
  System.out.println("==========================================");
  System.out.println(bounds);
  System.out.println("==========================================");
  System.out.println("Sending bounds to server ... ");
  */
  SendBounds(bounds);
  //rm.PassivelySendControlInfo(welcomeSocket, bounds);

  switch(dataType)
  {
     case StorageType.Float:

              ReceiveData();
              this.bytesToFloats(dstBlock.getFloats(),dstByteArray,dstByteArray.length);//Loading dstBlock with float values from dstByteArray.
              break;

       case StorageType.Short:
              ReceiveData();
              this.bytesToShorts(dstBlock.getShorts(),dstByteArray,dstByteArray.length);//Loading dstBlock with float values from dstByteArray.
              break;

       case StorageType.Int:

              System.out.println("Int not implemented");
              break;

     case StorageType.Double:

              System.out.println("Double not implemented");
              break;

     case StorageType.Byte:
							//System.out.println("About to receive data ... ");
              ReceiveData();
              //System.out.println("Data received ...Done ");
              System.arraycopy(dstByteArray,0,dstBlock.getBytes(),0,dstByteArray.length);
              break;

     default:
              new InternalFaultError(SDBError.Fatal,"Unimplemented storage type in RemoteBFDS.pssublockUDP()");
              break;
   }
}

/** this method will extract data from datasource on remote server using UDT or UDP logic.
    */
protected void getStencilBlocks(DSBlockStencil stencil)
{
     int dataType = stencil.getStorageType();
     int size = stencil.size();
     int offset;

     switch(dataType)
     {
        case StorageType.Float:

                 System.out.println("Float not implemented");
                 break;

          case StorageType.Short:

//                 System.out.println("Short not implemented");
                 ReceiveData();
                 stencil.setShortData(dstByteArray);
                 break;

          case StorageType.Int:

                 System.out.println("Int not implemented");
                 break;

        case StorageType.Double:

                 System.out.println("Double not implemented");
                 break;

        case StorageType.Byte:
        
                  ReceiveData();
                  stencil.setData(dstByteArray);
                 break;

        default:
                 new InternalFaultError(SDBError.Fatal,"Unimplemented storage type in RemoteBFDS.pssublockUDP()");
                 break;
      }
}


/** this method will extract data from datasource on remote server using UDT or UDP logic.
    */
protected void getStencilBlocks(DataBlockCollection dstBlock)
{
     int dataType = dstBlock.getStorageType();
     int offset;

     switch(dataType)
     {

        case StorageType.Float:

                 System.out.println("Float not implemented");
                 break;

          case StorageType.Short:

                 System.out.println("Short not implemented");
                 break;

          case StorageType.Int:

                 System.out.println("Int not implemented");
                 break;

        case StorageType.Double:

                 System.out.println("Double not implemented");
                 break;

        case StorageType.Byte:
								 /*####	 Not used?	###*/
                 offset = 0;                     
                 for(int i=0;i<dstBlock.size();i++)
                 {
				                  //System.out.println(i+" - In getStencilBlocks(DataBlockCollection dstBlock): Receiving data from server");
										ReceiveData();
				                  //System.out.println(i+" - In getStencilBlocks(DataBlockCollection dstBlock): Done with receiving data");
                
                  	int blockLength = dstByteArray.length/dstBlock.size();//size of dstByteArray/number of blocks.
                    System.arraycopy(dstByteArray,0,dstBlock.get(i).getBytes(),0,blockLength);//copying the dataStore into "tmp" array.
                 }
                 break;

        default:
                 new InternalFaultError(SDBError.Fatal,"Unimplemented storage type in RemoteBFDS.pssublockUDP()");
                 break;
      }
}


/** Builds the bounds string to be sent to the RemoteUDPServer.
  */
public void buildBounds(ISBounds srcBounds)
{
	  lower = srcBounds.getLowerArray();//reference to the lower array of srcBounds.
	  upper = srcBounds.getUpperArray();//reference to the upper array of srcBounds.
	  	
	  int k = lower.length + upper.length;//variable to hold the sum of dimensions.
	
	  bounds = "#,"+ k + ",";
	  
	  //lower bounds
	  for(int i=0;i<lower.length;i++)
	  {
	     bounds += lower[i] + ",";
	  }
	
	  //upper bounds
	  for(int i=0;i<upper.length-1;i++)
	  {
	           bounds += upper[i] + ",";
	  }
	
	  bounds += upper[upper.length-1];
}

/** Builds the stencil space bounds & partitioning dimensions string to be sent to the RemoteBlockServer.
*/
private static String buildStencilConstructorInfo(DSStencil stencil)
{
    DSPartitionedStencil pStencil = (DSPartitionedStencil) stencil;
    
    int [] partitioningElementDimensions = pStencil.getPartitionDimensions().getCoords();
    
    ISBounds spaceBounds = pStencil.getSpaceBounds(); 
     
    //Space dimensions.
    lowerSpaceArray = spaceBounds.getLowerArray();//reference to the lower array of srcBounds.
    upperSpaceArray = spaceBounds.getUpperArray();//reference to the upper array of srcBounds.

    int dim = lowerSpaceArray.length; //space dimensionality

    bounds = "$,"+ dim + ",";//$ is the control character used for Stencils.
          
    //lower space bounds
    for(int i=0;i<dim;i++)
    {
        bounds += lowerSpaceArray[i] + ",";
    }
     
    //upper space bounds
    for(int i=0;i<dim;i++)
    {
             bounds += upperSpaceArray[i] + ",";
    }
     
    //partitioning dimensions
    for(int i=0;i<dim;i++)
    {
             bounds += partitioningElementDimensions[i] + ",";
    }
     

    return(bounds);
}

public long transmitTime()
{
    return time;
}


/** private method, convert bytes into floats.
It takes numBytes from bytes array starting from index 0, convert
them into floats, and put them into floatDst array starting at index offset
*/
public static void bytesToFloats( float[] floatDst,byte[] bytes, int numBytes )
{
    int bits;
    int i=0,j=0;
    while(i< numBytes){
		   bits=   	(  (bytes[i++] & 0xFF)  << 24 ) |
		  				 	(  (bytes[i++] & 0xFF)  << 16 ) |
		  					(  (bytes[i++] & 0xFF)  << 8  ) |
		  					(  (bytes[i++] & 0xFF));

		   floatDst[j] = Float.intBitsToFloat(bits);
		   j++;
    }
}

/** private method, convert bytes into shorts.
It takes numBytes from bytes array starting from index 0, convert
them into shorts, and put them into shortDst array starting at index offset
*/
public static void bytesToShorts( short[] shortDst, byte[] bytes, int numBytes )
{

   ShortBuffer buffer = ByteBuffer.wrap(bytes).asShortBuffer();
   
   buffer.get(shortDst, 0, numBytes / 2); // 2 is sizeof(short)

}


/** private method, convert bytes into floats.
It takes numBytes from bytes array starting from index 0, convert
them into floats, and put them into floatDst array starting at index offset
*/
static void bytesToFloatsOldVersion( float[] floatDst, int offset, byte[] bytes, int numBytes )
{
    int bits;
    int j = offset;
    int i=0;
    while(i< numBytes){

   bits=   (  (bytes[i++] & 0xFF)  << 24 ) |
  (  (bytes[i++] & 0xFF)  << 16 ) |
  (  (bytes[i++] & 0xFF)  << 8  ) |
  (  (bytes[i++] & 0xFF));

   floatDst[j] = Float.intBitsToFloat(bits);
   //System.out.println(" convert a float " + floatDst[j] );
   j++;
    }
}


    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. Note that the DataBlock is assumed to already have a valid data
        array.
        NOTE: DataBlock must be a CompositeBlock
    */
    /**** Commented by Sridhar on March 1, 2006
    private void subCompositeblock(CompositeBlock dstBlock,
                                   ISBounds dstBounds, ISBounds srcBounds)
    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: " +
                    " Unimplemented method!" );
    }
    ****/

    /** Returns a spatial subset of the DataSource as delimited by the ISBounds argument.
        The DataSource data is copied when filling the DataBlock.
        This method returns a BasicBlock.  Need more discussion on this method
    */
public DataBlock subblock(ISBounds db)
{
    BasicBlock basicBlock = new BasicBlock( db, this.recordDesc );
    this.subBasicBlock( basicBlock, basicBlock.getBounds(), db );
    return basicBlock;
}


/** For debugging only. */
private String toBinaryString(int value, int fieldWidth)
{
    int mask=(int) (Math.pow(2,fieldWidth)-1);
    value=value & mask;

    String binaryString=Integer.toBinaryString(value);
    int i,numPadChars=fieldWidth-binaryString.length();

    for(i=0;i<numPadChars;i++){
        binaryString="0"+binaryString;
    }

  	return binaryString;
}


/** request server for for subblock
*/
//private class ServerSubblock extends Thread {
//		ServerSubblock()
//		{
//		   super();
//		}
//		public void  run()
//		{
//		   try
//		  {
//		      fileServer.subblock( currentSubBounds);
//		  }
//		   catch(RemoteException e)
//		  {
//		      System.out.println(" In client thread, can't find remote object " + e.toString() );
//		  }
//		}
//}


/** File transfer client program.
*/
public void ReceiveData() //throws Exception
{
	int recvTryTimes = 2;
	while(true){
			if((fileSize = GetDataSize()) == 0) //something wrong when receiving the file size
			{
		      rm.PassivelySendControlInfo(welcomeSocket, "FAIL", ipStr);
		      //System.out.println("Failed to receive data size - exiting 928");
					System.exit(1); //this will make the rt script to request another server thread by killing the old one
				//But this will not work for the last test in which case the server will just time out slowly
			}
			
			int ackStatus = rm.PassivelySendControlInfo(welcomeSocket, "OK", ipStr);
			if(ackStatus == 1){
	      //System.out.println("Failed to tell server OK for data size - exiting 935");
				System.exit(1);
			}
		   //System.out.println("File Size Received:"+fileSize);

      dstByteArray = new byte[fileSize];
			
		//this means that the first time failed, and we don't want to duplicate the time 
		//used to send control msg in the "valid" cdstime, so, wait until here (start is 
		//the one of previous trial)
  		if(recvTryTimes==1){ 
     		end = System.currentTimeMillis();//used to calculate wasted time
			System.out.println("Waste_Time " + (end - start));
		}
		
     	start = System.currentTimeMillis();//start of transfer.
      int status = receiver.ReceiveData(dstByteArray, fileSize);
  		recvTryTimes--;
  		//recvTryTimes is used to prevent infinite loop
      if(status==1 && recvTryTimes!=0){
			   System.out.println("Receive failed");
      		rm.PassivelySendControlInfo(welcomeSocket, "FAIL", ipStr);
				System.out.println("FAIL sent to server");
      		continue;
      }else{
      		if(status==0){
						end = System.currentTimeMillis();//end of transfer.
						//System.out.println("Receive Succeeded");
						rm.PassivelySendControlInfo(welcomeSocket, "OK", ipStr);
						//System.out.println("OK sent to server");
	      		break;
				}
				else{//all efforts have been tried
					System.out.println("Client opting out due to the receiveData error and two tries!!!");
					System.exit(1); //this will make the rt script to restart the test
				}
      }
   }
//   System.out.println("Transfer_Time " + (end - start));
}//end of ReceiveData.

/** TCP Control Traffic to Server about Stencil's partition list.
*/
public static void ClientToServerStencilPartitionList(byte[] b, int length) // throws IOException
{
	//System.out.println("Sending to server the partion list in bytes");
   try {
   
      connectionSocket = welcomeSocket.accept();
      
   } catch (SocketTimeoutException e) {

      new SDBError(SDBError.Fatal, "WelcomeSocket Timed Out");
   }catch (IOException e) {

      new SDBError(SDBError.Fatal, e.toString());
   }
   
   try{
	   OutputStream outToServer = connectionSocket.getOutputStream();
	   outToServer.write(b,0,length);//send the getPartitionListBytes to the Server.
	   outToServer.flush();
	 }catch(IOException ioe){
       
      new SDBError(SDBError.Fatal, ioe.toString());
	 }
   
   try{
   	connectionSocket.close();
   }
   catch(IOException ioe){
      
      new SDBError(SDBError.Recoverable, ioe.toString());
   }
}


/** Send the given byte to the master thread on the server.
		####################	Baoqiang Yan	###################
		Changed to return the actual received ACK to allow more
		specific control depending on the status of server
		#######################################################
		
		      if(ack != 0){
      
         System.out.println("Non zero acknowledge from server in SendControlCode()");
         ###################################################################
         		We should stop the client since the non-zero ACK means the remote 
         		controller generateed some error and stopped
         ###################################################################
       }

*/
public static int SendControlCode(byte b)
{
   String ctrl_info;
   OutputStream os;
   InputStream is;
   int ack = 0;

   try{
	   clientAlarmSocket = new Socket(ip, Globals.serverAlarmPort);
	   //System.out.println("Client Alarm Socket created for "+ip+" at "+Globals.serverAlarmPort);
	   
	   os = clientAlarmSocket.getOutputStream();
	   
	   os.write(b);
	   os.flush();
   	   //System.out.println("Control Code is sent: "+ (char)b);
   }catch(IOException ioe){
   		System.out.println("Error while sending control code:"+ioe.toString());
   		//this means the control code was not sent, exit
   		ack = -1;
   }
   
   try {
   
      clientAlarmSocket.setSoTimeout(10 * 60 *1000); //the clear cache takes long time
	    is = clientAlarmSocket.getInputStream();
      ack =is.read();
   		//System.out.println("ACK Code received: "+ ack);
   } 
   catch (SocketException e) {

      new SDBError(SDBError.Fatal, "clientAlarmSocket SocketException");
      //this means control code was sent out but no ACK back
      ack = -1;
   }
   catch (IOException e) {

      new SDBError(SDBError.Fatal, "clientAlarmSocket IOException");
      //this means control code was sent out but no ACK back
      ack = -1;
   }
		
	 try{   
	   	clientAlarmSocket.close();
   		//System.out.println("Client Alarm Socket is closed. ");
   }
   catch (IOException e) {
      new SDBError(SDBError.Fatal, "clientAlarmSocket close exception");
   }

   return ack;
}

/** Clear the filesystem cache on the server. */
public void clearCache()
{
    //System.out.println("Clearing Server Cache");
    
    // Must prepare for connection to server.
    initNetwork();
    
    
    int ack = 0; 
    int retrys = 0;
    while( (ack=SendControlCode((byte)'C')) != 0 && (retrys < allowedRetryTimes)) //get the server to clear Server Cache
    {
    	System.out.println("Last clear cache request was not successfully served, requesting again!");
    	retrys++;
    }
    /*###################
    Added by Baoqiang Yan
    ###################*/
    switch(ack){
        case 1:
	        	System.out.println("There is something wrong with the server clear cache script!");
	        	System.out.println("Server cache not cleared after " + allowedRetryTimes + " trials!");
	        	System.out.println("This test continued without server cache cleared!");
	      	break;
	      case -1:
	      		//Maybe the Control Code is not sent or sent but no ACK from the remote controller
	      		System.out.println("This test continued without knowing if server cache is cleared!");
	      		//System.out.println("Client is backing out due to the communication problem!");
	      		//System.exit(0);
					break;
				default:
					break;
    }
}
		/** GraniteRemoteClient main.
		*/
		public static void main(String[]  args)
		{
		
		   GraniteRemoteClient dsrc =(GraniteRemoteClient) DataSource.createRemoteClient(args[0],args[0]+".xfdl",args[1]);
		   
		   lower = new int[]{Integer.parseInt(args[2]),Integer.parseInt(args[3]),Integer.parseInt(args[4])};
		   upper = new int[]{Integer.parseInt(args[5]),Integer.parseInt(args[6]),Integer.parseInt(args[7])};
		
		   dsrc.activate();//sets up the clienttoserver control socket.
			 /*
			 System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!");
			 System.out.println("!!!Server is activated!!!");
			 System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!");
			 */
		   RecordDescriptor rec = dsrc.getRecordDescriptor();
		
		   //constructing bounds string to pass on to server.
		   int k = lower.length + upper.length;//sum of dimensions.
		   bounds = "#,"+ k + ",";
		   for(int i=0;i<lower.length;i++)
		   {
		      bounds += lower[i] + ",";
		   }
		   for(int i=0;i<upper.length-1;i++)
		   {
		      bounds += upper[i] + ",";
		   }
		   bounds += upper[upper.length-1];
			
		   ISBounds tbounds = new ISBounds(lower,upper);
		   BasicBlock theBlock = new BasicBlock( tbounds , rec );
		
		   try
		   {  dsrc.psubblockUDP(theBlock, tbounds, tbounds );//calling psubblock method to make network call.
		   }
		   catch ( Exception exception )
		   {
		      exception.printStackTrace();
		    }
		
		   //System.out.println( theBlock);
			 dsrc.deactivate();
		   /*****
		   int[] lower1 = { 2,2,2};
		   int[] upper1 = {4,4,4};
		   ISBounds tbounds1 = new ISBounds( new IntegerIndexSpaceID(lower1), new IntegerIndexSpaceID(upper1) );
		
		   BasicBlock theBlock1 = new BasicBlock( tbounds1 , rec );
		   try
		   {
		      dsrc.psubblockUDP(theBlock1, tbounds, tbounds1 );//calling psubblock method to make network call.
		   }
		   catch ( Exception exception )
		   {
		      exception.printStackTrace();
		    }
		   System.out.println( theBlock1);
		   ****/
		
		   /*
		   System.exit(0);
		   */
		}
}
 