/*
    IndexBlock.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
*/

/**
 * The IndexBlock class is developed to store the index data info of 
 * an ARRS leaf node. 
 */


package edu.unh.sdb.datasource;

//import java.io.*;

import edu.unh.sdb.common.*;

public class IndexBlock 
{
   private float[] indexes = null;
   private int low;
   private int high;
  
   /** Constructor takes three arguments */ 
   public IndexBlock( float[] indexes, int low, int high )
   {
      this.indexes = indexes;
      this.low = low;
      this.high = high;
   }

   /** Get the index block */
   public float[] getIndexes()
   {
      return this.indexes;
   }

   /** Get the low index */
   public int getLowIndex()
   {
      return this.low;
   }

   /** Get the high index */
   public int getHighIndex()
   {
      return this.high;
   }

   /** Set low index */
   public void setLowIndex( int low ) 
   {
      this.low = low;
   }

   /** Set low index */
   public void setHighIndex( int high )
   {
      this.high = high;
   }

   /** Get the data corresponding to the index block */
   public void getCellData(Indexable theData, Indexable pds)
   {
      int j; 
      Datum data;
      ISIterator iter = new ISIterator( pds.copyBounds() );
       //System.out.println("low: "+this.low); //-------
       //System.out.println("high: "+this.high); //-------
      iter.init();
      for( int i = this.low; i <= this.high; i++ ) 
      {  //System.out.println("index value: "+this.indexes[i]); //------
         j = (int)(this.indexes[ i ]);
         data = theData.datum( new IntegerIndexSpaceID( 0, j ));
         //System.out.println("data: "+data.toString()); //-------
         ((DataBlock)pds).setDatum( iter, data );
         if ( iter.valid() )
            iter.next();
      }   
   }

   /** Get the data source corresponding to the index block */
  /* public void getDataSource(float[] ds, DataSource index_ds, 
                             ISBounds dstBounds)
   {
      int i;
      FloatDatum data;
      ISIterator iter = new ISIterator(index_ds.getBounds());
      for(i = this.low, iter.init(); i <= this.high; i++) {
         data = new FloatDatum(ds[this.indexes[i]]);
         ((BinaryFileDataSource)index_ds).setDatum(iter, data);
         if (iter.valid())
            iter.next();
      }   
   }
*/
   /** Return Datums corresponding to the index block */
   public Datum[] getDatums(Indexable theData)
   {
      Datum[] data = new Datum[this.high-this.low+1];
      for(int i = this.low; i <= this.high; i++) {
         data[i] = theData.datum(new IntegerIndexSpaceID(0, i));
      } 
      return data;
   }

   /** Get Datums corresponding to the index block */
   public void getDatums(Indexable theData, Datum[] data)
   {
      data = this.getDatums(theData); 
   }
}
    
