package edu.unh.sdb.datasource;

import edu.unh.sdb.common.SDBIterator;

/**
 *
 * @author dlipsa
 * This interface contains methods needed by the CacheMaker to create cache
 * for ObliquePlaneIterator
 */
public interface ObliquePlaneIteratorInterface extends SDBIterator
{
   /** 
    * Returns the dimensions of the projection of the iteration 
    * parallelogram to the orthogonal plane that contains the maximum 
    * projection area
    */
   public int [] getPlaneShapeArray();
   
   /** Return a copy of the normals array.*/   
   public float [] getNormalArray();   
   
   
   /**
    * Returns an array of ints representing the dimensions of the space being
    * iterated over. This space is calculated as the bounding box of the
    * path of the plane through space, so not all elements in this space are
    * necessarily visited by the iterator. 
    */   
   public int[] getSpaceShapeArray();
   
   
   /**
    * Returns an ISBounds representing the dimensions of the space being
    * iterated over. This space is calculated as the bounding box of the
    * path of the plane through space, so not all elements in this space are
    * necessarily visited by the iterator. 
    */
   public ISBounds getSpaceBounds();   
   
   public DataSource getDataSource ();
   public void setDataSource (DataSource dataSource);
   
   
   /** Indicate to this iterator whether the underlying cache has blocks shaped according to 
    *  the plane normal.
    */
   void setSPCacheBlocks(boolean spBlockShape);
}
