//
//  OneDDataSink.java
//  Unstructured
//
//  Created by Philip Rhodes on Thu Oct 23 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//

package edu.unh.sdb.datasource;

import java.io.*;

import edu.unh.sdb.common.*;

/** A 1D Datasink. If this class is used, then an IndexSpaceID cannot
    be passed to the write() method. Datums are simply written one 
    after the other.
*/

public class OneDDataSink extends DataSink {


	protected int currentOffset;


	public OneDDataSink(String filename, RecordDescriptor rd){
	
        super(filename, rd);
        
		this.filename=new String(filename);	
		this.recordDescriptor=rd;
		this.numFields=rd.size();
		this.bounds=null;
		this.dimensionality=1;
		
		this.ordering=null;
		this.axisWeights=new long[]{1};
		this.datumSize=this.recordDescriptor.byteSize();
		
	}

	public void activate(){
	
        super.activate();
        
		this.currentOffset=0;
	}


	public void write(Datum d){
	
		float value;
		int storageType;
		int i, n=this.numFields;
		
		try {
		
			this.raFile.seek(this.currentOffset);
			
			for(i=0; i<n; i++){
			
				storageType=recordDescriptor.getAttribute(i).getStorageType();
				
				value=d.getFloat(i);
				
				switch(storageType){
				
					case StorageType.Float:
						this.raFile.writeFloat(value);
						this.currentOffset+=4;
						break;
					case StorageType.Double:
						this.raFile.writeDouble(value);
						this.currentOffset+=8;
						break;
					case StorageType.Int:
						this.raFile.writeInt((int)value);
						this.currentOffset+=4;
						break;
					case StorageType.Short:
						this.raFile.writeShort((short)value);
						this.currentOffset+=2;
						break;
					default:
						new InternalFaultError(SDBError.Fatal,
                    "Unsupported storage type code in OneDDataSink.write(Datum):"+storageType);
						break;
				}
			}
			
			
		} catch (IOException e){
			new FileError(SDBError.Recoverable, "IO Error in write().");
		}
	}
    
	public void write(Datum d, IndexSpaceID id){
        
		
        
        super.write(d,id);
		
		this.currentOffset=(this.computeOffset(id)+1)*this.datumSize;
    }

    /** For OneDDataSinks, this method returns a bounds constructed from
        the current position of the file pointer. The file pointer may not
        be at the end of the file, so this value is not always meaningful. */
	public ISBounds getBounds(){
	
		return new ISBounds(new IntegerIndexSpaceID(this.currentOffset));
	}


}
