/** RayCellIterator
 * 
 * ------------------- Edit history (as  of 01/09/05)
 * 10/21/05 rdb: removed unused variables  
 * 01/09/05 rdb: -changed unimplemented data access methods to passthroughs to 
 *                the contained DataBlock (this.currentBlock)
 *               
 */
package edu.unh.sdb.datasource;
//import java.io.*;
import edu.unh.sdb.common.*;

/** 
This class simulates a ray passing througth a cubic index space. 
Assume that the distances between adjacent indexes are 1 ( unite length). 
this cell iterator is also a small cubic datablock containing 8 index 

    8           7
    -------------
   -           _-
  -           - -
 ------------   -
 - 4      3 -   - 6 
 -          - -  
 - 1       2--  
 ------------

This iterator maintain two pieces of information

1. the current dataBlock
2. the ray intersection position within the block

  -------------------- Edit history as of 12/05/04 ---------------------------
  12/05/04 rdb: deleted size(). inherits from DataCollection; it was a dummy anyway.
*/ 

public class RayCellIterator extends DataBlock
{
    //starting postion of the ray
    private float ray_x, tempX;
    private float ray_y, tempY;
    private float ray_z, tempZ;

    // ray intersection location within the cell
    private float localX;
    private float localY;
    private float localZ;

    private IntegerIndexSpaceID baseIndexID, tempBaseID,upperID, unitID;

    private BasicBlock currentBlock;

    private ISBounds  iteratorBounds, currentBounds;
    private ISBounds  sourceBounds;

    //private DataBlock sourceBlock; 

    private Indexable source;

    private float dx,dy,dz;

    private float step;
    
    private int[] unit = {1,1,1};

    /** construct the interator by giving voxel,starting point,ending point,concreteBlock
     */
    public RayCellIterator( float startx, float starty, float startz, 
                             float dx,     float dy,     float dz,    float step,  
                             Indexable source )
    {
      this.source    = source;

   
      this.ray_x  = startx;
      this.ray_y  = starty;
      this.ray_z  = startz;
      this.dx     = dx;
      this.dy     = dy;
      this.dz     = dz;

      this.step         = step;
      this.localX      = 0;
      this.localY      = 0;
      this.localZ      = 0;
      tempBaseID = new IntegerIndexSpaceID(0,0,0); // use to tempray holding the isid
    }

    /** initialize the itertor
        move to the first cell block which intersects with the ray, set default axisordering as
        {2,1,0}
     */
    public  void init()
    {
   	  int ordarr3D[] = {2,1,0};
      AxisOrdering order3D = new AxisOrdering(ordarr3D);
      baseIndexID = new IntegerIndexSpaceID( (int)ray_x, (int)ray_y, (int)ray_z );
      unitID  = new IntegerIndexSpaceID(unit);
      upperID = (IntegerIndexSpaceID)baseIndexID.copyID();
      upperID.add(unitID);
      sourceBounds = source.copyBounds();
      this.iteratorBounds = new ISBounds(tempBaseID, unitID);
       
      this.currentBounds = new ISBounds(baseIndexID, upperID);
      this.currentBlock = new BasicBlock(this.currentBounds, source.getRecordDescriptor(), order3D);
      source.subblock( currentBlock, iteratorBounds , currentBounds);   
    }

    /** initialize the itertor
        move to the first cell block which intersects with the ray
     */
    public  void init(AxisOrdering order)
    {
      baseIndexID = new IntegerIndexSpaceID( (int)ray_x, (int)ray_y, (int)ray_z );
      unitID  = new IntegerIndexSpaceID(unit);

      upperID = (IntegerIndexSpaceID)baseIndexID.copyID();
      upperID.add(unitID);
      sourceBounds = source.copyBounds();

      this.iteratorBounds = new ISBounds(tempBaseID, unitID);
      //System.out.println("iterator bounds is " + iteratorBounds.toString() );
       
      this.currentBounds = new ISBounds(baseIndexID, upperID);

      this.currentBlock = new BasicBlock(this.currentBounds, source.getRecordDescriptor(), order);
      source.subblock( currentBlock, iteratorBounds , currentBounds);   
    }


    /**Return true if this ISIterator can be incremented.
     *  move to next ray sample location
     */
    public boolean hasMoreElements()
    {
          tempX = this.ray_x + dx * step;
          tempY = this.ray_y + dy * step;
          tempZ = this.ray_z + dz * step;

          tempBaseID.setCoord(0, (int)tempX);
          tempBaseID.setCoord(1, (int)tempY);
          tempBaseID.setCoord(2, (int)tempZ);
          //tempBaseID.add(unitID);

          if ( sourceBounds.containsExcludeUpper( tempBaseID) )
             return true;
          else
             return false;
     }


   /** move the iterator to the next possiton, and get the data from the source 
    */
   public void next()
   {
       //if( currentBounds.equals( new ISBounds( new IntegerIndexSpaceID(0,0,0), new IntegerIndexSpaceID(1,1,1) )
       //System.out.println(" RayCellIterator next function  Boundls is : "+ currentBounds);
       if( this.nextBounds() )
           source.subblock( currentBlock, iteratorBounds , currentBounds);   
   }


   // local_x is the local position within the iteration block
   /** return true if iterator moves to next spacial postion. return false if it stays
       where it is
    */
   public boolean  nextBounds()
   {
       this.ray_x = this.ray_x + dx * step;
       this.ray_y = this.ray_y + dy * step;
       this.ray_z = this.ray_z + dz * step;

       this.localX = this.ray_x - (int)this.ray_x;
       this.localY = this.ray_y - (int)this.ray_y;
       this.localZ = this.ray_z - (int)this.ray_z; 

       if( ((int)ray_x)== baseIndexID.getCoord(0) &&
           ((int)ray_y)== baseIndexID.getCoord(1) &&
           ((int)ray_z)== baseIndexID.getCoord(2)    ) 
            
            return false; // block does not move, stays where it is
       
        baseIndexID.setCoord(0, (int) this.ray_x);
        baseIndexID.setCoord(1, (int) this.ray_y);
        baseIndexID.setCoord(2, (int) this.ray_z);
        
        //System.out.println(" RayCellIterator nextBounds, next ray location " + this.ray_x + " " + this.ray_y + " " + this.ray_z ); 
        this.currentBounds.moveTo(baseIndexID);  
        return true;
   }


   /** configure the cache system for iteration
       if the indexable is a datablock, there is nothing to worry about. Dont need cache in this case
       if the indexable is a datasouce, the iterator has to tell the datasource how to cache.
       return true is the cache is configured successfully, otherwise false;
   */

/*   public boolean configureCache()
   {
      if(source instanceof DataBlock)
        return false;
      else if( source instanceof DataSource)
       {
        DataSource.configureCache(
        return true;
       }
      else;
      return true; 
  }
*/
   /** get an array of ISBounds within the iteration enumeration window
       Enumeration window is the number of iteration steps that can be predicted in advance
       it's possible that two element in the array have the same ISBounds
   */
   public ISBounds[] getEnumerWindow( int windowSize )
   {
      float x, y, z;
      x = this.ray_x;
      y = this.ray_y;
      z = this.ray_z;
       
      ISBounds[] window = new ISBounds[windowSize];
      IntegerIndexSpaceID[] bid = new IntegerIndexSpaceID[windowSize];
      for( int i =0; i< windowSize; i++)
      {
          x = x + dx * step;
          y = y + dy * step;
          z = z + dz * step;
          bid[i] = new IntegerIndexSpaceID( (int)x, (int)y, (int)z);          
          window[i] = new ISBounds(3);
          window[i].moveTo(bid[i]);
      }
      return window;
    }

   public float[] getLocalIntersection()
   {
       float[] intersection = new float[3];
       intersection[0] =this.localX;
       intersection[1] =this.localY;
       intersection[2] =this.localZ;
       return intersection;
    }

/** have the implement some methods to make compiler happy
   ---------------------------------------------------------------------------
*/
    
    /** Given an IndexSpaceID, return the corresponding datum.*/
    public  Datum datum(IndexSpaceID did)
    { 
       return this.currentBlock.datum( did);
    }
    
    /** Given an array of coordinates, copy the corresponding datum
        into the first parameter.*/
    public  void datum(Datum d,int[] coordArray)
    {
       this.currentBlock.datum( d, coordArray );
    }
    
    public void datum(Datum datum,IndexSpaceID id, RecordSpec spec)
    {
       this.currentBlock.datum( datum, id, spec );
    }
    public void datum(Datum datum,IndexSpaceID id, FieldIDMapper mp)
    {
       this.currentBlock.datum( datum, id, mp );
    }
    public void datum(Datum datum,IndexSpaceID id)
    {
       this.currentBlock.datum( datum, id );
    }

    /** Given an IndexSpaceID, return the corresponding datum with attributes
        specified in the RecordSpec. */
    public  Datum datum(IndexSpaceID did, RecordSpec spec)
    {
       return this.currentBlock.datum( did, spec );
    }
    
    /** Set the location specified by the IndexSpaceID to the given datum.*/
    public void setDatum(IndexSpaceID isid, Datum d)
    {
       this.currentBlock.setDatum( isid, d );
    }
    
    
    /** Return a new DataBlock that is a subset of this datablock. The Data
        is shared between the two DataBlocks.*/
    public  DataBlock subblock(ISBounds db)
    {
       return this.currentBlock.subblock( db );
    }
 
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds)
    {
       this.currentBlock.subblock( theBlock, dstBounds, srcBounds );
    }


    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The FieldIDMapper argument determines the mapping between 
        datums in this DataBlock and datums in the argument DataBlock. */    
    public void subblock( DataBlock theBlock, ISBounds dstBounds, 
                          ISBounds srcBounds, FieldIDMapper mp)
    {
       this.currentBlock.subblock( theBlock, dstBounds, srcBounds, mp );
    }
 
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The RecordSpec argument specifies a subset of the attributes
        of this DataSource to be written into the DataBlock datums.*/
    public void subblock( DataBlock theBlock, ISBounds dstBounds, 
                          ISBounds srcBounds, RecordSpec spec)
    {
       this.currentBlock.subblock( theBlock, dstBounds, srcBounds, spec );
    }
 
    /** Returns the Values of the current DataBlock.
     * 
     */
    public Values getValues() 
    {
       return this.currentBlock.getValues();
    }

//----------------------------------------------------------------------------------------------------------
   /**  use trilinear interpoloation to get the value at a given space location within
    *   the the given voxel */ 
/*   public float getRaySample(int attribute_id) 
    {
          return  this.interpolateValue( localX, localY, localZ, attribute_id );
    }
*/


   /** return interpolated values of all the attribute, stored in a array*/
  public float[] interpolateValues()
    {
          Datum d000 = currentBlock.datum(new IntegerIndexSpaceID(0,0,0) );
          Datum d100 = currentBlock.datum(new IntegerIndexSpaceID(1,0,0) );
          Datum d110 = currentBlock.datum(new IntegerIndexSpaceID(1,1,0) );
          Datum d010 = currentBlock.datum(new IntegerIndexSpaceID(0,1,0) );
          Datum d001 = currentBlock.datum(new IntegerIndexSpaceID(0,0,1) );
          Datum d101 = currentBlock.datum(new IntegerIndexSpaceID(1,0,1) );
          Datum d111 = currentBlock.datum(new IntegerIndexSpaceID(1,1,1) );
          Datum d011 = currentBlock.datum(new IntegerIndexSpaceID(0,1,1) );

          int numberOfAttributes = this.currentBlock.getNumAttributes();
          float[] samples = new float[numberOfAttributes];
          for(int i =0;i<numberOfAttributes; i++)
           {
              samples[i] = triliner(  this.localX, this.localY, this.localZ,
                                      d000.getFloat(i), d100.getFloat(i), d110.getFloat(i),
                                      d010.getFloat(i), d001.getFloat(i), d101.getFloat(i),
                                      d111.getFloat(i), d011.getFloat(i)
                                    );
            }
         return samples;
    }

  /** trilinear interpolation. Store 8 values in a array, first 4 values---front face start
   *     at 000,100,110,010, the last 4 values map to 001,101,111,011)
   *     fx,fy,fz : the fraction position between the two points
   *     
   *     vx00 = v000 + fx * (v100-v000)
   *     vx01 = v001 + fx * (v101-v001)
   *     vx10 = v010 + fx * (v110-v010)
   *     vx11 = v011 + fx * (v111-v011)
   *
   *     vxy0 = vx00 + fx * (vx10 - vx00)
   *     vxy1 = vx01 + fx * (vx11 - vx01)
   *
   *     vxyz = vxy0 + fz * (vxy1 - vxy0)
   */
   public float triliner(float fx, float fy, float fz, 
                           float v000, float v100, float v110, float v010,
                           float v001,float v101, float v111, float v011  
                         )
   {
        float vx00 = v000 + fx * (v100-v000);
        float vx01 = v001 + fx * (v101-v001);
        float vx10 = v010 + fx * (v110-v010);
        float vx11 = v011 + fx * (v111-v011);
   
        float vxy0 = vx00 + fy * (vx10 - vx00);
        float vxy1 = vx01 + fy * (vx11 - vx01);
        float vxyz = vxy0 + fz * (vxy1 - vxy0);
        return vxyz;
   }

}
