/*******************************************************************************
 * RemoteBFDS.java Copyright 2000 University of New Hampshire Durham, NH 03824
 *  
 ******************************************************************************/

// ------------------ Edit history (as of 01/10/05) --------------------
// 12/04/05 rdb: Revised to inherit public subblock methods from parent.
// 10/21/05 rdb: removed unused variables
// 05/08/05 rdb: reformatted source
// 01/19/05 rdb: replaced getFloatAttributes() with getFloats()
//                    and getByteAttributes() with getBytes()
//          rdb: still lots of type limitations
// 01/10/05 rdb: removed direct allocation of children of Datum
//
// ---------------------------------------------------------------------
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

import java.io.*;
import java.rmi.*;
import java.net.*;

/**
 * This class will read data from a remote datasource
 */

public class RemoteBFDS extends PhysicalDataSource
{
   private Server fileServer = null;
   private String fileServerURL = null;

   private Socket dataSocket;
   private DataInputStream input;

   private String remoteXmlFile, remoteSourceName;

   private ISBounds currentSubBounds = null;

   private byte[] tmpBuffer;
   private int tmpBufferSize;

   //for performance tests
   private long t1 = 0;
   private long t2 = 0;
   private long time = 0;

   /**
    * crate a RemoteBFDS, set the initial file offset at 0
    */
   public RemoteBFDS( String name, RecordDescriptor recordDesc,
                                 ISBounds theBounds, AxisOrdering ordering,
                                 String URL, String serverURL )
   {
      super( name, recordDesc, theBounds, ordering, URL ); // call inherited
                                                           // constructor
      this.fileServerURL = serverURL;

      // this buffer is used to read data from the socket
      this.tmpBufferSize = 2048;
      this.tmpBuffer = new byte[tmpBufferSize];
   }

   /**
    * Set remote file server specified by the argument string. The string should
    * be format as "//hostname/server_object_name" For example, a file server,
    * named SAMPLE-SERVER is running on zheng.cs.unh.edu The argument string
    * should look like "//zheng.cs.unh.edu/SAMPLE-SERVER"
    */
   public void setFileServer( String serverURL )
   {
      this.fileServerURL = serverURL;
   }

   public void init()
   {
      System.setSecurityManager( new RMISecurityManager() );
      System.out.println( "Security Manager loaded" );

      //get the remote object from the registry
      try
      {
         this.fileServer = (Server) Naming.lookup( this.fileServerURL );
         System.out.println( "Got remote object, test remote object " );
         System.out.println( " 1 + 2 = " + this.fileServer.sum( 1, 2 ) );
         System.out.println( " testing done, remote file server is running" );
      } catch ( RemoteException exc )
      {
         System.out.println( "Error in lookup: " + exc.toString() );
      } catch ( java.net.MalformedURLException exc )
      {
         System.out.println( "Malformed URL: " + exc.toString() );
      } catch ( java.rmi.NotBoundException exc )
      {
         System.out.println( "NotBound: " + exc.toString() );
      }

   }

   /** Open the binary data file for this datasource. */
   public void activate()
   {
      super.activate();
      try
      {
         this.fileServer.setDataSource( remoteXmlFile, remoteSourceName );

         // setup client socket
         dataSocket = new Socket( "zheng.cs.unh.edu", 1700 );

         input = new DataInputStream( dataSocket.getInputStream() );
      } catch ( RemoteException exc )
      {
         System.out.println( "Error in lookup: " + exc.toString() );
      } catch ( IOException exc )
      {
         System.out.println( "socket IO exception" + exc.toString() );
      } catch ( Exception exc )
      {
         System.out.println( exc.toString() );
      }
   }

   /** set the xml file for the remote datasource */
   public void setRemoteDataSource( String xmlFile, String dsName )
   {
      this.remoteXmlFile = xmlFile;
      this.remoteSourceName = dsName;
   }

   /** Return the Datum corresponding to the given IndexSpaceID. */
   public Datum datum( IndexSpaceID did )
   {
      Datum returnVal = this.getRecordDescriptor().createDatum();

      this.datum( returnVal, did );
      return returnVal;
   }

   public void datum( Datum datum, IndexSpaceID did )
   {
   }

   /**
    * Get the Datum and fill its values in the passed Datum at positions
    * specified by the corresponding RecordSpec.
    */
   public void datum( Datum datum, IndexSpaceID did, RecordSpec recSpec )
   {
   }

   /**
    * Get the Datum and fill its values in the passed Datum at positions
    * specified by the corresponding FieldIDMapper.
    */
   public void datum( Datum datum, IndexSpaceID did, FieldIDMapper mp )
   {
   }

   /**
    * Return the Datum corresponding to the given IndexSpaceID. The Datum will
    * only contain attributes specified by the RecordSpec argument.
    */
   public Datum datum( IndexSpaceID did, RecordSpec spec )
   {
      Datum returnVal = null;
      return returnVal;
   }

   // basicBlock implementation java 1.4
   // dstBlock is considered as composite, the calling binaryFileDataSource is
   // considered as component
   void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds,
                                 ISBounds srcBounds, FieldIDMapper mp )
   {
      new SDBError( SDBError.Recoverable, "RemoteBFDS.subBasicBlock:: "
                                    + " Unimplemented method!" );
   }

   // compositeBlock implementation
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
                                 ISBounds srcBounds, FieldIDMapper mp )
   {
      new SDBError( SDBError.Recoverable, "RemoteBFDS.subBasicBlock:: "
                                    + " Unimplemented method!" );
   }

   /**
    * Unimplemented
    * 
    * @param dstBlock
    * @param dstBounds
    * @param srcBounds
    * @param recSpec
    */
   void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds,
                                 ISBounds srcBounds, RecordSpec recSpec )
   {
      new SDBError( SDBError.Recoverable, "RemoteBFDS.subBasicBlock:: "
                                    + " Unimplemented method!" );
   }

   // composite implementation
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
                                 ISBounds srcBounds, RecordSpec recSpec )
   {
      new SDBError( SDBError.Recoverable, "RemoteBFDS.subBasicBlock:: "
                                    + " Unimplemented method!" );
   }

   /**
    * Reads a subblock defined by srcBounds into the region of the Datablock
    * argument defined by dstBounds. The BasicBlock is assumed to be created and
    * a valid data array is already allocated. NOTE: dstBlock can be BasicBlock
    * or CompositeBlock. data is copied from file to BasicBlock or
    * CompositeBlock based on its data type.
    *
   void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds )
   {
       psubBasicBlock( (BasicBlock) dstBlock, dstBounds, srcBounds ); // PJR
   }/********/

   /**
    * this method will invoke remote datasource server extract data from the
    * disk, and push it into the java socket
    */
   protected void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds,
                                 ISBounds srcBounds )
   {
      int dataType = dstBlock.getStorageType();

      this.currentSubBounds = srcBounds;
      // requests the server push data into the socket
      ServerSubblock serverrequest = new ServerSubblock();
      serverrequest.start();

      //System.out.println("Client requested server for action");

      switch ( dataType )
      {
         case StorageType.Float:
            //byte[] tmp = new byte[srcBounds.volume()
                                   //* dstBlock.getRecordDescriptor().byteSize()];

            //System.out.println(" number of bytes to read :" + tmp.length );
            //System.out.println(" available bytes in the stream :" +
            // this.input.available() );
            //System.out.println(" ### Number of byte read :" +
            // this.input.read(tmp) );
            t1 = System.currentTimeMillis();
            this.readFloats( dstBlock.getFloats() );
            t2 = System.currentTimeMillis();
            time += ( t2 - t1 );
            //System.out.println(" &&&&&&&&&&&&& RemoteBFDS finishes reading
            // floats from a socket");
            break;
         case StorageType.Short:
            System.out.println( "Short not implemented" );
            break;
         case StorageType.Int:
            System.out.println( "Int not implemented" );
            break;
         case StorageType.Double:
            System.out.println( "Double not implemented" );
            break;
         case StorageType.Byte:
            byte[] dstByteArray = dstBlock.getBytes();
            t1 = System.currentTimeMillis();
            this.readBytes( dstByteArray, dstByteArray.length );
            t2 = System.currentTimeMillis();
            time += ( t2 - t1 );
            System.out.println( " &&&&&&&&&&&&&   RemoteBFDS finishes reading " +
            		                                         "data from a socket" );
            break;

         default:
            new InternalFaultError( SDBError.Fatal,
                          "Unimplemented storage type in RemoteBFDS.pssublock()" );
            break;
      }
   }

   /**
    * read an array of bytes from the socket. It makes sure the number of bytes
    * being read == the length. The function will block until length number of
    * bytes being read
    */
   private void readBytes( byte[] data, int length )
   {
      //int length = data.length;
      int offset = 0;
      int count = 0;
      try
      {
         while ( offset < length )
         {
            offset += this.input.read( data, offset, length - offset );
            count++ ;
         }
      } catch ( IOException e )
      {
         System.out.println( " IO exeception " + e.toString() );
      }
      //System.out.println("number of reads for a block" + count );
      //return offset;
   }

   /**
    * read an array of floats from the socket. It reads bytes from the socket
    * buffer and put it into a small temp byte buffer with its size as a
    * multiple of 4. Then we convert the bytes into floats and copy them into
    * dst float array. This function will block until the dstArray is full
    */
   private void readFloats( float[] dstArray )
   {
      //int size = dstArray.length;
      int byteSize = dstArray.length * 4;
      int numLoop = byteSize / tmpBufferSize;
      int leftover = byteSize % tmpBufferSize;
      int floatOffset = 0;

      for ( int i = 0; i < numLoop; i++ )
      {
         this.readBytes( tmpBuffer, tmpBufferSize );
         //convert to floats
         RemoteBFDS.bytesToFloats( dstArray, floatOffset, tmpBuffer, tmpBufferSize );
         floatOffset += ( tmpBufferSize / 4 );
      }
      //take care of the leftover

      this.readBytes( tmpBuffer, leftover );
      RemoteBFDS.bytesToFloats( dstArray, floatOffset, tmpBuffer, leftover );
   }

   public long transmitTime()
   {
      return time;
   }

   /**
    * private method, convert bytes into floats. It takes numBytes from bytes
    * array starting from index 0, convert them into floats, and put them into
    * floatDst array starting at index offset
    */
   static void bytesToFloats( float[] floatDst, int offset, byte[] bytes,
                                 int numBytes )
   {
      int bits;
      int j = offset;
      int i = 0;
      while ( i < numBytes )
      {

         bits =  ( ( bytes[i++ ] & 0xFF ) << 24 )
               | ( ( bytes[i++ ] & 0xFF ) << 16 )
               | ( ( bytes[i++ ] & 0xFF ) << 8 )
               | ( ( bytes[i++ ] & 0xFF ) );

         floatDst[j] = Float.intBitsToFloat( bits );
         //System.out.println(" convert a float " + floatDst[j] );
         j++ ;
      }
   }

   /**
    * Reads a subblock defined by srcBounds into the region of the Datablock
    * argument defined by dstBounds. Note that the DataBlock is assumed to
    * already have a valid data array. NOTE: DataBlock must be a CompositeBlock
    */
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
                                 ISBounds srcBounds )
   {
      new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: "
                                    + " Unimplemented method!" );
   }

   /**
    * request server for for subblock
    */
   private class ServerSubblock extends Thread
   {
      ServerSubblock()
      {
         super();
      }

      public void run()
      {
         try
         {
            fileServer.subblock( currentSubBounds );
         } catch ( RemoteException e )
         {
            System.out.println( " In client thread, can't find remote object "
                                          + e.toString() );
         }
      }
   }

   public static void main( String[] args )
   {
      RemoteBFDS ds = (RemoteBFDS) DataSource.createRemote( "foo",
                                    "/home/csg/xtang/file/simplefile.bin.xml",
                                    "//zheng.cs.unh.edu/SAMPLE-SERVER" );
      ds.init();
      ds.setRemoteDataSource( "/home/csg/xtang/file/simplefile.bin.xml", "foo" );
      ds.activate();

      RecordDescriptor rec = ds.getRecordDescriptor();
      int[] lower = { 0, 0, 0 };
      int[] upper = { 2, 2, 2 };
      ISBounds tbounds = new ISBounds( new IntegerIndexSpaceID( lower ),
                                    new IntegerIndexSpaceID( upper ) );

      BasicBlock theBlock = new BasicBlock( tbounds, rec );
      ds.subBasicBlock( theBlock, tbounds, tbounds );
      System.out.println( theBlock );

      int[] lower1 = { 2, 2, 2 };
      int[] upper1 = { 4, 4, 4 };
      ISBounds tbounds1 = new ISBounds( new IntegerIndexSpaceID( lower1 ),
                                    new IntegerIndexSpaceID( upper1 ) );

      BasicBlock theBlock1 = new BasicBlock( tbounds1, rec );
      ds.subBasicBlock( theBlock1, tbounds, tbounds1 );
      System.out.println( theBlock1 );
   }
}
