/************************************************************
   RemoteUDPClient.java
   Copyright 2005 University of New Hampshire
   Durham, NH 03824

************************************************************/

// ------------------ Edit history (as of 01/10/05) --------------------
// 02/12/12 rdb: Added impl of subBasicBlock( ISB, ISB, ISB ) (which
//               just call psubblockUDP), because this is now an abstract
//               class in parent.
// 06/06/08 rdb: comment unused variables
//               made references to static variables/methods "correct"; only
//                   used class name.
// 01/19/05 rdb: replaced getFloatAttributes() with getFloats()
//                    and getByteAttributes() with getBytes()
//          rdb: still lots of type limitations
// 01/10/05 rdb: removed direct allocation of children of Datum
// 03/20/05 srk: added the remote functionality using UDP based
//               	  file transfer support.
// 05/05/05 srk: added a variable packetSize to hold the packet size.
//
// ---------------------------------------------------------------------

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

import java.io.*;
//import java.nio.*;

import java.rmi.*;
import java.net.*;
import java.util.*;



/** This class will read data  from a remote datasource using UFTP logic.
*/
public class RemoteUDPClient extends PhysicalDataSource{


    private Server fileServer =null;
    private String fileServerURL = null;
    //private byte[] bArray;

    //private ByteBuffer datumBuffer;
    //private FloatBuffer fdatumBuffer;
    //private Socket      dataSocket;
    private DataInputStream input;

    private String remoteXmlFile;
    private String remoteSourceName;

    private ISBounds currentSubBounds = null;

    private byte[] tmpBuffer;
    private int tmpBufferSize;

    //for performance tests
    //private long t1 =0;
    //private long t2 =0;
    private long time =0;

	//port numbers.
	public static int serverDataPort = 1111;//Data source port.
	public static int serverControlPort = 2223;//Control source/sink port.
	//public static int serverHelloPort = 6789;//Server Hello Port.
	public static int serverAlarmPort = 3333;//Server Alarm port for abnormal exit.
	public static int clientDataPort = 5555;//Data sink port.
	public static int clientControlPort = 4445;//Control sink/source port.


	//sockets used.
	public static ServerSocket welcomeSocket;
	public static Socket connectionSocket;
	public static Socket clientSocket;
	public static Socket clientAlarmSocket;
	public static DatagramSocket ds;
	public static InetAddress ip;

	//some global variables used.
	public static String fileName;// = "8gigbytes.xfdl";//name of file being transferred.
	public static int fileSize;//size of file being transferred.
	public static int buffer_size;//data packet size.
	public static int packetSize;//holds the packet size.
	public static int times = 0;//variable used to store the number of expected packets.
	public static int rempart = 0;//variable to hold the size of last partial packet, if any.
	public static int number_packets = 0;//variable to hold number of packets received.
	public static int[] lostPacketArray = new int[100000];//array to store missing packet sequence numbers.
	public static int lostPacketCounter;//variable to hold a count of lost packets.
	public static String missedPacket;//variable to store missed packet integers as a string.
	public static int globalVal;//variable used by the alarm thread to check the client activity.
	public static long start, end;//variables to hold start & end times of file transfer.
	public static int statusFlag = 0;//flag that tells which stage of main thread is running.
	public static byte[] dstByteArray;//byte array into which the requested subblock will get into.
	public static int[] lower, upper;//integer arrays used for storing the bounds.
	public static String bounds;//variable used to hold a string of lower & upper bounds.


	/** List of remote servers.
	*/
	//public static byte ip[] = new byte[]{(byte) 130,(byte) 74,(byte) 96,(byte) 49};//scorpion.cs.olemiss.edu (server).
	//public static byte ip[] = new byte[]{(byte) 130,(byte) 74,(byte) 87,(byte) 188};//My machine somewhere (server).
	//public static byte ip[] = new byte[]{(byte) 127,(byte) 0,(byte) 0,(byte) 1};//Localhost (server).
	//public static byte ip[] = new byte[]{(byte) 130,(byte) 74,(byte) 96,(byte) 30};//jocasta.mine.nu (server).
	//public static byte ip[] = new byte[]{(byte) 130,(byte) 74,(byte) 120,(byte) 71};//sunset.olemiss.edu (server).
	//public static byte ip[] = new byte[]{(byte) 130,(byte) 74,(byte) 96,(byte) 5};//onyx.cs.olemiss.edu (server).
	//public static byte ip[] = new byte[]{(byte) 132,(byte) 177,(byte) 4,(byte) 119};//zheng.cs.unh.edu (server).
	//public static byte ip[] = new byte[]{(byte) 132,(byte) 177,(byte) 4,(byte) 137};//scott.cs.unh.edu (server).


	/** RemoteUDPClient class constructor.
     */
	public RemoteUDPClient(String name, RecordDescriptor recordDesc, ISBounds theBounds,
		      AxisOrdering ordering, String URL, String serverURL)
	{
		super(name,recordDesc,theBounds,ordering,URL );// call inherited constructor
		this.fileServerURL = serverURL;
		// this buffer is used to read data from the socket
		this.tmpBufferSize = 61440;//setting this value to UDP packet size.
        this.tmpBuffer = new byte[tmpBufferSize];
        //System.out.println("***URL*** " + URL);
        remoteSourceName = URL;
	}

	/** Sets the filename for transfer.
	 */
	public static void setFileName(String s)
	{
		fileName = s;
	}

    /** Open the server socket on the client side.
    */
    public void activate()
    {
		super.activate();
		try
	    {
			//InterPacketDelay t = new InterPacketDelay();//object used to set interpacket delay.

			/*
			//Remote command execution daemon code starts.
				Runtime runtime = Runtime.getRuntime();
				String s1 = "ssh sridhar@";
			*/
				String s2 = remoteSourceName.substring(6,25);
			/*
				String s3 = " ./runserver";
				String remoteCommand = s1 + s2 + s3;
				System.out.println("***remoteCommand*** " + remoteCommand);
				Process proc = runtime.exec(remoteCommand);
			//Remote command execution daemon code ends.

			t.wait(10);
			*/

			ds = new DatagramSocket(clientDataPort);//binding datagram socket to clientDataPort --> Data input channel.
			ds.setSoTimeout(20000);
			//System.out.println("Socket Timeout: " + ds.getSoTimeout());
			////System.out.println("Opening the welcome socket.");
			welcomeSocket = new ServerSocket(clientControlPort);//binding welcomeSocket to clientControlPort --> Control channel from client to Server.
			//welcomeSocket.setSoTimeout(0);//default value.

			//following lines of code are to connect to ServerStarter.
			/*
				Socket helloSocket = new Socket(InetAddress.getByAddress(ip),serverHelloPort);//clientsocket connecting to zheng.cs.unh.edu (server)
				BufferedReader inFromServer = new BufferedReader(new InputStreamReader(helloSocket.getInputStream()));
				String ctrl_info;
				ctrl_info = inFromServer.readLine();
				helloSocket.close();
				ClientToServerControlInfo("Wait");
				System.out.println("Synchronizing with server");
			*/


			ip = InetAddress.getByName(s2);//getting the ip address from URL.

			ServerToClientControlInfo();//tcp connection to synchronize client with the server.
			System.out.println("Done...");
			//System.out.println("File Name set by CacheTestUDP: " + fileName);
			////System.out.println("Client sending metadata file \"" + fileName + "\"" + " to request \"" + fileName.substring(0,fileName.length()-5) + "\"");
			ClientToServerControlInfo("File Name: " + remoteSourceName.substring(25,remoteSourceName.length()));//tcp file name control info [now doing this from CacheTestUDP]
	    }
		catch (RemoteException exc)
	    {
			System.out.println("Error in lookup: " + exc.toString());
	    }
		catch ( IOException exc)
	    {
			System.out.println("Socket IO exception" + exc.toString() );
	    }
		catch ( Exception exc)
	    {
			System.out.println( exc.toString());
	    }
    }


	/** Open welcomeSocket for Control Information.
	*/
	public static void openWelcomeSocket() throws Exception
	{
		////System.out.println("Opening the welcome socket.");
		welcomeSocket = new ServerSocket(clientControlPort);
		welcomeSocket.setSoTimeout(10800000);
	}


	/** Close welcomeSocket for Control Information.
	*/
	public static void closeWelcomeSocket() throws Exception
	{
		////System.out.println("Closing the welcome socket.");
		welcomeSocket.close();
	}


	/** Close the server socket on the client side.
	*/
	public void deactivate()
	{
		super.deactivate();
		////System.out.println("Closing the welcome socket.");
		try
		{
			welcomeSocket.close();
		}
		catch ( Exception exc)
		{
			System.out.println( exc.toString());
	    }
	}


	/**Method to provide a reference to the welcomeSocket on client.
	*/
	public static void getSocketRef(ServerSocket s)
	{
		s = welcomeSocket;
	}


	/**Method to tell server the bounds for basic block construction.
	*/
	public static void setBasicBlockBounds(String s) throws Exception
	{
		ClientToServerControlInfo(s);
	}

	/** set the xml file for the remote datasource.
	*/
    public void setRemoteDataSource( String xmlFile, String dsName )
    {
	this.remoteXmlFile = xmlFile;
	this.remoteSourceName = dsName;
    }


    /** Return the Datum corresponding to the given IndexSpaceID.
    */
    public    Datum datum( IndexSpaceID did )
    {
        Datum returnVal = this.getRecordDescriptor().createDatum();

        this.datum( returnVal, did);
        return returnVal;
    }


    /** Get the Datum and fill its values in the passed Datum at positions
	*  specified by the corresponding IndexSpaceID coordinates.
    */
    public void datum(Datum datum, IndexSpaceID did){

		bounds = "@,3,0,0,0";//setting bounds so as not to change the client code.
		pdatumUDP(datum, did);//call to the remote datum fetch method.
    }


    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
    public void datum(Datum datum, IndexSpaceID did, RecordSpec recSpec){
    }


    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum(Datum datum, IndexSpaceID did, FieldIDMapper mp){
    }


    /** Return the Datum corresponding to the given IndexSpaceID. The
        Datum will only contain attributes specified by the RecordSpec
        argument.
    */
    public Datum datum(IndexSpaceID did, RecordSpec spec){
        Datum returnVal= null;
        return returnVal;
    }


	/** this method will extract data from datasource on remote server using UFTP logic.
     */
	protected  void pdatumUDP(Datum datum, IndexSpaceID did)
	{
		int dataType = datum.getStorageType();
		//int offset;

		switch(dataType)
		{
			case StorageType.Float:

					System.out.println("Float not implemented");
					break;
			case StorageType.Short:

					System.out.println("Short not implemented");
					break;

			case StorageType.Int:

					System.out.println("Int not implemented");
					break;

			case StorageType.Double:

					System.out.println("Double not implemented");
					break;

			case StorageType.Byte:

					TheClient();
					datum.setBytes(dstByteArray);

					/***** 	Commented the below logic, the reason being, since this is a datum case,
							every packet should be a known number of bytes & can be transferred in
							a packet at the most.
					/*
					if(rempart == 0)//incase the fileSize is an exact multiple of packet size.
					{
					 	for(int i=0;i<times;i++)
					   	{
							System.arraycopy(dstByteArray,offset,datum.getBytes(),offset,packetSize);//copying the dataStore into "tmp" array.
							offset += packetSize;
						}
					}
					else//incase the fileSize is not an exact multiple of packet size.
					{
						if(times == 1)//case of only one partial packet.
						{
							System.arraycopy(dstByteArray,0,datum.getBytes(),0,rempart);//copying the last partial packet into "tmp" array.
						}
						else//many packets with a last partial packet.
						{
							for(int i=0;i<times-1;i++)
							{
								System.arraycopy(dstByteArray,offset,datum.getBytes(),offset,packetSize);//copying the dataStore into "tmp" array.
								offset += packetSize;
							}
							//System.out.println("times: " + times + " Offset: " + offset + " rempart: " + rempart);
							System.arraycopy(dstByteArray,offset,datum.getBytes(),offset,rempart);//copying the last partial packet into "tmp" array.
						}
					}
					*/
				    break;

			default:

				    new InternalFaultError(SDBError.Fatal,"Unimplemented storage type in RemoteBFDS.pssublockUDP()");
				    break;
		}
	}

    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The FieldIDMapper argument determines the mapping between
        datums in the datasource and datums in the DataBlock. Note that the DataBlock is
        assumed to already have a valid data array. the dataBlock can be basicBlock or
        compositeBlock
    */
    public void subblock( DataBlock dstBlock,
                          ISBounds dstBounds, ISBounds srcBounds,
                          FieldIDMapper mp)
    {
       if ( dstBlock instanceof BasicBlock)
          subBasicBlock( (BasicBlock)dstBlock,dstBounds,srcBounds,mp);
       else if ( dstBlock instanceof CompositeBlock)
          subCompositeBlock( (CompositeBlock)dstBlock, dstBounds, srcBounds,mp);
       else  // what's the next
       {
          new SDBError( SDBError.Recoverable, "RemoteBFDS.subbblock:: " +
          " Unimplemented method. Unknown DataBlock type" );
       }
    }


    // compositeBlock implementation
    void subCompositeBlock( CompositeBlock dstBlock,
                                    ISBounds dstBounds, ISBounds srcBounds,
                                    FieldIDMapper mp)
    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: " +
                    " Unimplemented method!" );
    }


    void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: " +
                    " Unimplemented method!" );
    }


    void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                                  RecordSpec recSpec )
    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: " +
                    " Unimplemented method!" );
    }


    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The RecordSpec argument specifies a subset of the DataSource
        attributes to be written into the DataBlock datums. Note that the DataBlock is
        assumed to already have a valid data array. the DataBlock can be BasicBlock or
        CompositeBlock
    */
    public void subblock( DataBlock dstBlock,
                          ISBounds dstBounds, ISBounds srcBounds,
                          RecordSpec recSpec)
    {
       if ( dstBlock instanceof BasicBlock)
          subBasicBlock( (BasicBlock)dstBlock,dstBounds,srcBounds, recSpec);
       else // should work for compositeBlock
          subCompositeblock( (CompositeBlock) dstBlock,dstBounds,srcBounds, recSpec);
    }


    /** Unimplemented
     * @param dstBlock
     * @param dstBounds
     * @param srcBounds
     * @param recSpec
     */
    void subBasicBlock( BasicBlock dstBlock,
                                ISBounds dstBounds, ISBounds srcBounds,
                                RecordSpec recSpec)
    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subBasicBlock:: " +
                    " Unimplemented method!" );
    }

    void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                                  FieldIDMapper mp )

    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subBasicBlock:: " +
                    " Unimplemented method!" );
    }


    // composite implementation
    private void subCompositeblock( CompositeBlock dstBlock,
                                    ISBounds dstBounds, ISBounds srcBounds,
                                    RecordSpec recSpec )
    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subBasicBlock:: " +
                    " Unimplemented method!" );
    }


    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The BasicBlock is assumed to be created and  a valid data array is already allocated.
        NOTE: dstBlock can be  BasicBlock or CompositeBlock. data is copied from file to BasicBlock or CompositeBlock
		based on its data type.
    */
    public void subblock( DataBlock dstBlock,
                          ISBounds dstBounds, ISBounds srcBounds)
    {
       buildBounds(srcBounds);

       if ( dstBlock instanceof BasicBlock){
		   try{
		   	psubblockUDP((BasicBlock)dstBlock,dstBounds,srcBounds ); // PJR //changed the method name to psubblockUDP.
		   }
		   catch ( Exception exception )
	   	   {
		   		exception.printStackTrace();
    	   }
	   }
       //subBasicBlock( (BasicBlock)dstBlock,dstBounds,srcBounds );
       else if ( dstBlock instanceof CompositeBlock ){
          subCompositeBlock( (CompositeBlock)dstBlock,dstBounds,srcBounds );
	   }
    }


    /** 
     *  Note:  The BasicBlock is assumed to be created and  a valid data 
     * array is already allocated.
     * 
     * This has been implemented as psubblockUDP
     */

    protected void subBasicBlock(BasicBlock dstBlock,
                               ISBounds dstBounds, 
                               ISBounds srcBounds)
    {
		   try{
		   	psubblockUDP(dstBlock,dstBounds,srcBounds ); 
		   }
		   catch ( Exception exception )
	   	   {
		   		exception.printStackTrace();
    	   }
    }

	/** this method will extract data from datasource on remote server using UFTP logic.
     */
    protected void psubblockUDP(BasicBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds) throws Exception
    {
		int dataType = dstBlock.getStorageType();
		int offset;
		this.currentSubBounds = srcBounds;

		switch(dataType)
		{
		case StorageType.Float:

		   TheClient();
		   if(times != 0 && rempart != 0)
		   {
		      RemoteUDPClient.bytesToFloats( dstBlock.getFloats(),
		            dstByteArray,((number_packets - 1) * packetSize) + rempart);
		      //call to the bytesToFloats method which populates
		      //floatDst array which inturn is referenced by dstBlock.
		   }
		   else if(times !=0 && rempart == 0)
		   {
		      RemoteUDPClient.bytesToFloats( dstBlock.getFloats(),
		            dstByteArray,number_packets * packetSize);
		      //call to the bytesToFloats method which populates
		      //floatDst array which inturn is referenced by dstBlock.
		   }
		   else
		   {
            RemoteUDPClient.bytesToFloats( dstBlock.getFloats(),
                  dstByteArray,rempart );
            //call to the bytesToFloats method which populates
		   }
		   break;

		case StorageType.Short:

		   System.out.println("Short not implemented");
		   break;

		case StorageType.Int:

		   System.out.println("Int not implemented");
		   break;

		case StorageType.Double:

		   System.out.println("Double not implemented");
		   break;

		case StorageType.Byte:

		   TheClient();
		   //logic to copy data from dstByteArray to dstBlock.
		   offset = 0;
		   if(rempart == 0)//incase the fileSize is an exact multiple of packet size.
		   {
		      for(int i=0;i<times;i++)
		      {
		         System.arraycopy(dstByteArray,offset,dstBlock.getBytes(),offset,packetSize);//copying the dataStore into "tmp" array.
		         offset += packetSize;
		      }
		   }
		   else//incase the fileSize is not an exact multiple of packet size.
		   {
		      if(times == 1)//case of only one partial packet.
		      {
		         System.arraycopy(dstByteArray,0,dstBlock.getBytes(),0,rempart);//copying the last partial packet into "tmp" array.
		      }
		      else//many packets with a last partial packet.
		      {
		         for(int i=0;i<times-1;i++)
		         {
		            System.arraycopy(dstByteArray,offset,dstBlock.getBytes(),offset,packetSize);//copying the dataStore into "tmp" array.
		            offset += packetSize;
		         }
		         //System.out.println("times: " + times + " Offset: " + offset + " rempart: " + rempart);
		         System.arraycopy(dstByteArray,offset,dstBlock.getBytes(),offset,rempart);//copying the last partial packet into "tmp" array.
		      }
		   }
		   break;

		default:
		   new InternalFaultError(SDBError.Fatal,"Unimplemented storage type in RemoteBFDS.pssublockUDP()");
		break;
		}
	}


    /**Builds the bounds string to be sent to the RemoteUDPServer.
      */
    public void buildBounds(ISBounds srcBounds)
    {
		lower = srcBounds.getLowerArray();//reference to the lower array of srcBounds.
		upper = srcBounds.getUpperArray();//reference to the upper array of srcBounds.

		int k = lower.length + upper.length;//variable to hold the sum of dimensions.
		//System.out.println("bounds: " + k);

		bounds = "#,"+ k + ",";

		//lower bounds
		for(int i=0;i<lower.length;i++)
		{
			bounds += lower[i] + ",";
		}

		//upper bounds
		for(int i=0;i<upper.length-1;i++)
		{
		   		bounds += upper[i] + ",";
		}

		bounds += upper[upper.length-1];//string variable "bounds" is used to let server know of sum of dimensions.

	   ////System.out.println("Bounds string: " + bounds);
	}


    /** read an array of bytes from the socket. It makes sure the number of bytes being read ==
     *  the length. The function will block until length number of bytes being read
     */
    private void readBytes(byte[] data, int length )
    {
		//int length = data.length;
		int offset =0;
		int count =0;
		try
	    {
			while( offset < length )
		    {
				offset += this.input.read(data,offset,length-offset);
				count ++;
		    }
	    }
		catch(IOException e )
	    {
			System.out.println(" IO exeception " + e.toString() );
	    }
		//System.out.println("number of reads for a block" + count );
		//return offset;
    }


    /** read an array of floats from the socket. It reads bytes from the socket buffer
	and put it into a small temp byte buffer with its size as a multiple of 4. Then
	we convert the bytes into floats and copy them into dst float array.
	This function will block until the dstArray is full
    */
    private void readFloats(float[] dstArray)
    {
		//int size = dstArray.length;
		int byteSize = dstArray.length * 4;
		int numLoop = byteSize/tmpBufferSize;
		int leftover = byteSize%tmpBufferSize;
		int floatOffset = 0;

		for(int i =0; i<numLoop ; i++)
		{
			this.readBytes(tmpBuffer, tmpBufferSize);
			//convert to floats
			RemoteBFDS.bytesToFloats( dstArray,floatOffset, tmpBuffer, tmpBufferSize);
			floatOffset += (tmpBufferSize/4);
		}
		//take care of the leftover

		this.readBytes(tmpBuffer,leftover);
		RemoteBFDS.bytesToFloats( dstArray, floatOffset, tmpBuffer, leftover);
    }

    public long transmitTime()
    {
        return time;
    }


	/** private method, convert bytes into floats.
	It takes numBytes from bytes array starting from index 0, convert
	them into floats, and put them into floatDst array starting at index offset
    */
    static void bytesToFloats( float[] floatDst,byte[] bytes, int numBytes )
    {
        int bits;
        int i=0,j=0;
        while(i< numBytes){

	    bits=	(  (bytes[i++] & 0xFF) 	<< 24 ) |
		(  (bytes[i++] & 0xFF) 	<< 16 ) |
		(  (bytes[i++] & 0xFF) 	<< 8  ) |
		(  (bytes[i++] & 0xFF));

	    floatDst[j] = Float.intBitsToFloat(bits);
	    //System.out.println(" convert a float " + floatDst[j] );
	    j++;
        }
    }


    /** private method, convert bytes into floats.
	It takes numBytes from bytes array starting from index 0, convert
	them into floats, and put them into floatDst array starting at index offset
    */
    static void bytesToFloatsOldVersion( float[] floatDst, int offset, byte[] bytes, int numBytes )
    {
        int bits;
        int j = offset;
        int i=0;
        while(i< numBytes){

	    bits=	(  (bytes[i++] & 0xFF) 	<< 24 ) |
		(  (bytes[i++] & 0xFF) 	<< 16 ) |
		(  (bytes[i++] & 0xFF) 	<< 8  ) |
		(  (bytes[i++] & 0xFF));

	    floatDst[j] = Float.intBitsToFloat(bits);
	    //System.out.println(" convert a float " + floatDst[j] );
	    j++;
        }
    }


    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. Note that the DataBlock is assumed to already have a valid data
        array.
        NOTE: DataBlock must be a CompositeBlock
    */
    /**** Commented by Sridhar on March 1, 2006
    private void subCompositeblock(CompositeBlock dstBlock,
                                   ISBounds dstBounds, ISBounds srcBounds)
    {
       new SDBError( SDBError.Recoverable, "RemoteBFDS.subCompositeBlock:: " +
                    " Unimplemented method!" );
    }
    ****/

    /** Returns a spatial subset of the DataSource as delimited by the ISBounds argument.
        The DataSource data is copied when filling the DataBlock.
        This method returns a BasicBlock.  Need more discussion on this method
    */
    public DataBlock subblock(ISBounds db)
    {
        BasicBlock basicBlock = new BasicBlock( db, this.recordDesc );
        this.subBasicBlock( basicBlock, basicBlock.getBounds(), db );
        return basicBlock;
    }


    /** For debugging only. */
    private String toBinaryString(int value, int fieldWidth)
	{
        int mask=(int) (Math.pow(2,fieldWidth)-1);
        value=value & mask;

        String binaryString=Integer.toBinaryString(value);
        int i,numPadChars=fieldWidth-binaryString.length();

        for(i=0;i<numPadChars;i++){
            binaryString="0"+binaryString;
        }

		return binaryString;
    }


    /** request server for for subblock
    */
    private class ServerSubblock extends Thread {

	ServerSubblock()
	{
	    super();
	}
	public void  run()
	{
	    try
		{
		    fileServer.subblock( currentSubBounds);
		}
	    catch(RemoteException e)
		{
		    System.out.println(" In client thread, can't find remote object " + e.toString() );
		}
	}
    }


/** File transfer client program.
*/
public static void TheClient() //throws Exception
{
	byte a[] = new byte[4];//packet seq_num decoding array.
	int loop_counter = 0; 


	/*////Remote command execution daemon code starts.
	Runtime runtime = Runtime.getRuntime();
	String remoteCommand = "ssh sridhar@scott.cs.unh.edu ./runserver";
	Process proc = runtime.exec(remoteCommand);
	//Remote command execution daemon code ends.
	*/

	try
	{
		//t.wait(10);//inducing a wait to synchronize client & server.
		ServerToClientControlInfo();//tcp connection to synchronize client with the server.
		ClientToServerControlInfo(bounds);//tcp datasource bounds info.
		ClientToServerControlInfo("Wait");//tcp wait until the server gets ready.

		while(true)
		{
			//t.wait(10);//inducing a wait to synchronize client & server.
			ServerToClientControlInfo();//tcp control information regarding the packet size(should not be equal to -1).
			//a nonzero positive value indicates the packet size being used for file transfer.

			start = System.currentTimeMillis();//start of transfer.

			if(buffer_size != -1)//tcp control information "-1" means end of transfer.
			{
				packetSize = buffer_size;//making a copy of packet size into packetSize.
				//t.wait(10);//inducing a wait to synchronize client & server.
				ServerToClientControlInfo();//tcp control information regarding the file size.
				//System.out.println(fileSize + " File Size");
				//System.out.println(buffer_size + " Packet Size");

				//logic to calculate the number of packets & store it in "times" variable.
				times = fileSize/buffer_size;
				rempart = fileSize % buffer_size;//size of last partial packet, if any.
				//System.out.println("Rempart: " + rempart);
				if(rempart != 0)
				{
					times += 1;
				}

				//System.out.println("Times: " + times);
				////System.out.println("**********************************");
				////System.out.println("Packet Size: " + buffer_size);
				byte[] receiveData = new byte[buffer_size+4];//array to receive packets from server.

				//declaring the length of dstByteArray.
				if(times != 0 && rempart != 0)
				{
					dstByteArray = new byte[((times-1) * buffer_size) + rempart];
				}
				else if(times != 0 && rempart == 0)
				{
					dstByteArray = new byte[times * buffer_size];
				}
				else
				{
					dstByteArray = new byte[rempart];
				}

				int currentSeqNum=0;//variable to store next expected packet sequence number.
				missedPacket = "";//variable to store missed packet integers as a string.
				lostPacketCounter = 0;//variable to count number of lost packets.
				//t.wait(10);//inducing a wait to synchronize client & server.
				ServerToClientControlInfo();//tcp connection for making the server wait for sometime so
											//that client can initialize itself for the file transfer.
				//System.out.println("Hey there client is synchronized with the server.");
				while(loop_counter != times)//Client expects to be running this loop "times" times.
				{
					try
					{
						DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
						ds.setSoTimeout(20000);
						ds.receive(receivePacket);
						//String p = new String(receivePacket.getData());

						//sequence number printing.
						//updating the loop counter to the latest UDP packet seq number.
						a[0] = receiveData[buffer_size];
						a[1] = receiveData[buffer_size+1];
						a[2] = receiveData[buffer_size+2];
						a[3] = receiveData[buffer_size+3];
						loop_counter = Seqnum_Decoder(a);//recording the packet sequence number into variable "loop_counter".

						//******************************
						globalVal = loop_counter;//alarm thread keeps watching "globalVal" for activity.
						//******************************

						//logic for figuring out missing packet sequence numbers.
						if(currentSeqNum == loop_counter)//expected packet seq number has arrived.
						{
							currentSeqNum++;//next expected packet sequence number.
						}
						else if(currentSeqNum < loop_counter)//forward jump in the arrived packet seq number.
						{
							//logic to store missing packet sequence numbers.
							for(int l=currentSeqNum;l<loop_counter;l++)
							{
								lostPacketArray[lostPacketCounter] = l;
								lostPacketCounter++;
							}
							currentSeqNum = loop_counter + 1;//next expected packet sequence number.
						}
						else//backward jump in the arrived packet sequence number.
						{
							////System.out.println("Hey there, this is an out of order packet!!");
							//search thru the lost packet list & delete this entry.
							for(int i=0;i<lostPacketCounter;i++)
							{
								if(lostPacketArray[i] == loop_counter)
								{
									for(int j=i;j<lostPacketCounter-1;j++)
									{
										lostPacketArray[j] = lostPacketArray[j+1];//left shift all the other lost packet entries.
									}
									lostPacketCounter--;//decrementing by one position for new lost packet entry.
									break;
								}
							}
						}

						number_packets++;//variable to store number of packets received.

						//logic to store the packet into "dstByteArray" array.
						if(loop_counter != times-1)
						{
							System.arraycopy(receiveData,0,dstByteArray,loop_counter * buffer_size,buffer_size);
						}
						else
						{
							if(rempart != 0)//case of last partial packet.
							{
								System.arraycopy(receiveData,0,dstByteArray,loop_counter * buffer_size,rempart);
							}
							else//case of last full packet.
							{
								System.arraycopy(receiveData,0,dstByteArray,loop_counter * buffer_size,buffer_size);
							}
						}
						loop_counter++;//variable to count the number of times this loop executes.
					}//end of try.
					catch(SocketException e)
					{
						////System.out.println("Lost last packet(s) case.");
						//System.out.println("Last packet received: " + globalVal);
						//System.out.println("Lost packet counter: " + lostPacketCounter);
						break;
					}//end of catch for SocketException e.
				}//end of inner while.

				statusFlag = 1;//first receive loop executed successfully.

				if(lostPacketCounter > 0)
				{
					//constructing a comma delimited string of missed packet sequence numbers to notify server.
					for(int k=0;k<lostPacketCounter;k++)
					{
						if(k != lostPacketCounter-1)
						{
							missedPacket += lostPacketArray[k] + ",";
						}
						else
						{
							missedPacket += lostPacketArray[k];
						}
					}
					//System.out.println("Missed packet string: " + missedPacket);
				}
				//t.wait(50);
				ServerToClientControlInfo();//tcp connection logic for synchronizing with server.
				////System.out.println("Client synchronized with the server.");
				ClientToServerControlInfo(missedPacket);//control packet to let server know about missing packets.
				//starting PacketRetransmission logic if missedPacket string is not null.
				if(!missedPacket.equals(""))
				{
					PacketRetransmission(missedPacket);
				}
				else
				{
					////System.out.println("No missing packets.");
				}

				////System.out.println("Number of packets received: " + number_packets);

				end = System.currentTimeMillis();//end of transfer.
				System.out.println("%%% Transfer_Time " + (end - start));
				//ClientToServerControlInfo("Time: " + (end-start));//control info to report server of total transfer time.
				////System.out.println("File size transferred: " + (double) (fileSize)/1000 + " KB");

				/*////checking to see if there are any lost packets.
				for(int i=0;i<times;i++)
				{
					flag = 0;
					globalVal = i;//activity variable for client alarm thread.
					for(int j=0;j<buffer_size;j++)
					{
						if(dataStore[i][j] != 0)
						{
							flag = 1;
							break;
						}
					}
					if(flag == 0)
					{
						System.out.println("Packet Loss.");
					}
				}
				*/

				////System.out.println("**********************************");
				////System.out.println();
			}//end of if.
			else
			{
				break;
			}//end of else.
		}//end of outer while.

		number_packets = 0;//reinitializing number_packets variable after the transfer.
	}//end of outer try.
	catch ( Exception exception )
	{
		//exception.printStackTrace();
		try
		{
			//special logic to handle the lost last packet(s) case.
			System.out.println("******Lost last packet(s) case.******");
			//System.out.println("Last packet received: " + globalVal);
			//System.out.println("Lost packet counter: " + lostPacketCounter);

			//logic to store missing packet sequence numbers.
			for(int l=globalVal+1;l<times;l++)
			{
				lostPacketArray[lostPacketCounter] = l;
				lostPacketCounter++;
			}

			if(lostPacketCounter > 0)
			{
				for(int k=0;k<lostPacketCounter;k++)
				{
					if(k != lostPacketCounter-1)
					{
						missedPacket += lostPacketArray[k] + ",";
					}
					else
					{
						missedPacket += lostPacketArray[k];
					}
				}
				//System.out.println("Missed packet string: " + missedPacket);
			}

			//t.wait(50);
			ServerToClientControlInfo();//tcp connection logic for synchronizing with server.
			////System.out.println("Client synchronized with the server.");
			ClientToServerControlInfo(missedPacket);//control packet to let server know about missing packets.
			//starting PacketRetransmission logic if missedPacket string is not null.
			if(!missedPacket.equals(""))
			{
				PacketRetransmission(missedPacket);
			}
			else
			{
				////System.out.println("No missing packets.");
			}

			////System.out.println("Number of packets received: " + number_packets);

			end = System.currentTimeMillis();//end of transfer.
			System.out.println("%%% Transfer_Time " + (end - start));
			//ClientToServerControlInfo("Time: " + (end-start));//control info to report server of total transfer time.
			////System.out.println("File size transferred: " + (double) (fileSize)/1000 + " KB");

			/*////checking to see if there are any lost packets.
			for(int i=0;i<times;i++)
			{
				flag = 0;
				globalVal = i;//activity variable for client alarm thread.
				for(int j=0;j<buffer_size;j++)
				{
					if(dataStore[i][j] != 0)
					{
						flag = 1;
						break;
					}
				}
				if(flag == 0)
				{
					System.out.println("Packet Loss.");
				}
			}
			*/

			////System.out.println("**********************************");
			////System.out.println();
			//t.wait(10);//inducing a wait to synchronize client & server.
			ServerToClientControlInfo();//tcp connection to read packet size.
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}//end of catch for Exception ex.
		//number_packets = 0;
    }//end of catch for Exception exception.
}//end of Client.



/** Interpacket delay parameter setter.
*/
public static class InterPacketDelay extends Thread
{
	public static void wait(int n)
	{
		try
		{
			Thread.sleep(n);
		}
		catch(InterruptedException exception)
		{
			exception.printStackTrace();
		}
	}
}



/** Logic to receive packet retransmissions.
*/
public static void PacketRetransmission(String s) //throws Exception
{
	int lostRetranPacketCounter = 0,loop_counter = 0,packetCounter=0,currentSeqNum,packetSeqnum=0;
	int startPos=0,endPos=0,tokenCounter=0;//"tokenCounter" holds the number of missed packets requiring to be retransmitted.
	StringTokenizer st = new StringTokenizer(s,",");//tokenizing the missedPacket string with ",".
	StringTokenizer st1 = new StringTokenizer(s,",");//tokenizing the missedPacket string with ",".
	//InterPacketDelay t = new InterPacketDelay();//object used to set interpacket delay.
	String missedRetranPacket = "";//variable to store missed retransmitted packet integers as a string.
	//logic to find out the number of missing packets.
	while(st.hasMoreTokens())
	{
		st.nextToken();
		tokenCounter++;
	}

	//System.out.println("Token Counter: " + tokenCounter);

	int[] lostPackets = new int[tokenCounter];//array to hold the lost packet numbers for retransmission logic below.
	//populating the lostPackets array with lost packet seq numbers.
	int x=0;
	while(st1.hasMoreTokens())
	{
		lostPackets[x] = Integer.parseInt(st1.nextToken());
		x++;
	}

	/*
	//checking the contents of lostPackets array.
	for(int i=0;i<x;i++)
	{
		System.out.println(lostPackets[i]);
	}
	*/

	int[] lostRetranPacketArray = new int[100000];//array to store missing retransmitted packet sequence numbers.

	byte[] receiveData = new byte[buffer_size+4];//array to receive packets from server.
	byte a[] = new byte[4];//packet seq_num decoding array.

	//System.out.println("# of missed packets: " + x);

	try
	{
		//System.out.println("Socket Timeout: " + ds.getSoTimeout());
		//t.wait(50);
		ServerToClientControlInfo();//tcp connection for making the server wait for sometime so
									//that client can initialize itself for the file transfer.
		while(loop_counter != tokenCounter)
		{
			try
			{
				DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
				ds.setSoTimeout(20000);
				ds.receive(receivePacket);

				currentSeqNum = lostPackets[loop_counter];//expected packet sequence number.

				//sequence number printing.
				//updating the loop counter to the latest UDP packet seq number.
				a[0] = receiveData[buffer_size];
				a[1] = receiveData[buffer_size+1];
				a[2] = receiveData[buffer_size+2];
				a[3] = receiveData[buffer_size+3];
				packetSeqnum = Seqnum_Decoder(a);//recording the packet sequence number into  variable "packetSeqnum".

				//******************************
				globalVal = packetSeqnum;//alarm thread keeps watching "globalVal" for activity.
				//******************************

				//System.out.println("Arrived Packet seq no: " + packetSeqnum);

				//logic for figuring out missing packet sequence numbers.
				if((currentSeqNum == packetSeqnum) && (packetSeqnum < lostPackets[x-1]))//expected packet seq number has arrived.
				{
					currentSeqNum = lostPackets[loop_counter+1];
				}
				else if(currentSeqNum < packetSeqnum)//forward jump in the arrived packet seq number.
				{
					//logic to store missing packet sequence numbers.
					for(int i=0;i<x;i++)
					{
						if(lostPackets[i] == currentSeqNum)
						{
							startPos = i;
						}
						else if(lostPackets[i] == packetSeqnum)
						{
							endPos = i;
						}
						else
						{
							//do nothing.
						}
					}
					for(int l=startPos;l<endPos;l++)
					{
						lostRetranPacketArray[lostRetranPacketCounter] = lostPackets[l];
						lostRetranPacketCounter++;
					}
					loop_counter = endPos;
					//setting the next expected sequence number to the next in lostPackets array as long as arrived packet is not the last.
					if(loop_counter != x-1)
					{
						currentSeqNum = lostPackets[loop_counter + 1];
					}
				}
				else if(currentSeqNum == lostPackets[x-1])
				{
					//case of the last lost packet being received.
				}
				else//backward jump in the arrived packet sequence number.
				{
					////System.out.println("Hey there, this is an out of order retransmitted packet!!" + packetSeqnum);
					//search thru the lost packet list & delete this entry.
					for(int i=0;i<lostRetranPacketCounter;i++)
					{
						if(lostRetranPacketArray[i] == packetSeqnum)
						{
							for(int j=i;j<lostRetranPacketCounter-1;j++)
							{
								lostRetranPacketArray[j] = lostRetranPacketArray[j+1];//left shift all the other lost packet entries.
							}
							lostRetranPacketCounter--;//decrementing by one position for new lost packet entry.
							break;
						}
					}
				}

				number_packets++;//variable to store number of packets received.

				//logic to store the packet into "dataStore" array.
				if(packetSeqnum != times-1)
				{
					System.arraycopy(receiveData,0,dstByteArray,packetSeqnum * buffer_size,buffer_size);
				}
				else
				{
					if(rempart != 0)//case of last partial packet.
					{
						System.arraycopy(receiveData,0,dstByteArray,packetSeqnum * buffer_size,rempart);
					}
					else//case of last full packet.
					{
						System.arraycopy(receiveData,0,dstByteArray,packetSeqnum * buffer_size,buffer_size);
					}
				}

				loop_counter++;//variable to count the number of times this loop executes.
				packetCounter++;//local variable to keep track of the number of retransmitted pkts received.
			}
			catch(SocketException e)
			{
				////System.out.println("Lost last packet(s) case.");
				//System.out.println("Last packet received: " + globalVal);
				//System.out.println("Lost packet counter: " + lostPacketCounter);
				break;
			}

		}//end of while.

		////System.out.println("Number of retransmissions received: " + packetCounter);

		//logic to handle retransmission of retransmitted packets.
		for(int k=0;k<lostRetranPacketCounter;k++)
		{
			if(k != lostRetranPacketCounter-1)
			{
				missedRetranPacket += lostRetranPacketArray[k] + ",";
			}
			else
			{
				missedRetranPacket += lostRetranPacketArray[k];
			}
		}
		//t.wait(50);
		ServerToClientControlInfo();//tcp connection logic for synchronizing with server.
		////System.out.println("Client synchronized with the server.");
		ClientToServerControlInfo(missedRetranPacket);//control packet to let server know about missing retransmitted packets.
		//starting PacketRetransmission logic if missedRetranPacket string is not null.
		if(!missedRetranPacket.equals(""))
		{
			PacketRetransmission(missedRetranPacket);//requesting a resend if missed packet string is not null.
		}
		else
		{
			////System.out.println("No missing packets.");
		}
	}//end of try.
	catch(Exception exception)
	{
		try
		{
			//logic to handle lost last retransmitted packet(s).
			////System.out.println("******Lost last retransmitted packet(s) case.******");
			for(int i=0;i<x;i++)
			{
				if(lostPackets[i] == globalVal)
				{
					startPos = i;
					break;
				}
			}
			for(int l=startPos;l<x;l++)
			{
				lostRetranPacketArray[lostRetranPacketCounter] = lostPackets[l];
				lostRetranPacketCounter++;
			}
			////System.out.println("Number of retransmissions received: " + packetCounter);

			//logic to handle retransmission of retransmitted packets.
			for(int k=0;k<lostRetranPacketCounter;k++)
			{
				if(k != lostRetranPacketCounter-1)
				{
					missedRetranPacket += lostRetranPacketArray[k] + ",";
				}
				else
				{
					missedRetranPacket += lostRetranPacketArray[k];
				}
			}
			//t.wait(50);
			ServerToClientControlInfo();//tcp connection logic for synchronizing with server.
			////System.out.println("Client synchronized with the server.");
			ClientToServerControlInfo(missedRetranPacket);//control packet to let server know about missing retransmitted packets.
			//starting PacketRetransmission logic if missedRetranPacket string is not null.
			if(!missedRetranPacket.equals(""))
			{
				PacketRetransmission(missedRetranPacket);//requesting a resend if missed packet string is not null.
			}
			else
			{
				////System.out.println("No missing packets.");
			}
		}//end of try.
		catch(Exception ex)
		{
			ex.printStackTrace();
		}//end of catch for Exception ex.
	}//end of catch for Exception exception.
}//end of PacketRetransmission.


/** Server to Client Control Information regarding
	file size & packet size.
*/
public static void ServerToClientControlInfo() throws Exception
{
	String ctrl_info;
	clientSocket = new Socket(ip, serverControlPort);
	BufferedReader inFromServer = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
	ctrl_info = inFromServer.readLine();

	//control information about File name.
	if(ctrl_info.charAt(0) == 'F')
	{
		fileName = ctrl_info.substring(11,ctrl_info.length());
	}
	//control information about File size.
	else if(ctrl_info.charAt(0) == 'S')
	{
		fileSize = Integer.parseInt(ctrl_info.substring(6,ctrl_info.length()));
	}
	//control information about packet size.
	else if(ctrl_info.charAt(0) == 'B')
	{
		buffer_size = Integer.parseInt(ctrl_info.substring(13,ctrl_info.length()));
	}
	//control information used to synchronize client with server.
	else
	{
		//empty else for doing nothing used to make the server wait
		//for the start of file transfer until the client initializes.
	}
	clientSocket.close();
}


/** TCP Control Traffic to Server about missing packets.
*/
public static void ClientToServerControlInfo(String s) throws Exception
{
	String ctrl_packet;
	connectionSocket = welcomeSocket.accept();
	DataOutputStream outToServer = new DataOutputStream(connectionSocket.getOutputStream());
	ctrl_packet = s + '\n';
	outToServer.writeBytes(ctrl_packet);
	connectionSocket.close();
	//System.out.println("welcomeSocket.isOpen()="+welcomeSocket.isOpen());
}


/** Server to Client Alarm Information reader.
*/
public static void ServerToClientAlarm() throws Exception
{
	String ctrl_info;
	clientAlarmSocket = new Socket(ip, serverAlarmPort);
	BufferedReader alarmFromServer = new BufferedReader(new InputStreamReader(clientAlarmSocket.getInputStream()));
	ctrl_info = alarmFromServer.readLine();
	//alarm thread waiting for client to report abnormal termination.
	if(ctrl_info.charAt(0) == 'O')
	{
		//logic to get conneted to server on the alarmPort
		//and terminate the server alarm.
	}
	clientAlarmSocket.close();
}


/** Finds the integer value of 32 bit binary string.
*/
public static int Seqnum_Decoder(byte a[])
{
	return((a[0] << 24) & 0xFF000000 |(a[1] << 16) & 0x00FF0000 | (a[2] << 8) & 0x0000FF00 | a[3] & 0x000000FF);
}


/** Method for restarting Alarm thread with different sleepTime quantum.
*/
public static void AlarmThreadStart()
{
	ClientAlarm alarmThread = new ClientAlarm(2000);
	alarmThread.start();
}


/** Thread class to control how long packet receiving has to be done.
*/
public static class ClientAlarm extends Thread
{
	private int sleepTime;//alarm wakes up after sleepTime milliseconds.
	//private Thread blinker = null;
	InterPacketDelay t = new InterPacketDelay();//object used to set interpacket delay.
	int oldGlobalVal, newGlobalVal,flag;//activity measurement variables & flag -> packet loss detection variable.

		/*
	//thread stops with this method call.
	public void stopThread()
	{
		blinker = null;
	}
		*/

	//ClientAlarm class constructor to initialize sleepTime.
	public ClientAlarm(int time)
	{
		sleepTime = time;
	}

	//thread starts with this method call.
	public void run()
	{
			/*
		if (blinker == null)
		{
			blinker = new Thread(this, "Clock");
		    blinker.start();
    	}
			*/
		//Thread thisThread = Thread.currentThread();
        //while (blinker == thisThread) {
		try
		{
			//System.out.println("Hey there !!! I am existing to control the main thread.");
			Thread.sleep(sleepTime);
			/*
				if(sleepTime == 30000)
				{
					System.out.println("Initial quantum of " + sleepTime + " (ms) expired, Sleeping again.");
				}
				else
				{
					System.out.println("Additional quantum of " + sleepTime + " (ms) expired.");
				}
			*/
			oldGlobalVal = globalVal;//recording the value of current packet sequence number.
         //time gap to check any transfer activity.
			RemoteUDPClient.InterPacketDelay.wait(100);
			newGlobalVal = globalVal;//recording the value of current packet sequence number after a wait for 100 ms.
			if((oldGlobalVal != newGlobalVal) || (statusFlag != 0))//incase of an activity.
			{
				AlarmThreadStart();//resetting the alarm thread.
			}
			else //if(statusFlag == 0) -> initial receive loop is stuck.
			{
				////System.out.println("Control taken over by Alarm Thread.");
				//special logic to handle the lost last packet(s) case.
				//logic to store missing packet sequence numbers.
				for(int l=globalVal+1;l<times;l++)
				{
					lostPacketArray[lostPacketCounter] = l;
					lostPacketCounter++;
				}

				if(lostPacketCounter > 0)
				{
					for(int k=0;k<lostPacketCounter;k++)
					{
						if(k != lostPacketCounter-1)
						{
							missedPacket += lostPacketArray[k] + ",";
						}
						else
						{
							missedPacket += lostPacketArray[k];
						}
					}
					//System.out.println("Missed packet string: " + missedPacket);
				}
				//t.wait(100);
				ServerToClientControlInfo();//tcp connection logic for synchronizing with server.
				////System.out.println("Client synchronized with the server.");
				ClientToServerControlInfo(missedPacket);//control packet to let server know about missing packets.

				if(!missedPacket.equals(""))
				{
					PacketRetransmission(missedPacket);
				}
				else
				{
					////System.out.println("No missing packets.");
				}


				////System.out.println("Number of packets received: " + number_packets);

				end = System.currentTimeMillis();//end of transfer.
				////System.out.println("%%% Transfer_Time " + (end - start));
				ClientToServerControlInfo("Time: " + (end-start));//control info to report server of total transfer time.
				////System.out.println("File size transferred: " + (double) (fileSize)/1000 + " KB");

				/*////checking to see if there are any lost packets.
				for(int i=0;i<times;i++)
				{
					flag = 0;
					globalVal = i;//activity variable for client alarm thread.
					for(int j=0;j<buffer_size;j++)
					{
						if(dataStore[i][j] != 0)
						{
							flag = 1;
							break;
						}
					}
					if(flag == 0)
					{
						////System.out.println("Packet Loss." + i);
					}
				}
				*/

				////System.out.println("**********************************");
				////System.out.println();
				ServerToClientControlInfo();//tcp connection to read a -1 buffer_size.
				//ServerToClientAlarm();//tcp connection logic for synchronizing with server's alarm port.
				//ClientToServerControlInfo("Normal Termination.");
      			number_packets = 0;//reinitializing number_packets variable after the transfer.
			}//end of else.
		}//end of try.
		catch(Exception exception)
		{
			exception.printStackTrace();
		}
	//}//end of while.
	}//end of run.
}


/** Thread class to start the Client.
*/
public static class ClientStarter extends Thread
{
   public void run()
   {
      try
      {
      	TheClient();//call to the method TheClient().
      }
      catch ( Exception exception )
      {
         exception.printStackTrace();
      }
   }//end of run.
}


/** RemoteUDPClient main.
*/
public static void main(String[]  args)
{
	/***** Commented main on March 1, 2006 -- By SRK.


	RemoteUDPClient dsrc =(RemoteUDPClient)DataSource.createRemoteUDPClient("simple.7.1.bin","simple.7.1.bin.xfdl","//zheng.cs.unh.edu/SAMPLE-SERVER");

	lower = new int[]{0,0};
	upper = new int[]{4,4};

	dsrc.activate();//sets up the clienttoserver control socket.

	RecordDescriptor rec = dsrc.getRecordDescriptor();

	//constructing bounds string to pass on to server.
	int k = lower.length + upper.length;//sum of dimensions.
	bounds = "#,"+ k + ",";
	for(int i=0;i<lower.length;i++)
	{
		bounds += lower[i] + ",";
	}
	for(int i=0;i<upper.length-1;i++)
	{
		bounds += upper[i] + ",";
	}
	bounds += upper[upper.length-1];

	ISBounds tbounds = new ISBounds(lower,upper);
	BasicBlock theBlock = new BasicBlock( tbounds , rec );

	try
	{	dsrc.psubblockUDP(theBlock, tbounds, tbounds );//calling psubblock method to make network call.
	}
	catch ( Exception exception )
	{
		exception.printStackTrace();
    }

	System.out.println( theBlock);
	*/

	/*****
	int[] lower1 = { 2,2,2};
	int[] upper1 = {4,4,4};
	ISBounds tbounds1 = new ISBounds( new IntegerIndexSpaceID(lower1), new IntegerIndexSpaceID(upper1) );

	BasicBlock theBlock1 = new BasicBlock( tbounds1 , rec );
	try
	{
		dsrc.psubblockUDP(theBlock1, tbounds, tbounds1 );//calling psubblock method to make network call.
	}
	catch ( Exception exception )
	{
		exception.printStackTrace();
    }
	System.out.println( theBlock1);
	****/

	/*
	System.exit(0);
	*/
}
}