package edu.unh.sdb.datasource;


/**
 * It allows to 'iterate' a {@link Slice} along is normal vector, calculating
 * the intersection with a {@link Volume} and reading data from a
 * {@link DataSource} to fill that intersection.
 * @author dlipsa
 */
public class SliceIterator
   extends Slice 
   implements ObliquePlaneIteratorInterface
{
   /*
    * Creates an iterator that traverses 'volume' using a plane that slides along
    * its normal. At each step the iterator calculates the intersection between 'plane'
    * and 'volume' and then slides the plane along its normal.
    * 'outsideColor' and 'insideColor' should be arrays of 4 bytes RGBA.
    */
   public SliceIterator (byte[] outsideColor, byte[] insideColor,
      Volume volume, Plane plane, DataSource dataSource, double step)
   {
      super (outsideColor, insideColor, volume, plane, dataSource);
      this.step = step;
   }

   @Override
   public void init ()
   {
      super.init ();
      Plane p = getPlane ();
      Volume v = getVolume ();
      p.startIteration (v);
      if (p.validPosition (v))
         read ();
   }
   
   public boolean next ()
   {
      Plane p = getPlane ();
      Volume v = getVolume ();
      boolean valid;
      p.next (step);
      valid = p.validPosition (v);
      if (valid)
         read ();
      else
         p.prev (step);      
      return valid;
   }
   
   public boolean prev ()
   {
      Plane p = getPlane ();
      Volume v = getVolume ();
      boolean valid;
      p.prev (step);
      valid = p.validPosition (v);
      if (valid)
         read ();
      else
         p.next (step);      
      return valid;
   }
   
   
   public boolean valid ()
   {
      return getPlane ().validPosition (getVolume ());
   }
   
   public void end ()
   {
   }
   
   public void setSPCacheBlocks(boolean spBlockShape){}

   public double getStep ()
   {
      return step;
   }

   private double step;

}
