package edu.unh.sdb.datasource;

import edu.unh.sdb.common.Datum;
import edu.unh.sdb.common.Values;
import java.util.EnumMap;
import javax.vecmath.Vector3d;


class TransferFunctionCT extends TransferFunction
{         
   public void classifyVoxel (DataBlock src, IndexSpaceID srcId,
      DataBlock dest, IndexSpaceID destId, Feature feature)
   {
      double opacity = classifyVoxel (src, srcId, 0, feature);
      assert (opacity <= 1 && opacity >= 0);
      dest.setDouble (destId, 0, opacity);
   }

   protected static double classifyVoxel (DataBlock src, IndexSpaceID srcId, 
      int attr, Feature feature)
   {      
      Level[] levels = featureToLevels.get (feature);
      short ctValue = src.getShort (srcId, attr);
      short fn, fnp1;
      double an, anp1;
      int i;
      for (i = 1; i < levels.length - 1; i++)
         if (ctValue < levels[i].ctValue)
            break;
      fn = levels[i-1].ctValue;
      an = levels[i-1].opacity;
      fnp1 = levels[i].ctValue;
      anp1 = levels[i].opacity;
      double gradientLength = gradient (src, srcId, attr).length ();
      double opacity = gradientLength * (anp1 * (ctValue - fn) / (fnp1 - fn) + 
         an * (fnp1 - ctValue) / (fnp1 - fn));
      opacity = (opacity > 1) ? 1 : opacity;
      return opacity;
   }
      
   public void shadeVoxel(DataBlock src, IndexSpaceID srcId,
      DataBlock dest, IndexSpaceID destId, Vector3d L, Vector3d H, double distFromPicturePlane)
   {
      double color = shadeVoxel (src, srcId, 0, L, H, distFromPicturePlane);
      logger.fine ("color = " + color);
      assert (color <= 1 && color >= 0);
      dest.setDouble (destId, 0, color);
   }
   
   public Type getType ()
   {
      return Type.CT;
   }
   public void convert (Datum datum, byte[] texture, int index)
   {
      short v = datum.getShort (0);
      
      // scale the value v to be in interval [0,255]
      v *=SCALE;
      
      texture[index++] = (byte) v;
      texture[index++] = (byte) v;
      texture[index++] = (byte) v;
      texture[index] = (byte) v;
   }
   
   public void convert (Values values, byte[] texture)
   {
      
      short[] data = values.getShorts ();
      byte v;
      
      for (int i = 0; i<data.length; i++)
      {
         
         v = (byte) (data[i] * SCALE);
         
         texture[4*i] = v;
         texture[4*i+1] = v;
         texture[4*i+2] = v;
         texture[4*i+3] = (byte) 255;
         
      }
      
   }
   
   public boolean isFirstBigger (byte[] t1, int i1, byte[] t2, int i2)
   {
      short first = t1[i1],  second = t2[i2];
      first &= 0x00ff;
      second &= 0x00ff;
      return first >= second;
   }
   
   public Datum createDatum ()
   {
      return Datum.create ((short) 0);
   }
   
   public void trilinearInterpolation (TransferFunction.Block block, Datum datum, 
      IndexSpaceID lower)
   {
      double t = TransferFunction.trilinearInterpolation (block, lower, 0);
      datum.setShort (0, (short) Math.round (t));
   }   
   
   private static final short CT_MIN = 0;
   private static final short CT_MAX = 4095;
   private static final float SCALE = (255.0f / (CT_MAX - CT_MIN));   
   private static final EnumMap<Feature, Level[]> featureToLevels;
   static 
   {
      featureToLevels = new EnumMap<Feature, Level[]>(Feature.class);
      featureToLevels.put (
         Feature.SKIN, new Level[] 
            // show skin opaque and everything else transparent
            {
               new Level(CT_MIN, 0),
               new Level((short)800, 0),   
               new Level((short)1224, 1),   //skin   
               new Level((short)1600, 0),  
               new Level(CT_MAX, 0)
            });
      
      featureToLevels.put (
         Feature.BONE, new Level[] 
            // show bone opaque and everything else transparent
            {
               new Level(CT_MIN, 0),
               new Level((short)100, 0),   //air
               new Level((short)1400, 0),   //skin   
               new Level((short)1800, 1),  // bone
               new Level(CT_MAX, 0)
            });      
      
      featureToLevels.put (
         Feature.SKIN_BONE, new Level[] 
            // show bone opaque and everything else transparent
            {
               new Level(CT_MIN, 0),
               new Level((short)100, 0),   //air
               new Level((short)1400, .1),   //skin   
               new Level((short)1800, 1),  // bone
               new Level(CT_MAX, 0)
            });      
      
   };
   
   private static class Level
   {
      Level (short ctValue, double opacity)
      {
         this.ctValue = ctValue;
         this.opacity = opacity;
      }
      
      short ctValue;
      double opacity;
   }   
}
