/************************************************************ 
   TypeConverter.java
   Copyright 2000 University of New Hampshire
   Durham, NH 03824

   Xuan Tang xtang@cs.unh.edu
   last update: 8/7/03
   updata comments: create
   
   ----------------- Edit history as of 12/12/04 -------------------
   06/01/05 rdb deprecated entire class; no longer need (already for a long time)
                since we went to the ByteBuffer stuff.
   12/12/04 rdb Added byte support, moved to datasource package, and made package access.
                Should return value functions be get<type> to be consistent with
                   other Granite functions? Pro: consistency. Con: this is a different
                   kind of method - it doesn't return different forms of a single
                   value, but the last value of a particular type that was converted
                   by this object. E.g., the int value of an instance of this object
                   is likely have a different value from the float value etc.
                Should this object BE like the other Granite objects, in that it
                   can return any return type.
                This class really should not be public; it's much too error-prone
                   and idiosyncratic. However, it is used in the datasource package.
   -----------------------------------------------------------------
************************************************************/


package edu.unh.sdb.datasource;
import edu.unh.sdb.common.*;
import java.nio.*;


/** This class is responsible for converting data from bytes to ints, floats, doubles..
 *   it's also responsible for handling big-endian vs. little endian 
 *  @deprecated This functionality is now  done with the ByteBuffers 
*/

class TypeConverter
{
     /** The result of the read. */
     private   float  floatResult;
     private   int    intResult;
     private   double doubleResult;
     private   short  shortResult;
     private   byte   byteResult;
     
     /** indicates number of bytes read.*/
     private int bytesRead = 0;

    public TypeConverter()
    {
        floatResult =0;
        intResult =0;
        shortResult =0;
        doubleResult =0;
        byteResult = 0;
    }

     public int intResult()
     {
        return intResult;
     }
     public float floatResult()
     {
        return floatResult;
     }
     public double doubleResult()
     {
        return doubleResult;
     }
     public short shortResult()
     {
        return shortResult;
     }
     public byte byteResult()
     {
        return byteResult;
     }
     


 /** Convert bytes into primitive data types assuming data is stored in big endian formate 
  * 
  * */
     public  void readValue( byte [] bytes, int index, int storageType, ByteOrder endian ){
        int bits;
        long lbits;
        switch (storageType){
            case StorageType.Float:
               this.bytesRead=4;
                if(  endian.equals(ByteOrder.BIG_ENDIAN) ){
                   bits=(  (bytes[index]   & 0xFF) 	<< 24 ) |
                        (  (bytes[index+1] & 0xFF) 	<< 16 ) |
                        (  (bytes[index+2] & 0xFF) 	<< 8  ) |
                        (  (bytes[index+3] & 0xFF));
                }
                else{
                   bits=(  (bytes[index]   & 0xFF) 	     ) |
                        (  (bytes[index+1] & 0xFF) 	<< 8 ) |
                        (  (bytes[index+2] & 0xFF) 	<< 16) |
                        (  (bytes[index+3] & 0xFF)      << 24);
                }
                this.floatResult=Float.intBitsToFloat(bits);
                //System.out.println("read float: " + r.floatResult);
                break;
            case StorageType.Double:
                this.bytesRead=8;
                if(endian.equals(ByteOrder.BIG_ENDIAN) ){
                  lbits=( (long) (bytes[index]	 & 0xFF) 	<< 56 ) |
                        ( (long) (bytes[index+1] & 0xFF) 	<< 48 ) |
                        ( (long) (bytes[index+2] & 0xFF) 	<< 40 ) |
                        ( (long) (bytes[index+3] & 0xFF)	<< 32 ) |
                        ( (long) (bytes[index+4] & 0xFF) 	<< 24 ) |
                        ( (long) (bytes[index+5] & 0xFF) 	<< 16 ) |
                        ( (long) (bytes[index+6] & 0xFF) 	<< 8  ) |
                        (        (bytes[index+7] & 0xFF));
                }
                else{
                  lbits=(        (bytes[index]	 & 0xFF) 	      ) |
                        ( (long) (bytes[index+1] & 0xFF) 	<< 8  ) |
                        ( (long) (bytes[index+2] & 0xFF) 	<< 16 ) |
                        ( (long) (bytes[index+3] & 0xFF)	<< 24 ) |
                        ( (long) (bytes[index+4] & 0xFF) 	<< 32 ) |
                        ( (long) (bytes[index+5] & 0xFF) 	<< 40 ) |
                        ( (long) (bytes[index+6] & 0xFF) 	<< 48 ) |
                        ( (long) (bytes[index+7] & 0xFF)        << 56);
                }
                this.doubleResult = Double.longBitsToDouble(lbits);
                //System.out.println("read double: "+ r.doubleResult);
                break;
            case StorageType.Int:
                this.bytesRead=4;
                if( endian.equals(ByteOrder.BIG_ENDIAN ) ){
                   bits=(  (bytes[index]   & 0xFF) 	<< 24 ) |
                        (  (bytes[index+1] & 0xFF) 	<< 16 ) |
                        (  (bytes[index+2] & 0xFF) 	<< 8  ) |
                        (  (bytes[index+3] & 0xFF));
                }
                else{
                  bits=	(  (bytes[index]   & 0xFF) 	      ) |
                        (  (bytes[index+1] & 0xFF) 	<< 8  ) |
                        (  (bytes[index+2] & 0xFF) 	<< 16 ) |
                        (  (bytes[index+3] & 0xFF)      << 24 );
                }

                this.intResult =  bits;
                //System.out.println("read int: "+ r.intResult);
                break;
            case StorageType.Short:
                this.bytesRead=2;
                if( endian.equals(ByteOrder.BIG_ENDIAN) ){
                   bits=(  (bytes[index]   & 0xFF) 	<< 8 ) |
                        (  (bytes[index+1] & 0xFF) );
                }
                else{
                   bits=(  (bytes[index]   & 0xFF) 	    ) |
                        (  (bytes[index+1] & 0xFF)      << 8);
                }
                this.shortResult= (short) bits;
                //System.out.println("read short: "+ r.shortResult);
                break;
            case StorageType.Byte:
               this.bytesRead=1;
               this.byteResult= bytes[index];
               //System.out.println("read byte: "+ r.byteResult);
               break;
            default:
                new IllegalArgumentError(SDBError.Fatal,
                    "Unsupported storage type code in BinaryFileDataSource.readValue:"+storageType);
        }
        if ( this.bytesRead == 0 )
           this.bytesRead = 0;
    }
}

