/*
    URCellIterator.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
*/
                                                                                
/**
 * The URCellIterator class implements an interface to iterate through  
 * an uniform cell space to visit each cell. 
 */

package edu.unh.sdb.datasource;
                                                                                
//import java.io.*;
                                                                                
import edu.unh.sdb.common.*;
                                                                                
public class URCellIterator extends CellIterator {
    //private URCellBlock cellblock;
    private CellIndexable cellblock;
    private ISIterator cIterator;
    private IntegerIndexSpaceID pos;

    /** Constructor takes an uniform cell block */
    //public URCellIterator(URCellBlock cblock)
    public URCellIterator(CellIndexable cblock)
    {
       super();
       this.cellblock = cblock;
       this.cIterator = new ISIterator(this.cellblock.copyBounds()); 
    }

    /** Initialize this URCellIterator object */
    public void init()
    {
       this.cIterator.init();
       if (this.cIterator.valid()) {
          this.valid = true;
          this.pos = (IntegerIndexSpaceID)this.cIterator.copyID();
          if (this.cellblock instanceof RegCellDataSource) {
             this.cell=(ARRSCell)((RegCellDataSource)this.cellblock).getCell( pos ); 
          }
          if (this.cellblock instanceof URCellBlock) {
             this.cell=(ARRSCell)((URCellBlock)this.cellblock).getCell( pos ); 
          }
       }
    }

    /** Moving the URCellIterator to the next position. */
    public boolean next()
    {
       if (this.valid == false)
          return false;
       this.cIterator.next();
       if (this.cIterator.valid()) {
          this.valid = true;
          this.pos = (IntegerIndexSpaceID)this.cIterator.copyID();
          if (this.cellblock instanceof RegCellDataSource) {
             this.cell=(ARRSCell)((RegCellDataSource)this.cellblock).getCell( pos ); 
          }
          if (this.cellblock instanceof URCellBlock) {
             this.cell=(ARRSCell)((URCellBlock)this.cellblock).getCell( pos ); 
          }
          return true;
       }

       else {
          this.valid = false;
          return false;
       }
    }
 
    /** Get a ARRS cell */
    public Cell getCell()
    {
       return this.cell;
    }

    /** Get the IndexSpaceID corresponding to the cell */
    public IndexSpaceID getIndexSpaceID()
    {
       return this.pos;
    }

    /** Get a uniform cell block */
    //public DataBlock getDataBlock()
    public CellIndexable getDataBlock()
    {
       return this.cellblock;
    }

    /** Return a String representing the object. */
    public String toString(){
        return this.toString("");
    }
                                                                                
   /** Return description of this ARRSCell */
    public String toString(String indent){
        String indent2=indent+Globals.toStringIndent;
        StringBuffer result=new StringBuffer(indent+"{ URCellIterator:\n");

        result.append(indent2+"valid = "+this.valid+"\n");
        result.append(indent2+"Cell info: \n");                                                                        
        result.append(this.cell.toString(indent2)+"\n");
        result.append(indent+"}");
        return new String(result);
    }
}

