/*
  programmer: Li Ye
*/

/**
 * VRIndexable provides different resolution view of the datasource.
 * It can be built on top of any datasource to view the datasource with
 * different resolution level. The resolution level is a integer number
 * which can be incremented and decrease by VRDS' function call.
 * Negative numbers mean the current resolution coarser than the datasource's
 * original resolution. It also means the current VRDS has less data points
 * than the original datasource.
 * Positive numbers' meaning is opposite.
*/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;

public class VRDataSource extends DataSource {

    protected VRIndexable vrIndexable;
 
    public VRDataSource(String dsname, DataSource theSource){
        super(dsname, theSource.getRecordDescriptor(), theSource.getBounds());    
        this.vrIndexable=new VRIndexable(theSource);
    }
    
    /** Activate this datasource*/
    public void activate(){
       ((DataSource)(this.vrIndexable.getSourceIndexable())).activate();
    }


    /** The DataSource will release any resources used during the 
      *  readling or processing of data.*/    
    public void deactivate(){
       ((DataSource)(this.vrIndexable.getSourceIndexable())).deactivate();
    }

    /** Return the Datum corresponding to the given IndexSpaceID. The 
      *  Datum will only contain attributes specified by the RecordSpec
      *  argument. 
      */
    public void datum(Datum dtm, IndexSpaceID did){
       this.vrIndexable.datum(dtm, did);
    }

    /** Return the Datum corresponding to the given IndexSpaceID. */
    public Datum datum(IndexSpaceID did){
        return this.vrIndexable.datum(did);
    }
            
    /** Return the Datum corresponding to the given IndexSpaceID. The
      *  Datum will only contain attributes specified by the RecordSpec
      *  argument.
    */
    public Datum datum(IndexSpaceID did, RecordSpec spec)
    {
        return this.vrIndexable.datum(did, spec);
    }
                                                                                
 
    /** PBR: Return by reference the Datum corresponding to the given 
     *  IndexSpaceID. Only datum fields referred to by the FieldIDMapper
     *  are modified.
     */
    public  void datum(Datum dtm, IndexSpaceID did, FieldIDMapper mp){
       this.vrIndexable.datum(dtm, did, mp);
    }
    
    
    /** PBR: Return by reference the Datum corresponding to the given 
     *  IndexSpaceID. Only datum fields referred to by the RecordSpec
     *  are modified.
     */
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec){
       this.vrIndexable.datum(dtm, did, recSpec);
    }


    public  DataBlock subblock(Datum [] d,ISBounds db)
    {
        return null;
    }
                                                                                
    
    /** Returns a DataBlock that is an attribute projection of a spatial
      *  subset of the DataSource.*/
    public  DataBlock subblock( ISBounds isb, RecordSpec spec )
    {
       return this.vrIndexable.subblock( isb, spec );
    }

    /** Returns a DataBlock that is a spatial subset of the DataSource.Override
      *  parent method for inefficiency.*/
    public DataBlock subblock(ISBounds db)
    {
       return this.vrIndexable.subblock(db);
    }


    /** Reads a subblock defined by srcBounds into the region of the 
      *  Datablock argument defined by dstBounds. */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, 
                         ISBounds srcBounds)
    {
       this.vrIndexable.subblock(theBlock,dstBounds, srcBounds);
    }


    /** Reads a subblock defined by srcBounds into the region of the Datablock 
      *  argument defined by dstBounds. The FieldIDMapper argument determines 
      *  the mapping between datums in the datasource and datums in the 
      *  DataBlock. */    
    public void subblock(DataBlock theBlock, ISBounds dstBounds, 
                         ISBounds srcBounds, FieldIDMapper mp)
    {
       this.vrIndexable.subblock(theBlock, dstBounds, srcBounds, mp);
    }
 

    /** Reads a subblock defined by srcBounds into the region of the Datablock 
      *  argument defined by dstBounds. The RecordSpec argument specifies a 
      *  subset of the DataSource attributes to be written into the DataBlock 
      *  datums.*/
    public void subblock(DataBlock theBlock, ISBounds dstBounds, 
                         ISBounds srcBounds, RecordSpec spec)
    {
       this.vrIndexable.subblock(theBlock, dstBounds, srcBounds, spec);
    }

    /** Increase Resolution Level by 1 */
    public boolean incr() {
       return this.vrIndexable.incr();
    }
    
    /** Decrease Resolution Level by 1 */
    public boolean decr() {
       return this.vrIndexable.decr();
    }
    
    public DataSource getDataSource(){
       return (DataSource) this.vrIndexable.getSourceIndexable();
    }
}
