//
//  AreaTriangleCellApproximator.java
//  Unstructured
//
//  Created by Philip Rhodes on Thu Apr 24 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//
// ------------------- Edit history (as of 01/10/05) -------------------------
// 
// 01/10/05 rdb: when creating Datums, return a type that matches input.
//
// ---------------------------------------------------------------------------

package edu.unh.sdb.lattice;

import edu.unh.sdb.common.*;

public class AreaTriangleCellApproximator extends UnstructuredApproximator {

    public AreaTriangleCellApproximator(){
    
    }
    
    public Datum approximate(Cell c, Point p){
    
        Datum[] data=c.getData();
		int  numAttributes=data[0].size();
		float[] attributes=new float[numAttributes];
        Point [] points=c.getVertexCoords();
		float [] pcoords=p.getCoords();
		float [] area=new float[3];
		float weight,totalArea=0;
		
        
		float	s0,s1;
		float 	t0,t1;
		
		float [][] vcoords=new float[3][];
		
		for (int i=0; i<3; i++){
			
			vcoords[i]=points[i].getCoords();
		}
		
		
		for ( int j,i=0; i<3; i++){
		
			j = (i+1) % 3;
			
			s0 = vcoords[i][0] - pcoords[0];
			s1 = vcoords[i][1] - pcoords[1];

			t0 = vcoords[j][0] - pcoords[0];
			t1 = vcoords[j][1] - pcoords[1];
			
			area[(i+2)%3]=(s0*t1 -s1*t0)/2;
			
			totalArea+=area[(i+2)%3];
		}

		
		for ( int i=0; i<3; i++){
			weight=(area[i]/totalArea);
			for ( int j=0; j<numAttributes; j++){
				attributes[j] += data[i].getFloat(j) * weight;
			}
		}

		//rdb return new FloatDatum(attributes);
		Datum retDatum = (Datum) data[0].clone(); // create matching type
		retDatum.setFloats( attributes );
		return retDatum;
    }
    
    public Datum approximate(Cell c, Point p, RecordSpec s){
        
        return this.approximate(c,p).subset(s);
    }
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }

    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        return new String(indent+"AreaTriangleCellApproximator:");
    }


}
