/* 
	ArrayCell.java
	Copyright 2000 University of New Hampshire
	Durham, NH 03824
	-------------------- Edit history as of 03/04/05 ----------------------------
	03/04/05 rdb: getDimensions error. It used getCoord() on first Point in array
	              to initialize min and max arrays. But, getCoord() returns a
	              reference. So, 1) we were modifying the coord and 2) min and
	              max were the same array.
*/

package edu.unh.sdb.lattice;

import edu.unh.sdb.common.*;

/** This class defines cells used with regular geometries and rectilinear 
    topologies.
    Vertices are numbered using a bit pattern constructed as follows:
    
        bit 0 is 0 if the vertex is in the lower position of axis 0, 1 otherwise.
        bit 1 is 0 if the vertex is in the lower position of axis 1, 1 otherwise.
        bit 2 is 0 if the vertex is in the lower position of axis 2, 1 otherwise.
            :
            :
        bit n is 0 if the vertex is in the lower position of axis n, 1 otherwise.
*/

public class ArrayCell extends Cell 
{
   /**
    * Construct a array cell with the corners defined by the "vertices" with
    * attribute values at those corners defined by "data".
    * An array cell is topologically rectilinear.
    * 
    * The dimensionality of the cell is inferred from the dimensionality of the
    * corner vertices.
    * 
    * There is no test made here to insure that there are the correct number of
    * vertices and data elements. This should be added.
    * @param vertices
    * @param data
    */
   public ArrayCell( Point [] vertices, Datum [] data )
   {
        super(vertices, data);
   }
    
   /**
    * Construct an ArrayCell.
    * @deprecated The dimensionality argument is superfluous.
    */
   public ArrayCell( Point [] vertices, Datum [] data, int dimensionality )
   {
        super(vertices, data);
   }
    
    
   /**
    * determine if the argument Cell intersects with this Cell.
    */
    public boolean intersects( Cell c )
    {
        new InternalFaultError( SDBError.Fatal,
                                         "ArrayCell.intersects not implemented");
        return false;
    }
    
 
    /**
     * return an array that represents the sizes of each dimension of the cell.
     */
    public float [] getDimensions()
    {
    /* This is a terrible implementation that needs to go away as soon as 
       we fix ArrayCell so that its vertices are properly ordered. */
    
        int  d = this.dim();
        float[] vertexCoords;
        int numVertices = this.coords.length;
        
        // declare and intitialize max and min
        // This assumes that getCoords returns new storage, not references to its
        // internal array. (rdb: this was an incorrect assumption)
        //float[] min = coords[0].getCoords();
        //float[] max = coords[0].getCoords();
        float[] min = ((Point)coords[0].clone()).getCoords();
        float[] max = ((Point)coords[0].clone()).getCoords();
               
        for ( int i = 1; i < numVertices; i++)
        {
            vertexCoords = coords[i].getCoords();
            
            for ( int j=0; j<d; j++)
            {
                min[j] = (vertexCoords[j] < min[j]) ? vertexCoords[j] : min[j];
                max[j] = (vertexCoords[j] > max[j]) ? vertexCoords[j] : max[j];
            }
        }
        
        // Let's re-use the max array to return the sizes
        for ( int j=0; j<d ; j++ )
            max[j] -= min[j];
    
        return max;
    }
    
    /** Return a String representing the object. */
    public String toString()
    {
        return this.toString("");
    }
    
    /** Return a String representing the object indented by the 
        argument string. 
     */
    public String toString(String indent)
    {
        StringBuffer result = new StringBuffer( indent + "ArrayCell: " );
        result.append( super.toString( "" ));
        return new String( result );
    }
}
