//
//  GIterator.java
//  Unstructured
//
//  Created by Philip Rhodes on Wed Jul 17 2002.
//  Copyright (c) 2001 University of New Hampshire. All rights reserved.
//
//------------------- Edit history --------------------------------------
// 10/21/05 rdb: removed unused variables
//-----------------------------------------------------------------------

package edu.unh.sdb.lattice;
//import java.io.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;


/** Child class of Point. GIterator can iterate through a Domain*/
public class GIterator extends Point implements Cloneable {

    private AxisOrdering ordering;
    //private float[] axes;
    private boolean valid=false;
    private float[] steps;
    private int dim;
    private float[] ubounds;
    private float[] lbounds;
    
    /** Constructor taking an GBounds, which specifies the IndexSpace */
    public GIterator(GBounds db) {
	
        int i,d=db.dim();
        this.dim=d;
        this.ordering = new AxisOrdering(d);
        this.axes = new float[d];
        this.steps = new float[d];
		
        for(i=0; i<d; i++){
            steps[i]=1.0f;
        }
        
        this.init(db);
    }
	
	/** Constructor taking an GBounds, which specifies the Domain,
	and the AxisOrdering which specifies the axis incrementation order 
	*/
    public GIterator(GBounds db, AxisOrdering ord) {
        int i, d=db.dim();
        this.dim=d;
        this.ordering = ord.copy();
        this.axes = new float[d];
        this.steps = new float[d];
		
        for(i=0;i<d;i++){
            steps[i]=1.0f;
        }
 
        this.init(db);
    }
    
    /** Constructor taking an GBounds, which specifies the Domain 
    and the steps array allowing GIterator increment step other than 1 */
    public GIterator(GBounds db, float [] steps) {
        this.dim=db.dim();
        this.ordering = new AxisOrdering(this.dim);
        this.axes = new float[this.dim];
        this.steps = steps.clone();
        
        this.init(db);
    }
	/** Constructor taking an GBounds, AxisOrdering and steps array */
    public GIterator(GBounds db, AxisOrdering ord, float [] steps) {
        this.dim=db.dim();
        this.ordering = ord;
        this.axes = new float[this.dim];
        this.steps = steps.clone();
        this.init(db);
    }
    
    /**Return a Point corresponding to the current position.*/
    public Point copyID(){
        return new Point(this.axes);
    }
    /** Return a copy of this GIterator
    
    */
    public Object clone(){
        GIterator cpy;
        GBounds db = new GBounds(new Point(this.lbounds),new Point(this.ubounds));
        cpy = new GIterator(db,this.ordering);
        float axes[] = new float[this.axes.length];
        for (int i=0;i<this.axes.length;i++){
            axes[i]=this.axes[i];
        }
        cpy.axes=axes;
        cpy.steps = this.steps.clone();
        cpy.valid = this.valid;
        return cpy;
    
    }
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        int i;
        String s=new String(indent+"<GIterator: "+axes[0]);
        
        for (i=1; i<this.axes.length; i++){
            s = s+","+axes[i];
        }
        s = s + ">";
	    return s;
	
	}
    
    /**Change the IndexSpace to which this GIterator is bound */
    public void setBounds(GBounds bound){
        this.init(bound);
    }
	
	
    private void init(GBounds bounds) {
        int i;
        this.ubounds = new float[this.dim];
        this.lbounds = new float[this.dim];
        
        for (i=0;i<this.dim;i++) {
            this.ubounds[i] = bounds.getUpper().getCoord(i);
            this.lbounds[i] = bounds.getLower().getCoord(i);
            this.axes[i] = lbounds[i];
        }
        
        valid=true;
        
    }
    
    /**Initilize this ISiterator*/
    public void init() {
        int i;

        for (i=0;i<this.dim;i++) {
            this.axes[i] = lbounds[i];
        }
        valid=true;
    }
    
    private void resetAxis(int axisNumber){
        this.setCoord(axisNumber, this.lbounds[axisNumber]);
    }
    
    /** Get the value corresponding to the specified axis.*/
    public float getCoord(int axisNumber){
        if ((axisNumber < 0) || (axisNumber>this.dim)) {
            new DimensionalityError(SDBError.Fatal, "Argument specifies non-existing axis.");
            return 0;
        }
        else {
            return axes[axisNumber];
        }
    }

    private boolean incAxis(int axisNumber, float value){
    
        if (this.ubounds[axisNumber]<value || this.lbounds[axisNumber]>value){
            return false;
        }
        
        axes[axisNumber] = value;
        return true;
    }
    
    /** Set the specified axis to the specified value.*/
    public void setCoord(int axisNumber, float value){
    
        if (this.lbounds[axisNumber]>value || this.ubounds[axisNumber]<value){
            new BoundsError(SDBError.Recoverable, "Argument is out of bounds.");
            return;
        }
        
        axes[axisNumber] = value;
    }
    
    /** Returns the dimensionality of this GIterator. */
//    public int dim() {
//        return this.dim;
//    }
    
    /**Return true if the current position is a valid position in the IndexSpace.*/
    public boolean valid(){
        return this.valid;
    }
    
    /**Increment the GIterator to the next position.*/
    public boolean next() {
        int i=0,axis=0;
        float value=0.0f;
        
        if (this.valid == false) return false;
        for (i=0;i<this.dim;i++) {
            axis=this.ordering.getAxis(this.dim - i - 1);
            value=this.getCoord(axis)+steps[axis];
            if (this.incAxis(axis, value)) {
                return true;
            }
            else {
                if (i==this.dim-1){
                    this.valid=false;
                    for (int j=0;j<this.dim;j++){
                        this.axes[j] = this.ubounds[j];
                    }
                    return false;
                }
                this.resetAxis(axis);
            }
        }
        return false;
    }
    
    /**Return true if this GIterator can be incremented.*/
    public boolean hasMoreElements(){
        
        int i,axis,axis0;
        
        boolean equal=true;
        for (i=1;i<this.dim();i++){
            axis = this.ordering.getAxis(this.dim() - i -1);
            if (this.getCoord(axis) !=  this.ubounds[axis]) {
                equal = false;
                break;
            }
        }
        axis0 = this.ordering.getAxis(this.dim() - 1);
        if ((equal) && ( this.getCoord(axis0)+steps[axis0] 
                        > ubounds[axis0] )) return false;
        
        
        for (i=0;i<this.dim;i++)  {
            if ((this.getCoord(i)>ubounds[i]) || (this.getCoord(i)<lbounds[i])){
                return false;
            }
        } 
        
        return true;
    }
	
	/** Returns true if this Point is contained by the 
		GBounds parameter, or false otherwise. 
		
		If the GBounds has a diifferent dimensionality than
		this Point, then a DimensionalityException is 
		thrown.
	*/
	public boolean containedBy(GBounds db){
	
		
		Point id0=db.getLower();
		Point id1=db.getUpper();
		
		return this.valid && id0.lessThanOrEqual(this) && id1.greaterThanOrEqual(this); 
	}

	
	
	
	
	
	static public void main(String[] args){
	
		
		Point lower1 = new	Point(0,0);
		Point upper1 = new	Point(4,3);
		//Point lower2 = new	Point(0,0);
		//Point upper2 = new	Point(1,2);

		GBounds GBounds1;
         //GBounds GBounds2;
		//int isorderingarray[]={1,0};
		int dsorderingarray[]={0,1};
		//AxisOrdering isordering=new AxisOrdering(isorderingarray);
		AxisOrdering dsordering=new AxisOrdering(dsorderingarray);
		GBounds1 = new GBounds(lower1,upper1);
		//GBounds2 = new GBounds(lower2,upper2);
		float steps[] = {1,1}; 
        GIterator iter1 = new GIterator(GBounds1,dsordering,steps);
        
        for (int i=0;i<30;i++){
            System.out.println(iter1+"  ");
            iter1.next();
        
        }
        /*
        iter1.init();
        System.out.println(iter1);
        while (iter1.valid()){
            
            System.out.println(iter1);
            iter1.next();
        }
//        for(iter1.init();iter1.valid();iter1.next()){
//            System.out.println(iter1);
//        }

*/
        System.out.println("end");

	} //end main()




		
	

}
