/* 
	Geometry.java
	Copyright 2000 University of New Hampshire
	Durham, NH 03824
*/

package edu.unh.sdb.lattice; 
import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;

/**
 * This public class is responsible for mapping a DataID to a location in the geometric space.
 *  For regular data, it can do this mapping with a simple calculation. For
 *  unstructured data, it must directly access the DataSet to retrieve the attributes being
 *  used as dimensions.
 */
public abstract class Geometry{
	
    protected int dimensionality;
	protected GBounds bounds;
    protected Partitioning partition;
	protected boolean isCellGeometry;
	
	/** This constructor creates a new domain appropriate
	 *  for the dataset described by the argument.
	 */
	public Geometry(GBounds bounds, Partitioning p, boolean isCellGeometry){

        this.bounds=(GBounds) bounds.clone();
        this.dimensionality=this.bounds.dim();
        this.partition=p;
		this.isCellGeometry=isCellGeometry;
	}
	

	/** returns dimensionality of the geometry */
	public final int dim(){
	
		return this.dimensionality;
	}
	
	/** returns true is this geometry is a cell geometry,
		false otherwise. */
	public boolean isCellGeometry(){
	
		return this.isCellGeometry;
	}
	
	
	/** returns the geometry bounds*/
	public final GBounds getBounds(){
	
		return this.bounds;
	}
    
    public final Partitioning getPartitioning(){
    
        return this.partition;
    }
    
    public IndexSpaceID map(Point p){
    
        return this.partition.map(p);
    }
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        String indent2=indent+Globals.toStringIndent;
        StringBuffer result=new StringBuffer(indent+"dimensionality="+this.dimensionality+"\n");
        
        result.append(indent+"bounds=");
        result.append(bounds+"\n");
        result.append(indent+"partitioning=\n");
        result.append(this.partition.toString(indent2)+"\n");
        
        return new String(result);
    }

}

