//
//  PartitionedSplitCellGroupGrid.java
//  Unstructured
//
//  Created by Philip Rhodes on Mon Aug 12 2002.
//  Copyright (c) 2001 University of New Hampshire. All rights reserved.
//

package edu.unh.sdb.lattice;

import edu.unh.sdb.datasource.*; 
import edu.unh.sdb.common.*;


public class PartitionedSplitCellGroupGrid extends SplitCellGroupGrid {
    
    public PartitionedSplitCellGroupGrid(Partitioning partitioning, AxisOrdering ordering,  String basePointInfoFileName, String baseVIndicesFileName, RecordSpec vIndicesRS, RecordSpec pointRS, RecordSpec dataRS){
    
	
		super(partitioning, ordering, basePointInfoFileName, baseVIndicesFileName, vIndicesRS, pointRS, dataRS);

    }
          
	
	protected void makeCellGroups(){
        
		ISBounds pBounds=this.partitioning.getISBounds();

        this.groups=new CellGroup[pBounds.volume()];
        
        GBoundsIterator partitionIterator=this.partitioning.getPartitionIterator();
        ISIterator tIter=new ISIterator(pBounds);
        

        for(partitionIterator.init(); partitionIterator.valid(); partitionIterator.next(),tIter.next()){
			
			groups[(int)AxisOrdering.computeOffset(tIter, this.axisWeights)]=new PartitionedFileArrayCellGroup(partitionIterator.copyBounds(),
																						dataFileName(this.baseVIndicesFileName,tIter)+".tri.xfdl",
																						this.vIndicesRS,
																						dataFileName(this.basePointInfoFileName,tIter)+".data.xfdl",
																						this.pointRS, 
																						this.dataRS);
        }
		
	}
	
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        int i,numCellGroups=this.groups.length;
        String indent2=indent+Globals.toStringIndent; 
        StringBuffer result=new StringBuffer(indent+"{ PartitionedSplitCellGroupGrid: contains ");

        
        result.append(numCellGroups);
        result.append("cell Groups\n");
        
        result.append(indent2+"geometric bounds=");
        result.append(this.partitioning.getGBounds().toString(indent2)+"\n");
        result.append(indent2+"topological bounds=");
        result.append(this.partitioning.getISBounds().toString(indent2)+"\n");

        for(i=0;i<numCellGroups;i++){
            result.append(indent2+i);
            result.append(") ");
            result.append(this.groups[i].toString(indent2));
            result.append("\n");
        }
        
        result.append(indent2+"}");
        return new String(result);
    }

}
