//
//  RegularCellPartitioning.java
//  MRLatticeTest
//
//  Created by Philip Rhodes on Sun Feb 16 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//
//------------------- Edit history --------------------------------------
// 10/21/05 rdb: removed unused variables
//-----------------------------------------------------------------------

package edu.unh.sdb.lattice;

import edu.unh.sdb.datasource.*; 
import edu.unh.sdb.common.*; 


public class RegularCellPartitioning extends Partitioning {

    protected float [] geomPartitionDimensions;

    public RegularCellPartitioning(ISBounds topoBounds, GBounds geomBounds){
    
        super(topoBounds,geomBounds);
        
        int i,d=this.dimensionality;
        
        this.geomPartitionDimensions=new float[d];
        
        
        //IndexSpaceID tDimensions=this.indexBounds.getDimensions();
        //Point gDimensions=this.geomBounds.getDimensions();
        
        for(i=0;i<d;i++){
            //this.geomPartitionDimensions[i]=(gDimensions.getCoord(i)) / (tDimensions.getCoord(i)-1);
			this.geomPartitionDimensions[i]=(this.geomUpperBounds[i]-this.geomLowerBounds[i])/
											(this.indexUpperBounds[i]-this.indexLowerBounds[i]);
        }

    }
    
    /** Map the point onto the partitioning If p falls on a partition boundary, we choose the
        partition with the lower index. */
    public final IndexSpaceID map(Point p){

        return this.mapLow(p); 
    }
    
    
    public final ISBounds map(GBounds gBounds){
    
        return new ISBounds(this.mapHigh(gBounds.getLower()),this.mapLow(gBounds.getUpper()));
    }

    /** Map the point onto the partitioning. If p falls on a partition boundary, we choose the
        partition with the lower index. */
    public final IndexSpaceID mapLow(Point p){
    
        int i;
        int []  cellIndices=new int[this.dimensionality];
        float cellIndexFloat;
        //IndexSpaceID tDimensions=this.indexBounds.getDimensions();
        int upperBound;
        //float axis;
                
        for(i=0;i<this.dimensionality;i++){
        
            //axis=p.getCoord(i); // rdb axis never used??
        
            cellIndexFloat=	(p.getCoord(i) - this.geomLowerBounds[i])/
							this.geomPartitionDimensions[i] +
							this.indexLowerBounds[i];

            cellIndices[i] = (cellIndexFloat < 1.0) ? 0 : (int)Math.ceil(cellIndexFloat-1.0);
            
            //upperBound=tDimensions.getCoord(i);
			upperBound=this.indexUpperBounds[i];
            cellIndices[i] = ( cellIndices[i] >= upperBound ) ? upperBound  : cellIndices[i]  ;

        }
        
        return new IntegerIndexSpaceID(cellIndices);
    }

    
    /** Map the point onto the grid. If p falls on a partition boundary, we choose the
        partition with the Higher index. */    
    public final IndexSpaceID mapHigh(Point p){
    
        int i;
        int []  cellIndices=new int[this.dimensionality];
        
        for(i=0;i<this.dimensionality;i++){
            
            cellIndices[i]=(int) ((p.getCoord(i) - this.geomLowerBounds[i])/
								  this.geomPartitionDimensions[i] +
								  this.indexLowerBounds[i]
								 );

        }
        
        return new IntegerIndexSpaceID(cellIndices);
    }

    public  GBoundsIterator getPartitionIterator(){
	
		GBounds iteratorSpaceBounds=new GBounds(
											new Point(this.geomLowerBounds),
											new Point(this.geomUpperBounds)
										);
 	
        return new GBoundsIterator(	iteratorSpaceBounds,
									this.geomPartitionDimensions);
    }
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        String indent2=indent+Globals.toStringIndent;
        int i,d=this.dimensionality;
        
        StringBuffer result=new StringBuffer(indent+"{ RegularCellPartitioning:\n");
        result.append(super.toString(indent2)+"\n");
        
        result.append(indent2+"geometric partition dimensions=<");
        for(i=0;i<d-1;i++){
            result.append(this.geomPartitionDimensions[i]+", ");
        }
        result.append(this.geomPartitionDimensions[i]+">\n");
        result.append(indent+"}");
        return new String(result);
    }

    

}



