//
//  RegularPartitioning.java
//  Unstructured
//
//  Created by Philip Rhodes on Sat Oct 05 2002.
//  Copyright (c) 2002 University of New Hampshire. All rights reserved.
//

package edu.unh.sdb.lattice;

import edu.unh.sdb.datasource.*; 
import edu.unh.sdb.common.*; 


public class RegularPartitioning extends Partitioning {

    protected float [] geomPartitionDimensions;

    public RegularPartitioning(ISBounds topoBounds, GBounds geomBounds){
    
        super(topoBounds,geomBounds);
        
        int i,d=this.dimensionality;
        
        this.geomPartitionDimensions=new float[d];
        
        
        //IndexSpaceID tDimensions=this.indexBounds.getDimensions();
        //Point gDimensions=this.geomBounds.getDimensions();
        
        for(i=0;i<d;i++){
		
            // this.geomPartitionDimensions[i]=(gDimensions.getCoord(i)) / (tDimensions.getCoord(i));
			
			this.geomPartitionDimensions[i]=(this.geomUpperBounds[i] - this.geomLowerBounds[i]) / 
											(this.indexUpperBounds[i] - this.indexLowerBounds[i] + 1);
        }

    }
    
    /** Map the point onto the partitioning If p falls on a partition boundary, we choose the
        partition with the lower index. */
    public final IndexSpaceID map(Point p){

        return this.mapLow(p); // Let's use mapLow as the default mapping.
    }
    
    
    public final ISBounds map(GBounds gBounds){
    
        return new ISBounds(this.mapHigh(gBounds.getLower()),this.mapLow(gBounds.getUpper()));
    }

    /** Map the point onto the partitioning. If p falls on a partition boundary, we choose the
        partition with the lower index. */
 /*   public final IndexSpaceID mapLow(Point p){
    
        int i;
        int []  cellIndices=new int[this.dimensionality];
        float cellIndexFloat;
        float axis;
                
        for(i=0;i<this.dimensionality;i++){
        
            axis=p.getCoord(i);
        
            cellIndexFloat=p.getCoord(i)/(this.geomPartitionDimensions[i]);

            cellIndices[i] = (cellIndexFloat < 1.0) ? 0 : (int)Math.ceil(cellIndexFloat-1.0);
        }
        
        return new IntegerIndexSpaceID(cellIndices);
    }*/
	
    
    /** From RegularCellPartitioning.java */
    public final IndexSpaceID mapLow(Point p){
    
        int i;
        int []  cellIndices=new int[this.dimensionality];
        float cellIndexFloat;
        //float axis;
        //IndexSpaceID tDimensions=this.indexBounds.getDimensions();
        int upperBound,lowerBound;
                
        for(i=0;i<this.dimensionality;i++){
        
            //rdb axis never used??
           //axis=p.getCoord(i);
        
            cellIndexFloat=(p.getCoord(i)-this.geomLowerBounds[i])/
							this.geomPartitionDimensions[i] +
							this.indexLowerBounds[i];

            cellIndices[i] = (cellIndexFloat < 1.0) ? 0 : (int)Math.ceil(cellIndexFloat-1.0);
            
            //upperBound=tDimensions.getCoord(i);
			upperBound=this.indexUpperBounds[i];
            cellIndices[i] = ( cellIndices[i] >= upperBound ) ? upperBound  : cellIndices[i];
			
			
			lowerBound=this.indexLowerBounds[i];
			cellIndices[i] = ( cellIndices[i] <= lowerBound ) ? lowerBound  : cellIndices[i];

        }
        
        return new IntegerIndexSpaceID(cellIndices);
    }

    
    /** Map the point onto the grid. If p falls on a partition boundary, we choose the
        partition with the Higher index. */    
    public final IndexSpaceID mapHigh(Point p){
    
        int i;
        int []  cellIndices=new int[this.dimensionality];
		int upperBound,lowerBound;
		
        for(i=0;i<this.dimensionality;i++){
            
            cellIndices[i]=(int) ((p.getCoord(i)-this.geomLowerBounds[i])/
									this.geomPartitionDimensions[i] +
									this.indexLowerBounds[i]
								 );


			upperBound=this.indexUpperBounds[i];
            cellIndices[i] = ( cellIndices[i] >= upperBound ) ? upperBound  : cellIndices[i];
			
			
			lowerBound=this.indexLowerBounds[i];
			cellIndices[i] = ( cellIndices[i] <= lowerBound ) ? lowerBound  : cellIndices[i];




        }
        
        return new IntegerIndexSpaceID(cellIndices);
    }




    public  GBoundsIterator getPartitionIterator(){
    
		GBounds iteratorSpaceBounds=new GBounds(
											new Point(this.geomLowerBounds),
											new Point(this.geomUpperBounds)
										);
		int i, d=this.dimensionality;
		int[] dimensions=new int[d];
		
		
		for(i=0; i<d; i++){
		
			dimensions[i] = (this.indexUpperBounds[i] - this.indexLowerBounds[i] + 1);
		}
		
		
        return new GBoundsIterator(	iteratorSpaceBounds, dimensions);
    }


    public  GBoundsIterator getPartitionIterator(GBounds subSpace, GBoundsIterator iter){
    
		iter.set(subSpace, this.geomLowerBounds, geomPartitionDimensions);
        
        return iter;
    }
    
	
	
	
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){

        String indent2=indent+Globals.toStringIndent;
        int i,d=this.dimensionality;
        
        StringBuffer result=new StringBuffer(indent+"{ RegularPartitioning:\n");
        result.append(super.toString(indent2)+"\n");
        
        result.append(indent2+"geometric partition dimensions=<");
        for(i=0;i<d-1;i++){
            result.append(this.geomPartitionDimensions[i]+", ");
        }
        result.append(this.geomPartitionDimensions[i]+">\n");
        result.append(indent+"}");
        return new String(result);
    }



    public static void main(String argv){
    
        ISBounds topoBounds=new ISBounds(new int[]{2,2});
        GBounds  gBounds= new GBounds(new float[]{0,0}, new float[]{10,10});
        
        RegularPartitioning p=new RegularPartitioning(topoBounds, gBounds);
        
        GBoundsIterator iter=p.getPartitionIterator();

        
        for(iter.init(); iter.valid(); iter.next()){
        
            System.out.println(iter);
        }
        
        
        System.out.println("\n*********\n");
        topoBounds = new ISBounds(new int[]{4,4});
        gBounds = new GBounds(new float[]{10,10});
        
        p=new RegularPartitioning(topoBounds, gBounds);
        
        GBounds subSpace=new GBounds(new float[] {3,3}, new float[]{8,8});
        
        p.getPartitionIterator(subSpace, iter);


        for(iter.init(); iter.valid(); iter.next()){
        
            System.out.println(iter);
        }
    
    }
    

}



