//  UnstructuredCell.java
//  Unstructured
//
//  Created by Philip Rhodes on Wed Jul 17 2002.
//  Copyright (c) 2001 University of New Hampshire. All rights reserved.
//
// -------------------- Edit history since 01/10/05 -----------------------------
//
// 01/10/05 rdb: Replaced FloatDatum with IntDatum for index information
//
// ------------------------------------------------------------------------------

package edu.unh.sdb.lattice;

import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;

public abstract class UnstructuredCell extends Cell {
    
    //protected Cell 	neighbors[]; // will be CellIDs
	IndexSpaceID [] vIndices;

	protected UnstructuredCell(){
	
		super(null, null);
		this.vIndices=null;
	}

    public UnstructuredCell(Point [] coords, Datum [] data){

        super(coords, data);
    }

	public UnstructuredCell(int numAttributes, int dimensionality, int numVertices){
	
		super(numAttributes, dimensionality, numVertices);
		
		this.vIndices=new IntegerIndexSpaceID[numVertices];
		
		for(int i=0; i<numVertices; i++){
			
			this.vIndices[i]=new IntegerIndexSpaceID(new int[dimensionality]);
		}
		
	}


    public UnstructuredCell(  DataSource vIndicesDS, DataSource coordsDS, 
							RecordSpec vIndicesRS, RecordSpec coordsRS, RecordSpec dataRS,
							int dimensionality, int numVertices,
							IndexSpaceID cellID){
	
		super(null, null);
		
		int j;
		//rdb Datum vIndicesDatum = new FloatDatum(numVertices);
		Datum vIndicesDatum = vIndicesDS.getRecordDescriptor().createDatum();
		this.vIndices       = new IntegerIndexSpaceID[numVertices];
		Point[] vertices    = new Point[numVertices];
		Datum[] data        = new Datum[numVertices];
		Datum pointData;

		vIndicesDatum = vIndicesDS.datum(cellID).subset(vIndicesRS);
		
		for(j=0; j<numVertices; j++){
			
			//rdb this.vIndices[j] = new IntegerIndexSpaceID((int)vIndicesDatum.getFloat(j));
			this.vIndices[j] = new IntegerIndexSpaceID( vIndicesDatum.getInt(j) );
			
			// Retrieve both point and data info
			pointData   = coordsDS.datum( this.vIndices[j] );
			vertices[j] = new Point( pointData.subset(coordsRS) );
			data[j]     = pointData.subset( dataRS );                
		}
		
        //super(vertices,data);
		
		super.setVertexCoords(vertices);
		super.setData(data);
    }



	public void assign(	Indexable vIndicesDS, Indexable coordsDS, 
						RecordSpec vIndicesRS, RecordSpec coordsRS, RecordSpec dataRS,
						int dimensionality, int numVertices, 
						IndexSpaceID cellID){

		int j;
		Datum vIndicesDatum= new FloatDatum(numVertices);
		this.vIndices=new IntegerIndexSpaceID[numVertices];
		Point[] vertices=new Point[numVertices];
		Datum[] data=new FloatDatum[numVertices];
		Datum pointData;


		vIndicesDatum=vIndicesDS.datum(cellID).subset(vIndicesRS);

		for(j=0; j<numVertices; j++){
			
			//rdb this.vIndices[j] = new IntegerIndexSpaceID((int)vIndicesDatum.getFloat(j));
			this.vIndices[j] = new IntegerIndexSpaceID( vIndicesDatum.getInt(j) );
			
			// Retrieve both point and data info
			pointData=coordsDS.datum(this.vIndices[j]);
			
			vertices[j]=new Point(pointData.subset(coordsRS));
			data[j]=pointData.subset(dataRS);                
		}
		
		this.setDimensionality(dimensionality);

		super.setVertexCoords(vertices);
		super.setData(data);
	}
	
	IndexSpaceID [] getVertexIndices(){
	
		return this.vIndices;
	}
    
}

