//
//  WriteableInCoreArrayCellGroupGrid.java
//  Unstructured
//
//  Created by Philip Rhodes on Sep 22 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//
//------------------- Edit history --------------------------------------
// 10/21/05 rdb: removed unused variables
//-----------------------------------------------------------------------

package edu.unh.sdb.lattice;

import edu.unh.sdb.datasource.*; 
import edu.unh.sdb.common.*; 
import java.io.*;

public class WriteableInCoreArrayCellGroupGrid extends CellGroupGrid {


	//protected int [] vIndices;
	//protected float [] coords;
	//protected float [] data;
	//protected int numAttributes;
	protected int numCells;
protected int lastIndex=-1;

	public WriteableInCoreArrayCellGroupGrid(Partitioning partitioning, AxisOrdering ordering){
	
		super(partitioning, ordering);

/*
		this.vIndices= (vIndices == null) ? new int[1] : vIndices;
		
		this.coords=coords;
		this.data=data;
		this.numAttributes=numAttributes;
*/

//System.out.println(partitioning);
		this.makeCellGroups();
		this.numCells=0;
	}






	protected void makeCellGroups(){
	
		ISBounds pBounds=this.partitioning.getISBounds();
        
        this.groups=new CellGroup[pBounds.volume()];

        GBoundsIterator partitionIterator=this.partitioning.getPartitionIterator();
		ISIterator tIter=this.partitioning.getISIterator();
		

        for(partitionIterator.init(); partitionIterator.valid(); partitionIterator.next(), tIter.next()){
			
			groups[(int)AxisOrdering.computeOffset(tIter, this.axisWeights)]=new WriteableInCoreArrayCellGroup(
																	partitionIterator.copyBounds()
																);
			           
        }
			
			
	}



	public int write(String baseFileName){
	
	
		this.writeFDLFile(baseFileName);
		
		
		String subgridFileName=new String(baseFileName + ".subgrid");
		
		ISBounds pBounds=this.partitioning.getISBounds();
		int i,d=pBounds.dim();
		
		ISIterator iter=new ISIterator(pBounds);
		RecordDescriptor rd=new RecordDescriptor(new AttributeDescriptor [] {new AttributeDescriptor(StorageType.Int, "cellIndex")});
		
		

		OneDDataSink ds=new OneDDataSink(
							subgridFileName, 
							rd);
							//new ISBounds(new IntegerIndexSpaceID(1 + d + pBounds.volume() + this.getNumCells())));
							
		ds.activate();
		
		
		Datum datum=new FloatDatum(1);
		IndexSpaceID dimensions=pBounds.getDimensions();
		int [] dimensionsArray=dimensions.getCoords(new int[d]);
		
		datum.setFloat(0, d);
		ds.write(datum);
		
		for(i=0; i<d; i++){
		
			datum.setFloat(0, dimensionsArray[i]);
			ds.write(datum);
		}
				
		
		for(iter.init(); iter.valid(); iter.next()){
		
			((WriteableInCoreArrayCellGroup) groups[(int)AxisOrdering.computeOffset(iter, this.axisWeights)]).write(ds);
		
		}
		
        ds.deactivate();
        
		return 0;
	}



	void dump(int [] vIndices, float [] coordFloats, float [] dataFloats, int numDataAttributes) {


		ISBounds pBounds=this.partitioning.getISBounds();
		
		ISIterator iter=new ISIterator(pBounds);

		for(iter.init(); iter.valid(); iter.next()){
		
			((WriteableInCoreArrayCellGroup) groups[(int)AxisOrdering.computeOffset(iter, this.axisWeights)]).dump(vIndices, coordFloats, dataFloats, numDataAttributes, this.dimensionality);
		
		}
	}
	
	int getTotalGroupNumCells(){
	
		ISBounds pBounds=this.partitioning.getISBounds();
		int numCells=0;
		
		ISIterator iter=new ISIterator(pBounds);


		for(iter.init(); iter.valid(); iter.next()){
		
			numCells += ((WriteableInCoreArrayCellGroup) groups[(int)AxisOrdering.computeOffset(iter, this.axisWeights)]).getNumCells();
		
		}
	
		return numCells;
	}

    public void writeFDLFile(String baseFileName){
		
		try {
		
			String fdlFileName=new String(baseFileName+".subgrid.xfdl");
			String subgridFileName=new String(baseFileName + ".subgrid");
			
			ISBounds pBounds=this.partitioning.getISBounds();
			int d=pBounds.dim();
			
			FileWriter fw=new FileWriter(fdlFileName);
			
			
			
			fw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
			fw.write("<!DOCTYPE FileDescriptor SYSTEM \"/Users/rhodes/Documents/Res/dtd/fdl.dtd\">\n");
			fw.write("<FileDescriptor filename= \""+subgridFileName+"\"  filetype=\"binary\" recordsize=\"4\">\n");
			
			fw.write("<Bounds lower= \"0\"    upper= \""+(1 + d + pBounds.volume() + this.getTotalGroupNumCells() - 1)+"\"/>\n");
	
		
			fw.write("<Field fieldname= \"v0\"   fieldtype= \"int\"/>\n");
	
			fw.write("</FileDescriptor>\n");
		
			fw.close();
		} catch (IOException e){
		
			new FileError(SDBError.Recoverable,"Couldn't write fdl file.");
		
		}
	
	
	}

	public void register(TriangleCell c, int cellIndex){

		ISBounds tBoundingBox;
		GBounds gBoundingBox;
		ISIterator tIterator;
		boolean foundIntersection=false, registered=false;

//System.out.println("Subgrid registering cell="+c);
		
		gBoundingBox=c.boundingBox();
		tBoundingBox=this.partitioning.map(gBoundingBox);
		tIterator=new ISIterator(tBoundingBox);
		
		for(tIterator.init(); tIterator.valid(); tIterator.next()){
		
			registered = ((WriteableInCoreArrayCellGroup)this.mapToCellGroup(tIterator)).register(c, cellIndex);

			foundIntersection = foundIntersection || registered;						
		}
		
		if(foundIntersection){
			
			this.numCells++;
		} else {
			System.out.println("******Subgrid:Couldn't register cell="+c);
		}

	}


	
	public Cell findEnclosingCell(Point p){
	
		new InternalFaultError(SDBError.Recoverable, "Method not implemented.");
		
        return null;

	}


	public int getNumCells(){
	
		return this.numCells;
	}

	
}
