/*
	Created by Baoqiang Yan to facilitate the control info communication
	Used in the cluster code, but it is probably good to separate it out.
	
	Changes:
	1. Made SendControlInfo return an integre to denote status 
*/
package edu.unh.sdb.remote;

import java.net.*;
import java.io.*;
import java.util.*;

public class RemoteMessenger
{
		InetAddress peerIP = null;
      boolean moribund = false;
      //int maxSocketTime = 30 * 60 * 1000;  // half an hour
    /** TCP Control Traffic to client.
     *	We need to add the desired IP so that it only send
     *	to the target address.
     */
    public synchronized int PassivelySendControlInfo(ServerSocket ssock, String s, String targetIP)
    {
    
      //System.out.println("Starting PassivelySendControlInfo:");
       String ctrlPacket;
			 try{
		       //System.out.println("SendControlInfo(): calling accept()");
		       //System.out.println("listening at " + ssock.getInetAddress()+"-"+ssock.getLocalPort());
		       Socket connectionSocket = null;
		       while(true){
			       connectionSocket = ssock.accept();
			       //System.out.println("SendControlInfo(): came back from accept(), connectionSocket.isBound()=" + connectionSocket.isBound());
                
      
			       if(connectionSocket.getInetAddress().getHostAddress().equalsIgnoreCase(targetIP) || connectionSocket.getInetAddress().getHostName().equalsIgnoreCase(targetIP))
			       //if(connectionSocket.getInetAddress().getHostName().equalsIgnoreCase(targetIP))

			      		break;//picks up the client address and assigns to "ip".]
                     
                     
			       connectionSocket = null;
			     }
		       DataOutputStream outToClient = new DataOutputStream(connectionSocket.getOutputStream());
		       //System.out.println("DataOutputStream is created");
		       ctrlPacket = s + '\n';
		       outToClient.writeBytes(ctrlPacket);
		       //System.out.println("Control msg \""+s + "\" was sent..");
		       connectionSocket.close();
             
             //System.out.println("Leaving PassivelySendControlInfo: no error.");
		       return 0;
	     }catch(IOException ioe){
        
            if( !moribund){

					System.out.println("SendControlInfo() in RemoteMessenger Socket - IO exception" + ioe.toString());	
            } else {
            
               //System.out.println("While Moribund: SendControlInfo() in RemoteMessenger Socket - IO exception" + ioe.toString());
            }
            
            //System.out.println("Leaving PassivelySendControlInfo: exception occurred.");
            return 1;
			 }
    }


    /** Client To Server Control Information for missing packets.
    */
    public synchronized String ActivelyReceiveControlInfo(String peerIP, int peerPort)
    {
        InetAddress ip = null;
        try{
         	ip = InetAddress.getByName(peerIP);
        }catch(UnknownHostException uhe){
					System.out.println("ActivelyReceiveControlInfo() in RemoteMessenger - UnknownHostException: " + uhe.toString());
					return null;	
		  }

        return ActivelyReceiveControlInfo(ip, peerPort);
    }


    public synchronized String ActivelyReceiveControlInfo(InetAddress peerIP, int peerPort)
    {
      //System.out.println("Starting ActivelyReceiveControlInfo:");
      String ctrl_info = null;
      for(int i=0; i<3 && !moribund; i++){
		  try{
	         //System.out.println("Getting control info from "+peerIP+ " at port " + peerPort);
	         //Socket sock = new Socket(peerIP, peerPort);
	         
					 SocketAddress sockaddr = new InetSocketAddress(peerIP.getHostName(), peerPort);
					 Socket sock = new Socket();
					 sock.connect(sockaddr, 2000);	         
	         //System.out.println("Connected to "+peerIP+ " at port " + peerPort);
            
	         sock.setSoTimeout(10 * 60 * 1000);
	         //System.out.println("Creating bufferreader object ...");
	         BufferedReader inFromClient = new BufferedReader(new InputStreamReader(sock.getInputStream()));
	         
	         //this.isListening = true; // Now we're listening for client input.
	         ctrl_info = inFromClient.readLine();
   	      //System.out.println("ctrl_info received: " + ctrl_info);
	         sock.close();
	         return ctrl_info;
        }catch(IOException ioe){
        
            if( !moribund){
            
					System.out.println("ActivelyReceiveControlInfo() in RemoteMessenger - Socket IO exception: " + ioe.toString());
               ioe.printStackTrace();
		try{
			Thread.sleep(1000);
		}catch(Exception e){System.out.println("Error while sleeping in remote messenger");}
            }
		  }
		}
     //System.out.println("cntrl_info received: "+ ctrl_info);
     //System.out.println("Leaving ActivelyReceiveControlInfo:");
     return ctrl_info;
    }

	 //ReceiveControlInfo
	 //Only accept msg from certain nodes, not any computer
    public String PassivelyReceiveControlInfo(ServerSocket ssock, Set allowedNodes)
    {
        String msg = null;
		  try{
		      Socket sock = null;
	         ssock.setSoTimeout(0); //wait forever until the whole application is done
		      while(true){
			      sock = ssock.accept();
			      //System.out.println("SendControlInfo(): came back from accept(), connectionSocket.isBound()=" + connectionSocket.isBound());
			      if(allowedNodes.contains(peerIP=sock.getInetAddress()))
			     		break;//picks up the client address and assigns to "ip".
			      sock = null;
			   }

	         sock.setSoTimeout(30*1000); //wait at most 30 seconds fot reading the msg
	         //System.out.println("Creating bufferreader object ...");
	         BufferedReader inFromClient = new BufferedReader(new InputStreamReader(sock.getInputStream()));
	         
	         //this.isListening = true; // Now we're listening for client input.
	         msg = inFromClient.readLine();
   	      //System.out.println("ctrl_info received: " + ctrl_info);
	         sock.close();
        }catch(IOException ioe){
        
            if( !moribund){
         
					System.out.println("ReceiveControlInfo() in RemoteMessenger - Socket IO exception" + ioe.toString());	
            }
			  }
        //System.out.println("cntrl_info received: "+ ctrl_info);
        return msg;
    }
	
	 //The hashtable contain pairs of ip address and the port number where to send msg to.
	 //The content of this hashtable should be the same as the one used in ReceiveBroadcastMsg
	 //Because only server side needs to actually use port number (Am I gibberishing?)
	 //Also, different nodes may use different port numbers for broadcasting, though here in a
	 //cluster environment, the same port number would be used.
    public int ActivelySendControlInfo(String msg, String targetIP, int targetPort)
    {
        InetAddress ip = null;
        try{
         	ip = InetAddress.getByName(targetIP);
        }catch(UnknownHostException uhe){
					System.out.println("ActivelySendControlInfo() in RemoteMessenger - UnknownHostException: " + uhe.toString());
					return 1;	
		  }
		  return ActivelySendControlInfo(msg, ip, targetPort);
//		  try{
//				//To make sure fairness on the order to send msg, we use an Iterator on Set Object here
//				Iterator keyIterator = targetNodes.keySet().iterator();
//				InetAddress targetIP 	= null;
//				int 			targetPort	= 0;
//				Socket sock = null;
//	       	DataOutputStream outToClient = null;
//	       	String outmsg = msg;
//				
//				while(keyIterator.hasNext()){
//					targetIP = (InetAddress) keyIterator.next();
//					targetPort = targetNodes.get(targetIP).intValue();
//					sock = new Socket(targetIP, targetPort);
//					sock.setSoTimeout(30 * 1000);
//
//
//		       	outToClient = new DataOutputStream(sock.getOutputStream());
//		       	outmsg = outmsg + '\n';
//		       	outToClient.writeBytes(outmsg);
//					outToClient.close();
//					sock.close();
//				}
//				return 0;
//        }catch(IOException ioe){
//					System.out.println("SendBroadcastMsg() in RemoteMessenger - Socket IO exception" + ioe.toString());
//					return 1;
//		  }
    }

    
    // Original
//     public int ActivelySendControlInfo(String msg, InetAddress targetIP, int targetPort)
//     {
//        
// 		  try{
// 				Socket sock = null;
// 	       	DataOutputStream outToClient = null;
// 	       	String outmsg = msg;
// 				
// 					sock = new Socket(targetIP, targetPort);
// 					sock.setSoTimeout(30 * 1000);
// 
// 
// 		       	outToClient = new DataOutputStream(sock.getOutputStream());
// 		       	outmsg = outmsg + '\n';
// 		       	outToClient.writeBytes(outmsg);
// 					outToClient.close();
// 					sock.close();
// 					return 0;
//         }catch(IOException ioe){
//         
//             if( !moribund){
//             
// 					System.out.println("ActivelySendControlInfo() in RemoteMessenger - Socket IO exception" + ioe.toString());
//                ioe.printStackTrace();
//             }
// 				return 1;
// 		  }
//     }

    public int ActivelySendControlInfo(String msg, InetAddress targetIP, int targetPort)
    {
        
         for(int i=0; i<3 && !moribund; i++){        
           try{
               Socket sock = null;
               DataOutputStream outToClient = null;
               String outmsg = msg;
               
                  sock = new Socket(targetIP, targetPort);
                  sock.setSoTimeout(30 * 1000);
   
   
                  outToClient = new DataOutputStream(sock.getOutputStream());
                  outmsg = outmsg + '\n';
                  outToClient.writeBytes(outmsg);
                  outToClient.close();
                  sock.close();
                  return 0;
           }catch(IOException ioe){
           
               if( !moribund){
               
                  System.out.println("ActivelySendControlInfo() in RemoteMessenger - Socket IO exception" + ioe.toString());
                  ioe.printStackTrace();
                try{
                        Thread.sleep(1000);
                }catch(Exception e){System.out.println("Error while sleeping in remote messenger");}

               }
           }
        }
        return 1;
    }

    
   // PJR
//   public int ActivelySendControlInfo(String msg, InetAddress targetIP, int targetPort)
//   {
//
//      Exception e = null;
//      
//      for(int i=0; i<3 && !moribund; i++){
//
//         try{
//            
//            Socket sock = null;
//            DataOutputStream outToClient = null;
//            String outmsg = msg;
//
//            sock = new Socket(targetIP, targetPort);
//            sock.setSoTimeout(30 * 1000);
//
//
//            outToClient = new DataOutputStream(sock.getOutputStream());
//            outmsg = outmsg + '\n';
//            outToClient.writeBytes(outmsg);
//            outToClient.close();
//            sock.close();
//            return 0;
//            
//         }catch(IOException ioe){
//          
//            try {Thread.sleep(1000);} catch(Exception tie) {}
//            
//            e = ioe;
//         }
//      }
//      
//      
//      if( !moribund){
//         
//         System.out.println("ActivelySendControlInfo() in RemoteMessenger - 3 Failures" + e.toString());
//         e.printStackTrace();
//      }
//      
//      return 1;
//   }
    
    
    
    public InetAddress getRequestAddress(){
    	return peerIP;
    }
    
    /** Tell this object that its thread is about to be interrupted. */
    public void setMoribund(){
    
      this.moribund = true;
    }
}
