/*
	To facilitate the data transfer using TCP for among compute nodes
	within a cluster. It also gives us more options doing data transfer.
	
   Author:  Baoqiang Yan
   Date:    2.25.2007
*/

package edu.unh.sdb.remote;

import java.net.*;
import java.io.*;

public class TCPReceiver extends DataReceiver
{
  Socket sock;
	InetSocketAddress servAdd;
	DataInputStream s_in;
	
	//the client ip is used to make sure the connected client is the right one
	public TCPReceiver()
	{
		serverIP 				= null;
		serverDataPort	= 0;
		sock = null;
		servAdd = null;
		s_in = null;
	}

	public int connect(){
     try{
	     sock = new Socket();
	     servAdd = new InetSocketAddress(serverIP, serverDataPort);
	     //System.out.println("Trying to connect to "+serverIP+" at port "+serverDataPort);
	     sock.connect(servAdd, 0);
	     sock.setSoTimeout( 30 * 1000); //half minute
	     s_in = new DataInputStream(sock.getInputStream());
	     return 0;
	   }catch(IOException e ){
         System.out.println("Exception while connecting to "+serverIP+" at dataport "+serverDataPort);
         e.printStackTrace();
         freeResource();
         return 1;
     }
	}
		
	public void freeResource(){
		 try{
			 if(s_in!=null) s_in.close();
		   if(sock!=null)	sock.close();
		 }catch(Exception e){
		 	 System.out.println("Exception while closing sockets");
		 	 System.exit(1);
		 }
	}
	
	//The logic to tell client about size is left to the app
	public synchronized int ReceiveData(byte dst[], int size){
			if(s_in==null) {
				int retryTimes = 5, count=0;
				int status = 1;
				while(status==1 && count<retryTimes){
					try{
						Thread.sleep(1000);
					}catch(Exception e){System.out.println("Interrupted while sleeping");}
					status = connect();
					count++;
				}
				if(s_in==null){
					System.out.println("Failed after 5 trials, giving up!");
					System.exit(1);
				}
			}
			//System.out.println("TCP Data Receiver Connected to " + serverIP+ " at "+serverDataPort);
      int offset =0;
      //int count =0;
      try{
         while( offset < size )
         {
            offset += s_in.read(dst,offset,size-offset);
            //count ++;
         }
         //System.out.println("TCP Data Receiver received " + size + "bytes");
         return 0;
      }catch(IOException e ){
         System.out.println(" IO exeception " + e.toString() );
         System.out.println(" ------stack trace----- " );
         e.printStackTrace();
         System.out.println(" ---------------------- " );
         freeResource();
         return 1;
      }
	}	
}