/*
 * UDPSender.java
 * -------------------------- edit history ----------------------------
 * 06/06/08 rdb: Changed invocations of the static UDPSender.InterPacketDelay.wait
 *               method to use correct static reference.
 */
package edu.unh.sdb.remote;
import java.net.*;
import java.io.*;
import java.util.*;

//import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;

public class UDPSender extends Sender
{

	protected int clientDataPort;
	protected int clientControlPort;
	
	protected InetAddress ip;
	protected ServerSocket serverCtrlSocket;
  	protected DatagramSocket ds;

	protected RemoteMessenger rm;
	protected String controlMsg;	

  	protected byte seq_num[]; //array for packet seq nos.

	private int rempart;
  	protected int totalPackets;//variable to hold the total number of packets to be sent.
	
	protected int buffer_size;
	protected int finalDelay;
	protected int initialDelay;
	protected int delayLength;
	
	public UDPSender(){

		this.clientDataPort = Globals.clientDataPort;
		this.clientControlPort = Globals.clientControlPort;
		this.serverDataPort = Globals.serverDataPort;
		//field in super class Sender 
		this.clientIP = null;
		this.ip = null;

		this.serverCtrlSocket = null;

    	this.rm = null;
 		this.controlMsg = null;

		this.seq_num = new byte[4];
		
		this.rempart=0;
		this.totalPackets=0;
		
		this.buffer_size = 0;
		this.finalDelay = 0;
		this.initialDelay = 0;
		this.delayLength = 0;		
	  
	}

	public void setRemoteMessenger(RemoteMessenger rm){
		this.rm = rm;
	}
	
	public void setServerCtrlSocket(ServerSocket ssock){
		this.serverCtrlSocket = ssock;
	}
	
	public void freeResource(){
		//we only need to take care of the data socket since the RM and control socket
		//is freed by the Server itself
		try{
			this.ds.close();
		}catch(Exception e){System.out.println("Error closing the datagram socket in UDPSender");}
	}
	
	public void setServerDataPort(int port){
		this.serverDataPort = port;
		try{
			this.ds = new DatagramSocket(serverDataPort);
			this.ds.setSoTimeout(10 * 60 * 1000);// 10 minute timeout
	  	}
	  	catch(SocketException se){se.printStackTrace();}
	  	//rdb catch(IOException ioe){ioe.printStackTrace();}
	}
	
	public void setClientIP(String newIP){
		this.clientIP = newIP;
		System.out.println("ClientIp is set to "+newIP);
		try{
			this.ip = InetAddress.getByName(newIP);
		}
		catch(UnknownHostException uhe){uhe.printStackTrace();}
	}

	public void setClientIP(InetAddress newAddress){

		this.ip = newAddress;		
		//this.clientIP = newAddress.getHostAddress();
		this.clientIP = newAddress.getHostName();
		System.out.println("ClientIp is set to "+clientIP);
		//System.exit(0);
	}
	
	public void setBufSize(int size){
		buffer_size = size;
	}
	
	public void setInitialDelay(int delay){
		initialDelay = delay;
	}

	public void setDelayLength(int dl){
		delayLength = dl;
	}

	public void setFinalDelay(int fd){
		finalDelay = fd;
	}
	
	/** Fills seq_num[] with (byte) Integer values of group of
	        8 bits from 32 bit binary string of input integer.
	*/
	public static void Seqnum_Encoder(byte seq_num[], int n)
	{
	   seq_num[0] = (byte) ((n & 0xFF000000) >>> 24);//most significant 8 bits.
	   seq_num[1] = (byte) ((n & 0x00FF0000) >>> 16);//next most significant 8 bits.
	   seq_num[2] = (byte) ((n & 0x0000FF00) >>> 8);//next least significant 8 bits.
	   seq_num[3] = (byte)  (n & 0x000000FF);//least significant 8 bits.
	}
	
	public String GetMissingPackets(){
		return rm.ActivelyReceiveControlInfo(clientIP, clientControlPort);
	}
	
  /**Sends data to client.
  */
  public int SendData(byte bytes[], int size)
  {
     int lowerlimit,upperlimit;//indexes for file read position based on packet size.
     int counter;//packet sequence number counter.
     //InterPacketDelay t = new InterPacketDelay();//object used to set interpacket delay.

     //initializing variables.
     lowerlimit = 0;
     upperlimit = 0;
     rempart = 0;
     counter = 0;
     initialDelay = 1;

     //calling the class constructor.
     //RemoteBlockStencilServer serv = new RemoteBlockStencilServer(buffer_size, initialDelay, delayLength, finalDelay);

     byte bytes_with_seq[] = new byte[buffer_size+4];//array with data & seq number at the tail used for packets.

     //System.out.println("Calling  rm.SendControlInfo(serverCtrlSocket, File Size) in SendData().");
     //Already sent the size before calling this function
     //rm.SendControlInfo(serverCtrlSocket, "Size: " + bytes.length);//control packet to let client know of file size.
     //System.out.println("Back from  rm.SendControlInfo(serverCtrlSocket, File Size) in SendData().");


     //whole number of packets into which the file can be broken into.
     int times = bytes.length/buffer_size;
     totalPackets = times;//assigning the total number of packets to totalPackets global variable.

     //if packet size is less than File size.
     if(buffer_size < bytes.length)
     {
        for(int i=0;i<times;i++)
        {
           upperlimit += buffer_size;
           Seqnum_Encoder(seq_num, counter);//encoding the sequence number for a packet.
           //reading a part of "bytes" array into data portion of "bytes_with_seq" array.
           System.arraycopy(bytes,lowerlimit,bytes_with_seq,0,buffer_size);
           //appending a seq no to the tail of a packet.
           System.arraycopy(seq_num,0,bytes_with_seq,buffer_size,4);

           //sending data packets.
           try{
           		/*
           	   if(ds!=null) System.out.println("ds is not null");
           	   if(bytes_with_seq!=null) System.out.println("bytes_with_seq is not null");
           	   DatagramPacket dp = new DatagramPacket(bytes_with_seq,0,buffer_size+4,ip,clientDataPort);
           	   if(dp!=null) System.out.println("Data packet is not null");
           	   
           	   System.out.println(ip);
           	   System.out.println(clientDataPort);
           	   */
	           ds.send(new DatagramPacket(bytes_with_seq,0,buffer_size+4,ip,clientDataPort));
	       }
	       catch(IOException ioe){ioe.printStackTrace();}
           counter ++;//incrementing counter value for next packet.
           lowerlimit = upperlimit;//shifting the reading position by buffer_size.

           if((delayLength != 0) && (counter == delayLength))
           {
              initialDelay = finalDelay;//assignment statement for finalDelay.
           }
           UDPSender.InterPacketDelay.wait(initialDelay);//delay induced between packets.
        }
        if((bytes.length % buffer_size) != 0)
        {
           totalPackets = times + 1;//assigning the total number of packets to totalPackets global variable.
           rempart = bytes.length - lowerlimit;//last fractional packet length.
           //seq number logic.
           Seqnum_Encoder(seq_num, counter);//encoding the sequence number for a packet.
           //reading a part of "bytes" array into data portion of "bytes_with_seq" array.
           System.arraycopy(bytes,lowerlimit,bytes_with_seq,0,rempart);
           //filling the extra spaces in the last packet with *'s.
           if(rempart < buffer_size)
           {
              for(int k=rempart;k<buffer_size;k++)
              {
                 bytes_with_seq[k] = (byte) '*';
              }
           }
           //appending a seq no to the tail of a packet.
           System.arraycopy(seq_num,0,bytes_with_seq,buffer_size,4);

           //sending data packets.
           try{
	           ds.send(new DatagramPacket(bytes_with_seq,0,buffer_size+4,ip,clientDataPort));
	       }
	       catch(IOException ioe){ioe.printStackTrace();}
           counter++;
           UDPSender.InterPacketDelay.wait(initialDelay);//delay induced between packets.
        }
                  
                  
        //System.out.println("Calling ClientToServerControlInfo(missedPacket) in SendData()");
        //controlMsg = rm.ReceiveControlInfo(clientIP, clientControlPort);//connection with client to find out if there are any missing packets.
        controlMsg = GetMissingPackets();//connection with client to find out if there are any missing packets.
        //System.out.println("Back From ClientToServerControlInfo(missedPacket) in SendData()");
		if(!(controlMsg.equals("")))
					PacketRetransmission(bytes, controlMsg);
     }
     else//packet size is greater than or equal to file size.
     {
        totalPackets = 1;//assigning the total number of packets to totalPackets global variable.
        Seqnum_Encoder(seq_num, counter);//encoding the sequence number for a packet.
                                //reading a part of "bytes" array into data portion of "bytes_with_seq" array.
        System.arraycopy(bytes,0,bytes_with_seq,0,bytes.length);
        
        if(bytes.length < buffer_size) //namely, bytes.length!=buffer_size
        {
           //filling the extra spaces in the last packet with *'s.
           for(int k=bytes.length;k<buffer_size-bytes.length;k++)
           {
              bytes_with_seq[k] = (byte) '*';
           }
        }
        
        //appending a seq no to the tail of a packet.
        System.arraycopy(seq_num,0,bytes_with_seq,buffer_size,4);
        
        //sending data packet.
        try{
	        ds.send(new DatagramPacket(bytes_with_seq,0,buffer_size+4,ip,clientDataPort));
	    }
	    catch(IOException ioe){ioe.printStackTrace();}
        counter++;
        UDPSender.InterPacketDelay.wait(initialDelay);//delay induced between packets.
           
        //System.out.println("Calling ClientToServerControlInfo(missedPacket) in SendData()");
        //controlMsg = rm.ReceiveControlInfo(clientIP, clientControlPort);//connection with client to find out if there are any missing packets.
        controlMsg = GetMissingPackets();//connection with client to find out if there are any missing packets.
        if(!(controlMsg.equals("")))
					PacketRetransmission(bytes, controlMsg);
        //System.out.println("Back From ClientToServerControlInfo(missedPacket) in SendData()");
     }
     
          
     rm.PassivelySendControlInfo(serverCtrlSocket, "Buffer Size: -1", clientIP);//end of transfer control word.
     return 0;
  }//end of SendData.
  
  
  /** Logic for retransmitting lost packets.
     */
  protected  void PacketRetransmission(byte[] bytes, String s)
  {
  
     System.out.println("In PacketRetransmission()");
     System.out.println("Missed packets: "+s);
  
     StringTokenizer st = new StringTokenizer(s,",");//tokenizing the missedPacket string with ",".
     byte bytes_with_seq[] = new byte[buffer_size+4];//array with data & seq number at the tail used for packets.
     //rdb InterPacketDelay t = new InterPacketDelay();//object used to set interpacket delay.
     int counter =0,missedPacketNum=0;//variable "counter" used to record number of retransmissions.
     //rm.SendControlInfo(serverCtrlSocket, "Ok");//client synchronization logic using a tcp connection wait.
                                              //System.out.println("Client synchronized with the server.");
     while (st.hasMoreTokens())
     {
        
        missedPacketNum = Integer.parseInt(st.nextToken());//finding out the missed packet sequence number.
        Seqnum_Encoder(seq_num, missedPacketNum);//encoding packet sequence number.
        
        if(missedPacketNum != totalPackets - 1)//if the missed packet is not the last packet.
        {
           //reading missed packet data from "bytes" array into data portion of "bytes_with_seq" array.
           System.arraycopy(bytes,(missedPacketNum * buffer_size),bytes_with_seq,0,buffer_size);
        }
        else//if the missed packet is the last packet.
        {
           if((rempart < buffer_size) && (rempart > 0))//case of last fractional packet.
           {
              //reading the data portion of last packet into data portion of "bytes_with_seq" array.
              System.arraycopy(bytes,(missedPacketNum * buffer_size),bytes_with_seq,0,rempart);
              //filling the extra spaces in the last packet with *'s.
              for(int k=rempart;k<buffer_size;k++)
              {
                 bytes_with_seq[k] = (byte) '*';
              }
           }
           else if((rempart == 0) && (missedPacketNum == totalPackets-1))//case of last full packet.
           {
              //reading the entire last packet into data portion of "bytes_with_seq" array.
              System.arraycopy(bytes,(missedPacketNum * buffer_size),bytes_with_seq,0,buffer_size);
           }
           else
           {
              //do nothing.
           }
        }
        
        //appending a seq no to the tail of a packet.
        System.arraycopy(seq_num,0,bytes_with_seq,buffer_size,4);
        try{
	        ds.send(new DatagramPacket(bytes_with_seq,0,buffer_size+4,ip,clientDataPort));
	      }
	    	catch(IOException ioe){ioe.printStackTrace();}
        counter++;//variable to record number of retransmissions.
        UDPSender.InterPacketDelay.wait(finalDelay);//interpacket delay.
     }//end of while.
      ////System.out.println("Number of retransmissions sent: " + counter);
     //rm.SendControlInfo(serverCtrlSocket, "Ok");//client synchronization logic using a tcp connection wait.
                                        ////System.out.println("Client synchronized with the server.");
                                        //t.wait(50);//inducing a wait purposefully to let the client send some contol information about missing packets.
     //controlMsg = rm.ReceiveControlInfo(clientIP, clientControlPort);//connection with client to find out if there are any missing packets.
     controlMsg = GetMissingPackets();//connection with client to find out if there are any missing packets.
     if(!(controlMsg.equals("")))
			  PacketRetransmission(bytes, controlMsg);
  }
  

   /** Interpacket delay parameter setter.
   */
   public static class InterPacketDelay extends Thread
   {
      public static void wait(int n){       
         try {
            Thread.sleep(n);
         }
         catch(InterruptedException exception)
         {
               exception.printStackTrace();
         }
      }
   }

}