  //++++++++++++++ Arguments ++++++++++++++++++++++++
  // an package class to store the arguments that have been read.
  // it contains the Vector of the argument values as well as a type
  // string that represents the types of the read arguments using
  // the same conventions as the parameter specifications of the Commands.
  // rdb 08/10/01
  //
  //---------------------------------------------------------
  // 05/04/08  rdb: made Vectors generic
  // 10/25/01  added getArgs method
  
  package edu.unh.sdb.util.CommandProcessor;

  //import java.io.*;
  import java.util.Vector;

  class Arguments
  {
    private      StringBuffer  types;   // specif of argument types
    private      Vector<Object>        args;    // the arguments read so far

    //--------- Arguments( ) -----------
    public      Arguments( )
    {
       args = new Vector<Object>();
       types = new StringBuffer();
    }
    //----------- size( ) ---------------
    public int size( ) 
    {
       return this.args.size( );
    }
    //----------- get( int  ) --------------- 
    public Object get( int index ) 
    {
       return this.args.get( index );
    }
    //----------- getArgs( ) --------------- 
    public Vector getArgs( ) 
    {
       return this.args;
    }
    //----------- getType( int  ) --------------- 
    // return type of i-th argument
    public char getType( int index ) 
    {
       if ( index < this.types.length() )
          return this.types.charAt( index );
       else
          return '?';
    }
    //----------- add( Object ) ---------------
    public void add( Object newArg ) 
    {
       this.args.add( newArg );
       this.types.append( determineType( newArg ));
    }
    //------------- determineType( arg ) ------------------------
    private char determineType( Object arg )
    {
       if      ( arg instanceof Integer )
          return 'i';
       else if ( arg instanceof Float )
          return 'f';
       else if ( arg instanceof Double )
          return 'd';
       else if ( arg instanceof String )
          return 's';
       else 
          return '?';
    }
    //----------- setElementAt( Object arg, int i ) ---------------
    public void setElementAt( Object arg, int i ) 
    {
       if ( i >= 0  &&  i < args.size() ) 
       {
          this.args.setElementAt( arg, i);
          this.types.setCharAt( i, determineType( arg ) );
       }
       else
          System.err.println("Arguments.setElementAt: Bad index: " + i );
    }
    //----------- removeElementAt( int i ) ---------------
    public void removeElementAt( int i ) 
    {
       if ( i >= 0  &&  i < args.size() ) 
       {
          this.args.removeElementAt( i );
          this.types.deleteCharAt( i );
       }
       else
          System.err.println("Arguments.removeElementAt: Bad index: " + i );
    }
    //----------- toString( ) ---------------
    public String toString( ) 
    {
       return "<" + this.types + ":" + this.args + ">";
    }
  }
