//----------------------- CommandAction class ----------------------------
//
// CommandAction is conceptually an abstract class, whose main role is 
// to provide an execute method definition, but its constructor also allows
// specification of a output writer for error message and the default
// execute method is very useful as a debugging and prototyping tool.
// 
// Consequently the class is NOT defined as "abstract" and so can be
// instantiated.
//
// History
//  05/29/01  "abstract" designation dropped from class definition.

package edu.unh.sdb.util.CommandProcessor;
import java.io.*;
import java.util.Vector;

public class CommandAction
//public abstract class CommandAction
{
  protected PrintWriter writer = null;

  //------------ CommandAction constructors -----------------------------------
  //
  //------------------ CommandAction() -------------------------
  //--- default ctor uses System.err for error output
  public CommandAction()
  {
     writer = new PrintWriter( new OutputStreamWriter( System.err ), true );
  }

  //------------------ CommandAction( PrintWriter ) ------------------
  //--- or, user can specify an output for errors (which can be null)
  public CommandAction( PrintWriter outfile )
  {
     writer = outfile;
  }

  //------------------ CommandAction( PrintStream ) ------------------
  //--- or, user can specify an output for errors (which can be null)
  public CommandAction( PrintStream outfile )
  {
     writer = new PrintWriter( outfile );
  }

  //------------ execute ---------------------------------------------------
  //
  // The default execute method just sends its arguments to the error output
  //
  public boolean execute( Vector args )
  {
     if ( writer != null )
     {
         writer.println( "CommandAction.execute --" );
         writer.println( args );
     }
     return true;
  }
}
