///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//          U N I V E R S I T Y    O F    N E W    H A M P S H I R E         //
//                                                                           //
//                       COMPUTER  SCIENCE  DEPARTMENT                       //
//                                                                           //
//                    Graphics And Visualization Research                    //
//                                                                           //
//          --------------------------------------------------------         //
//  File: ControlModule.h                                                    //
//                                                                           //
//  The ControlModule is responsible for controlling all data transfer       //
//  between Visit and Granite.                                               // 
//                                                                           //
//                                                                           //
//  Programmer:  Jon Frain                                                   //
//  Creation Date: 7/21/09                                                   //
//                                                                           //
//  History:                                                                 //
//      7-27-09 - Code specific to meshes, metadata, .. were stripped        // 
//                out and given to the RemoteDataSource class                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef _CONTROL_MODULE_H_
#define _CONTROL_MODULE_H_

#include "GraniteRemoteIF.h"

extern "C" {
   #include <SimCodeGlobals.h>
}
#include <VisItDataInterface_V1.h>
#include <VisItControlInterface_V1.h>
#include <SimHelperFunc.h>
#include "RemoteDataSource.h"

#include <vector>

class ControlModule
{
public:  // constructors
            ControlModule(GraniteRemoteIF *r_IF);
    virtual ~ControlModule();

public:  // api
    static int                       CreateDataSource(char* name, char* path);
    static VisIt_MeshData*           GetMesh(int dsID, int domain, 
                                             const char* name);
    static VisIt_SimulationMetaData* GetMetaData(int dsID);
    static VisIt_ScalarData*         GetScalar(int dsID, int domain, 
                                               const char *name);
    static VisIt_DomainList*         GetDomainList(int dsID);

private:  // methods


private:  // members
    static GraniteRemoteIF* remoteConn;
    static vector<RemoteDataSource*> dsVec;

private: // disable default and copy constructor and operator= 
         // - disabling to force passing in of a remote connection 
    ControlModule();
    ControlModule( ControlModule& );
    ControlModule& operator=( ControlModule& ); 
};


#endif // _CONTROL_MODULE_H_
