package edu.unh.sdb.util.VisitGraniteFramework.GraniteServer;

// ++++++++++++++++++++++++++ GraniteServer ++++++++++++++++++++++++++++
// Intended to allow parser to be invoked from any application, not
// just the "main" method of this file.
//
//Generated from template using MakeParser program.
// MakeParser Version 3. 11/01/01
//------------------------- Edit history ---------------------------------
// 08/07/09 jpf: Commented out all buffer relate code as the buffer
//               data was never being sent
//               Also was forced to add calls to buffer clear() in order
//               to prevent overflow exceptions
// 08/06/09 jpf: Modified outputbuffer
// 08/04/09 jpf: Added method for getting number of attributes 
// 07/29/09 jpf: Added methods for querying dimensions of datasource 
// 10/27/09 jpf: Added methods for getting/setting byte order (endianness)
// 11/10/05 rdb: Got incomplete socket code to comple, so that file version works
//               *********** Only file input version works! **********************
// 10/10/05 rdb: Started socket version (not completed)
// 06/17/05 rdb: Finished file input version
//------------------------------------------------------------------------
//

/** GraniteServer
 * 06/17/05 
 *    Argument is command file to be read; results of commands are printed (although
 *          not all the data is always printed)
 * 10/10/05
 *   -If command line argument is an integer, open that socket #
 *    else it should be a file name to be used for non-socket testing
 *   -Reads commands over a socket passed to constructor
 *   -Writes return data as a Java ByteBuffer object with following conventions:
 *       byte
 *         0 - acknowledgement code
 *                   0 operation successful
 *                   1-255 error codes
 *         1 - type of binary data in buffer
 *                  <use Granite StorageType codes>
 *         2 - byte order
 *                   0 BigEndian
 *                   1 LittleEndian
 *         3 - remaining buffer count.
 *               if the data to be returned is larger than the available buffers,
 *               it is returned in multiple writes to the socket. Each write includes
 *               the 8 byte header. 255 is used to indicate 255 or more remaining
 *               buffers.
 *        4-7  integer count of data objects in the remainder of the buffer
 *               it is in the same byte order as the data. This is not the byte
 *               size of the buffer, but number of data objects in it.
 *         >8  data
 *     Issues:
 *        - Should there be a size field for the buffer size as well as for the
 *           data in the buffer. I don't think so, since only the needed data will
 *           be sent out the socket and I think the read will be able to know
 *           what that length is.
 *        - should the data length be in bytes?
 */

import java.io.*;
import java.util.Date;
import java.util.Vector;
import java.util.ArrayList;
import java.nio.*;
import java.net.*;
import javax.net.*;

import edu.unh.sdb.util.CommandProcessor.*;
import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.lattice.*;

//--------------------------GraniteServer class ----------------------
public class GraniteServer
{
   //---------------- "main" variables -----------------------------   
   private static String        inFile     = null;
   private static int           portNumber = -1;

   //--------------- variables generated by MakeParser ------------------
   // for running standalone, keep a reference to an instance available.
   private static GraniteServer theApp = null;
   
   //-------- "Global" instance variables --------------------------------------
   // define the parser data member
   private CommandProcessor parser = null;

   private ArrayList<Object> graniteObjects = new ArrayList<Object>();
   
   //---- Last command information
   byte     resultCode = 0;            // 0 => good, anything else is error code
   Vector   errorArgs = null;          // args array for last command
   String   errorMsg  = null;          // set to error message if cmd failed
   
   // a pointer to the RemoteServer
   private RemoteServer remoteServer = null;

   //---- socket connection information --------------------
   private  InputStream   inStream  = null;
   private  Socket        client    = null;
   private  ServerSocket  server    = null;
   private  PrintWriter   ack       = null;
   
   private  ByteBuffer    outBuffer = null;
   private  FloatBuffer   floatBuf = null;
   private  DoubleBuffer  doubleBuf = null;
   private  IntBuffer     intBuf = null;
   private  ShortBuffer   shortBuf = null;
   
   private  int           bufDataByteSize = 0;     // # bytes of data in buffer
   private  int           bufferSize  = 0;     // Total capacity of the ByteBuffer
   private  int           bufHeaderSize = 8;   // the size of the buffer header

   //----------------- main -----------------------------------------
   public static void main( String argv[] ) throws IOException
   {
      parseArgs( argv );
      
      if ( GraniteServer.portNumber < 0 )
      {        
         //*********** direct tests **************************
         theApp = new GraniteServer( GraniteServer.inFile );
          theApp.readAllCommands();
      }
      else
      {
         //************** following is just a guess and not tested ************
         Socket client       = new Socket();
         theApp              = new GraniteServer( client );
         ServerSocket server = theApp.getServerSocket();
         new GraniteServer( client );
      }
   }

   //----------------- parseArgs( String[] ) -------------------------------
   public static void parseArgs( String argv[] )
   {
      if ( argv.length == 0 )
      {
         System.out.println( "Usage: java GraniteServer <inputFile>" );
         System.out.println( "or   : java GraniteServer socket#" );
         System.exit( 0 );
      }
      try 
      {
          GraniteServer.portNumber = Integer.parseInt( argv[ 0 ] );
      } catch (Exception e) 
      {
      }
      
      if ( GraniteServer.portNumber < 0 )
      {
         GraniteServer.inFile = argv[ 0 ];
      }
   }

   //----------------- GraniteServer -----------------------------------------
   public GraniteServer()
   {
      this.ack = new PrintWriter( System.out, false );

      final int initialBufferSize = 16384;
      
      this.outBuffer = ByteBuffer.allocate( initialBufferSize );
      this.doubleBuf = this.outBuffer.asDoubleBuffer();
      this.floatBuf  = this.outBuffer.asFloatBuffer();
      this.intBuf    = this.outBuffer.asIntBuffer();
      this.shortBuf  = this.outBuffer.asShortBuffer();
      this.bufferSize = initialBufferSize;
      this.bufDataByteSize = initialBufferSize - this.bufHeaderSize;

      this.parser = new CommandProcessor();

      System.out.println("  Creating commands");
      createCommands();
   }

   private void createCommands()
   {
      // create the commands
      try
      {
         //%&%------- Insert defineCommand calls here -----------------------
         parser.defineCommand( "dsCreate", "ss", new GraniteServerAction.DSCreate(this ));
         parser.defineCommand( "dsc", "ss", new GraniteServerAction.DSCreate(this ));
         parser.defineCommand( "tsCreate", "ss", new GraniteServerAction.TSCreate(this ));
         parser.defineCommand( "tsc", "ss", new GraniteServerAction.TSCreate(this ));
         parser.defineCommand( "numSteps", "i", new GraniteServerAction.NumSteps(this ));
         parser.defineCommand( "ns", "i", new GraniteServerAction.NumSteps(this ));
         parser.defineCommand( "getStep", "ii", new GraniteServerAction.GetStep(this ));
         parser.defineCommand( "step", "ii", new GraniteServerAction.GetStep(this ));
         parser.defineCommand( "getTime", "ii", new GraniteServerAction.GetTime(this ));
         parser.defineCommand( "time", "ii", new GraniteServerAction.GetTime(this ));
         parser.defineCommand( "subblock", "ii*", new GraniteServerAction.Subblock(this ));
         parser.defineCommand( "sub", "ii*", new GraniteServerAction.Subblock(this ));
         parser.defineCommand( "getDouble", "iii*", new GraniteServerAction.GetDouble(this ));
         parser.defineCommand( "gd", "iii*", new GraniteServerAction.GetDouble(this ));
         parser.defineCommand( "getFloat", "iii*", new GraniteServerAction.GetFloat(this ));
         parser.defineCommand( "gf", "iii*", new GraniteServerAction.GetFloat(this ));
         parser.defineCommand( "getInt", "iii*", new GraniteServerAction.GetInt(this ));
         parser.defineCommand( "gi", "iii*", new GraniteServerAction.GetInt(this ));
         parser.defineCommand( "getShort", "iii*", new GraniteServerAction.GetShort(this ));
         parser.defineCommand( "gs", "iii*", new GraniteServerAction.GetShort(this ));
         parser.defineCommand( "getByte", "iii*", new GraniteServerAction.GetByte(this ));
         parser.defineCommand( "gb", "iii*", new GraniteServerAction.GetByte(this ));
         parser.defineCommand( "getDoubles", "ii*", new GraniteServerAction.GetDoubles(this ));
         parser.defineCommand( "gds", "ii*", new GraniteServerAction.GetDoubles(this ));
         parser.defineCommand( "getFloats", "ii*", new GraniteServerAction.GetFloats(this ));
         parser.defineCommand( "gfs", "ii*", new GraniteServerAction.GetFloats(this ));
         parser.defineCommand( "getInts", "ii*", new GraniteServerAction.GetInts(this ));
         parser.defineCommand( "gis", "ii*", new GraniteServerAction.GetInts(this ));
         parser.defineCommand( "getShorts", "ii*", new GraniteServerAction.GetShorts(this ));
         parser.defineCommand( "gss", "ii*", new GraniteServerAction.GetShorts(this ));
         parser.defineCommand( "getBytes", "ii*", new GraniteServerAction.GetBytes(this ));
         parser.defineCommand( "gbs", "ii*", new GraniteServerAction.GetBytes(this ));
         parser.defineCommand( "getDoublesByAttr", "i",
                                       new GraniteServerAction.GetDoublesByAttr(this ));
         parser.defineCommand( "gda", "i", new GraniteServerAction.GetDoublesByAttr(this ));
         parser.defineCommand( "getFloatsByAttr", "i", 
                                       new GraniteServerAction.GetFloatsByAttr(this ));
         parser.defineCommand( "gfa", "i", new GraniteServerAction.GetFloatsByAttr(this ));
         parser.defineCommand( "getIntsByAttr", "i", new GraniteServerAction.GetIntsByAttr(this ));
         parser.defineCommand( "gia", "i", new GraniteServerAction.GetIntsByAttr(this ));
         parser.defineCommand( "getShortsByAttr", "i", 
                                       new GraniteServerAction.GetShortsByAttr(this ));
         parser.defineCommand( "gsa", "i", new GraniteServerAction.GetShortsByAttr(this ));
         parser.defineCommand( "getBytesByAttr", "i", new GraniteServerAction.GetBytesByAttr(
                                      this ));
         parser.defineCommand( "gba", "i", new GraniteServerAction.GetBytesByAttr(this ));
         parser.defineCommand( "getDoublesByPoint", "i",
                                       new GraniteServerAction.GetDoublesByPoint(this ));
         parser.defineCommand( "gdp", "i", new GraniteServerAction.GetDoublesByPoint(this ));
         parser.defineCommand( "getFloatsByPoint", "i",
                                       new GraniteServerAction.GetFloatsByPoint(this ));
         parser.defineCommand( "gfp", "i", new GraniteServerAction.GetFloatsByPoint(this ));
         parser.defineCommand( "getIntsByPoint", "i", new GraniteServerAction.GetIntsByPoint(this ));
         parser.defineCommand( "gip", "i", new GraniteServerAction.GetIntsByPoint(this ));
         parser.defineCommand( "getShortsByPoints", "i",
                                       new GraniteServerAction.GetShortsByPoint(this ));
         parser.defineCommand( "gsp", "i", new GraniteServerAction.GetShortsByPoint(this ));
         parser.defineCommand( "getBytesByPoint", "i", new GraniteServerAction.GetBytesByPoint(this ));

         parser.defineCommand( "gbo", "i", new GraniteServerAction.GetByteOrder(this ));
         parser.defineCommand( "getByteOrder", "i", new GraniteServerAction.GetByteOrder(this ));
         parser.defineCommand( "sbo", "is", new GraniteServerAction.SetByteOrder(this ));
         parser.defineCommand( "setByteOrder", "is", new GraniteServerAction.SetByteOrder(this ));

         parser.defineCommand( "getDimensions", "i", new GraniteServerAction.GetDimensions(this));
         parser.defineCommand( "getNumDimensions", "i", new GraniteServerAction.GetNumDimensions(this));

         parser.defineCommand( "getNumAttributes", "i", new GraniteServerAction.GetNumAttributes(this));
         parser.defineCommand( "getAttributeNames", "i", new GraniteServerAction.GetAttributeNames(this));

         parser.defineCommand( "gbp", "i", new GraniteServerAction.GetBytesByPoint(this ));
         parser.defineCommand( "interpDouble", "iif*", new GraniteServerAction.InterpDouble( this ));
         parser.defineCommand( "id", "iif*", new GraniteServerAction.InterpDouble(this ));
         parser.defineCommand( "interpFloat", "iif*", new GraniteServerAction.InterpFloat(this ));
         parser.defineCommand( "if", "iif*", new GraniteServerAction.InterpFloat(this ));
         parser.defineCommand( "interpInt", "iif*", new GraniteServerAction.InterpInt(this ));
         parser.defineCommand( "ii", "iif*", new GraniteServerAction.InterpInt(this ));
         parser.defineCommand( "interpShort", "iif*",  new GraniteServerAction.InterpShort(this ));
         parser.defineCommand( "is", "iif*", new GraniteServerAction.InterpShort(this ));
         parser.defineCommand(  "interpByte", "iif*",  new GraniteServerAction.InterpByte(this ));
         parser.defineCommand( "ib", "iif*", new GraniteServerAction.InterpByte(this ));
         parser.defineCommand( "interpDoubles", "if*", new GraniteServerAction.InterpDoubles(this ));
         parser.defineCommand( "ids", "if*", new GraniteServerAction.InterpDoubles(this ));
         parser.defineCommand( "interpFloats", "if*", new GraniteServerAction.InterpFloats(this ));
         parser.defineCommand( "ifs", "if*", new GraniteServerAction.InterpFloats(this ));
         parser.defineCommand( "interpInts", "if*", new GraniteServerAction.InterpInts(this ));
         parser.defineCommand( "iis", "if*", new GraniteServerAction.InterpInts(this ));
         parser.defineCommand( "interpShorts", "if*", new GraniteServerAction.InterpShorts(this ));
         parser.defineCommand( "iss", "if*", new GraniteServerAction.InterpShorts(this ));
         parser.defineCommand( "interpBytes", "if*", new GraniteServerAction.InterpBytes(this ));
         parser.defineCommand( "ibs", "if*", new GraniteServerAction.InterpBytes(this ));
         parser.defineCommand( "newIterator", "ii*", new GraniteServerAction.NewIterator(this ));
         parser.defineCommand( "ni", "ii*", new GraniteServerAction.NewIterator(this ));
         parser.defineCommand( "next", "i", new GraniteServerAction.Next(this ));
         parser.defineCommand( "init", "i", new GraniteServerAction.Init(this ));
         parser.defineCommand( "valid", "i", new GraniteServerAction.Valid(this ));
         parser.defineCommand( "newBoundsIterator", "ii*",
                                       new GraniteServerAction.NewBoundsIterator(this ));
         parser.defineCommand( "nbi", "ii*", new GraniteServerAction.NewBoundsIterator(this ));
         
         // metadata commands
         parser.defineCommand( "getBounds", "i", new GraniteServerAction.GetBounds(this ));
         parser.defineCommand( "b",         "i", new GraniteServerAction.GetBounds(this ));
         parser.defineCommand( "numAttrs", "i", new GraniteServerAction.NumAttributes(this ));
         parser.defineCommand( "na",       "i", new GraniteServerAction.NumAttributes(this ));
         parser.defineCommand( "storageTypes", "i", new GraniteServerAction.GetStorageTypes(this ));
         parser.defineCommand( "st",           "i", new GraniteServerAction.GetStorageTypes(this ));
         
         //%&%---------- end defineCommand calls ----------------------------
      } 
      catch ( CommandDefException cde )
      {
         System.err.println( cde.getMessage() );
         System.err.print( "*** Error in command specification " );
         if ( cde.command() != null )
            System.err.print( "for: " + cde.command() );
         if ( cde.parmSpec() != null )
            System.err.print( "with parameter specification: " + cde.parmSpec() );
         System.err.println();
      }
   }

   /* ====================================================================== */
   /**
    *  Constructs a GraniteServer with the specified RemoteServer.
    *  If a remote server is present, then the granite server sends
    *  a copy of all output to the RemoteServer, which presumably is
    *  a client application on the other end of a socket.
    */

   public GraniteServer( RemoteServer r )
   {
       this();

       remoteServer = r;
   }

   /* ====================================================================== */

   //----------------- GraniteServer( String ) ------------------------------
   public GraniteServer( String infile )
   {
      this(); // invoke default constructor
      if ( !setInput( infile ) )
      {
         // we could throw an IO Exception here ??
         System.err.println( "File can't be opened: " + infile );
         System.err.println( "Using standard input" );
         theApp.setInput( System.in );
      }
   }

   //----------------- GraniteServer( Server ) ------------------------------
   /**
    * Create a GraniteServer using a socket interface
    */
   public GraniteServer( Socket client )
   {
      this(); // invoke default constructor
      this.client = client;
      
      try
      {
         this.server = ServerSocketFactory.getDefault()
                                          .createServerSocket( client.getPort() );
         this.inStream = this.client.getInputStream();
         this.client.connect( this.server.getLocalSocketAddress() );
         theApp.setInput( this.inStream );
         //this.ack = new PrintWriter( this.client.getOutputStream() );
      }
      catch ( IOException ioex )
      {
         System.err.println( "GranteServer construction failure: " + ioex );
         System.exit( -1 );
      }
   }
   
   /**
    * Listener class: acquired from Michel Charpentier
    * @author rdb
    *
    */
   class Listener extends Thread 
   {
       private final ServerSocket serv;

       public Listener () throws java.io.IOException 
       {
           serv = new ServerSocket( GraniteServer.portNumber );
       }

       public void run () 
       {
           System.out.println( new Date() + ": server running" );
           while ( true ) 
           {
               try {
                   Socket s = serv.accept();
                   s.setKeepAlive( true );
                   new ConnectionHandler( s ).start();
               } catch ( java.io.IOException e ) {
                   System.out.println( "Can't create socket: " + e.getMessage());
                   return;
               }
           }
       }
   }

   /**
    * ConnectionHandler class: acquired from Michel Charpentier
    * @author rdb
    *
    */
   class ConnectionHandler extends Thread 
   {
       private Socket socket;

       public ConnectionHandler (Socket s) 
       {
           socket = s;
       }

       public void run () 
       {
           InetAddress info = socket.getInetAddress();
           String name = info.getHostName() + "(" + info.getHostAddress() + ")";
           System.out.println( new Date() + ": listening to "+name );
           try {
               BufferedReader in =  new BufferedReader
                   ( new InputStreamReader( socket.getInputStream() ));
               PrintWriter out = new PrintWriter
                   ( new BufferedWriter
                    ( new OutputStreamWriter( socket.getOutputStream())), true );
               out.println( "Hello " + name );
               try {
                   while ( true ) {
                       String line;
                       try {
                           line = in.readLine();
                       } catch (java.io.IOException e) {
                           line = null;
                       }
                       if ( line == null )
                           break;
                       System.out.println( name + ":" + line );
                       out.println( "You said: " + line.toUpperCase());
                   }
               } finally {
                   System.out.println( "Connection with " + name + " closed" );
                   in.close();
                   out.close();
                   socket.close();
               }
           } catch ( java.io.IOException e ) {
               System.err.println( "Something went wrong with " + name + ": "+
                                  e.getMessage());
           }
       }
   }

   /**
    * Return the client socket if this is a socket instantiation, else null.
    */
   Socket getClientSocket() 
   {
      return this.client;
   }

   /**
    * Return the server socket if this is a socket instantiation, else null.
    *
    */
   ServerSocket getServerSocket() 
   {
      return this.server;
   }

   //++++++++++++++++++Application methods +++++++++++++++++++++++++++++++

   /* ====================================================================== */
   /**
    *  Process a single command.  How to handle the return value?
    *  Array of bytes?  Separate API?
    */
   public boolean processCommand( String command ) throws IOException
   {
       boolean success = false;

       System.out.printf( "processCommand: Got command '%s'\n",command );

       try
       {
           return parser.parseCommand( command );
       }
       catch( CommandException e )
       {
           System.err.printf( "Command Error '%s'\n", e );
       }

       return success;
   }

   /* ====================================================================== */;

   /**
    * log information about an error on last command
    */
   void logError( Vector args, String msg )
   {
      logError( args, msg, null );
   }
   
   /**
    * log information about an error on last command
    */
   void logError( Vector args, String msg, Exception ex )
   {
      this.errorArgs = (Vector)args.clone();
      this.errorMsg = msg;
      if ( ex == null )
      {
         System.out.println( "logError: " +  msg );
      }
      else
      {
         System.out.println( "logError: Exception " 
                                       + ex.getClass().getName() + ":" 
                                       + ex.getMessage() + " -- "
                                       + msg );
         ex.printStackTrace();
      }
   }
   
   /**
    * Log the information being sent.
    * 
    * @param cmdStatus
    * @param type
    * @param len
    */
   void logSendAction( byte cmdStatus, String type, int len )
   {
      this.resultCode = cmdStatus;
      // temporary test
      if ( cmdStatus == 0 )
      {
         System.out.println( "Sending " + len + " " + type );
         //this.ack.println( "Sending " + len + " " + type );
      }
      else
         System.out.println( "****Failure to send " + type + "*****");
         //this.ack.println( "****Failure to send " + type + "*****");
   }
   
   /**
    * Extract an Bounds of dimension 'dim' from the argument array beginning at 
    * 'start' position. The bounds is specified by 2 arrays of length 'dim' 
    * (lower and upper bounds). 
    * 
    * @param args
    * @param dim
    * @param start
    * @return ISBounds
    */
   ISBounds getBoundsFromArgs( Vector args, int dim, int start )
   {
      ISBounds bounds = null;
      
      try
      {
         int[] lower = getIntsFromArgs( args, dim, start );
         int[] upper = getIntsFromArgs( args, dim, start+dim );
         bounds = new ISBounds( lower, upper );              
      }
      catch ( RuntimeException ex )
      {
         throw ex;
      }
      return bounds;
   }
   
   /**
    * Extract an ISID of dimension 'd' from the argument array beginning at 
    * 'start' position.
    * @param args
    * @param dim
    * @param start
    * @return ISID
    */
   IndexSpaceID getISIDFromArgs( Vector args, int dim, int start )
   {

      int[] pos = new int[ dim ];
      int arg = start;
      
      for ( int i = 0; i < dim; i++, arg++ ) {
         pos[i] = ((Integer) args.get( arg )).intValue();
      }
      return new IntegerIndexSpaceID( pos );
   }
   
   /**
    * Extract an arbitrary (float) position  of dimension 'd' from the argument 
    * array beginning at 'start' position.
    * @param args
    * @param dim
    * @param start
    * @return ISID
    */
   float[] getLocFromArgs( Vector args, int dim, int start )
   {
      float[] pos = new float[ dim ];
      int arg = start;
      
      for ( int i = 0; i < dim; i++, arg++ )
         pos[i] = ((Integer) args.get( arg )).floatValue();
      return pos;
   }
   
   /**
    * get a Cell from the specified DataCollection that encompasses the
    * position included in the args vector beginning at the start position.
    * 
    * This is not a real implementation. It is meaningless.
    * 
    * @param dc   - data collection
    * @param args
    * @param start
    * @return
    */
   ArrayCell getCell( DataCollection dc, Vector args, int start )
   {
      float[] position = getLocFromArgs( args, dc.dim(), start );
      //return new ArrayCell( new Point[ dc.dim() ], new FloatDatum [ dc.dim() ] );
      return null;
   }
   
   /**
    * Extract an array of length 'num' from the argument array beginning at 
    * 'start' position.
    * @param args
    * @param num
    * @param start
    * @return int[] of length 'dim'
    */
   int[] getIntsFromArgs( Vector args, int num, int start )
   {
      int[] pos = new int[ num ];
      int arg = start;
      
      for ( int i = 0; i < num; i++, arg++ )
         pos[i] = ((Integer) args.get( arg )).intValue();
      return pos;
   }
   
   /**
    * Add a new Granite object to database. A Granite object can be a
    * DataSource, DataBlock, ISIterator, or ISBoundsIterator. Returns the
    * Granite Id for that object.
    */
   int addGraniteObject( Object newObject )
   {
      this.graniteObjects.add( newObject );
      return this.graniteObjects.size() - 1;
   }

   /**
    * Find and return the Granite object associated with the id.
    * 
    * @param graniteId
    */
   Object findGraniteObject( int id )
   {
      if ( id >= 0 && id < graniteObjects.size() )
         return graniteObjects.get( id );
      else 
         return null;
   }
   
   //+++++++++++++++++++ send data methods +++++++++++++++++++++++++++++
   
   /**
    * Send value back to client.
    */
   public void sendData( byte retCode, byte value )
   {
//       logSendAction( retCode, "byte", 1 );
//       putHeader( retCode, StorageType.Byte, 0, 1 );
//       if ( retCode == 0 )
//          this.outBuffer.put( value );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send values back to client.
    */
   public void sendData( byte retCode, byte[] value )
   {
//       logSendAction( retCode, "bytes", value.length );
//       int sType        = StorageType.Byte;
//       int bufDataSize  = this.bufDataByteSize;
//       int extraBuffers = 0;
//       int lastCopySize = 0;
//       int copyFrom     = 0;
      
//       int dataLen      = 0;
//       if ( retCode == 0 )
//       {
//          extraBuffers = ( dataLen - 1 ) / bufDataSize;
//          lastCopySize = dataLen - extraBuffers * bufDataSize;
//       }
            
//       for ( int i = 0; i < extraBuffers; i++ )
//       {
//          putHeader( retCode, sType, extraBuffers, bufDataSize );
//          this.outBuffer.put( value, copyFrom, bufDataSize );
//          copyFrom += bufDataSize; 
//          extraBuffers--;
//          writeBuffer();
//       }
//       // Now write last buffer
//       putHeader( retCode, sType, 0, lastCopySize );
//       if ( retCode == 0 )
//          this.outBuffer.put( value, copyFrom, lastCopySize );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send value back to client.
    */
   public void sendData( byte retCode, double value )
   {
//       logSendAction( retCode, "double", 1 );
//       putHeader( retCode, StorageType.Double, 0, 1 );
//       if ( retCode == 0 )
//          this.doubleBuf.put( value );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send values back to client.
    */
   public void sendData( byte retCode, double[] value )
   {
//       logSendAction( retCode, "doubles", value.length );
//       int sType        = StorageType.Double;
//       int bufDataSize  = this.bufDataByteSize;
//       int extraBuffers = 0;
//       int lastCopySize = 0;
//       int copyFrom     = 0;
//       int dataLen      = 0;
      
//       if ( retCode == 0 )
//       {
//          extraBuffers = ( dataLen - 1 ) / bufDataSize;
//          lastCopySize = dataLen - extraBuffers * bufDataSize;
//       }
            
//       for ( int i = 0; i < extraBuffers; i++ )
//       {
//          putHeader( retCode, sType, extraBuffers, bufDataSize );
//          this.doubleBuf.put( value, copyFrom, bufDataSize );
//          copyFrom += bufDataSize; 
//          extraBuffers--;
//          writeBuffer();
//       }
//       // Now write last buffer
//       putHeader( retCode, sType, 0, lastCopySize );
//       if ( retCode == 0 )
//          this.doubleBuf.put( value, copyFrom, lastCopySize );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send value back to client.
    */
   public void sendData( byte retCode, float value )
   {
//       System.out.println( "float: " + value );
//       logSendAction( retCode, "float", 1 );
//       putHeader( retCode, StorageType.Float, 0, 1 );
//       if ( retCode == 0 )
//          this.floatBuf.put( value );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send values back to client.
    */
   public void sendData( byte retCode, float[] value )
   {

 //      int initLimit = floatBuf.limit();
//       // number of bytes = 4 times number of floats
//       int floatLength = value.length * 4;

// //       if (value.length > floatBuf.limit()) {
// //          this.outBuffer       = ByteBuffer.allocate( value.length * 4 );
// //          this.doubleBuf       = this.outBuffer.asDoubleBuffer();
// //          this.floatBuf        = this.outBuffer.asFloatBuffer();
// //          this.intBuf          = this.outBuffer.asIntBuffer();
// //          this.shortBuf        = this.outBuffer.asShortBuffer();
// //          this.bufferSize      = value.length;
// //          this.bufDataByteSize = value.length - this.bufHeaderSize;
// //       }

//       logSendAction( retCode, "floats", value.length );
//       int sType = StorageType.Float;
      
//       int bufDataSize  = this.bufDataByteSize / StorageType.byteSize( sType );
//       int extraBuffers = 0;
//       int lastCopySize = 0;
//       int copyFrom     = 0;
//       int dataLen      = value.length;




//       if ( retCode == 0 )
//       {
//          extraBuffers = ( dataLen - 1 ) / bufDataSize;
//          lastCopySize = dataLen - extraBuffers * bufDataSize;
//       }

//       System.out.println("     value.length = " + value.length) ;
//       System.out.println("     extraBuffers = " + extraBuffers  + " lastCopySize = " + lastCopySize);
//       System.out.println("     floatBuf.remaining() = " + floatBuf.remaining());

//       for ( int i = 0; i < extraBuffers; i++ )
//       {
//          floatBuf.clear();
//          System.out.println("     i = " + i);
//          putHeader( retCode, sType, extraBuffers, bufDataSize );
//          try {
//             this.floatBuf.put( value, copyFrom, bufDataSize );
//          }
//          catch (java.nio.BufferOverflowException e){
//              System.out.println("*** java.nio.BufferOverflowException caught in FloatBuffer.put() method " +
//                                 " in GraniteServer.sendData(byte retcode, float[] value)***");
//              System.out.println(" bufDataSize = " + bufDataSize + 
//                                 "  FloatBuffer.remaining() = " + floatBuf.remaining() +
//                                 "  FloatBuffer.limit() = " + floatBuf.limit());
//              System.exit(0);
//          }
//          copyFrom += bufDataSize; 
//          extraBuffers--;
//          writeBuffer();
//       }


//       System.out.println("  writing last buffer!!!!");
//       // Now write last buffer
//       floatBuf.clear();
//       putHeader( retCode, sType, 0, lastCopySize );
//       if ( retCode == 0 )
//          this.floatBuf.put( value, copyFrom, lastCopySize );

//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send value back to client.
    */
   public void sendData( byte retCode, int value )
   {
//       System.out.println( "int: " + value );
//       logSendAction( retCode, "int", 1 );
//       putHeader( retCode, StorageType.Int, 0, 1 );
//       if ( retCode == 0 )
//          this.intBuf.put( value );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send values back to client.
    */
   public void sendData( byte retCode, int[] value )
   {
 //      logSendAction( retCode, "int", value.length );
//       int sType = StorageType.Int;
      
//       int bufDataSize  = this.bufDataByteSize;
//       int extraBuffers = 0;
//       int lastCopySize = 0;
//       int copyFrom     = 0;
//       int dataLen      = 0;

//       // Only write data if good return code. 
//       if ( retCode == 0 )
//       {
//          extraBuffers = ( dataLen - 1 ) / bufDataSize;
//          lastCopySize = dataLen - extraBuffers * bufDataSize;
//       }
      
//       for ( int i = 0; i < extraBuffers; i++ )
//       {
//          putHeader( retCode, sType, extraBuffers, bufDataSize );
//          this.intBuf.put( value, copyFrom, bufDataSize );
//          copyFrom += bufDataSize; 
//          extraBuffers--;
//          writeBuffer();
//       }
//       // Now write last buffer
//       putHeader( retCode, sType, 0, lastCopySize );

//       if ( retCode == 0 )
//          this.intBuf.put( value, copyFrom, lastCopySize );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send value back to client.
    */
   public void sendData( byte retCode, short value )
   {
//       logSendAction( retCode, "short", 1 );
//       putHeader( retCode, StorageType.Short, 0, 1 );
//       if ( retCode == 0 )
//          this.shortBuf.put( value );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   /**
    * Send values back to client.
    */
   public void sendData( byte retCode, short[] value )
   {
 //      logSendAction( retCode, "shorts", value.length );
//       int sType = StorageType.Short;
      
//       int bufDataSize  = this.bufDataByteSize;
//       int extraBuffers = 0;
//       int lastCopySize = 0;
//       int copyFrom     = 0;
//       int dataLen      = 0;
      
//       if ( retCode == 0 )
//       {
//          extraBuffers = ( dataLen - 1 ) / bufDataSize;
//          lastCopySize = dataLen - extraBuffers * bufDataSize;
//       }
      
//       for ( int i = 0; i < extraBuffers; i++ )
//       {
//          putHeader( retCode, sType, extraBuffers, bufDataSize );
//          this.shortBuf.put( value, copyFrom, bufDataSize );
//          copyFrom += bufDataSize; 
//          extraBuffers--;
//          writeBuffer();
//       }
//       // Now write last buffer
//       putHeader( retCode, sType, 0, lastCopySize );
//       if ( retCode == 0 )
//          this.shortBuf.put( value, copyFrom, lastCopySize );
//       writeBuffer();

      if( remoteServer != null )
          remoteServer.sendMessageToClient( (int)retCode, value );
   }

   //++++++++++++++++++Accessor methods +++++++++++++++++++++++++++++++
   //----------------- boolean setInput( InputStream ) --------------------
   //  set the input file for parsing the language
   public void setInput( InputStream inFile )
   {
      this.inStream = inFile;
      this.parser.newReader( new InputStreamReader( inFile ) );
   }

   //----------------- boolean setInput( String ) --------------------
   //  set the input file for parsing the language
   public boolean setInput( String inFileName )
   {
      InputStreamReader inFile = null;
      try
      {
         this.inStream = new FileInputStream( inFileName );
         inFile = new InputStreamReader( this.inStream );
         this.parser.newReader( inFile );
      } 
      catch ( IOException e )
      {
      }
      return inFile != null;
   }
   
   /**
    * put the header information into the buffer.
    * 
    * Really need the boolean to be an byte code
    *
    */
   public void putHeader( byte retCode, int dataType, int extraBufs, int len  )
   {

      outBuffer.clear();

      byte order = 0;  
      if ( this.outBuffer.order() == ByteOrder.LITTLE_ENDIAN )
         order = 1;
 
      try {     
         if ( retCode == 0 )
            this.outBuffer.put( (byte) 0 );
         else
            this.outBuffer.put( (byte) 1 );
         
         this.outBuffer.put( (byte) dataType );
         this.outBuffer.put( order );

         if ( extraBufs < 256  )
            this.outBuffer.put( (byte) extraBufs );
         else 
            this.outBuffer.put( (byte) 255 );

         this.outBuffer.putInt( len );
      }
      catch (java.nio.BufferOverflowException e) {
         System.out.println("*** java.nio.BufferOverflowException caught in FloatBuffer.put() method " +
                            " in GraniteServer.putHeader ***");
         System.out.println("  outBuffer.remaining() = " + outBuffer.remaining() +
                            "  outBuffer.limit() = " + outBuffer.limit());
         System.exit(0);          
      }
   }
   /**
    * write the current buffer to the socket.
    * Not implemented.
    *
    */  
   public void writeBuffer()
   {
      System.out.println( "Writing a buffer at position: " + this.outBuffer.position() );
      //this.ack.println( "Writing a buffer at position: " + this.outBuffer.position() );
      this.outBuffer.rewind();
      // write buffer 
   }

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   //----------------- readAllCommands( ) ---------------------------
   public void readAllCommands() throws IOException
   {
       BufferedReader r = new BufferedReader( new FileReader(inFile) );

       try
       {
           String line = r.readLine();

           while( line != null )
           {
               System.err.printf("readAllCommands: about to process line '%s'\n",line);

               processCommand( line );

               line = r.readLine();
           }
       }
       catch( Exception e )
       {
           System.err.printf("got exception '%s'\n", e );
           System.exit(-1);
       }

       //old code
      //this.parser.readAllCommands();
   }

   //----------------- nextCommand( ) ---------------------------
   // read and execute one command
   public boolean nextCommand() throws IOException
   {
      return this.parser.nextCommand();
   }
}
