#include <jni.h>

#include <fstream>
#include <cstdlib>
#include <iostream>
#include <udt.h>

#include "RemoteUDTServer.h"

using namespace std;

extern "C" {
void* monitor(void*);

JNIEXPORT jint JNICALL 
Java_RemoteUDTServer_nativeSendData(JNIEnv *env, jclass cls, jint port, jbyteArray buf, jint size)
{
    //Try to get a direct pointer to the java direct byte buffer
    jbyte *cppPtr = env->GetByteArrayElements(buf, 0); //not (env, buf)
    if(cppPtr == NULL){
    	return 0; /* Exception occurred */
    }
		cout << "Size:  " << size << "\n";
		//return 1;
    //If successfully got the pointer, send the data after notifying the peer 
    //of the data size
    //char data[size];
    //memcpy(data, cppPtr, size);
    char* data = (char*)(&(*cppPtr));
    
   UDTSOCKET serv = UDT::socket(AF_INET, SOCK_STREAM, 0);
//
//#ifdef WIN32
//   int mss = 1052;
//   UDT::setsockopt(serv, 0, UDT_MSS, &mss, sizeof(int));
//#endif
//
   sockaddr_in my_addr;
   my_addr.sin_family = AF_INET;
   my_addr.sin_port = htons(port);
   my_addr.sin_addr.s_addr = INADDR_ANY;
   memset(&(my_addr.sin_zero), '\0', 8);
	 
   if (UDT::ERROR == UDT::bind(serv, (sockaddr*)&my_addr, sizeof(my_addr)))
   {
      cout << "bind: " << UDT::getlasterror().getErrorMessage() << endl;
      return 0;
   }
	
   cout << "server is ready at port: " << port << endl;

   UDT::listen(serv, 1);

   sockaddr_in their_addr;
   int namelen = sizeof(their_addr);

   UDTSOCKET fhandle;

   if (UDT::INVALID_SOCK == (fhandle = UDT::accept(serv, (sockaddr*)&their_addr, &namelen)))
   {
      cout << "accept: " << UDT::getlasterror().getErrorMessage() << endl;
      return 0;
   }

   UDT::close(serv);

   pthread_create(new pthread_t, NULL, monitor, &fhandle);

   if (UDT::ERROR == UDT::send(fhandle, (char*)&size, sizeof(jint), 0))
   {
      cout << "send data size: " << UDT::getlasterror().getErrorMessage() << endl;
      return 0;
   }


   UDT::TRACEINFO trace;
   UDT::perfmon(fhandle, &trace);
	 int sentDataSize = 0;
	 int handle;
   // send the file
   if (UDT::ERROR == (sentDataSize=UDT::send(fhandle, data, size, 0, &handle)))
   {
      cout << "send data: " << UDT::getlasterror().getErrorMessage() << endl;
      return 0;
   }
   cout << sentDataSize <<" bytes sent to client" << endl;
   //ReleaseByteArrayElements(buf, data, JNI_ABORT);
   UDT::perfmon(fhandle, &trace);
   cout << "speed = " << trace.mbpsSendRate << " mbps" << endl;

   UDT::close(fhandle);
    
    env->ReleaseByteArrayElements(buf, cppPtr, 0);
    return 1;
}

void* monitor(void* s)
{
   UDTSOCKET u = *(UDTSOCKET*)s;

   UDT::TRACEINFO perf;

   cout << "SendRate(Mb/s) RTT(ms) FlowWindow PktSndPeriod(us) RecvACK RecvNAK" << endl;

   while (true)
   {
      sleep(1);
      if (UDT::ERROR == UDT::perfmon(u, &perf))
      {
         cout << "perfmon: " << UDT::getlasterror().getErrorMessage() << endl;
         break;
      }

      cout << perf.mbpsSendRate << "\t" 
           << perf.msRTT << "\t" 
           << perf.pktFlowWindow << "\t" 
           << perf.usPktSndPeriod << "\t" 
           << perf.pktRecvACK << "\t" 
           << perf.pktRecvNAK << endl;
   }

   return NULL;
}

};
