/**
 * @author Sam Vohr
 *
 * This is a simple program using granite that creates smaller data sets
 * from a larger one. The purpose of this is to get an understanding of how
 * granite works as a user.
 *
 */

package selection; 

import java.io.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;

public class SamSelection
{

  //private static int[]  stepsArg = {1,1,1};

  public static void printUsage( )
  {
    System.err.println ("java SamSelection [fdl file for src] [num of res]"); // number of resolution
    System.err.println ("(need both granite.jar and . in classpath)");
  }

  public static void main( String[] args )
  {
    DataSource ds     = null;
    FDLDescriptor fdl = null;

    String fileName = null;
    int numOfRes = 0;

    // get the file to be read.
    if ( args.length > 1 ) {
      fileName = args[0];
      numOfRes = Integer.parseInt(args[1]);
    } else {
      System.err.println ("Missing Arguments!");
      printUsage();
      System.exit(1);
    }

    fdl = FDLDescriptor.create(fileName);
    ds = openDataSource ( "thing", fdl ); 
    ds.activate();


    // build arrays for filenames and files 
    String [] outFileNames = new String [numOfRes];
    String [] outFDLNames  = new String [numOfRes];

    FileOutputStream [] outFile = new FileOutputStream [numOfRes];

    String outFileName = fileName.substring(0,fileName.lastIndexOf('.'));
    
    // create files for the new resolutions
    for (int i=1; i <= numOfRes; i++) {
      try {
        outFDLNames[i-1] = outFileName + "-" + i + ".fdl";
        outFileNames[i-1] = outFileName + "-" + i + ".dtd";

//        System.out.println("outFDLNames[i-1]: " + outFDLNames[i-1]);
//        System.out.println("outFileNames[i-1]: " + outFileNames[i-1]);
        
        outFile[i-1] = new FileOutputStream(outFileNames[i-1]);
      } catch (Exception e) {
        System.err.println ("Cannot open output file.");
        System.exit(1);
      }
    }
    
    ISBounds bounds = ds.getBounds(); // all of the data set.
    System.out.println("inside main, ds bounds is: " + bounds);
    int [] arr = ds.getDimensionsArray();
    for(int i=0; i<arr.length; i++)
    	System.out.println(arr[i]);
    
    System.out.println("------shape--------");
    
    int [] shape = {bounds.getDimension(0), bounds.getDimension(1), 1};
    // I think that this is the correct shape.
    for(int i=0; i<shape.length; i++)
    	System.out.println(shape[i]);
    
    ISBoundsIterator sliceIter = new ISBoundsIterator (ds.getBounds(), shape);
    System.out.println("sliceIter: " + sliceIter);
    
    Datum dat = ds.createDatum();
    
    // Store the slices for error calculation.
    int sliceBufferSize = (int)(Math.pow(2,numOfRes)) + 1;
    DataBlock [] sliceBuffer = new DataBlock [sliceBufferSize]; // 0 to 2^2+1=5
    
    for ( int i=0; sliceIter.valid(); sliceIter.next(), i++ ) {
    	
      ISBounds slice = sliceIter.copyBounds();
      System.out.println("slice:" + slice);
      DataBlock sliceData = new BasicBlock (slice, ds.getRecordDescriptor());
      
      // need to save 2^numOfRes slices in order to do interpolation
      // of error calculation. 
      int sliceNum = i % (int)(Math.pow(2,numOfRes));
      System.out.println("i: " + i + ", sliceNum: " + sliceNum);
      
      if (sliceNum == 0) {
          if (i != 0) {
              // store the slice in the final position
              sliceBuffer[numOfRes - 1] = sliceData;
              
              // calculate the error... later
          }
      }
      sliceBuffer[sliceNum] = sliceData;
      //---------------------- don't need above? by Can -----------------
//      for(int k=0; k<sliceBuffer.length; k++)
//    	  System.out.println("sliceBuffer["+k+"]: " + sliceBuffer[k]);

      for ( int j=1; j <= numOfRes; j++ ) {

        // only build iterators for things we need
        // need every 2^j slice, where j is the level of resolution
        if ( (i % (int)(Math.pow(2,j))) == 0 ) {
          
          int [] skip = {j,j,1};
          ISIterator iter = new ISIterator (slice, skip);

          System.out.println("i: " + i + ", skip: {" + j + ", " + j + ", 1}" );
          // go through the data set, and output 
          for ( ; iter.valid(); iter.next() )
          {
            try {
              ds.datum( dat, iter );
//              System.out.println("dat: " + dat + ", iter: " + iter);
              outFile[j-1].write( dat.getBytes() );
            } catch (Exception e) {
              System.err.println ("Error in writing resolution " + j);
              System.exit(1);
            }
          }
        }
      }
    }

    // close everything
    try {
      for ( int i=0; i<numOfRes; i++ )
        outFile[i].close();
    } catch (Exception e) {
      System.err.println("unable to close file");
    }

    // write the FDL XML Files
    try {
      for ( int i=0; i<numOfRes; i++ ) {
        FDLDescriptor fd = (FDLDescriptor)fdl.clone();
        XFDLWriter xfdlWriter = new XFDLWriter ();
        
        System.out.println("BEFORE, fd bounds: " + fd.getBounds() + ", 2^i+1: " + (int)(Math.pow(2,i+1))) ;
        fd.setBounds( shrinkByN(fd.getBounds(),(int)(Math.pow(2,i+1))) );
        System.out.println("AFTER, fd bounds: " + fd.getBounds() ) ;
        fd.setDataFileURL ( outFileNames[i] );
        
        xfdlWriter.setOutputFile (outFDLNames[i]);
        
        System.out.println("outFileName: " + outFileNames[i] + "\n outFDLName: " + outFDLNames[i]);
        xfdlWriter.writeFDL (fd);
        xfdlWriter.close();
      }
    } 
    catch (Exception e) {
      System.err.println("Couldn't write FDL file.");
      System.exit( -1 );
    }

/*    // Write the new FDL file
    try {
      FDLDescriptor fd = (FDLDescriptor)fdl.clone();
      fd.setBounds( shrinkByN(fd.getBounds(),2) );
      XFDLWriter xfdlWriter = new XFDLWriter ();
      xfdlWriter.setOutputFile (outFDLName);
      xfdlWriter.writeFDL (fd);
      xfdlWriter.close();
    } 
    catch (Exception e) {
      System.err.println("Couldn't write FDL file.");
      System.exit( -1 );
    }
*/
    // done.
    
    return;
  }

  /**
   * utility function borrowed from BasicDemo 
   */
  private static DataSource openDataSource( String dsName, FDLDescriptor fdlD )
  {
    DataSource ds = null;
    try
    {
      ds = DataSource.create( dsName, fdlD );
    } catch ( Exception ex )
    {
      System.err.println( "Open error: " + ex.getMessage() );
      System.exit( -1 );
    }
    return ds;
  }


  /**
   * returns an ISBounds with each dimension reduced by half
   */
  private static ISBounds shrinkByN (ISBounds b, int n)
  {
    int [] dims = b.getDimensionsArray();

    for (int i=0; i < dims.length ; i++) 
    {
//    	System.out.println("dims[i]:" + dims[i]);
    	dims[i] /= n;
//    	System.out.println("dims[i] /= n:" + dims[i]);
    }

    return new ISBounds (dims);
  }
}

